//
// ConfigurationTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.application;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigDecimal;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.application.ConfigReader;
import com.dakengo.libs.application.Configuration;
import com.dakengo.libs.application.ConfigVariableValueProvider;
import com.dakengo.libs.application.Constants;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.KoRules;
import com.dakengo.tests.utils.BaseConfiguredTests;

/**
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class ConfigurationTests extends BaseConfiguredTests
{
   private class EmptyProvider implements ConfigVariableValueProvider
   {
      @Override
      public String getValue()
      {
         return ""; //$NON-NLS-1$
      }
   }

   private class FooProvider implements ConfigVariableValueProvider
   {
      @Override
      public String getValue()
      {
         return "fubar"; //$NON-NLS-1$
      }
   }

   private class BazProvider implements ConfigVariableValueProvider
   {
      @Override
      public String getValue()
      {
         return "$FOO"; //$NON-NLS-1$
      }
   }

   public ConfigurationTests() throws IOException
   {
   }

   @Before
   public void setUp()
   {
      m_config2 = new Configuration();
   }

   @Test
   public void testExpandVariables()
   {
      HashMap<String,ConfigVariableValueProvider> providers = new HashMap<String,ConfigVariableValueProvider>();
      String actual = Configuration.expandVariables( "", '$', providers ); //$NON-NLS-1$
      assertThat( "input = ''", actual, is( "" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      actual = Configuration.expandVariables( "abc", '$', providers ); //$NON-NLS-1$
      assertThat( "input = 'abc'", actual, is( "abc" ) ); //$NON-NLS-1$ //$NON-NLS-2$

      String[] input = new String[]
      {
         "", "abc", "$FOO", "$BAR", "$BAZ", "a$FOO", "a$BAR", "a$BAZ", "$FOOz", "$BARz", "$BAZz", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$
         "$A", "a$DEF", "a$FOO$BAR$BAZb", "a$X$FOO", "$", "$$", "$$$", "$$$$", "$$A", "$$FOO",    //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$
         "a$$b", "a$$$b"                                                                          //$NON-NLS-1$ //$NON-NLS-2$
      };
      String[] expected = new String[]
      {
         "", "abc", "fubar", "", "$FOO", "afubar", "a", "a$FOO", "fubarz", "z", "$FOOz",          //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$
         "$A", "a$DEF", "afubar$FOOb", "a$Xfubar", "", "$", "$", "$$", "$A", "$FOO",              //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$
         "a$b", "a$$b"                                                                            //$NON-NLS-1$ //$NON-NLS-2$
      };
      assertThat( input.length, is( expected.length ) );
      providers.put( "FOO", new FooProvider() );    //$NON-NLS-1$
      providers.put( "BAR", new EmptyProvider() );  //$NON-NLS-1$
      providers.put( "BAZ", new BazProvider() );    //$NON-NLS-1$
      for (int i = 0; i < input.length; i++)
         assertThat( String.format( "input = '%s'", input[i] ), Configuration.expandVariables( input[i], '$', providers ), is( expected[i] ) ); //$NON-NLS-1$
   }

   @Test
   public void testInstanceCalculation()
   {
      Integer[] expected = new Integer[] { 1 };
      Collection<Integer> actuals = Configuration.computeInstanceCounts( 1, 1 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 1, 1, 1 };
      actuals = Configuration.computeInstanceCounts( 3, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 1, 1, 1, 1 };
      actuals = Configuration.computeInstanceCounts( 4, 4 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 1 };
      actuals = Configuration.computeInstanceCounts( 1, 2 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 1, 1, 1 };
      actuals = Configuration.computeInstanceCounts( 3, 4 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 1, 1, 1 };
      actuals = Configuration.computeInstanceCounts( 3, 6 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 2 };
      actuals = Configuration.computeInstanceCounts( 2, 1 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 3 };
      actuals = Configuration.computeInstanceCounts( 3, 1 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 1, 1, 1 };
      actuals = Configuration.computeInstanceCounts( 3, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 2, 1, 1 };
      actuals = Configuration.computeInstanceCounts( 4, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 2, 2, 1 };
      actuals = Configuration.computeInstanceCounts( 5, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 2, 2, 2 };
      actuals = Configuration.computeInstanceCounts( 6, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 3, 2, 2 };
      actuals = Configuration.computeInstanceCounts( 7, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 3, 3, 2 };
      actuals = Configuration.computeInstanceCounts( 8, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );

      expected = new Integer[] { 3, 3, 3 };
      actuals = Configuration.computeInstanceCounts( 9, 3 );
      assertThat( actuals.toArray( expected ), is( expected ) );
   }

   @Test
   public void testDefaults() throws URISyntaxException
   {
      String file_sep = System.getProperty( "file.separator" ); //$NON-NLS-1$
      assertThat( m_config2.getDateFormat().toPattern(), is( Configuration.DEFAULT_DATE_FORMAT ) );
      assertThat( m_config2.getEngineName(), is( Configuration.DEFAULT_ENGINE_NAME ) );
      assertThat( m_config2.getFilenameFormat(), is( Configuration.DEFAULT_FILENAME_FORMAT ) );
      assertThat( m_config2.getKomi(), is( Game.DEFAULT_KOMI ) );
      assertThat( m_config2.getKoRule(), is( Configuration.DEFAULT_KO_RULE ) );
      String path = System.getProperty( Constants.APP_HOME_PROP );
      String expected_path = String.format( "%s%setc%slayers.json", path, file_sep, file_sep ); //$NON-NLS-1$
      assertThat( m_config2.getLayersFile().getAbsolutePath(), is( expected_path ) );
      expected_path = String.format( "%s%sdklog.txt", path, file_sep ); //$NON-NLS-1$
      assertThat( m_config2.getOutputDirectory().getAbsolutePath(), is( System.getProperty( "user.dir" ) ) ); //$NON-NLS-1$
      SimpleDateFormat formatter = new SimpleDateFormat( Configuration.DEFAULT_DATE_FORMAT );
      Date today = new Date();
      expected_path = String.format( "%s%s%s.sgf", m_config2.getOutputDirectory(), file_sep, formatter.format( today ) ).replace( "/", file_sep ); //$NON-NLS-1$ //$NON-NLS-2$
      assertThat( m_config2.getOutputFile().getAbsolutePath(), is( expected_path ) );
      assertTrue( m_config2.getPlayerData( "nosuchplayer" ).isEmpty() ); //$NON-NLS-1$
      List<URI> uris = m_config2.getPluginURIs();
      assertThat( uris.size(), is( 1 ) );
      expected_path = String.format( "file://%s%s/plugins", (m_drive_letter_pattern.matcher( path ).matches() ? "/" : ""), path ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
      assertThat( m_config2.getScoring(), is( Configuration.DEFAULT_SCORING_SYSTEM ) );
      assertThat( m_config2.allowSuicide(), is( false ) );
   }

   @Test
   public void testInvalidValues()
   {
      assertThrows(
         new Runnable() { @Override public void run() { m_config2.setEngineName( "" ); } },      //$NON-NLS-1$
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { m_config2.setScoring( "" ); } },         //$NON-NLS-1$
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { m_config2.setKomi( "abc" ); } },         //$NON-NLS-1$
         ParseException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { m_config2.setKoRule( "blah" ); } },      //$NON-NLS-1$
         ParseException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { m_config2.setScoring( "ha ha ha" ); } }, //$NON-NLS-1$
         ParseException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { m_config2.setAllowSuicide( "xyz" ); } }, //$NON-NLS-1$
         ParseException.class, null );
   }

   @Test
   public void testVariableExpansion()
   {
      Board.setSize( 9 );
      String file_sep = System.getProperty( "file.separator" ); //$NON-NLS-1$
      m_config2.setOutputDirectory( String.format( "$app-home%s$size%s$date", file_sep, file_sep ) ); //$NON-NLS-1$
      SimpleDateFormat formatter = new SimpleDateFormat( "yyyyMMdd" ); //$NON-NLS-1$
      String today_str = formatter.format( new Date() );
      String dkhome_dir = System.getProperty( Constants.APP_HOME_PROP );
      String expected_path = String.format( "%s%s9%s%s", dkhome_dir, file_sep, file_sep, today_str ); //$NON-NLS-1$
      assertThat( m_config2.getOutputDirectory().getAbsolutePath(), is( expected_path ) );
      Board.setSize( 13 );
      expected_path = String.format( "%s%s13%s%s", dkhome_dir, file_sep, file_sep, today_str ); //$NON-NLS-1$
      assertThat( m_config2.getOutputDirectory().getAbsolutePath(), is( expected_path ) );
   }

   @Test
   public void testTestConfig() throws URISyntaxException
   {
      assertThat( m_config.getEngineName(), is( Configuration.DEFAULT_ENGINE_NAME ) );
      List<URI> uris = m_config.getPluginURIs();
      assertThat( uris.size(), is( 2 ) );
      String path = System.getProperty( Constants.APP_HOME_PROP ).replace( "\\", "/" ); //$NON-NLS-1$ //$NON-NLS-2$
      String expected_path = String.format( "file://%s%s/build/dist/plugins", (m_drive_letter_pattern.matcher( path ).matches() ? "/" : ""), path  ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
      assertThat( uris.get( 0 ).toString(), is( expected_path ) );
      assertThat( uris.get( 1 ).toString(), is( "classpath://*" ) ); //$NON-NLS-1$
   }

   @Test
   public void testLoadSampleConfig() throws IOException, URISyntaxException
   {
      String file_sep = System.getProperty( "file.separator" ); //$NON-NLS-1$
      Board.setSize( 9 );
      String path = System.getProperty( Constants.APP_HOME_PROP );
      assertNotNull( path );
      File home_dir = new File( path );
      File etc_dir = new File( home_dir, "etc" ); //$NON-NLS-1$
      File config_file = new File( etc_dir, "sample-config.json" ); //$NON-NLS-1$
      try (FileReader input = new FileReader( config_file ))
      {
         m_config2 = ConfigReader.read( input );
      }
      assertThat( m_config2.getDateFormat().toPattern(), is( "MM/dd/yyyy" ) ); //$NON-NLS-1$
      assertThat( m_config2.getEngineName(), is( "MyTestGoEngine" ) ); //$NON-NLS-1$
      assertThat( m_config2.getEngineVersion(), is( "1.0" ) ); //$NON-NLS-1$
      assertThat( m_config2.getFilenameFormat(), is( "$engine-name$engine-version.$file-ext" ) ); //$NON-NLS-1$
      assertThat( m_config2.getKomi(), is( new BigDecimal( "7.5" ) ) ); //$NON-NLS-1$
      assertThat( m_config2.getKoRule(), is( KoRules.POSITIONAL ) );
      String expected_path = String.format( "%s%ssomedir%sapp-layers.json", path, file_sep, file_sep ); //$NON-NLS-1$
      assertThat( m_config2.getLayersFile().getAbsolutePath(), is( expected_path ) );
      SimpleDateFormat formatter = new SimpleDateFormat( "MM/dd/yyyy" ); //$NON-NLS-1$
      Date today = new Date();
      expected_path = String.format( "%s/9/%s", path, formatter.format( today ) ).replace( "/", file_sep ); //$NON-NLS-1$ //$NON-NLS-2$
      assertThat( m_config2.getOutputDirectory().getAbsolutePath(), is( expected_path ) );
      expected_path = String.format( "%s%s%s%s.sgf", expected_path, file_sep, m_config2.getEngineName(), m_config2.getEngineVersion() ).replace( "/", file_sep ); //$NON-NLS-1$ //$NON-NLS-2$
      assertThat( m_config2.getOutputFile().getAbsolutePath(), is( expected_path ) );
      assertThat( m_config2.getScoring(), is( Configuration.DEFAULT_SCORING_SYSTEM ) );
      assertThat( m_config2.allowSuicide(), is( false ) );
      List<URI> uris = m_config2.getPluginURIs();
      assertThat( uris.size(), is( 3 ) );
      assertThat( uris.get( 0 ).toString(), is( "classpath://*" ) ); //$NON-NLS-1$
      String user_dir = System.getProperty( "user.home" ).replace( "\\", "/" ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
      expected_path = String.format( "file://%s%s/.myapp/plugins/", (m_drive_letter_pattern.matcher( user_dir ).matches() ? "/" : ""), user_dir ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
      assertThat( uris.get( 1 ).toString(), is( expected_path ) );
      path = path.replace( "\\", "/" ); //$NON-NLS-1$ //$NON-NLS-2$
      expected_path = String.format( "file://%s%s/plugins", (m_drive_letter_pattern.matcher( path ).matches() ? "/" : ""), path ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
      assertThat( uris.get( 2 ).toString(), is( expected_path ) );

      Map<String,JsonElement> data = m_config2.getPlayerData( "PlayerType1" ); //$NON-NLS-1$
      JsonElement value = data.get( "player-type" ); //$NON-NLS-1$
      assertThat( value.getAsString(), is( "PlayerType1" ) ); //$NON-NLS-1$
      value = data.get( "attr1" ); //$NON-NLS-1$
      assertThat( value.getAsInt(), is( 102 ) );
      value = data.get( "attr2" ); //$NON-NLS-1$
      assertThat( value.getAsString(), is( "ab cd" ) ); //$NON-NLS-1$

      data = m_config2.getPlayerData( "PlayerType2" ); //$NON-NLS-1$
      value = data.get( "player-type" ); //$NON-NLS-1$
      assertThat( value.getAsString(), is( "PlayerType2" ) ); //$NON-NLS-1$
      value = data.get( "stuff" ); //$NON-NLS-1$
      JsonObject obj = value.getAsJsonObject();
      assertThat( obj.get( "foo" ).getAsJsonPrimitive().getAsString(), is( "bar" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      value = data.get( "more_stuff" ); //$NON-NLS-1$
      obj = value.getAsJsonObject();
      assertThat( obj.get( "xyz" ).getAsJsonPrimitive().getAsInt(), is( 123 ) ); //$NON-NLS-1$
      assertThat( obj.get( "abc" ).getAsJsonPrimitive().getAsInt(), is( 456 ) ); //$NON-NLS-1$
   }

   private Configuration m_config2;
   private Pattern       m_drive_letter_pattern = Pattern.compile( "^[a-zA-Z]:.*" ); //$NON-NLS-1$
}
