//
// GtpTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.application;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.io.IOException;
import java.io.StringReader;
import java.math.BigDecimal;

import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.application.Configuration;
import com.dakengo.libs.application.GtpCommand;
import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.application.GtpEngine;
import com.dakengo.libs.application.Resources;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.libs.record.Record;
import com.dakengo.tests.utils.BaseConfiguredTests;
import com.dakengo.tests.utils.DkAssertUtils;
import com.dakengo.tests.utils.GtpReplyValidator;
import com.dakengo.tests.utils.GtpRecordValidator;
import com.dakengo.tests.utils.TestPropVisitor;

public class GtpTests extends BaseConfiguredTests
{
   public GtpTests() throws IOException
   {
   }

   @Before
   public void setUp()
   {
      Board.setSize( Board.DEFAULT_SIZE );
      String player_type = "test"; //$NON-NLS-1$
      m_context = new GameContext( m_config, m_plugin_mgr, player_type, "" ); //$NON-NLS-1$
      assertNotNull( m_context.getPlayer() );
      assertNotNull( m_context.getScorer() );
      m_engine = new GtpEngine( m_plugin_mgr, m_handler, m_store );
   }

   @Test
   public void testCommand()
   {
      GtpCommand cmd = GtpCommand.parse( "cmdname" ); //$NON-NLS-1$
      assertThat( cmd.getID(), is( GtpCommand.NO_ID ) );
      assertThat( cmd.getName(), is( "cmdname" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().size(), is( 0 ) );

      cmd = GtpCommand.parse( "  cmdname   " ); //$NON-NLS-1$
      assertThat( cmd.getID(), is( GtpCommand.NO_ID ) );
      assertThat( cmd.getName(), is( "cmdname" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().size(), is( 0 ) );

      cmd = GtpCommand.parse( "10 cmdname" ); //$NON-NLS-1$
      assertThat( cmd.getID(), is( 10L ) );
      assertThat( cmd.getName(), is( "cmdname" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().size(), is( 0 ) );

      cmd = GtpCommand.parse( "cmdname arg1" ); //$NON-NLS-1$
      assertThat( cmd.getID(), is( GtpCommand.NO_ID ) );
      assertThat( cmd.getName(), is( "cmdname" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().size(), is( 1 ) );
      assertThat( cmd.getArguments().get( 0 ), is( "arg1" ) ); //$NON-NLS-1$

      cmd = GtpCommand.parse( "cmdname arg1 arg2" ); //$NON-NLS-1$
      assertThat( cmd.getID(), is( GtpCommand.NO_ID ) );
      assertThat( cmd.getName(), is( "cmdname" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().size(), is( 2 ) );
      assertThat( cmd.getArguments().get( 0 ), is( "arg1" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().get( 1 ), is( "arg2" ) ); //$NON-NLS-1$

      cmd = GtpCommand.parse( "  5   cmdname  arg1  arg2   " ); //$NON-NLS-1$
      assertThat( cmd.getID(), is( 5L ) );
      assertThat( cmd.getName(), is( "cmdname" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().size(), is( 2 ) );
      assertThat( cmd.getArguments().get( 0 ), is( "arg1" ) ); //$NON-NLS-1$
      assertThat( cmd.getArguments().get( 1 ), is( "arg2" ) ); //$NON-NLS-1$

      String[] bad_cmds = new String[]{ "", " ", "   ", "5", "5  " }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
      for (String text : bad_cmds)
      {
         final String tmp = text;
         assertThrows(
            new Runnable() { @Override public void run() { GtpCommand.parse( tmp ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testAdminCommands()
   {
      String eol = System.getProperty( "line.separator" ); //$NON-NLS-1$
      String[] admin_cmds = new String[]
      {
         "boardsize", "clear_board", "dk-playout", "final_score", "fixed_handicap", "genmove",  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
         "known_command", "komi", "list_commands", "loadsgf", "name", "place_free_handicap",    //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
         "play", "protocol_version", "quit", "set_free_handicap", "showboard", "version"        //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
      };
      StringBuilder buffer = new StringBuilder();
      int count = 0;
      for (String name : admin_cmds)
      {
         count++;
         buffer.append( name );
         if (count < admin_cmds.length)
            buffer.append( eol );
      }
      GtpCommand cmd = GtpCommand.parse( "list_commands" ); //$NON-NLS-1$
      m_handler.setExpected( "list_commands", false, GtpCommand.NO_ID, buffer.toString() ); //$NON-NLS-1$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "list_commands foo" ); //$NON-NLS-1$
      m_handler.setExpected( "list_commands", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "protocol_version" ); //$NON-NLS-1$
      m_handler.setExpected( "protocol_version", false, GtpCommand.NO_ID, GtpEngine.PROTOCOL_VERSION ); //$NON-NLS-1$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "protocol_version foo" ); //$NON-NLS-1$
      m_handler.setExpected( "protocol_version", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "name" ); //$NON-NLS-1$
      m_handler.setExpected( "name", false, GtpCommand.NO_ID, "Dakengo" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "name foo" ); //$NON-NLS-1$
      m_handler.setExpected( "name", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "version" ); //$NON-NLS-1$
      m_handler.setExpected( "version", false, GtpCommand.NO_ID, m_config.getEngineVersion() ); //$NON-NLS-1$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "version foo" ); //$NON-NLS-1$
      m_handler.setExpected( "version", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      assertSubsetCommands( admin_cmds );

      Configuration orig_config = new Configuration( m_config );
      try
      {
         m_config.setEngineName( "unittests" ); //$NON-NLS-1$
         m_config.setEngineVersion( "1.0" ); //$NON-NLS-1$
   
         cmd = GtpCommand.parse( "name" ); //$NON-NLS-1$
         m_handler.setExpected( "name", false, GtpCommand.NO_ID, "unittests" ); //$NON-NLS-1$ //$NON-NLS-2$
         m_engine.execute( cmd, m_context );
   
         cmd = GtpCommand.parse( "version" ); //$NON-NLS-1$
         m_handler.setExpected( "version", false, GtpCommand.NO_ID, "1.0" ); //$NON-NLS-1$ //$NON-NLS-2$
         m_engine.execute( cmd, m_context );
      }
      finally
      {
         m_config = orig_config;
      }
   }

   @Test
   public void testBoardCommands()
   {
      Game game = m_context.getGame();
      short pnt = (short) (Board.getTotal() - 1);
      game.play( pnt, null );

      GtpCommand cmd = GtpCommand.parse( "1 boardsize 9" ); //$NON-NLS-1$
      m_handler.setExpected( "boardsize", false, 1, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( Board.getSize(), is( 9 ) );
      assertThat( game.getBoard().getCount( Colors.EMPTY ), is( Board.getTotal() ) );
      assertThat( game.getTurnNumber(), is( 1 ) );
      assertFalse( game.isOver() );

      pnt = (short) (Board.getTotal() - 1);
      game.play( pnt, null );
      cmd = GtpCommand.parse( "2 boardsize 4" ); //$NON-NLS-1$
      m_handler.setExpected( "boardsize", true, 2, Resources.getString( "msg.2030" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( Board.getSize(), is( 9 ) );
      DkAssertUtils.assertBoardPoints( game.getBoard(), new short[] { pnt }, null );
      assertThat( game.getTurnNumber(), is( 2 ) );
      assertFalse( game.isOver() );

      cmd = GtpCommand.parse( "3 boardsize 1" ); //$NON-NLS-1$
      m_handler.setExpected( "boardsize", true, 3, Resources.getString( "msg.2030" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( Board.getSize(), is( 9 ) );
      DkAssertUtils.assertBoardPoints( game.getBoard(), new short[] { pnt }, null );
      assertThat( game.getTurnNumber(), is( 2 ) );
      assertFalse( game.isOver() );

      cmd = GtpCommand.parse( "boardsize" ); //$NON-NLS-1$
      m_handler.setExpected( "boardsize", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( Board.getSize(), is( 9 ) );
      DkAssertUtils.assertBoardPoints( game.getBoard(), new short[] { pnt }, null );
      assertThat( game.getTurnNumber(), is( 2 ) );
      assertFalse( game.isOver() );

      cmd = GtpCommand.parse( "4 boardsize 9" ); //$NON-NLS-1$
      m_handler.setExpected( "boardsize", false, 4, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( Board.getSize(), is( 9 ) );
      assertThat( game.getBoard().getCount( Colors.EMPTY ), is( Board.getTotal() ) );
      assertThat( game.getTurnNumber(), is( 1 ) );
      assertFalse( game.isOver() );

      game.play( pnt, null );
      cmd = GtpCommand.parse( "5 clear_board" ); //$NON-NLS-1$
      m_handler.setExpected( "clear_board", false, 5, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( Board.getSize(), is( 9 ) );
      assertThat( game.getBoard().getCount( Colors.EMPTY ), is( Board.getTotal() ) );
      assertThat( game.getTurnNumber(), is( 1 ) );
      assertFalse( game.isOver() );

      cmd = GtpCommand.parse( "101 clear_board foo" ); //$NON-NLS-1$
      m_handler.setExpected( "clear_board", true, 101, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( Board.getSize(), is( 9 ) );
      assertFalse( game.isOver() );
   }

   @Test
   public void testMiscCommands()
   {
      Game game = m_context.getGame();

      GtpCommand cmd = GtpCommand.parse( "54 komi 8.5" ); //$NON-NLS-1$
      m_handler.setExpected( "komi", false, 54, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( game.getKomi(), is( new BigDecimal( "8.5" ) ) ); //$NON-NLS-1$

      cmd = GtpCommand.parse( "1052574910653 komi 7.a" ); //$NON-NLS-1$
      m_handler.setExpected( "komi", true, 0xf5125a88bdL, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertThat( game.getKomi(), is( new BigDecimal( "8.5" ) ) ); //$NON-NLS-1$
   }

   @Test
   public void testHandicapCommands()
   {
      Board board = m_context.getGame().getBoard();

      GtpCommand cmd = GtpCommand.parse( "fixed_handicap 4" ); //$NON-NLS-1$
      m_handler.setExpected( "fixed_handicap", false, GtpCommand.NO_ID, "d4 d16 q4 q16" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      DkAssertUtils.assertBoardPoints( board, Convert.stringsToPoints( "d4", "d16", "q4", "q16" ), null ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$

      cmd = GtpCommand.parse( "fixed_handicap" ); //$NON-NLS-1$
      m_handler.setExpected( "fixed_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "fixed_handicap 2 3" ); //$NON-NLS-1$
      m_handler.setExpected( "fixed_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "fixed_handicap abc" ); //$NON-NLS-1$
      m_handler.setExpected( "fixed_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "fixed_handicap 100" ); //$NON-NLS-1$
      m_handler.setExpected( "fixed_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2041" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      
      cmd = GtpCommand.parse( "fixed_handicap 2" ); //$NON-NLS-1$
      m_handler.setExpected( "fixed_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2033" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "set_free_handicap" ); //$NON-NLS-1$
      m_handler.setExpected( "set_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2031" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "set_free_handicap a1 zz" ); //$NON-NLS-1$
      m_handler.setExpected( "set_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "set_free_handicap a1 b2" ); //$NON-NLS-1$
      m_handler.setExpected( "set_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2033" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      board.clear();
      cmd = GtpCommand.parse( "set_free_handicap a1 b2" ); //$NON-NLS-1$
      m_handler.setExpected( "set_free_handicap", false, GtpCommand.NO_ID, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      DkAssertUtils.assertBoardPoints( board, Convert.stringsToPoints( "a1", "b2" ), null ); //$NON-NLS-1$ //$NON-NLS-2$

      cmd = GtpCommand.parse( "place_free_handicap" ); //$NON-NLS-1$
      m_handler.setExpected( "place_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "place_free_handicap 2 3" ); //$NON-NLS-1$
      m_handler.setExpected( "place_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "place_free_handicap xyz" ); //$NON-NLS-1$
      m_handler.setExpected( "place_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "place_free_handicap 10203" ); //$NON-NLS-1$
      m_handler.setExpected( "place_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2041" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "place_free_handicap 3" ); //$NON-NLS-1$
      m_handler.setExpected( "place_free_handicap", true, GtpCommand.NO_ID, Resources.getString( "msg.2033" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      board.clear();
      cmd = GtpCommand.parse( "place_free_handicap 3" ); //$NON-NLS-1$
      m_handler.setExpected( "place_free_handicap", false, GtpCommand.NO_ID, "t17 t18 t19" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
   }

   @Test
   public void testPlayCommand()
   {
      GtpCommand cmd = GtpCommand.parse( "play B m2" ); //$NON-NLS-1$
      m_handler.setExpected( "play", false, GtpCommand.NO_ID, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      cmd = GtpCommand.parse( "play W M3" ); //$NON-NLS-1$
      m_handler.setExpected( "play", false, GtpCommand.NO_ID, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      String[] bad_cmds = new String[] { "play", "play W", "play X a1", "play W zz" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
      for (String text : bad_cmds)
      {
	      cmd = GtpCommand.parse( text );
	      m_handler.setExpected( "play", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
	      m_engine.execute( cmd, m_context );
      }

      Game game = m_context.getGame();
      assertFalse( game.isOver() );
      Board board = game.getBoard();
      assertThat( board.getCount( Colors.BLACK ), is( 1 ) );
      assertThat( board.getCount( Colors.WHITE ), is( 1 ) );
      assertThat( board.getColor( Convert.stringToPoint( "m2" ) ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( board.getColor( Convert.stringToPoint( "m3" ) ), is( Colors.WHITE ) ); //$NON-NLS-1$

      TestPropVisitor visitor = new TestPropVisitor( true );
      Record rec = m_context.getRecord();
      rec.visit( visitor, true );
      visitor.assertValue( PropertyID.MOVE, 0, "BLACK:m2" ); //$NON-NLS-1$
      visitor.assertValue( PropertyID.MOVE, 1, "WHITE:m3" ); //$NON-NLS-1$
   }

   @Test
   public void testGenMoveCommand()
   {
      GtpCommand cmd = GtpCommand.parse( "genmove B" ); //$NON-NLS-1$
      m_handler.setExpected( "genmove", false, GtpCommand.NO_ID, "a1" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      String[] bad_cmds = new String[] { "genmove", "genmove 123", "genmove W 123" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
      for (String text : bad_cmds)
      {
         cmd = GtpCommand.parse( text );
         m_handler.setExpected( "genmove", true, GtpCommand.NO_ID, Resources.getString( "msg.2044" ) ); //$NON-NLS-1$ //$NON-NLS-2$
         m_engine.execute( cmd, m_context );
      }

      cmd = GtpCommand.parse( "genmove W" ); //$NON-NLS-1$
      m_handler.setExpected( "genmove", false, GtpCommand.NO_ID, "a2" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );

      Game game = m_context.getGame();
      assertFalse( game.isOver() );
      Board board = game.getBoard();
      assertThat( board.getCount( Colors.BLACK ), is( 1 ) );
      assertThat( board.getCount( Colors.WHITE ), is( 1 ) );
      assertThat( board.getColor( Convert.stringToPoint( "a1" ) ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( board.getColor( Convert.stringToPoint( "a2" ) ), is( Colors.WHITE ) ); //$NON-NLS-1$

      TestPropVisitor visitor = new TestPropVisitor( true );
      Record rec = m_context.getRecord();
      rec.visit( visitor, true );
      visitor.assertValue( PropertyID.MOVE, 0, "BLACK:a1" ); //$NON-NLS-1$
      visitor.assertValue( PropertyID.MOVE, 1, "WHITE:a2" ); //$NON-NLS-1$

      Colors color = Colors.BLACK;
      int i = 2;
      for ( ; i <= 10; i++)
      {
         cmd = GtpCommand.parse( "genmove " + color ); //$NON-NLS-1$
         color = Game.otherPlayer( color );
         m_handler.setExpected( "genmove", false, GtpCommand.NO_ID, Convert.pointToString( (short) i ) ); //$NON-NLS-1$
         m_engine.execute( cmd, m_context );
      }
      cmd = GtpCommand.parse( "genmove " + color ); //$NON-NLS-1$
      color = Game.otherPlayer( color );
      m_handler.setExpected( "genmove", false, GtpCommand.NO_ID, "pass" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      cmd = GtpCommand.parse( "play " + color + " pass" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_handler.setExpected( "genmove", false, GtpCommand.NO_ID, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( cmd, m_context );
      assertTrue( game.isOver() );
   }

   @Test
   public void testLoop() throws IOException
   {
      String eol = System.getProperty( "line.separator" ); //$NON-NLS-1$
      StringBuilder buffer = new StringBuilder();
      String[] cmds = new String[]
      {
         "boardsize 19", "clear_board", "fixed_handicap 2", "play W a1", "genmove B", "play W a3", "genmove B" //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
      };
      for (String cmd : cmds)
      {
         buffer.append( cmd );
         buffer.append( eol );
      }
      m_store.setExpectedGameProps( "Test Player", Configuration.DEFAULT_OPPONENT_NAME, 19 ); //$NON-NLS-1$
      m_handler.setAnyTextOK( false );
      m_engine.loop( new StringReader( buffer.toString() ), m_context );
      assertTrue( m_handler.wasInvoked() );
      GameContext context = m_context;
      Game game = context.getGame();
      assertFalse( game.isOver() );
      short last_pnt = Convert.stringToPoint( "a4" ); //$NON-NLS-1$
      assertThat( game.getLastPoint(), is( last_pnt ) );
      assertThat( game.getPlayer(), is( Colors.WHITE ) );
      int turn_num = 5;
      assertThat( game.getTurnNumber(), is( turn_num++ ) );
      DkAssertUtils.assertBoardPoints( game.getBoard(), Convert.stringsToPoints( "d4", "a2", "a4", "q16" ), Convert.stringsToPoints( "a1", "a3" ) ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$

      m_engine.execute( GtpCommand.parse( "play W pass" ), m_context ); //$NON-NLS-1$
      assertFalse( game.isOver() );
      assertThat( game.getLastPoint(), is( last_pnt ) );
      assertThat( game.getPlayer(), is( Colors.BLACK ) );
      assertThat( game.getTurnNumber(), is( turn_num++ ) );

      m_engine.execute( GtpCommand.parse( "play b pass" ), m_context ); //$NON-NLS-1$
      assertTrue( game.isOver() );
      assertThat( game.getLastPoint(), is( last_pnt ) );
      assertThat( game.getPlayer(), is( Colors.WHITE ) );
      assertThat( game.getTurnNumber(), is( turn_num++ ) );

      m_handler.setExpected( "quit", false, GtpCommand.NO_ID, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( GtpCommand.parse( "quit" ), m_context ); //$NON-NLS-1$

      m_store.validate();

      m_store.setExpectedGameProps( Configuration.DEFAULT_OPPONENT_NAME, "Test Player", 19 ); //$NON-NLS-1$
      context.reset();
      m_handler.setAnyTextOK( false );
      cmds = new String[]
      {
         "clear_board", "play b a1", "genmove w", "play b a3", "genmove w", "play b a5", "genmove w", "play b pass", "genmove w" //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$
      };
      buffer = new StringBuilder();
      for (String cmd : cmds)
      {
         buffer.append( cmd );
         buffer.append( eol );
      }
      m_engine.loop( new StringReader( buffer.toString() ), m_context );
      game = context.getGame();
      assertTrue( m_handler.wasInvoked() );
      assertTrue( game.isOver() );
      assertThat( game.getLastPoint(), is( Convert.stringToPoint( "a6" ) ) ); //$NON-NLS-1$
      assertThat( game.getPlayer(), is( Colors.BLACK ) );
      assertThat( game.getTurnNumber(), is( 9 ) );
      DkAssertUtils.assertBoardPoints( game.getBoard(), Convert.stringsToPoints( "a1", "a3", "a5" ), Convert.stringsToPoints( "a2", "a4", "a6" ) ); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$

      m_handler.setExpected( "clear_board", false, GtpCommand.NO_ID, "" ); //$NON-NLS-1$ //$NON-NLS-2$
      m_engine.execute( GtpCommand.parse( "clear_board" ), m_context ); //$NON-NLS-1$

      m_store.validate();
   }

   private void assertSubsetCommands( String[] cmd_names )
   {
      String msg = Resources.getString( "msg.2045" ); //$NON-NLS-1$
      for (String name : cmd_names)
      {
         GtpCommand cmd = GtpCommand.parse( "known_command " + name ); //$NON-NLS-1$
         m_handler.setExpected( "known_command", false, GtpCommand.NO_ID, msg ); //$NON-NLS-1$
         m_engine.execute( cmd, m_context );
      }
   }

   private GameContext        m_context;
   private GtpEngine          m_engine;
   private GtpReplyValidator  m_handler = new GtpReplyValidator();
   private GtpRecordValidator m_store   = new GtpRecordValidator();
}
