//
// LexerTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.base;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Test;

import com.dakengo.libs.base.Lexer;

public class LexerTests
{
   @Test
   public void testBasic() throws IOException
   {
      Collection<ImmutablePair<Integer,String>> tokens = new ArrayList<ImmutablePair<Integer,String>>();
      tokens.add( LEFT_PAREN );
      tokens.add( RIGHT_BRACKET );
      tokens.add( FOO );
      int[] expected_ids = new int[]
      {
         LEFT_PAREN.getLeft(), Lexer.TEXT_TOKEN_ID, RIGHT_BRACKET.getLeft(),
         Lexer.TEXT_TOKEN_ID, FOO.getLeft(), Lexer.INTEGER_TOKEN_ID
      };
      String[] expected_text = new String[]
      {
         "(", ")[", "]", "a2", "b", "1"
      };
      String text = "()[]a2b1";

      m_lexer = new Lexer( new StringReader( text ), tokens, "" );
      for (int i = 0; i < expected_ids.length; i++)
         assertToken( expected_ids[i], expected_text[i] );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      m_lexer.setInput( new StringReader( text ) );
      m_lexer.enable( Lexer.INTEGER_TOKEN_ID, false );
      m_lexer.enable( Lexer.DECIMAL_TOKEN_ID, false );
      expected_ids = new int[]
      {
         LEFT_PAREN.getLeft(), Lexer.TEXT_TOKEN_ID, RIGHT_BRACKET.getLeft(),
         Lexer.TEXT_TOKEN_ID, FOO.getLeft(), Lexer.TEXT_TOKEN_ID
      };
      for (int i = 0; i < expected_ids.length; i++)
         assertToken( expected_ids[i], expected_text[i] );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      m_lexer.setInput( new StringReader( text ) );
      m_lexer.enable( LEFT_PAREN.getLeft(), false );
      m_lexer.enable( FOO.getLeft(), false );
      m_lexer.enable( Lexer.INTEGER_TOKEN_ID, true );
      expected_text = new String[] { "()[", "]", "a2b1" };
      expected_ids = new int[]
      {
         Lexer.TEXT_TOKEN_ID, RIGHT_BRACKET.getLeft(), Lexer.TEXT_TOKEN_ID
      };
      for (int i = 0; i < expected_ids.length; i++)
         assertToken( expected_ids[i], expected_text[i] );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      m_lexer.setInput( new StringReader( text ) );
      m_lexer.enable( LEFT_PAREN.getLeft(), true );
      expected_text = new String[] { "(", ")[", "]", "a2b1" };
      expected_ids = new int[]
      {
         LEFT_PAREN.getLeft(), Lexer.TEXT_TOKEN_ID, RIGHT_BRACKET.getLeft(), Lexer.TEXT_TOKEN_ID
      };
      for (int i = 0; i < expected_ids.length; i++)
         assertToken( expected_ids[i], expected_text[i] );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      assertThat( m_lexer.isEnabled( 1001 ), is( false ) );
   }

   @Test
   public void testComplex() throws IOException
   {
      Collection<ImmutablePair<Integer,String>> tokens = new ArrayList<ImmutablePair<Integer,String>>();
      tokens.add( LEFT_PAREN );
      tokens.add( RIGHT_BRACKET );
      tokens.add( NIL );
      String text = "(nil]";
      m_lexer = new Lexer( new StringReader( text ), tokens, "" );
      assertToken( LEFT_PAREN.getLeft(), "(" );
      assertToken( NIL.getLeft(), "nil" );
      assertToken( RIGHT_BRACKET.getLeft(), "]" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      m_lexer.setInput( new StringReader( text ) );
      m_lexer.enable( LEFT_PAREN.getLeft(), false );
      m_lexer.enable( RIGHT_BRACKET.getLeft(), false );
      assertToken( Lexer.TEXT_TOKEN_ID, "(" );
      assertToken( NIL.getLeft(), "nil" );
      assertToken( Lexer.TEXT_TOKEN_ID, "]" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );
   }

   @Test
   public void testEscape() throws IOException
   {
      Collection<ImmutablePair<Integer,String>> tokens = new ArrayList<ImmutablePair<Integer,String>>();
      tokens.add( LEFT_BRACKET );
      Reader input = new StringReader( "\\[" );
      m_lexer = new Lexer( input, tokens, "\\" );
      assertToken( Lexer.TEXT_TOKEN_ID, "[" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "\\\\[" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, "\\" );
      assertToken( LEFT_BRACKET.getLeft(), "[" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "a\\[b[c" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, "a[b" );
      assertToken( LEFT_BRACKET.getLeft(), "[" );
      assertToken( Lexer.TEXT_TOKEN_ID, "c" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "a\\\rb" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, "ab" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertThat( m_lexer.hasMore(), is( false ) );
   }

   @Test
   public void testWhitespace() throws IOException
   {
      String text = "a\tb\na   b\\\na\r";
      Collection<ImmutablePair<Integer,String>> tokens = new ArrayList<ImmutablePair<Integer,String>>();
      tokens.add( BAR );
      tokens.add( FOO );
      m_lexer = new Lexer( new StringReader( text ), tokens, "\\" );
      assertToken( BAR.getLeft(), "a" );
      assertToken( Lexer.WS_TOKEN_ID, "\t" );
      assertToken( FOO.getLeft(), "b" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertToken( Lexer.EOL_TOKEN_ID, "\n" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );
      assertToken( BAR.getLeft(), "a" );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( FOO.getLeft(), "b" );
      assertThat( m_lexer.getLineNumber(), is( 2 ) );
      assertToken( BAR.getLeft(), "a" );
      assertThat( m_lexer.getLineNumber(), is( 3 ) );
      assertToken( Lexer.TEXT_TOKEN_ID, "\r" );
      assertThat( m_lexer.getLineNumber(), is( 3 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      m_lexer.setInput( new StringReader( text ) );
      m_lexer.enable( FOO.getLeft(), false );
      m_lexer.enable( BAR.getLeft(), false );
      m_lexer.enable( Lexer.WS_TOKEN_ID, false );
      m_lexer.enable( Lexer.EOL_TOKEN_ID, false );
      assertToken( Lexer.TEXT_TOKEN_ID, "a\tb\na   ba\r" );
      assertThat( m_lexer.getLineNumber(), is( 3 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      m_lexer.setInput( new StringReader( "a\n\nb" ) );
      m_lexer.enable( BAR.getLeft(), false );
      m_lexer.enable( FOO.getLeft(), false );
      ImmutablePair<Integer,String> token = m_lexer.getNext();
      assertThat( token.getRight(), is( "a\n\nb" ) );
      assertThat( m_lexer.getLineNumber(), is( 3 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      m_lexer.setInput( new StringReader( "a\r\nb" ) );
      m_lexer.enable( BAR.getLeft(), false );
      m_lexer.enable( FOO.getLeft(), false );
      token = m_lexer.getNext();
      assertThat( token.getRight(), is( "a\r\nb" ) );
      assertThat( m_lexer.getLineNumber(), is( 2 ) );
      assertThat( m_lexer.hasMore(), is( false ) );

      Collection<ImmutablePair<Integer,String>> no_tokens = new ArrayList<ImmutablePair<Integer,String>>();
      try (InputStreamReader isr = new InputStreamReader( getClass().getResourceAsStream( "/data/linebreaks.txt" ) ))
      {
         m_lexer = new Lexer( isr, no_tokens, "\\" );
         m_lexer.enable( Lexer.INTEGER_TOKEN_ID, false );
         m_lexer.enable( Lexer.WS_TOKEN_ID, false );
         assertToken( Lexer.TEXT_TOKEN_ID, "line 1" );
         assertThat( m_lexer.getLineNumber(), is( 1 ) );
         assertToken( Lexer.EOL_TOKEN_ID, "\r\n" );
         assertThat( m_lexer.getLineNumber(), is( 1 ) );
         assertToken( Lexer.TEXT_TOKEN_ID, "\r line 2" );
         assertThat( m_lexer.getLineNumber(), is( 2 ) );
         assertToken( Lexer.EOL_TOKEN_ID, "\r\n" );
         assertThat( m_lexer.getLineNumber(), is( 2 ) );
         assertToken( Lexer.TEXT_TOKEN_ID, "line 3 <" );
         assertThat( m_lexer.getLineNumber(), is( 3 ) );
         assertToken( Lexer.EOL_TOKEN_ID, "\r\n" );
         assertThat( m_lexer.getLineNumber(), is( 3 ) );
         assertToken( Lexer.TEXT_TOKEN_ID, "line 4 xxx" );
         assertThat( m_lexer.getLineNumber(), is( 5 ) );
         assertToken( Lexer.EOL_TOKEN_ID, "\r\n" );
         assertThat( m_lexer.getLineNumber(), is( 5 ) );
         assertThat( m_lexer.hasMore(), is( false ) );
      }
   }

   @Test
   public void testPushBack() throws IOException
   {
      Reader input = new StringReader( "[ab]" );
      Collection<ImmutablePair<Integer,String>> tokens = new ArrayList<ImmutablePair<Integer,String>>();
      tokens.add( LEFT_BRACKET );
      tokens.add( RIGHT_BRACKET );
      m_lexer = new Lexer( input, tokens, "" );

      assertToken( LEFT_BRACKET.getLeft(), "[" );
      m_lexer.pushBack( LEFT_BRACKET );
      assertToken( LEFT_BRACKET.getLeft(), "[" );
      ImmutablePair<Integer,String> token = assertToken( Lexer.TEXT_TOKEN_ID, "ab" );
      m_lexer.pushBack( token );
      assertToken( Lexer.TEXT_TOKEN_ID, "ab" );
      assertToken( RIGHT_BRACKET.getLeft(), "]" );
      assertThat( m_lexer.getLineNumber(), is( 1 ) );

      assertThat( m_lexer.hasMore(), is( false ) );
   }

   @Test
   public void testNumbers() throws IOException
   {
      Reader input = new StringReader( "1 2.3 4a 1. +0 -0 0.1" );
      m_lexer = new Lexer( input, new ArrayList<ImmutablePair<Integer,String>>(), "" );
      assertToken( Lexer.INTEGER_TOKEN_ID, "1" );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.DECIMAL_TOKEN_ID, "2.3" );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.TEXT_TOKEN_ID, "4a" );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.DECIMAL_TOKEN_ID, "1." );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.INTEGER_TOKEN_ID, "+0" );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.INTEGER_TOKEN_ID, "-0" );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.DECIMAL_TOKEN_ID, "0.1" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "1 2" );
      m_lexer.setInput( input );
      m_lexer.enable( Lexer.WS_TOKEN_ID, false );
      assertToken( Lexer.TEXT_TOKEN_ID, "1 2" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "1.2" );
      m_lexer.setInput( input );
      m_lexer.enable( Lexer.DECIMAL_TOKEN_ID, false );
      assertToken( Lexer.TEXT_TOKEN_ID, "1.2" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "1 2" );
      m_lexer.setInput( input );
      m_lexer.enable( Lexer.WS_TOKEN_ID, true );
      m_lexer.enable( Lexer.DECIMAL_TOKEN_ID, true );
      m_lexer.enable( Lexer.INTEGER_TOKEN_ID, false );
      assertToken( Lexer.DECIMAL_TOKEN_ID, "1" );
      assertToken( Lexer.WS_TOKEN_ID, " " );
      assertToken( Lexer.DECIMAL_TOKEN_ID, "2" );
      assertThat( m_lexer.hasMore(), is( false ) );
   }

   @Test
   public void testNumberLike() throws IOException
   {
      Reader input = new StringReader( "2007-12-30" );
      m_lexer = new Lexer( input, new ArrayList<ImmutablePair<Integer,String>>(), "" );
      assertToken( Lexer.TEXT_TOKEN_ID, "2007-12-30" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "1+2" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, "1+2" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "-1+2" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, "-1+2" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "++12.5" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, "++12.5" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( ".14.3" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, ".14.3" );
      assertThat( m_lexer.hasMore(), is( false ) );

      input = new StringReader( "1.2.3" );
      m_lexer.setInput( input );
      assertToken( Lexer.TEXT_TOKEN_ID, "1.2.3" );
      assertThat( m_lexer.hasMore(), is( false ) );
   }

   @Test
   public void testOverrideBuiltinTokens() throws IOException
   {
      Reader input = new StringReader( " \n\\ \\\n" );   // final EOL is discarded because it is escaped
      Collection<ImmutablePair<Integer,String>> tokens = new ArrayList<ImmutablePair<Integer,String>>();
      tokens.add( MY_WS );
      tokens.add( MY_EOL );
      m_lexer = new Lexer( input, tokens, "\\" );
      assertToken( MY_WS.getLeft(), " " );
      assertToken( MY_EOL.getLeft(), "\n" );
      assertToken( Lexer.TEXT_TOKEN_ID, " " );
      assertThat( m_lexer.hasMore(), is( false ) );
   }

   private ImmutablePair<Integer,String> assertToken( int id, String text ) throws IOException
   {
      assertThat( m_lexer.isEnabled( id ), is( true ) );
      assertThat( m_lexer.hasMore(), is( true ) );
      ImmutablePair<Integer,String> token = m_lexer.getNext();
      assertThat( text, token.getLeft(), is( id ) );
      assertThat( token.getRight(), is( text ) );
      return token;
   }

   private static final ImmutablePair<Integer,String> LEFT_PAREN    = ImmutablePair.of( 0, "(" );
   private static final ImmutablePair<Integer,String> LEFT_BRACKET  = ImmutablePair.of( 1, "[" );
   private static final ImmutablePair<Integer,String> RIGHT_BRACKET = ImmutablePair.of( 2, "]" );
   private static final ImmutablePair<Integer,String> FOO           = ImmutablePair.of( 3, "b" );
   private static final ImmutablePair<Integer,String> BAR           = ImmutablePair.of( 4, "a" );
   private static final ImmutablePair<Integer,String> NIL           = ImmutablePair.of( 5, "nil" );
   private static final ImmutablePair<Integer,String> MY_WS         = ImmutablePair.of( 6, " " );
   private static final ImmutablePair<Integer,String> MY_EOL        = ImmutablePair.of( 7, "\n" );

   private Lexer m_lexer;
}
