//
// BoardTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.game;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.application.Configuration;
import com.dakengo.libs.application.PluginUtils;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.PointVisitor;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Constants;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.Locations;
import com.dakengo.libs.game.PointLinkedList;
import com.dakengo.libs.game.PrisonerFinder;
import com.dakengo.libs.game.Scorer;
import com.dakengo.tests.utils.BaseConfiguredTests;
import com.dakengo.tests.utils.DkAssertUtils;

/**
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class BoardTests extends BaseConfiguredTests
{
   class TestPointVisitor implements PointVisitor
   {
      public TestPointVisitor()
      {
      }

      public int getCount()
      {
         return m_count;
      }

      public void clear()
      {
         m_count = 0;
      }

      public boolean contains( Colors color, int point )
      {
         for (int i = 0; i < m_points.length; i++)
         {
            if (m_points[i] == (short) point && m_colors[i] == color)
               return true;
         }
         return false;
      }

      @Override
      public boolean accept( Board board, short point )
      {
         m_points[m_count] = point;
         m_colors[m_count] = board.getColor( point );
         m_count++;
         return true;
      }

      private short[]  m_points = new short[Board.DEFAULT_SIZE * Board.DEFAULT_SIZE];
      private Colors[] m_colors = new Colors[Board.DEFAULT_SIZE * Board.DEFAULT_SIZE];
      private int      m_count  = 0;
   }

   public BoardTests() throws IOException
   {
   }

   @Before
   public void setUp()
   {
      Board.setSize( SIZE );
      m_board = new Board();
      m_board.setColor( BLACK_POINTS, Colors.BLACK );
      m_board.setColor( WHITE_POINTS, Colors.WHITE );
      DkAssertUtils.assertBoardPoints( m_board, BLACK_POINTS, WHITE_POINTS );
      initScorer( Configuration.DEFAULT_SCORING_SYSTEM );
   }

   @Test
   public void testPointLinkedList()
   {
      int[] storage1 = new int[SIZE * SIZE];
      PointLinkedList black = new PointLinkedList( storage1, PointLinkedList.BLACK_VALUE );
      assertThat( black.isEmpty(), is( true ) );
      assertThat( black.size(), is( 0 ) );
      assertThat( black.getTail(), is( -1 ) );
      assertThat( black, is( new PointLinkedList( storage1, PointLinkedList.BLACK_VALUE ) ) );
      assertThat( black, is( not( new PointLinkedList( storage1, PointLinkedList.WHITE_VALUE ) ) ) );

      black.reserve();
      assertThat( black.isEmpty(), is( false ) );
      assertThat( black.size(), is( storage1.length ) );
      assertThat( black.getHead(), is( 0 ) );
      assertThat( black.getTail(), is( storage1.length - 1 ) );
      int i = 0;
      for (int pnt = black.getHead(); pnt >= 0; pnt = black.getNext( pnt ), i++)
      {
         assertThat( pnt, is( i ) );
         if (pnt == 0)
         {
            assertThat( black.getPrev( pnt ), is( -1 ) );
            assertThat( black.getNext( pnt ), is( pnt + 1 ) );
         }
         else if (pnt == SIZE * SIZE - 1)
         {
            assertThat( black.getPrev( pnt ), is( pnt - 1 ) );
            assertThat( black.getNext( pnt ), is( -1 ) );
         }
         else
         {
            assertThat( black.getPrev( pnt ), is( pnt - 1 ) );
            assertThat( black.getNext( pnt ), is( pnt + 1 ) );
         }
      }

      black.remove( 5 );
      assertThat( black.isEmpty(), is( false ) );
      assertThat( black.size(), is( storage1.length - 1 ) );
      assertThat( black.getHead(), is( 0 ) );
      assertThat( black.getTail(), is( storage1.length - 1 ) );
      for (int pnt = black.getHead(); pnt >= 0; pnt = black.getNext( pnt ))
      {
         assertThat( pnt, is( not( 5 ) ) );
         if (pnt == 4)
         {
            assertThat( black.getPrev( pnt ), is( 3 ) );
            assertThat( black.getNext( pnt ), is( 6 ) );
         }
         else if (pnt == 6)
         {
            assertThat( black.getPrev( pnt ), is( 4 ) );
            assertThat( black.getNext( pnt ), is( 7 ) );
         }
         else if (pnt == black.getTail())
         {
            assertThat( black.getPrev( pnt ), is( pnt - 1 ) );
            assertThat( black.getNext( pnt ), is( -1 ) );
         }
         else
         {
            assertThat( black.getPrev( pnt ), is( pnt - 1 ) );
            assertThat( black.getNext( pnt ), is( pnt + 1 ) );
         }
      }

      black.remove( black.getHead() );
      assertThat( black.isEmpty(), is( false ) );
      assertThat( black.size(), is( storage1.length - 2 ) );
      assertThat( black.getHead(), is( 1 ) );
      assertThat( black.getTail(), is( storage1.length - 1 ) );

      black.remove( black.getTail() );
      assertThat( black.isEmpty(), is( false ) );
      assertThat( black.size(), is( storage1.length - 3 ) );
      assertThat( black.getHead(), is( 1 ) );
      assertThat( black.getTail(), is( storage1.length - 2 ) );

      int[] storage2 = new int[SIZE * SIZE];
      PointLinkedList black2 = new PointLinkedList( storage2, PointLinkedList.BLACK_VALUE );
      assertThat( black2, is( not( black ) ) );
      assertThat( black2.isEmpty(), is( true ) );
      assertThat( black2.size(), is( 0 ) );
      assertThat( black2.getTail(), is( -1 ) );

      black2.insert( 7 );
      assertThat( black2.isEmpty(), is( false ) );
      assertThat( black2.size(), is( 1 ) );
      assertThat( black2.getHead(), is( 7 ) );
      assertThat( black2.getTail(), is( 7 ) );

      black2.insert( 3 );
      assertThat( black2.isEmpty(), is( false ) );
      assertThat( black2.size(), is( 2 ) );
      assertThat( black2.getHead(), is( 7 ) );
      assertThat( black2.getTail(), is( 3 ) );

      black2.insert( 10 );
      assertThat( black2.isEmpty(), is( false ) );
      assertThat( black2.size(), is( 3 ) );
      assertThat( black2.getHead(), is( 7 ) );
      assertThat( black2.getTail(), is( 10 ) );

      black2.insert( 5 );
      assertThat( black2.isEmpty(), is( false ) );
      assertThat( black2.size(), is( 4 ) );
      assertThat( black2.getHead(), is( 7 ) );
      assertThat( black2.getTail(), is( 5 ) );

      int pnt = black2.getHead();
      assertThat( pnt, is( 7 ) );
      assertThat( black2.getPrev( pnt ), is( -1 ) );
      assertThat( black2.getNext( pnt ), is( 3 ) );

      pnt = black2.getNext( pnt );
      assertThat( pnt, is( 3 ) );
      assertThat( black2.getPrev( pnt ), is( 7 ) );
      assertThat( black2.getNext( pnt ), is( 10 ) );

      pnt = black2.getNext( pnt );
      assertThat( pnt, is( 10 ) );
      assertThat( black2.getPrev( pnt ), is( 3 ) );
      assertThat( black2.getNext( pnt ), is( 5 ) );

      pnt = black2.getNext( pnt );
      assertThat( pnt, is( 5 ) );
      assertThat( black2.getPrev( pnt ), is( 10 ) );
      assertThat( black2.getNext( pnt ), is( -1 ) );

      black2.clear();
      assertThat( black2.isEmpty(), is( true ) );
      assertThat( black2.size(), is( 0 ) );
      assertThat( black2.getTail(), is( -1 ) );

      black.clear();
      assertThat( black, is( black2 ) );
      black.insert( 8 );
      black2.insert( 8 );
      assertThat( black, is( black2 ) );
      black.insert( 20 );
      black.insert( 3 );
      black2.insert( 3 );
      black2.insert( 20 );
      assertThat( black, is( black2 ) );
      black.remove( 8 );
      assertThat( black, is( not( black2 ) ) );
      black2.remove( 8 );
      assertThat( black, is( black2 ) );
   }

   @Test
   public void testAreaScoring()
   {
      ImmutablePair<Integer,Integer> scores = m_scorer.computeScores( m_board, null );
      assertThat( scores.getLeft(), is( 4 ) );
      assertThat( scores.getRight(), is( 6 ) );

      Board copy = new Board( m_board );
      copy.setColor( (short) 12, Colors.EMPTY );
      scores = m_scorer.computeScores( copy, null );
      assertThat( scores.getLeft(), is( 3 ) );
      assertThat( scores.getRight(), is( 8 ) );

      Board.setSize( 9 );
      Board board2 = new Board();
      scores = m_scorer.computeScores( board2, null );
      assertThat( scores.getLeft(), is( 0 ) );
      assertThat( scores.getRight(), is( 0 ) );

      // The following 9x9 board was borrowed from a paper by Robert Jasiek
      // entitled "Introduction to Area Scoring, version 2".
      short[] b_points = new short[]
      {
         2, 7, 11, 12, 16, 17, 19, 21, 25, 27, 28, 30, 31, 34, 35,
         37, 39, 41, 44, 45, 46, 48, 49, 50, 51, 52, 53, 58, 68,
         77, 78, 79, 80
      };
      short[] w_points = new short[]
      {
         1, 3, 4, 6, 9, 10, 13, 14, 15, 20, 22, 24, 29, 32, 33,
         38, 42, 43, 47, 54, 55, 56, 57, 59, 60, 61, 62,
         64, 66, 67, 71, 73, 76
      };
      board2.setColor( b_points, Colors.BLACK );
      board2.setColor( w_points, Colors.WHITE );
      scores = m_scorer.computeScores( board2, null );
      assertThat( scores.getLeft(), is( 37 ) );
      assertThat( scores.getRight(), is( 41 ) );

      board2 = new Board();
      board2.setColor( (short) 42, Colors.BLACK );
      scores = m_scorer.computeScores( board2, null );
      assertThat( scores.getLeft(), is( Board.getTotal() ) );
      assertThat( scores.getRight(), is( 0 ) );

      board2.setColor( (short) 20, Colors.WHITE );
      scores = m_scorer.computeScores( board2, null );
      assertThat( scores.getLeft(), is( 1 ) );
      assertThat( scores.getRight(), is( 1 ) );
   }

   @Test
   public void testAreNeighbors()
   {
      Board.setSize( 19 );

      int[] lhs = new int[] 
      {
         0, 0, 20, 0, 1, 0, 19,
         25, 25, 5, 25, 7, 25, 43, 25, 45, 25, 6, 25, 44, 25, 24, 25, 26,
         0, 360, 0, 342, 0, 18,
         38, 37, 38, 39,
         18, 19, 18, 17,
         17, 359, 17, 36
      };
      int[] rhs = new int[]
      {
         0, 20, 0, 1, 0, 19, 0,
         25, 5, 25, 7, 25, 43, 25, 45, 25, 6, 25, 44, 25, 24, 25, 26, 25,
         360, 0, 342, 0, 18, 0,
         37, 38, 39, 38,
         19, 18, 17, 18,
         359, 17, 36, 17
      };
      boolean[] expected = new boolean[]
      {
         false, false, false, true, true, true, true,
         false, false, false, false, false, false, false, false, false, true, true, true, true, true, true, true, true,
         false, false, false, false, false, false,
         false, false, true, true,
         false, false, true, true,
         false, false, true, true
      };
      assertThat( lhs.length, is( rhs.length ) );
      assertThat( lhs.length, is( expected.length ) );
      for (int i = 0; i < lhs.length; i++)
      {
         String msg = String.format( "lhs %d rhs %d", lhs[i], rhs[i] ); //$NON-NLS-1$
         assertThat( msg, Board.areNeighbors( (short) lhs[i], (short) rhs[i] ), is( expected[i] ) );
      }
   }

   // TODO: re-enable these test cases once we've decided how to implement
   // atari detection
   //
/*
   @Test
   public void testAtari()
   {
      // This test is primarily intended to demonstrate and test how
      // the check for opponent atari is supposed to work. E.g., black
      // makes a move that places a white group in atari.
      //
      // These test cases came from http://senseis.xmp.net/?SpotTheAtari
      //

      Board.setSize( 9 );
      PrisonerFinder finder = new PrisonerFinder( true );

      // White stone at E6.
      //
      Board board = new Board();
      short[] w_pnts = new short[] { 23, 24, 33, 41, 49 };
      short[] b_pnts = new short[] { 31, 32, 40, 50 };
      String msg = "white stone at E6"; //$NON-NLS-1$
      board.setColor( w_pnts, Colors.WHITE );
      board.setColor( b_pnts, Colors.BLACK );
      finder.find( board, (short) 50, Colors.BLACK );
      assertThat( msg, finder.getPrisoners().size(), is( 0 ) );
      assertThat( msg, finder.getAtariGroupCount(), is( 1 ) );
      Set<Short> group = finder.getAtariGroups().get( 0 );
      assertThat( msg, group.contains( (short) 41 ), is( true ) );

      // White stones at D4 and D5.
      //
      board.clear();
      w_pnts = new short[] { 30, 31, 41, 42 };
      b_pnts = new short[] { 21, 22, 29, 32, 40 };
      msg = "white stones at D4 and D5"; //$NON-NLS-1$
      board.setColor( w_pnts, Colors.WHITE );
      board.setColor( b_pnts, Colors.BLACK );
      finder.find( board, (short) 40, Colors.BLACK );
      assertThat( msg, finder.getPrisoners().size(), is( 0 ) );
      assertThat( msg, finder.getAtariGroupCount(), is( 1 ) );
      group = finder.getAtariGroups().get( 0 );
      assertThat( msg, group.size(), is( 2 ) );
      assertThat( msg, group.contains( (short) 30 ), is( true ) );
      assertThat( msg, group.contains( (short) 31 ), is( true ) );

      // White stones D4 and D5; white stone at E6
      //
      board.clear();
      w_pnts = new short[] { 11, 20, 23, 30, 31, 33, 41 };
      b_pnts = new short[] { 21, 22, 29, 39, 40, 42, 50 };
      msg = "white stones D4 and D5; white stone at E6"; //$NON-NLS-1$
      board.setColor( w_pnts, Colors.WHITE );
      board.setColor( b_pnts, Colors.BLACK );
      finder.find( board, (short) 40, Colors.BLACK );
      assertThat( msg, finder.getPrisoners().size(), is( 0 ) );
      assertThat( msg, finder.getAtariGroupCount(), is( 2 ) );
      Iterator<Set<Short>> group_iter = finder.getAtariGroups().iterator();
      assertThat( msg, group_iter.hasNext(), is( true ) );
      group = group_iter.next();
      int count = group.size();
      if (count == 1)
      {
         assertThat( msg, group.contains( (short) 41 ), is( true ) );
      }
      else if (count == 2)
      {
         assertThat( msg, group.contains( (short) 30 ), is( true ) );
         assertThat( msg, group.contains( (short) 31 ), is( true ) );
      }
      else
      {
         Assert.fail( "atari group count was not 1 or 2" ); //$NON-NLS-1$
      }
      assertThat( msg, group_iter.hasNext(), is( true ) );
      group = group_iter.next();
      assertThat( msg, group.size(), is( not( count ) ) );
      if (group.size() == 1)
      {
         assertThat( msg, group.contains( (short) 41 ), is( true ) );
      }
      else if (group.size() == 2)
      {
         assertThat( msg, group.contains( (short) 30 ), is( true ) );
         assertThat( msg, group.contains( (short) 31 ), is( true ) );
      }
      else
      {
         Assert.fail( "atari group count was not 1 or 2" ); //$NON-NLS-1$
      }

      // Black stones C4, C5, D4, D6, E4, E5, and E6
      //
      board.clear();
      w_pnts = new short[] { 12, 13, 20, 23, 29, 33, 38, 42, 48, 49, 50 };
      b_pnts = new short[] { 10, 11, 21, 22, 30, 32, 39, 40, 41, 47, 51, 56, 59, 60 };
      msg = "black stones C4, C5, D4, D6, E4, E5, and E6"; //$NON-NLS-1$
      board.setColor( w_pnts, Colors.WHITE );
      board.setColor( b_pnts, Colors.BLACK );
      finder.find( board, (short) 23, Colors.WHITE );
      assertThat( msg, finder.getPrisoners().size(), is( 0 ) );
      assertThat( msg, finder.getAtariGroupCount(), is( 1 ) );
      group = finder.getAtariGroups().get( 0 );
      assertThat( msg, group.size(), is( 7 ) );
      assertThat( msg, group.contains( (short) 21 ), is( true ) );
      assertThat( msg, group.contains( (short) 22 ), is( true ) );
      assertThat( msg, group.contains( (short) 30 ), is( true ) );
      assertThat( msg, group.contains( (short) 32 ), is( true ) );
      assertThat( msg, group.contains( (short) 39 ), is( true ) );
      assertThat( msg, group.contains( (short) 40 ), is( true ) );
      assertThat( msg, group.contains( (short) 41 ), is( true ) );
   }
*/

   @Test
   public void testHash()
   {
      Board board = new Board();
      board.setColor( (short) 12, Colors.BLACK );
      assertThat( board.getZobrist(), is( not( 0L ) ) );
      assertThat( board.hashCode(), is( not( 0 ) ) );

      board = new Board( m_board );
      long hash = board.getZobrist();
      assertThat( hash, is( m_board.getZobrist() ) );
      board.setColor( (short) 20, Colors.WHITE );
      board.setColor( (short) 21, Colors.BLACK );
      hash ^= Constants.WHITE_POS_VALUES[20];
      hash ^= Constants.BLACK_POS_VALUES[21];
      assertThat( board.getZobrist(), is( hash ) );

      Board.setSize( 19 );
      board = new Board();
      assertThat( board.getZobrist(), is( 0L ) );
      board.setColor( (short) 65, Colors.BLACK );
      assertThat( board.getZobrist(), is( Constants.BLACK_POS_VALUES[65] ) );
      board.clear();
      assertThat( board.getZobrist(), is( 0L ) );
      assertThat( board.hashCode(), is( 0 ) );
      board.setColor( (short) 101, Colors.WHITE );
      assertThat( board.getZobrist(), is( Constants.WHITE_POS_VALUES[101] ) );
   }

   @Test
   public void testLibertyFinding()
   {
      PrisonerFinder finder = new PrisonerFinder();

      short[] pnts = new short[]
      {
         0, 0, 20, 20, 5, 5, 11, 11, 23, 23
      };
      boolean[] has_lib = new boolean[]
      {
         true, true, true, true, true, true, false, true, true, true
      };
      boolean[] is_eye = new boolean[]
      {
         false, false, false, false, false, false, false, false, false, false
      };
      assertThat( pnts.length, is( has_lib.length ) );
      assertThat( pnts.length, is( is_eye.length ) );
      Colors color = Colors.WHITE;
      for (int i = 0; i < pnts.length; i++)
      {
         color = Game.otherPlayer( color );
         String msg = color.toString() + " " + pnts[i]; //$NON-NLS-1$
         finder.find( m_board, pnts[i], color );
         assertThat( msg, finder.hasLiberty(), is( has_lib[i] ) );
         msg += " eye check"; //$NON-NLS-1$
         assertThat( msg, finder.isEye(), is( is_eye[i] ) );
      }

      Board copy = new Board( m_board );
      copy.setColor( (short) 12, Colors.WHITE );
      is_eye[is_eye.length - 3] = true;
      color = Colors.WHITE;
      for (int i = 0; i < pnts.length; i++)
      {
         color = Game.otherPlayer( color );
         String msg = color.toString() + " " + pnts[i]; //$NON-NLS-1$
         finder.find( copy, pnts[i], color );
         assertThat( msg, finder.hasLiberty(), is( has_lib[i] ) );
         msg += " eye check"; //$NON-NLS-1$
         assertThat( msg, finder.isEye(), is( is_eye[i] ) );
      }

      copy.setColor( (short) 7, Colors.BLACK );
      color = Colors.WHITE;
      for (int i = 0; i < pnts.length; i++)
      {
         color = Game.otherPlayer( color );
         String msg = color.toString() + " " + pnts[i]; //$NON-NLS-1$
         finder.find( copy, pnts[i], color );
         assertThat( msg, finder.hasLiberty(), is( has_lib[i] ) );
         msg += " eye check"; //$NON-NLS-1$
         assertThat( msg, finder.isEye(), is( is_eye[i] ) );
      }

      copy.setColor( (short) 17, Colors.BLACK );
      is_eye[is_eye.length - 3] = false;
      color = Colors.WHITE;
      for (int i = 0; i < pnts.length; i++)
      {
         color = Game.otherPlayer( color );
         String msg = color.toString() + " " + pnts[i]; //$NON-NLS-1$
         finder.find( copy, pnts[i], color );
         assertThat( msg, finder.hasLiberty(), is( has_lib[i] ) );
         msg += " eye check"; //$NON-NLS-1$
         assertThat( msg, finder.isEye(), is( is_eye[i] ) );
      }

      copy.setColor( (short) 22, Colors.BLACK );
      has_lib[has_lib.length - 1] = false;
      is_eye[is_eye.length - 2] = true;
      color = Colors.WHITE;
      for (int i = 0; i < pnts.length; i++)
      {
         color = Game.otherPlayer( color );
         String msg = color.toString() + " " + pnts[i]; //$NON-NLS-1$
         finder.find( copy, pnts[i], color );
         assertThat( msg, finder.hasLiberty(), is( has_lib[i] ) );
         msg += " eye check"; //$NON-NLS-1$
         assertThat( msg, finder.isEye(), is( is_eye[i] ) );
      }

      copy.setColor( (short) 17, Colors.WHITE );
      is_eye[is_eye.length - 2] = false;
      is_eye[is_eye.length - 3] = true;
      color = Colors.WHITE;
      for (int i = 0; i < pnts.length; i++)
      {
         color = Game.otherPlayer( color );
         String msg = color.toString() + " " + pnts[i]; //$NON-NLS-1$
         finder.find( copy, pnts[i], color );
         assertThat( msg, finder.hasLiberty(), is( has_lib[i] ) );
         msg += " eye check"; //$NON-NLS-1$
         assertThat( msg, finder.isEye(), is( is_eye[i] ) );
      }
   }

   @Test
   public void testLocations()
   {
      Locations[] actuals = new Locations[Board.getTotal()];
      Locations[] expected = new Locations[]
      {
         Locations.CORNER, Locations.CORNER, Locations.CORNER, Locations.CORNER,
         Locations.EDGE, Locations.EDGE, Locations.EDGE, Locations.EDGE,
         Locations.INTERIOR, Locations.INTERIOR
      };
      List<Short> test_pnts = Arrays.asList( (short) 0, (short) 4, (short) 20, (short) 24, (short) 1, (short) 14, (short) 15, (short) 23, (short) 11, (short) 8 );
      assertThat( test_pnts.size(), is( expected.length ) );
      Board.getLocations( test_pnts, actuals );
      for (int i = 0; i < test_pnts.size(); i++)
      {
         short pnt = test_pnts.get( i );
         assertThat( "GetLocations() -- " + pnt, actuals[pnt], is( expected[i] ) ); //$NON-NLS-1$
         assertThat( "GetLocation() -- " + pnt, Board.getLocation( pnt ), is( expected[i] ) ); //$NON-NLS-1$
      }
   }

   @Test
   public void testLongBitmasks()
   {
      assertThat( 0x995dc01825410c3aL & 0x8000000000000000L, is( 0x8000000000000000L ) );
      assertThat( 0x995dc01825410c3aL >>> 4, is( 0x0995dc01825410c3L ) );
      assertThat( 0x8000000000000000L | 0x4000000000000000L, is( 0xC000000000000000L ) );
   }

   @Test
   public void testMiscOps()
   {
      Board copy = new Board( m_board );
      assertThat( copy, is( m_board ) );
      assertThat( copy.getColor( (short) 0 ), is( Colors.BLACK ) );
      assertThat( copy.getColor( (short) 24 ), is( Colors.BLACK ) );
      assertThat( copy.getColor( (short) 10 ), is( Colors.WHITE ) );
      assertThat( copy.getColor( (short) 20 ), is( Colors.EMPTY ) );
      for (short pnt = 0; pnt < Board.getTotal(); pnt++)
         assertThat( copy.getColor( pnt ), is( not( Colors.DONT_CARE ) ) );
      copy.setColor( (short) 3, Colors.WHITE );
      copy.setColor( (short) 24, Colors.EMPTY );
      assertThat( m_board.getColor( (short) 3 ), is( Colors.EMPTY ) );
      assertThat( m_board.getColor( (short) 24 ), is( Colors.BLACK ) );
      assertThat( copy, is( not( m_board ) ) );

      Board board2 = new Board( m_board );
      assertThat( board2.getZobrist(), is( m_board.getZobrist() ) );
      board2.setColor( (short) 24, Colors.EMPTY );
      assertThat( board2, is( not( m_board ) ) );
      assertThat( board2.getZobrist(), is( not( m_board.getZobrist() ) ) );
   }

   @Test
   public void testPrisonerFinder()
   {
      PrisonerFinder finder = new PrisonerFinder();

      finder.find( m_board, (short) 11, Colors.WHITE );
      Set<Short> prisoners = finder.getPrisoners();
      assertThat( prisoners.size(), is( 1 ) );
      assertThat( prisoners.contains( (short) 12 ), is( true ) );

      Board copy = new Board( m_board );
      short[] black_pnts = new short[] { 11, 5, 1, 2, 14 }; // pnt 11 is technically suicide, but this is just for testing purposes
      for (int i = 0; i < black_pnts.length; i++)
         copy.setColor( black_pnts[i], Colors.BLACK );
      finder.find( copy, (short) 8, Colors.BLACK );  // should produce two groups of prisoners
      prisoners = finder.getPrisoners();
      assertThat( prisoners.size(), is( 3 ) );
      assertThat( prisoners.contains( (short) 6 ), is( true ) );
      assertThat( prisoners.contains( (short) 7 ), is( true ) );
      assertThat( prisoners.contains( (short) 13 ), is( true ) );

      finder.find( m_board, (short) 13, Colors.WHITE );
      assertThat( finder.getPrisoners().isEmpty(), is( true ) );
      finder.find( m_board, (short) 0, Colors.BLACK );
      assertThat( finder.getPrisoners().isEmpty(), is( true ) );
   }

   @Test
   public void testTransforms()
   {
      assertThat( Board.flipPoint( 0 ), is( 4 ) );
      assertThat( Board.flipPoint( 1 ), is( 3 ) );
      assertThat( Board.flipPoint( 2 ), is( 2 ) );
      assertThat( Board.flipPoint( 3 ), is( 1 ) );
      assertThat( Board.flipPoint( 4 ), is( 0 ) );
      assertThat( Board.flipPoint( 20 ), is( 24 ) );
      assertThat( Board.flipPoint( 21 ), is( 23 ) );
      assertThat( Board.flipPoint( 22 ), is( 22 ) );
      assertThat( Board.flipPoint( 23 ), is( 21 ) );
      assertThat( Board.flipPoint( 24 ), is( 20 ) );

      // TODO: add more tests for other board sizes

      Board copy1 = m_board.transform( 1, false );
      assertThat( "source: 2,3 rotate 1", copy1.getColor( (short) 17 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 1,2 rotate 1", copy1.getColor( (short) 13 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 3,1 rotate 1", copy1.getColor( (short) 6 ),  is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 0,0 rotate 1", copy1.getColor( (short) 4 ),  is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 2,2 rotate 1", copy1.getColor( (short) 12 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 3,3 rotate 1", copy1.getColor( (short) 16 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 4,4 rotate 1", copy1.getColor( (short) 20 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 0,2 rotate 1", copy1.getColor( (short) 14 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 1,0 rotate 1", copy1.getColor( (short) 3 ),  is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 3,4 rotate 1", copy1.getColor( (short) 21 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 4,2 rotate 1", copy1.getColor( (short) 10 ), is( Colors.EMPTY ) ); //$NON-NLS-1$

      Board copy2 = m_board.transform( 2, false );
      assertThat( "source: 2,3 rotate 2", copy2.getColor( (short) 11 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 1,2 rotate 2", copy2.getColor( (short) 17 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 3,1 rotate 2", copy2.getColor( (short) 8 ),  is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 0,0 rotate 2", copy2.getColor( (short) 24 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 2,2 rotate 2", copy2.getColor( (short) 12 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 3,3 rotate 2", copy2.getColor( (short) 6 ),  is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 4,4 rotate 2", copy2.getColor( (short) 0 ),  is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 0,2 rotate 2", copy2.getColor( (short) 22 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 1,0 rotate 2", copy2.getColor( (short) 19 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 3,4 rotate 2", copy2.getColor( (short) 5 ),  is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 4,2 rotate 2", copy2.getColor( (short) 2 ),  is( Colors.EMPTY ) ); //$NON-NLS-1$

      Board copy3 = m_board.transform( 3, false );
      assertThat( "source: 2,3 rotate 3", copy3.getColor( (short) 7 ),  is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 1,2 rotate 3", copy3.getColor( (short) 11 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 3,1 rotate 3", copy3.getColor( (short) 18 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 0,0 rotate 3", copy3.getColor( (short) 20 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 2,2 rotate 3", copy3.getColor( (short) 12 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 3,3 rotate 3", copy3.getColor( (short) 8 ),  is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 4,4 rotate 3", copy3.getColor( (short) 4 ),  is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 0,2 rotate 3", copy3.getColor( (short) 10 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 1,0 rotate 3", copy3.getColor( (short) 21 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 3,4 rotate 3", copy3.getColor( (short) 3 ),  is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 4,2 rotate 3", copy3.getColor( (short) 14 ), is( Colors.EMPTY ) ); //$NON-NLS-1$

      Board copy4 = m_board.transform( -1, false );
      assertThat( copy4, is( copy3 ) );
      copy4 = copy4.transform( -1, false );
      assertThat( copy4, is( copy2 ) );
      copy4 = copy4.transform( -1, false );
      assertThat( copy4, is( copy1 ) );
      assertThat( copy4, is( not( copy2 ) ) );
      assertThat( copy4, is( not( copy3 ) ) );

      copy4 = m_board.transform( 0, true );
      assertThat( "source: 2,3 flip true", copy4.getColor( (short) 11 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 1,2 flip true", copy4.getColor( (short) 7 ),  is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 3,1 flip true", copy4.getColor( (short) 18 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( "source: 0,0 flip true", copy4.getColor( (short) 4 ),  is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 2,2 flip true", copy4.getColor( (short) 12 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 3,3 flip true", copy4.getColor( (short) 16 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 4,4 flip true", copy4.getColor( (short) 20 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( "source: 0,2 flip true", copy4.getColor( (short) 2 ),  is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 1,0 flip true", copy4.getColor( (short) 9 ),  is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 3,4 flip true", copy4.getColor( (short) 15 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( "source: 4,2 flip true", copy4.getColor( (short) 22 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
   }

   @Test
   public void testVisitors()
   {
      TestPointVisitor visitor = new TestPointVisitor();

      m_board.visitNeighbors( (short) 2, visitor );
      assertThat( visitor.getCount(), is( 3 ) );
      assertThat( visitor.contains( Colors.EMPTY, 1 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 3 ), is( true ) );
      assertThat( visitor.contains( Colors.WHITE, 7 ), is( true ) );

      visitor.clear();
      m_board.visitNeighbors( (short) 24, visitor );
      assertThat( visitor.getCount(), is( 2 ) );
      assertThat( visitor.contains( Colors.EMPTY, 23 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 19 ), is( true ) );

      visitor.clear();
      m_board.visitNeighbors( (short) 13, visitor );
      assertThat( visitor.getCount(), is( 4 ) );
      assertThat( visitor.contains( Colors.EMPTY, 14 ), is( true ) );
      assertThat( visitor.contains( Colors.BLACK, 12 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 8 ), is( true ) );
      assertThat( visitor.contains( Colors.BLACK, 18 ), is( true ) );

      visitor.clear();
      m_board.visitDiagonals( (short) 1, visitor );
      assertThat( visitor.getCount(), is( 2 ) );
      assertThat( visitor.contains( Colors.WHITE, 7 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 5 ), is( true ) );

      visitor.clear();
      m_board.visitDiagonals( (short) 6, visitor );
      assertThat( visitor.getCount(), is( 4 ) );
      assertThat( visitor.contains( Colors.EMPTY, 2 ), is( true ) );
      assertThat( visitor.contains( Colors.BLACK, 0 ), is( true ) );
      assertThat( visitor.contains( Colors.BLACK, 12 ), is( true ) );
      assertThat( visitor.contains( Colors.WHITE, 10 ), is( true ) );

      visitor.clear();
      m_board.visitDiagonals( (short) 0, visitor );
      assertThat( visitor.getCount(), is( 1 ) );
      assertThat( visitor.contains( Colors.WHITE, 6 ), is( true ) );

      visitor.clear();
      m_board.visitDiagonals( (short) 24, visitor );
      assertThat( visitor.getCount(), is( 1 ) );
      assertThat( visitor.contains( Colors.BLACK, 18 ), is( true ) );

      visitor.clear();
      m_board.visitPoints( Colors.WHITE, visitor );
      assertThat( visitor.getCount(), is( WHITE_POINTS.length ) );
      int i = 0;
      for (i = 0; i < WHITE_POINTS.length; i++)
         assertThat( visitor.contains( Colors.WHITE, WHITE_POINTS[i] ), is( true ) );

      visitor.clear();
      m_board.visitPoints( Colors.BLACK, visitor );
      assertThat( visitor.getCount(), is( BLACK_POINTS.length ) );
      for (i = 0; i < BLACK_POINTS.length; i++)
         assertThat( visitor.contains( Colors.BLACK, BLACK_POINTS[i] ), is( true ) );

      visitor.clear();
      m_board.visitPoints( Colors.EMPTY, visitor );
      assertThat( visitor.getCount(), is( Board.getTotal() - WHITE_POINTS.length - BLACK_POINTS.length ) );
      for (i = 0; i < BLACK_POINTS.length; i++)
         assertThat( visitor.contains( Colors.BLACK, BLACK_POINTS[i] ), is( false ) );
      for (i = 0; i < WHITE_POINTS.length; i++)
         assertThat( visitor.contains( Colors.WHITE, WHITE_POINTS[i] ), is( false ) );

      visitor.clear();
      m_board.visitPoints( Colors.DONT_CARE, visitor );
      assertThat( visitor.getCount(), is( Board.getTotal() ) );
      for (i = 0; i < BLACK_POINTS.length; i++)
         assertThat( visitor.contains( Colors.BLACK, BLACK_POINTS[i] ), is( true ) );
      for (i = 0; i < WHITE_POINTS.length; i++)
         assertThat( visitor.contains( Colors.WHITE, WHITE_POINTS[i] ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 4 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 8 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 11 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 15 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, 22 ), is( true ) );

      TestPointVisitor visitor2 = new TestPointVisitor();
      visitor.clear();
      m_board.visitChain( (short) 0, visitor, visitor2 );
      assertThat( visitor.getCount(), is( 0 ) );
      assertThat( visitor2.getCount(), is( 2 ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 1 ), is( true ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 5 ), is( true ) );

      visitor.clear();
      visitor2.clear();
      m_board.visitChain( (short) 12, visitor, visitor2 );
      assertThat( visitor.getCount(), is( 0 ) );
      assertThat( visitor2.getCount(), is( 1 ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 11 ), is( true ) );

      visitor.clear();
      visitor2.clear();
      m_board.visitChain( (short) 6, visitor, visitor2 );
      assertThat( visitor.getCount(), is( 1 ) );
      assertThat( visitor.contains( Colors.WHITE, (short) 7 ), is( true ) );
      assertThat( visitor2.getCount(), is( 5 ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 1 ), is( true ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 2 ), is( true ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 5 ), is( true ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 8 ), is( true ) );
      assertThat( visitor2.contains( Colors.EMPTY, (short) 11 ), is( true ) );

      visitor.clear();
      visitor2.clear();
      m_board.visitChain( (short) 4, visitor, visitor2 );
      assertThat( visitor.getCount(), is( 7 ) );
      assertThat( visitor.contains( Colors.EMPTY, (short) 1 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, (short) 2 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, (short) 3 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, (short) 8 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, (short) 9 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, (short) 14 ), is( true ) );
      assertThat( visitor.contains( Colors.EMPTY, (short) 19 ), is( true ) );
      assertThat( visitor2.getCount(), is( 0 ) );
   }

   private void initScorer( String name )
   {
      assertFalse( name.isEmpty() );
      if (m_system.equalsIgnoreCase( name ))
         return;
      m_system = name;
      m_scorer = m_plugin_mgr.getPlugin( Scorer.class, PluginUtils.getScoringCapability( m_system ) );
      assertNotNull( m_scorer );
   }

   private static final int SIZE = 5;

   // The 5x5 test board:
   //
   //  . . . . X
   //  . . O X .
   //  . O X O .
   //  . O . O .
   //  X . O . .
   //
   private static final short[] BLACK_POINTS = new short[] { 0, 12, 18, 24 };
   private static final short[] WHITE_POINTS = new short[] { 6, 7, 10, 13, 16, 17 };

   private Board  m_board;
   private String m_system = ""; //$NON-NLS-1$
   private Scorer m_scorer;
}
