//
// ChainTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.game;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections.map.MultiValueMap;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Chain;
import com.dakengo.libs.game.ChainFinder;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;

/**
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class ChainTests
{
   @Before
   public void setUp()
   {
      Board.setSize( SIZE );
      m_board = new Board();
      m_board.setColor( Convert.stringsToPoints( CHAIN1_POINTS ), Colors.BLACK );
      m_board.setColor( Convert.stringsToPoints( CHAIN3_POINTS ), Colors.BLACK );
      m_board.setColor( Convert.stringsToPoints( CHAIN5_POINTS ), Colors.BLACK );
      m_board.setColor( Convert.stringsToPoints( CHAIN6_POINTS ), Colors.BLACK );
      m_board.setColor( Convert.stringsToPoints( CHAIN2_POINTS ), Colors.WHITE );
      m_board.setColor( Convert.stringsToPoints( CHAIN4_POINTS ), Colors.WHITE );
      String[] anchor_text = new String[] { "e8", "b7", "g8", "c3", "g4", "h3" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
      m_chains = new Chain[anchor_text.length];
      m_anchors = new short[anchor_text.length];
      for (int i = 0; i < anchor_text.length; i++)
      {
         m_anchors[i] = Convert.stringToPoint( anchor_text[i] );
         m_chains[i] = new Chain( m_board, m_anchors[i] );
      }
   }

   @Test
   public void testComparisons()
   {
      assertChainLengths();
      assertThat( m_empty.getAnchor(), is( Board.INVALID_POINT ) );

      for (int i = 0; i < m_chains.length; i++)
      {
         assertThat( m_chains[i], is( not( m_empty ) ) );
         assertThat( m_chains[i].hashCode(), is( not( m_empty.hashCode() ) ) );
         assertThat( m_chains[i].connects( m_empty ), is( false ) );
         assertThat( m_chains[i].toString(), m_chains[i].getAnchor(), is( m_anchors[i] ) );

         for (short member : m_chains[i].getPoints())
         {
            assertThat( m_chains[i].contains( member ), is( true ) );
            assertThat( m_chains[i].connects( member ), is( true ) );
         }

         Set<Short> expected_libs = null;
         switch (i)
         {
            case 0:
               expected_libs = Convert.stringsToPoints( CHAIN1_LIBS );
               break;
            case 1:
               expected_libs = Convert.stringsToPoints( CHAIN2_LIBS );
               break;
            case 2:
               expected_libs = Convert.stringsToPoints( CHAIN3_LIBS );
               break;
            case 3:
               expected_libs = Convert.stringsToPoints( CHAIN4_A_LIBS );
               expected_libs.addAll( Convert.stringsToPoints( CHAIN4_B_LIBS ) );
               break;
            case 4:
               expected_libs = Convert.stringsToPoints( CHAIN5_A_LIBS );
               expected_libs.addAll( Convert.stringsToPoints( CHAIN5_B_LIBS ) );
               break;
            case 5:
               expected_libs = Convert.stringsToPoints( CHAIN6_A_LIBS );
               expected_libs.addAll( Convert.stringsToPoints( CHAIN6_B_LIBS ) );
               break;
            default:
               Assert.fail( "unexpected chain index" ); //$NON-NLS-1$
               break;
         }
         assertLiberties( m_chains[i].getAnchor(), expected_libs, m_chains[i].getLiberties() );

         for (int j = 0; j < m_chains.length; j++)
         {
            if (i == j)
               continue;
            assertThat( m_chains[i], is( not( m_chains[j] ) ) );
            assertThat( m_chains[i].hashCode(), is( not( m_chains[j].hashCode() ) ) );
            assertThat( m_chains[i].connects( m_chains[j] ), is( false ) );
         }
      }

      Chain[] new_chains = new Chain[m_chains.length];
      String[] anchors = new String[] { "e7", "b7", "g7", "c1", "f4", "h2" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
      for (int i = 0; i < anchors.length; i++)
         new_chains[i] = new Chain( m_board, Convert.stringToPoint( anchors[i] ) );
      for (int i = 0; i < m_chains.length; i++)
      {
         assertThat( new_chains[i].getCount(), is( m_chains[i].getCount() ) );
         assertThat( new_chains[i], is( m_chains[i] ) );
         assertThat( new_chains[i].hashCode(), is( m_chains[i].hashCode() ) );
         assertThat( new_chains[i].connects( m_chains[i] ), is( true ) );
         assertThat( new_chains[i].toString(), new_chains[i].getAnchor(), is( m_chains[i].getAnchor() ) );
         assertLiberties( m_chains[i].getAnchor(), m_chains[i].getLiberties(), new_chains[i].getLiberties() );
      }

      assertThat( "chain4 does not connect to chain5", m_chains[4].connects( m_chains[5] ), is( false ) ); //$NON-NLS-1$
   }

   @Test
   public void testFindChains()
   {
      ChainFinder finder = new ChainFinder();
      finder.find( m_board, Colors.BLACK, false );
      Map<Short,Chain> chains = finder.getChains();
      assertThat( chains.size(), is( 4 ) );
      assertThat( chains.containsKey( m_anchors[0] ), is( true ) );
      assertThat( chains.containsKey( m_anchors[2] ), is( true ) );
      assertThat( chains.containsKey( m_anchors[4] ), is( true ) );
      assertThat( chains.containsKey( m_anchors[5] ), is( true ) );

      finder.find( m_board, Colors.WHITE, false );
      chains = finder.getChains();
      assertThat( chains.size(), is( 2 ) );
      assertThat( chains.containsKey( m_anchors[1] ), is( true ) );
      assertThat( chains.containsKey( m_anchors[3] ), is( true ) );
   }

   @Test
   @SuppressWarnings( "unchecked" )
   public void testFindEyeSpacesDefaultBoard()
   {
      ChainFinder finder = new ChainFinder();

      finder.find( m_board, Colors.BLACK, true );
      MultiValueMap spaces = finder.getSpaces();
      assertThat( spaces.size(), is( 2 ) );
      assertThat( spaces.containsKey( m_chains[4].getAnchor() ), is( true ) );
      assertThat( spaces.containsKey( m_chains[5].getAnchor() ), is( true ) );
      assertThat( spaces.getCollection( m_chains[4].getAnchor() ), is( spaces.getCollection( m_chains[5].getAnchor() ) ) );
      Set<Short> expected = Convert.stringsToPoints( CHAIN5_B_LIBS );
      Collection<?> actuals = spaces.getCollection( m_chains[4].getAnchor() );
      assertThat( actuals.size(), is( 1 ) );
      assertThat( (Set<Short>) actuals.iterator().next(), is( expected ) );

      finder.find( m_board, Colors.WHITE, true );
      spaces = finder.getSpaces();
      assertThat( spaces.size(), is( 1 ) );
      assertThat( spaces.containsKey( m_chains[3].getAnchor() ), is( true ) );
      String[] chain4_exp_str = new String[] { "a2", "b2", "b1", "a1" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
      expected = Convert.stringsToPoints( chain4_exp_str );
      actuals = spaces.getCollection( m_chains[3].getAnchor() );
      assertThat( (Set<Short>) actuals.iterator().next(), is( expected ) );

      for (int i = 0; i < 3; i++)
      {
         finder.find( m_board, m_chains[i] );
         assertThat( finder.getSpaces().size(), is( 0 ) );
      }

      finder.find( m_board, m_chains[3] );
      spaces = finder.getSpaces();
      assertThat( spaces.size(), is( 1 ) );
      actuals = spaces.getCollection( m_chains[3].getAnchor() );
      assertThat( (Set<Short>) actuals.iterator().next(), is( expected ) );

      finder.find( m_board, m_chains[4] );
      spaces = finder.getSpaces();
      assertThat( spaces.size(), is( 1 ) );
      expected = Convert.stringsToPoints( CHAIN5_B_LIBS );
      actuals = spaces.getCollection( m_chains[4].getAnchor() );
      assertThat( (Set<Short>) actuals.iterator().next(), is( expected ) );

      finder.find( m_board, m_chains[5] );
      spaces = finder.getSpaces();
      assertThat( spaces.size(), is( 1 ) );
      actuals = spaces.getCollection( m_chains[5].getAnchor() );
      assertThat( (Set<Short>) actuals.iterator().next(), is( expected ) );
   }

   @Test
   @SuppressWarnings( "unchecked" )
   public void testFindSplitEyespaces()
   {
      // 9  . . . . . . . . .
      // 8  . O O O O O O O .
      // 7  . O . O . O . O .
      // 6  . O O O O O O O .
      // 5  . . . . . . . . .
      // 4  . . . . . . . . .
      // 3  X X X X X . . . .
      // 2  . . X . X . . . .
      // 1  . . X . X . . . .
      //    A B C D E F G H J
      //
      String[] b_pnts = new String[] { "c1", "e1", "c2", "e2", "a3", "b3", "c3", "d3", "e3" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$
      String[] w_pnts = new String[] { "b6", "c6", "d6", "e6", "f6", "g6", "h6", "b7", "b8", "c8", "d8", "e8", "f8", "g8", "h8", "f7", "d7", "h7" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$ //$NON-NLS-12$ //$NON-NLS-13$ //$NON-NLS-14$ //$NON-NLS-15$ //$NON-NLS-16$ //$NON-NLS-17$ //$NON-NLS-18$
      Board board = new Board();
      board.setColor( Convert.stringsToPoints( b_pnts ), Colors.BLACK );
      board.setColor( Convert.stringsToPoints( w_pnts ), Colors.WHITE );

      ChainFinder finder = new ChainFinder();
      finder.find( board, Colors.BLACK, true );
      assertThat( finder.getChains().size(), is( 1 ) );

      short anchor = Convert.stringToPoint( "e3" ); //$NON-NLS-1$
      Map<Short,Chain> chains = finder.getChains();
      assertThat( "chain e3 missing", chains.containsKey( anchor ), is( true ) ); //$NON-NLS-1$
      Chain chain = new Chain( board, anchor );
      assertThat( "chain e3", chains.get( anchor ), is( chain ) ); //$NON-NLS-1$
      MultiValueMap space_map = finder.getSpaces();
      assertThat( space_map.size(), is( 1 ) );
      Collection<?> spaces = space_map.getCollection( anchor );
      assertThat( spaces.size(), is( 2 ) );
      boolean found1 = false;
      boolean found2 = false;
      Iterator<?> iter = spaces.iterator();
      while (iter.hasNext())
      {
         Set<Short> pnts = (Set<Short>) iter.next();
         if (pnts.equals( Convert.stringsToPoints( new String[] { "a1", "b1", "a2", "b2" } ) )) //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
            found1 = true;
         else if (pnts.equals( Convert.stringsToPoints( new String[] { "d1", "d2" } ) )) //$NON-NLS-1$ //$NON-NLS-2$
            found2 = true;
         else
            Assert.fail( "unexpected space" ); //$NON-NLS-1$
      }
      assertThat( found1 && found2, is( true ) );

      finder.find( board, Colors.WHITE, true );
      assertThat( finder.getChains().size(), is( 1 ) );

      anchor = Convert.stringToPoint( "h8" ); //$NON-NLS-1$
      chains = finder.getChains();
      assertThat( "chain h8 missing", chains.containsKey( anchor ), is( true ) ); //$NON-NLS-1$
      chain = new Chain( board, anchor );
      assertThat( "chain h8", chains.get( anchor ), is( chain ) ); //$NON-NLS-1$
      space_map = finder.getSpaces();
      assertThat( space_map.size(), is( 1 ) );
      spaces = space_map.getCollection( anchor );
      assertThat( spaces.size(), is( 3 ) );
      found1 = true;
      found2 = true;
      boolean found3 = false;
      iter = spaces.iterator();
      while (iter.hasNext())
      {
         Set<Short> pnts = (Set<Short>) iter.next();
         if (pnts.equals( Convert.stringsToPoints( new String[] { "c7" } ) )) //$NON-NLS-1$
            found1 = true;
         else if (pnts.equals( Convert.stringsToPoints( new String[] { "e7" } ) )) //$NON-NLS-1$
            found2 = true;
         else if (pnts.equals( Convert.stringsToPoints( new String[] { "g7" } ) )) //$NON-NLS-1$
            found3 = true;
         else
            Assert.fail( "unexpected space" ); //$NON-NLS-1$
      }
      assertThat( found1 && found2 && found3, is( true ) );
   }

   private void assertChainLengths()
   {
      assertThat( "empty chain", m_empty.getCount(), is( 0 ) ); //$NON-NLS-1$
      assertThat( "chain 1: " + m_chains[0].toString(), m_chains[0].getCount(), is( CHAIN1_POINTS.length ) ); //$NON-NLS-1$
      assertThat( "chain 2: " + m_chains[1].toString(), m_chains[1].getCount(), is( CHAIN2_POINTS.length ) ); //$NON-NLS-1$
      assertThat( "chain 3: " + m_chains[2].toString(), m_chains[2].getCount(), is( CHAIN3_POINTS.length ) ); //$NON-NLS-1$
      assertThat( "chain 4: " + m_chains[3].toString(), m_chains[3].getCount(), is( CHAIN4_POINTS.length ) ); //$NON-NLS-1$
      assertThat( "chain 5: " + m_chains[4].toString(), m_chains[4].getCount(), is( CHAIN5_POINTS.length ) ); //$NON-NLS-1$
      assertThat( "chain 6: " + m_chains[5].toString(), m_chains[5].getCount(), is( CHAIN6_POINTS.length ) ); //$NON-NLS-1$
   }

   private void assertLiberties( short anchor, Set<Short> expected, Set<Short> actuals )
   {
      String anchor_str = Convert.pointToString( anchor );
      String exp_str = Convert.pointsToString( expected, "," ); //$NON-NLS-1$
      String act_str = Convert.pointsToString( actuals, "," ); //$NON-NLS-1$
      String msg = String.format( "chain %s libs expected: %s vs actual: %s", anchor_str, exp_str, act_str ); //$NON-NLS-1$
      assertThat( msg, actuals, is( expected ) );
   }

   //
   // 9  . . . . . . . . .
   // 8  . . . 1 1 . 3 . .
   // 7  . 2 . 1 1 . 3 . .
   // 6  . . . 1 . . 3 . .
   // 5  . . . . . . . . .
   // 4  . . . . 5 5 5 . .
   // 3  4 4 4 . 5 . . 6 .
   // 2  . . 4 . . 6 6 6 .
   // 1  . . 4 . . . . . .
   //    A B C D E F G H J
   //
   private static final String[] CHAIN1_POINTS = new String[] { "d8", "d7", "d6", "e8", "e7" };  // black //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
   private static final String[] CHAIN2_POINTS = new String[] { "b7" };                          // white //$NON-NLS-1$
   private static final String[] CHAIN3_POINTS = new String[] { "g6", "g7", "g8" };              // black //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
   private static final String[] CHAIN4_POINTS = new String[] { "a3", "b3", "c1", "c2", "c3" };  // white //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
   private static final String[] CHAIN5_POINTS = new String[] { "e3", "e4", "f4", "g4" };        // black //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
   private static final String[] CHAIN6_POINTS = new String[] { "f2", "g2", "h2", "h3" };        // black //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$

   private static final String[] CHAIN1_LIBS   = new String[] { "c6", "c7", "c8", "d5", "d9", "e6", "e9", "f7", "f8" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$
   private static final String[] CHAIN2_LIBS   = new String[] { "a7", "b8", "c7", "b6" };                               //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
   private static final String[] CHAIN3_LIBS   = new String[] { "f6", "f7", "f8", "g9", "h6", "h7", "h8", "g5" };       //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$
   private static final String[] CHAIN4_A_LIBS = new String[] { "d1", "d2", "d3", "c4", "b4", "a4" };                   //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
   private static final String[] CHAIN4_B_LIBS = new String[] { "a2", "b2", "b1" };                                     //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
   private static final String[] CHAIN5_A_LIBS = new String[] { "e2", "d3", "d4", "e5", "f5", "g5", "h4" };             //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
   private static final String[] CHAIN5_B_LIBS = new String[] { "f3", "g3" };                                           //$NON-NLS-1$ //$NON-NLS-2$
   private static final String[] CHAIN6_A_LIBS = new String[] { "e2", "f1", "g1", "h1", "j2", "j3", "h4" };             //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
   private static final String[] CHAIN6_B_LIBS = new String[] { "f3", "g3" };                                           //$NON-NLS-1$ //$NON-NLS-2$

   private static final int SIZE = 9;

   private Board   m_board;
   private Chain   m_empty   = new Chain();
   private Chain[] m_chains;
   private short[] m_anchors;
}
