//
// ConvertTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.game;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.util.Set;

import org.apache.commons.lang3.ArrayUtils;
import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.tests.utils.CollectionUtils;

/**
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class ConvertTests
{
   @Before
   public void setUp()
   {
      Board.setSize( 19 );
   }

   @Test
   public void testColors()
   {
      assertThat( Convert.colorToString( Colors.BLACK ), is( "black" ) ); //$NON-NLS-1$
      assertThat( Convert.colorToString( Colors.WHITE ), is( "white" ) ); //$NON-NLS-1$
      assertThat( Convert.colorToString( Colors.EMPTY ), is( "empty" ) ); //$NON-NLS-1$
      assertThat( Convert.colorToString( Colors.DONT_CARE ), is( "?" ) ); //$NON-NLS-1$

      assertThat( Convert.stringToColor( "b", 0 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( Convert.stringToColor( "black", 0 ), is( Colors.BLACK ) ); //$NON-NLS-1$
      assertThat( Convert.stringToColor( "w", 0 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( Convert.stringToColor( "white", 0 ), is( Colors.WHITE ) ); //$NON-NLS-1$
      assertThat( Convert.stringToColor( ".", 0 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( Convert.stringToColor( "empty", 0 ), is( Colors.EMPTY ) ); //$NON-NLS-1$
      assertThat( Convert.stringToColor( "?", 0 ), is( Colors.DONT_CARE ) ); //$NON-NLS-1$

      assertThat( Convert.colorToChar( Colors.BLACK ), is( 'B' ) );
      assertThat( Convert.colorToChar( Colors.WHITE ), is( 'W' ) );
      assertThat( Convert.colorToChar( Colors.EMPTY ), is( '.' ) );
      assertThat( Convert.colorToChar( Colors.DONT_CARE ), is( '?' ) );

      assertThat( Convert.charToColor( 'b', 0 ), is( Colors.BLACK ) );
      assertThat( Convert.charToColor( 'B', 0 ), is( Colors.BLACK ) );
      assertThat( Convert.charToColor( 'w', 0 ), is( Colors.WHITE ) );
      assertThat( Convert.charToColor( 'W', 0 ), is( Colors.WHITE ) );
      assertThat( Convert.charToColor( '.', 0 ), is( Colors.EMPTY ) );
      assertThat( Convert.charToColor( '?', 0 ), is( Colors.DONT_CARE ) );

      assertThrows(
         new Runnable() { @Override public void run() { Convert.charToColor( '*' ); } },
         ParseException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { Convert.stringToColor( "xxx" ); } }, //$NON-NLS-1$
         ParseException.class, null );
}

   @Test
   public void testPointStrings()
   {
      String[] std_text = new String[] { "B18", "s10", "c2", "T19", "t1" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
      short[] points = new short[] { (short) 36, (short) 332, (short) 39, (short) 360, (short) 342 };
      assertThat( std_text.length, is( points.length ) );
      for (int i = 0; i < std_text.length; i++)
      {
         String std = std_text[i];
         short actual = Convert.stringToPoint( std );
         assertThat( std, actual, is( points[i] ) );
         std = std.toLowerCase();
         assertThat( Convert.pointToString( actual ), is( std ) );
      }

      String pnt_text = "b12"; //$NON-NLS-1$
      Set<Short> points2 = Convert.stringToPoints( pnt_text, " " ); //$NON-NLS-1$
      assertThat( pnt_text, points2.size(), is( 1 ) );
      assertTrue( pnt_text, points2.contains( Convert.stringToPoint( pnt_text ) ) );

      pnt_text = " c12"; //$NON-NLS-1$
      points2 = Convert.stringToPoints( pnt_text, " " ); //$NON-NLS-1$
      assertThat( pnt_text, points2.size(), is( 1 ) );
      assertTrue( pnt_text, points2.contains( Convert.stringToPoint( "c12" ) ) ); //$NON-NLS-1$

      pnt_text = "d12 "; //$NON-NLS-1$
      points2 = Convert.stringToPoints( pnt_text, " " ); //$NON-NLS-1$
      assertThat( pnt_text, points2.size(), is( 1 ) );
      assertTrue( pnt_text, points2.contains( Convert.stringToPoint( "d12" ) ) ); //$NON-NLS-1$

      pnt_text = " B18 s10   c2 T19 t1  "; //$NON-NLS-1$
      points2 = Convert.stringToPoints( pnt_text, " " ); //$NON-NLS-1$
      assertThat( pnt_text, points2, is( CollectionUtils.arrayToSet( ArrayUtils.toObject( points ) ) ) );

      points2 = Convert.stringToPoints( "", " " ); //$NON-NLS-1$ //$NON-NLS-1$
      assertTrue( points2.isEmpty() );

      assertThrows(
         new Runnable() { @Override public void run() { Convert.stringToPoints( "a1", "" ); } },
         IllegalArgumentException.class, null );

      assertThrows(
         new Runnable() { @Override public void run() { Convert.stringToPoints( "aa", " " ); } },
         ParseException.class, null );

      String[] bad_text = new String[]
      {
         "aa", "12", "a1a", "1a", "a0", "a99", "z1", "z100"  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$
      };
      for (String text : bad_text)
      {
         final String str = text;
         assertThrows(
            new Runnable() { @Override public void run() { Convert.stringToPoint( str ); } },
            ParseException.class, null );
      }
   }
}
