//
// GameTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.game;

import java.util.HashSet;
import java.util.Set;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.GameException;
import com.dakengo.libs.game.KoRules;
import com.dakengo.libs.game.MoveTypes;
import com.dakengo.libs.game.Signature;
import com.dakengo.tests.utils.DkAssertUtils;

/**
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class GameTests
{
   @Rule
   public ExpectedException r_exception = ExpectedException.none();

   @Before
   public void setUp()
   {
      initGame( KoRules.POSITIONAL );
   }

   @Test
   public void testClassification()
   {
      // LEGAL move
      //
      String pnt_str = "t1"; //$NON-NLS-1$
      short move = Convert.stringToPoint( pnt_str );
      HashSet<Short> prisoners = new HashSet<Short>();
      MoveTypes type = m_game.classify( new Board( m_game.getBoard() ), Colors.BLACK, move, prisoners );
      assertThat( type, is( MoveTypes.LEGAL ) );
      assertThat( prisoners.size(), is( 0 ) );
      assertThat( m_game.isOver(), is( false ) );

      // OCCUPIED move
      //
      pnt_str = "d6"; //$NON-NLS-1$
      move = Convert.stringToPoint( pnt_str );
      type = m_game.classify( new Board( m_game.getBoard() ), Colors.BLACK, move, prisoners );
      assertThat( type, is( MoveTypes.OCCUPIED ) );
      assertThat( prisoners.size(), is( 0 ) );
      assertThat( m_game.isOver(), is( false ) );

      // SUICIDE move
      //
      pnt_str = "d5"; //$NON-NLS-1$
      move = Convert.stringToPoint( pnt_str );
      type = m_game.classify( new Board( m_game.getBoard() ), Colors.BLACK, move, prisoners );
      assertThat( type, is( MoveTypes.SUICIDE ) );
      assertThat( prisoners.size(), is( 0 ) );
      assertThat( m_game.isOver(), is( false ) );

      pnt_str = "t19"; //$NON-NLS-1$
      move = Convert.stringToPoint( pnt_str );
      type = m_game.classify( new Board( m_game.getBoard() ), Colors.BLACK, move, prisoners );
      assertThat( type, is( MoveTypes.SUICIDE ) );
      assertThat( prisoners.size(), is( 0 ) );
      assertThat( m_game.isOver(), is( false ) );

      // SELF_ATARI capturing move
      //
/*
      pnt_str = "d5"; //$NON-NLS-1$
      m_game.placeStones( Colors.BLACK, Convert.stringsToPoints( new String[] { "e6" } ) ); //$NON-NLS-1$
      move = Convert.stringToPoint( pnt_str );
      type = m_game.classify( new Board( m_game.getBoard() ), Colors.BLACK, move, prisoners );
      assertThat( type, is( MoveTypes.SELF_ATARI ) );
      assertThat( prisoners.size(), is( 1 ) );
      assertThat( prisoners.contains( Convert.stringToPoint( "e5" ) ), is( true ) ); //$NON-NLS-1$
*/

      // LEGAL move
      //
      pnt_str = "a19"; //$NON-NLS-1$
      move = Convert.stringToPoint( pnt_str );
      prisoners.clear();
      type = m_game.classify( new Board( m_game.getBoard() ), Colors.WHITE, move, prisoners );
      assertThat( type, is( MoveTypes.LEGAL ) );
      assertThat( prisoners.size(), is( 0 ) );
      assertThat( m_game.isOver(), is( false ) );

      // OPPONENT_ATARI move
      //
/*
      pnt_str = "b1"; //$NON-NLS-1$
      m_game.placeStones( Colors.WHITE, Convert.stringsToPoints( new String[] { "a1" } ) ); //$NON-NLS-1$
      move = Convert.stringToPoint( pnt_str );
      type = m_game.classify( new Board( m_game.getBoard() ), Colors.BLACK, move, prisoners );
      assertThat( type, is( MoveTypes.OPPONENT_ATARI ) );
      assertThat( prisoners.size(), is( 0 ) );
      assertThat( m_game.isOver(), is( false ) );

      // SELF_ATARI move
      //
      pnt_str = "P11"; //$NON-NLS-1$
      move = Convert.stringToPoint( pnt_str );
      type = m_game.classify( new Board( m_game.getBoard() ), Colors.WHITE, move, prisoners );
      assertThat( type, is( MoveTypes.SELF_ATARI ) );
      assertThat( prisoners.size(), is( 0 ) );
      assertThat( m_game.isOver(), is( false ) );
*/
   }

   @Test
   public void TestMoves()
   {
      assertThat( m_game.isOver(), is( false ) );
      assertThat( m_game.getTurnNumber(), is( 1 ) );
      assertThat( m_game.getPlayer(), is( Colors.BLACK ) );

      assertMove( Colors.BLACK, "p13", MoveTypes.LEGAL, 2 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "e3",  MoveTypes.LEGAL, 3 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "p11", MoveTypes.LEGAL, 4 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "n12", MoveTypes.LEGAL, 5 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "e6",  MoveTypes.LEGAL, 6 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "o12", MoveTypes.LEGAL, 7 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "d5",  MoveTypes.LEGAL, 8 ); //$NON-NLS-1$

      short pnt = Board.INVALID_POINT;
      m_game.play( pnt, null );
      assertThat( m_game.isOver(), is( false ) );
      assertThat( m_game.getTurnNumber(), is( 9 ) );
      assertThat( m_game.getPlayer(), is( Colors.BLACK ) );

      m_game.play( pnt, null );
      assertThat( m_game.isOver(), is( true ) );
      assertThat( m_game.getTurnNumber(), is( 10 ) );
   }

   @Test
   public void testKoViolation()
   {
      String[] pnts = new String[] { "p13", "e3", "p11", "n12", "e6", "o12", "d5" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
      for (String pnt_str : pnts)
         m_game.play( Convert.stringToPoint( pnt_str ), null );

      short pnt = Convert.stringToPoint( "e5" ); //$NON-NLS-1$
      r_exception.expect( GameException.class );
      m_game.play( pnt, null );
   }

   @Test
   public void testKoRuleNone()
   {
      initGame( KoRules.NONE );
      assertMove( Colors.BLACK, "e6", MoveTypes.LEGAL, 2 );                                                        //$NON-NLS-1$
      assertMove( Colors.WHITE, "b5", MoveTypes.LEGAL, 3 );                                                        //$NON-NLS-1$
      assertMove( Colors.BLACK, "d5", MoveTypes.LEGAL, 4 ); // captures WHITE at e5                                //$NON-NLS-1$
      assertMove( Colors.WHITE, "e5", MoveTypes.LEGAL, 5 ); // captures BLACK at d5, legal because of KoRules.NONE //$NON-NLS-1$
      assertMove( Colors.BLACK, "d5", MoveTypes.LEGAL, 6 ); // captures WHITE at e5, legal because of KoRules.NONE //$NON-NLS-1$
      assertMove( Colors.WHITE, "e5", MoveTypes.LEGAL, 7 ); // captures BLACK at d5, legal because of KoRules.NONE //$NON-NLS-1$
   }

   @Test
   public void testKoRuleBasic1()
   {
      initGame( KoRules.BASIC );
      assertMove( Colors.BLACK, "e6", MoveTypes.LEGAL, 2 );                          //$NON-NLS-1$
      assertMove( Colors.WHITE, "b5", MoveTypes.LEGAL, 3 );                          //$NON-NLS-1$
      assertMove( Colors.BLACK, "d5", MoveTypes.LEGAL, 4 ); // captures WHITE at e5  //$NON-NLS-1$
      short pnt = Convert.stringToPoint( "e5" );                                     //$NON-NLS-1$
      r_exception.expect( GameException.class );
      m_game.play( pnt, null );  // cannot capture e5
   }

   @Test
   public void testKoRuleBasic2()
   {
      initGame( KoRules.BASIC );
      assertMove( Colors.BLACK, "e6",  MoveTypes.LEGAL, 2 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "b5",  MoveTypes.LEGAL, 3 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "d5",  MoveTypes.LEGAL, 4 ); // captures WHITE at e5 //$NON-NLS-1$
      assertMove( Colors.WHITE, "f19", MoveTypes.LEGAL, 5 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "g19", MoveTypes.LEGAL, 6 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "e5",  MoveTypes.LEGAL, 7 ); // can recapture BLACK at d5 now because of basic ko //$NON-NLS-1$
      short pnt = Convert.stringToPoint( "d5" ); //$NON-NLS-1$
      r_exception.expect( GameException.class );
      m_game.play( pnt, null ); // cannot capture d5
   }

   @Test
   public void testKoRuleBasic3()
   {
      initGame( KoRules.BASIC );

      // the upper-left configuration used in the following test comes
      // from the example at http://senseis.xmp.net/?Superko
      //
      assertMove( Colors.BLACK, "t15", MoveTypes.LEGAL, 2 ); // just to get it to be WHITE's turn //$NON-NLS-1$
      assertMove( Colors.WHITE, "c19", MoveTypes.LEGAL, 3 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "a19", MoveTypes.LEGAL, 4 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "b19", MoveTypes.LEGAL, 5 ); // can recapture at b19 because of BASIC ko rule //$NON-NLS-1$
   }

   @Test
   public void testKoRulePositional1()
   {
      initGame( KoRules.POSITIONAL );
      assertMove( Colors.BLACK, "e6", MoveTypes.LEGAL, 2 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "b5", MoveTypes.LEGAL, 3 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "d5", MoveTypes.LEGAL, 4 ); // captures WHITE at e5 //$NON-NLS-1$
      short pnt = Convert.stringToPoint( "e5" ); //$NON-NLS-1$
      r_exception.expect( GameException.class );
      m_game.play( pnt, null );  // cannot capture e5
   }

   @Test
   public void testKoRulePositional2()
   {
      initGame( KoRules.POSITIONAL );
      assertMove( Colors.BLACK, "e6",  MoveTypes.LEGAL, 2 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "b5",  MoveTypes.LEGAL, 3 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "d5",  MoveTypes.LEGAL, 4 ); // captures WHITE at e5 //$NON-NLS-1$
      assertMove( Colors.WHITE, "f19", MoveTypes.LEGAL, 5 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "g19", MoveTypes.LEGAL, 6 ); //$NON-NLS-1$
      assertMove( Colors.WHITE, "e5",  MoveTypes.LEGAL, 7 ); // can recapture BLACK at d5 now because of pos. ko //$NON-NLS-1$
      short pnt = Convert.stringToPoint( "d5" ); //$NON-NLS-1$
      r_exception.expect( GameException.class );
      m_game.play( pnt, null ); // cannot capture d5
   }

   @Test
   public void testKoRulePositional3()
   {
      initGame( KoRules.POSITIONAL );

      // the upper-left configuration used in the following test comes
      // from the example at http://senseis.xmp.net/?Superko
      //
      assertMove( Colors.BLACK, "t15", MoveTypes.LEGAL, 2 ); // just to get it to be WHITE's turn //$NON-NLS-1$
      assertMove( Colors.WHITE, "c19", MoveTypes.LEGAL, 3 ); //$NON-NLS-1$
      assertMove( Colors.BLACK, "a19", MoveTypes.LEGAL, 4 ); //$NON-NLS-1$
      short pnt = Convert.stringToPoint( "b19" ); //$NON-NLS-1$
      r_exception.expect( GameException.class );
      m_game.play( pnt, null ); // cannot recapture at b19
   }

   @Test
   public void testSignature()
   {
      // The moves for this test were taken from src-tests/data/yi203.sgf
      //
      // The points for the expected signature points are: s17, b14, r6, p5, o9, k8
      //
      Board.setSize( 19 );
      Signature exp_sig = new Signature();
      Game game = new Game( KoRules.POSITIONAL, false, Game.DEFAULT_KOMI );
      assertThat( game.getSignature(), is( exp_sig ) );

      String[] moves = new String[]
      {
         "q16", "d17", "q4", "d3", "d5", "c5", "d4", "c4", "e3", "c6", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$
         "q10", "d2", "k4", "o17", "m17", "r17", "r16", "q17", "p16"   //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$
      };
      for (String text : moves)
         game.play( Convert.stringToPoint( text ), null );
      assertThat( game.getSignature(), is( exp_sig ) );
      short pnt = Convert.stringToPoint( "s17" ); //$NON-NLS-1$
      game.play( pnt, null );
      exp_sig.addPoint( pnt );
      assertThat( game.getSignature(), is( exp_sig ) );

      moves = new String[]
      {
         "o16", "o18", "d15", "e15", "c17", "e16", "c18", "e14", "d16", "d14" //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$
      };
      for (String text : moves)
         game.play( Convert.stringToPoint( text ), null );
      assertThat( game.getSignature(), is( exp_sig ) );
      pnt = Convert.stringToPoint( "b14" ); //$NON-NLS-1$
      game.play( pnt, null );
      exp_sig.addPoint( pnt );
      assertThat( game.getSignature(), is( exp_sig ) );
   }
   
   private void initGame( KoRules rules )
   {
      // The tests in this class are based on the board defined by
      // src-tests/data/test_board.sgf
      //

      Board.setSize( 19 );
      m_game = new Game( rules, false, Game.DEFAULT_KOMI );

      Set<Short> white_stones = Convert.stringsToPoints( WHITE_PLACED_STONES );
      m_game.placeStones( Colors.WHITE, white_stones );

      Set<Short> black_stones = Convert.stringsToPoints( BLACK_PLACED_STONES );
      m_game.placeStones( Colors.BLACK, black_stones );

      assertThat( m_game.isOver(), is( false ) );
      assertThat( m_game.getTurnNumber(), is( 1 ) );
      assertThat( m_game.getPlayer(), is( Colors.BLACK ) );
      assertThat( m_game.getLastPoint(), is( Board.INVALID_POINT ) );
      assertThat( m_game.getLastMoveType(), is( MoveTypes.OUT_OF_TURN ) );

      DkAssertUtils.assertBoardPoints( m_game.getBoard(), black_stones, white_stones );
   }

   private void assertMove( Colors color, String pnt_str, MoveTypes exp_type, int exp_turn )
   {
      short pnt = Convert.stringToPoint( pnt_str );
      String msg = color.toString() + " " + pnt_str; //$NON-NLS-1$
      assertThat( msg, m_game.play( pnt, null ), is( exp_type ) );
      assertThat( msg, m_game.getLastMoveType(), is( exp_type ) );
      if (pnt != Board.INVALID_POINT)
         assertThat( msg, m_game.getLastPoint(), is( pnt ) );
      assertThat( msg, m_game.isOver(), is( false ) );
      assertThat( msg, m_game.getTurnNumber(), is( exp_turn ) );
      assertThat( msg, m_game.getPlayer(), is( Game.otherPlayer( color ) ) );
      assertThat( msg, m_game.getBoard().getColor( pnt ), is( color ) );
   }

   private static final String[] BLACK_PLACED_STONES = new String[]
   {
      "e4", "f5", "m11", "m12", "n10", "n13", "o9", "o13", "p10", "q11", "q13", "r12", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$ //$NON-NLS-12$
      "a16", "b16", "c16", "c17", "b18", "c18", "d18", "d19"                           //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$
   };
   private static final String[] WHITE_PLACED_STONES = new String[]
   {
      "c5", "d4", "d6", "e5", "n11", "n12", "o10", "o11", "o12", "p12", "q12", "s19",  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$ //$NON-NLS-12$
      "t18", "a15", "b15", "c15", "d15", "d16", "a17", "d17", "e17", "a18", "e18",     //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$
      "b19", "e19"                                                                     //$NON-NLS-1$ //$NON-NLS-2$
   };

   private Game m_game;
}
