//
// LayersTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.game;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Test;

import com.dakengo.libs.application.Constants;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.Layers;
import com.dakengo.tests.utils.TestRunnable;

public class LayerTests
{
   @Test
   public void testExpandRect()
   {
      Board.setSize( 19 );

      Set<Short> actual = new HashSet<Short>();
      int count = Layers.expandRect( ImmutablePair.of( (short) 1, (short) 1 ), actual );
      assertThat( count, is( 1 ) );
      assertThat( actual.contains( (short) 1 ), is( true ) );

      actual.clear();
      Set<Short> expected = Convert.stringsToPoints( LAYER3 );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER3[0] ), Convert.stringToPoint( LAYER3[LAYER3.length - 1] ) ), actual );
      assertThat( count, is( LAYER3.length ) );
      assertThat( actual, is( expected ) );

      actual.clear();
      String[] copy = ArrayUtils.clone( LAYER3 );
      ArrayUtils.reverse( copy );
      expected = Convert.stringsToPoints( copy );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER3[LAYER3.length - 1] ), Convert.stringToPoint( LAYER3[0] ) ), actual );
      assertThat( count, is( LAYER3.length ) );
      assertThat( actual, is( expected ) );

      actual.clear();
      expected = Convert.stringsToPoints( LAYER2 );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER2[0] ), Convert.stringToPoint( LAYER2[LAYER2.length - 1] ) ), actual );
      assertThat( count, is( LAYER2.length ) );
      assertThat( actual, is( expected ) );

      actual.clear();
      copy = ArrayUtils.clone( LAYER2 );
      ArrayUtils.reverse( copy );
      expected = Convert.stringsToPoints( copy );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER2[LAYER2.length - 1] ), Convert.stringToPoint( LAYER2[0] ) ), actual );
      assertThat( count, is( LAYER2.length ) );
      assertThat( actual, is( expected ) );

      actual.clear();
      expected = Convert.stringsToPoints( LAYER4 );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER4[0] ), Convert.stringToPoint( LAYER4[LAYER4.length - 1] ) ), actual );
      assertThat( count, is( LAYER4.length ) );
      assertThat( actual, is( expected ) );

      actual.clear();
      copy = ArrayUtils.clone( LAYER4 );
      ArrayUtils.reverse( copy );
      expected = Convert.stringsToPoints( copy );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER4[LAYER4.length - 1] ), Convert.stringToPoint( LAYER4[0] ) ), actual );
      assertThat( count, is( LAYER4.length ) );
      assertThat( actual, is( expected ) );

      actual.clear();
      expected = Convert.stringsToPoints( LAYER1 );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER1[0] ), Convert.stringToPoint( LAYER1[LAYER1.length - 1] ) ), actual );
      assertThat( count, is( LAYER1.length ) );
      assertThat( actual, is( expected ) );

      actual.clear();
      copy = ArrayUtils.clone( LAYER1 );
      ArrayUtils.reverse( copy );
      expected = Convert.stringsToPoints( copy );
      count = Layers.expandRect( ImmutablePair.of( Convert.stringToPoint( LAYER1[LAYER1.length - 1] ), Convert.stringToPoint( LAYER1[0] ) ), actual );
      assertThat( count, is( LAYER1.length ) );
      assertThat( actual, is( expected ) );
   }

   @Test
   public void testReadPoints() throws IOException
   {
      String text = "[{size: 5, layers: [{names: [foo], points: [a1, b2, c3]}]}]";
      Map<String,Set<Short>> layers = Layers.load( new StringReader( text ), 7 );
      assertTrue( layers.isEmpty() );
      layers = Layers.load( new StringReader( text ), 5 );
      assertThat( layers.size(), is( 1 ) );
      Set<Short> pnts = layers.get( "foo" );
      assertThat( pnts.size(), is( 3 ) );
      assertThat( pnts, is( Convert.stringsToPoints( "c3", "b2", "a1" ) ) );

      final String bad_text = "[{layers: [{names: [foo], points: [a1, b2, c3]}], size: 5}]";
      assertThrows(
            "size property after layers property", //$NON-NLS-1$
            new TestRunnable() { @Override public void runTest() throws Exception { Layers.load( new StringReader( bad_text ), 5 ); } },
            ParseException.class, null );

      text = "[{size: 5, layers: [{names: [foo], points: [a1, b2, c3]}]}]";
      layers = Layers.load( new StringReader( text ), 7 );
      assertTrue( layers.isEmpty() );
      layers = Layers.load( new StringReader( text ), 5 );
      assertThat( layers.size(), is( 1 ) );
      pnts = layers.get( "foo" );
      assertThat( pnts, is( Convert.stringsToPoints( "c3", "b2", "a1" ) ) );

      text = "[{size: 5, layers: [{names: [foo], points: []}]}]";
      layers = Layers.load( new StringReader( text ), 5 );
      assertTrue( layers.isEmpty() );
   }

   @Test
   public void testReadRectangle() throws IOException
   {
      String text = "[{size: 5, layers: [{names: [foo], rectangle: [a1, b2]}]}]";
      Map<String,Set<Short>> layers = Layers.load( new StringReader( text ), 5 );
      Set<Short> pnts = layers.get( "foo" );
      assertThat( pnts, is( Convert.stringsToPoints( "a2", "a1", "b1", "b2" ) ) );
   }

   @Test
   public void testReadIntersect() throws IOException
   {
      String text = "[{size: 5, layers: [{names: [foo], points: [a1, b2]}, {names: [bar], points: [b2, c3]}, {names: [abc], intersect: [foo]}, {names: [def], intersect: [foo, bar]}, {names: [ghi], intersect: [abc]}]}]";
      Map<String,Set<Short>> layers = Layers.load( new StringReader( text ), 5 );
      assertThat( layers.size(), is( 5 ) );
      Set<Short> pnts = layers.get( "abc" );
      assertThat( pnts, is( Convert.stringsToPoints( "a1", "b2" ) ) );
      pnts = layers.get( "def" );
      assertThat( pnts, is( Convert.stringsToPoints( "b2" ) ) );
      pnts = layers.get( "ghi" );
      assertThat( pnts, is( Convert.stringsToPoints( "a1", "b2" ) ) );

      text = "[{size: 5, layers: [{names: [foo], points: [a1, b2]}, {names: [xyz], intersect: [foo, a1]}]}]";
      layers = Layers.load( new StringReader( text ), 5 );
      pnts = layers.get( "xyz" );
      assertThat( pnts, is( Convert.stringsToPoints( "a1" ) ) );
   }

   @Test
   public void testReadUnion() throws IOException
   {
      String text = "[{size: 5, layers: [{names: [foo], points: [a1, b2]}, {names: [bar], points: [b2, c3]}, {names: [abc], union: [foo]}, {names: [def], union: [foo, bar]}, {names: [ghi], union: [abc]}]}]";
      Map<String,Set<Short>> layers = Layers.load( new StringReader( text ), 5 );
      assertThat( layers.size(), is( 5 ) );
      Set<Short> pnts = layers.get( "abc" );
      assertThat( pnts, is( Convert.stringsToPoints( "a1", "b2" ) ) );
      pnts = layers.get( "def" );
      assertThat( pnts, is( Convert.stringsToPoints( "b2", "c3", "a1" ) ) );
      pnts = layers.get( "ghi" );
      assertThat( pnts, is( Convert.stringsToPoints( "a1", "b2" ) ) );

      text = "[{size: 5, layers: [{names: [foo], union: [a1, b2]}]}]";
      layers = Layers.load( new StringReader( text ), 5 );
      pnts = layers.get( "foo" );
      assertThat( pnts, is( Convert.stringsToPoints( "a1", "b2" ) ) );
   }

   @Test
   public void testBadLayers() throws IOException
   {
      String[] bad_layers = new String[]
      {
         "[{layers: [{names: [foo], points: [a1, b2, c3]}], size: 5}]",
         "[{size: 4, layers: [{names: [foo], points: [a1]}]}]"
      };
      for (String text : bad_layers)
      {
         final StringReader input = new StringReader( text );
         assertThrows(
            String.format( "'%s'", text ),
            new TestRunnable() { @Override public void runTest() throws Exception { Layers.load( input, 4 ); } },
            IllegalArgumentException.class, null );
      }

      bad_layers = new String[]
      {
         "{x: y}", "[{x: y}]", "[{layers: [x: y]}]",
         "[{size: a, layers: [{names: [foo], points: [a1]}]}]",
         "[{size: a, layers: [{names: [foo], points: [a1]}, {names: [foo], points: [b2]}]}]",
         "[{size: 5, layers: [{points: [a1]}]}]",
         "[{size: 5, layers: [{names: [], points: [a1]}]}]", "[{size: 5, layers: [{xyz: [foo], points: [a1]}]}]",
         "[{size: 5, layers: [{names: [foo], points: [abc]}]}]",
         "[{size: 5, layers: [{names: [foo], rectangle: [ab, cd]}]}]",
         "[{size: 5, layers: [{names: [foo], rectangle: [a1]}]}]",
         "[{size: 5, layers: [{names: [foo], rectangle: [a1, b2, c3]}]}]",
         "[{size: 5, layers: [{names: [foo], intersect: []}]}]",
         "[{size: 5, layers: [{names: [foo], intersect: [xyz]}]}]",
         "[{size: a, layers: [{names: [foo], points: [a1]}, {names: [bar], points: [b2]}, {names: [xyz], intersect: [foo, bar]}]}]",
         "[{size: 5, layers: [{names: [foo], union: []}]}]",
         "[{size: 5, layers: [{names: [foo], union: [xyz]}]}]"
      };
      for (String text : bad_layers)
      {
         final StringReader input = new StringReader( text );
         assertThrows(
            String.format( "'%s'", text ),
            new TestRunnable() { @Override public void runTest() throws Exception { Layers.load( input, 5 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testLoadLayerFile() throws IOException
   {
      String path = System.getProperty( Constants.APP_HOME_PROP );
      assertNotNull( path );
      File home_dir = new File( path );
      File etc_dir = new File( home_dir, "etc" );
      File layers_file = new File( etc_dir, "layers.json" );

      try (BufferedReader br = new BufferedReader( new FileReader( layers_file ) ))
      {
         Map<String,Set<Short>> layers = Layers.load( br, 13 );
         Board.setSize( 13 );
         assertThat( layers.size(), is( 3 ) );
         String[] names = new String[] { "tengen", "center", "ten-ten" };
         for (String name : names)
         {
            Set<Short> pnts = layers.get( name );
            assertThat( pnts, is( Convert.stringsToPoints( "g7" ) ) );
         }
      }

      try (BufferedReader br = new BufferedReader( new FileReader( layers_file ) ))
      {
         Map<String,Set<Short>> layers = Layers.load( br, 19 );
         Board.setSize( 19 );
         assertThat( layers.size(), is( 67 ) );
         Set<Short> pnts = layers.get( "four-three" );
         assertThat( pnts, is( Convert.stringsToPoints( "d3", "d17", "q3", "q17" ) ) );
         pnts = layers.get( "top-third-line" );
         assertThat( pnts, is( Convert.stringsToPoints( "c17", "d17", "e17", "f17", "g17", "h17", "j17", "k17", "l17", "m17", "n17", "o17", "p17", "q17", "r17" ) ) );
         pnts = layers.get( "north-west" );
         assertThat( pnts, is( Convert.stringsToPoints( "a16", "a17", "a18", "a19", "b16", "b17", "b18", "b19", "c16", "c17", "c18", "c19", "d16", "d17", "d18", "d19" ) ) );
      }

      try (BufferedReader br = new BufferedReader( new FileReader( layers_file ) ))
      {
         Map<String,Set<Short>> layers = Layers.load( br, 5 );
         assertTrue( layers.isEmpty() );
      }
   }

   // The following are all based on a 19x19 board.
   //
   private static final String[] LAYER1 = new String[] { "a7", "b7", "a6", "b6", "a5", "b5", "a4",
                                                         "b4", "a3", "b3", "a2", "b2", "a1", "b1" };               // (a1 b7)

   private static final String[] LAYER2 = new String[] { "e3", "e5", "g3", "e4", "g4", "f4", "f5", "f3", "g5" };   // (e3 g5)

   private static final String[] LAYER3 = new String[] { "b2", "b3", "b4" };                                       // (b2 b4)

   private static final String[] LAYER4 = new String[] { "c3", "d3", "e3" };                                       // (c3 e3)
}
