//
// MoveFilterTests.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.player;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.PrisonerFinder;
import com.dakengo.libs.player.BaseLegalMoveFilter;

public class MoveFilterTests
{
   class TestFilter extends BaseLegalMoveFilter
   {
      @Override
      protected void sort( Colors player, List<ImmutablePair<Short,Board>> candidates )
      {
         Collections.sort( candidates, new Comparator<ImmutablePair<Short,Board>>()
         {
            @Override
            public int compare( ImmutablePair<Short,Board> lhs, ImmutablePair<Short,Board> rhs )
            {
               return lhs.getLeft() - rhs.getLeft();
            }
         } );
      }
   }

   @Before
   public void setUp()
   {
      Board.setSize( SIZE );
      m_board = new Board();
      m_board.setColor( BLACK_POINTS, Colors.BLACK );
      m_board.setColor( WHITE_POINTS, Colors.WHITE );
   }

   @Test
   public void testBaseFilter()
   {
      TestFilter filter = new TestFilter();
      List<ImmutablePair<Short,Board>> candidates = filter.find( Colors.BLACK, m_board, Board.getTotal() );
      int[] exp_pnts = new int[]
      {
         1, 2, 3, 4, 5, 8, 9, 14, 15, 19, 20, 21, 22, 23
      };
      assertThat( candidates.size(), is( exp_pnts.length ) );
      for (int i = 0; i < exp_pnts.length; i++)
         assertCandidate( Colors.BLACK, (short) exp_pnts[i], candidates.get( i ) );

      candidates = filter.find( Colors.BLACK, m_board, 2 );
      assertThat( candidates.size(), is( 2 ) );
      assertCandidate( Colors.BLACK, (short) 1, candidates.get( 0 ) );
      assertCandidate( Colors.BLACK, (short) 2, candidates.get( 1 ) );

      exp_pnts = new int[]
      {
         1, 2, 3, 4, 5, 8, 9, 11, 14, 15, 19, 20, 21, 22, 23
      };
      candidates = filter.find( Colors.WHITE, m_board, exp_pnts.length );
      assertThat( candidates.size(), is( exp_pnts.length ) );
      for (int i = 0; i < exp_pnts.length; i++)
         assertCandidate( Colors.WHITE, (short) exp_pnts[i], candidates.get( i ) );
   }

   private void assertCandidate( Colors color, short exp_pnt, ImmutablePair<Short,Board> actual )
   {
      assertThat( actual.getLeft(), is( exp_pnt ) );
      m_finder.find( m_board, exp_pnt, color );
      Board new_board = new Board( m_board );
      new_board.setColor( exp_pnt, color );
      new_board.setColor( m_finder.getPrisoners(), Colors.EMPTY );
      assertThat( "point: " + Convert.pointToString( exp_pnt ), actual.getRight(), is( new_board ) );
   }

   private static final int SIZE = 5;

   // The 5x5 test board:
   //
   //  . . . . X
   //  . . O X .
   //  . O X O .
   //  . O . O .
   //  X . O . .
   //
   private static final short[] BLACK_POINTS = new short[] { 0, 12, 18, 24 };
   private static final short[] WHITE_POINTS = new short[] { 6, 7, 10, 13, 16, 17 };

   private Board          m_board;
   private PrisonerFinder m_finder = new PrisonerFinder(); 
}
