//
// TreeTests.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.player;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import org.junit.Test;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.player.BaseTreeNode;
import com.dakengo.libs.player.TreeCursor;
import com.dakengo.libs.player.TreePolicy;
import com.dakengo.libs.player.TreeStore;

public class TreeTests
{
   public class TestNode extends BaseTreeNode
   {
      public static final int PROPERTY_COUNT = 1;

      public TestNode( TreeStore store )
      {
         super( store );
      }

      public TestNode( TestNode other )
      {
         super( other );
      }

      public int getTestProperty()
      {
         return getProperty( BaseTreeNode.FIRST_SUBCLASS_PROP_INDEX );
      }

      public void setTestProperty( int value )
      {
         setProperty( BaseTreeNode.FIRST_SUBCLASS_PROP_INDEX, value );
      }

      @Override
      public TestNode cloneMe()
      {
         TestNode copy = new TestNode( this );
         return copy;
      }
   }

   public class TestPolicy implements TreePolicy<TestNode>
   {
      public TestPolicy()
      {
      }

      public void selectChild( int index )
      {
         assertThat( index, is( greaterThanOrEqualTo( 0 ) ) );
         m_child_index = index;
      }

      @Override
      public boolean shouldExpand( TreeCursor<TestNode> cursor )
      {
         return true;
      }

      @Override
      public int selectChild( TreeCursor<TestNode> cursor )
      {
         return m_child_index;
      }

      private int m_child_index;
   }

   @Test
   public void testStore()
   {
      TreeStore store = new TreeStore( 2, FANOUT );
      assertThat( store.getPropertyCount(), is( BaseTreeNode.FIRST_SUBCLASS_PROP_INDEX + 2 ) );
      assertThat( store.getFanOut(), is( FANOUT ) );
      assertThat( store.getData().length, is( 768 ) );

      int link_block = store.allocLinks();
      assertThat( link_block, is( 5 ) );
      store.allocNodes( 0, link_block );
      assertThat( store.getData()[link_block], is( 8 ) );
      assertThat( store.getData()[link_block + 1], is( 13 ) );
      assertThat( store.getData()[link_block + 2], is( 18 ) );
      assertThat( store.getData()[8], is( 0 ) );
      assertThat( store.getData()[13], is( 0 ) );
      assertThat( store.getData()[18], is( 0 ) );
      assertThat( store.getData().length, is( 768 ) );

      assertThat( store.allocLinks(), is( 23 ) );
      assertThat( store.allocLinks(), is( 26 ) );
      link_block = 26;
      assertThat( store.getData().length, is( 768 ) );

      store.allocNodes( 1, link_block );
      assertThat( store.getData()[link_block], is( 29 ) );
      assertThat( store.getData()[link_block + 1], is( 34 ) );
      assertThat( store.getData()[link_block + 2], is( 39 ) );
      assertThat( store.getData().length, is( 768 ) );
      assertThat( store.getData()[29], is( 1 ) );
      assertThat( store.getData()[34], is( 1 ) );
      assertThat( store.getData()[39], is( 1 ) );

      for (int i = 44; i < 767; i += FANOUT)
         link_block = store.allocLinks();
      assertThat( store.getData().length, is( 768 ) );
      store.allocNodes( 2, link_block );
      assertThat( store.getData()[link_block], is( 767 ) );
      assertThat( store.getData()[link_block + 1], is( 772 ) ); 
      assertThat( store.getData()[link_block + 2], is( 777 ) );
      assertThat( store.getData().length, is( 1536 ) );
      assertThat( store.getData()[0], is( -1 ) );
      assertThat( store.getData()[8], is( 0 ) );
      assertThat( store.getData()[13], is( 0 ) );
      assertThat( store.getData()[18], is( 0 ) );
      assertThat( store.getData()[29], is( 1 ) );
      assertThat( store.getData()[34], is( 1 ) );
      assertThat( store.getData()[39], is( 1 ) );
      assertThat( store.getData()[767], is( 2 ) );
      assertThat( store.getData()[772], is( 2 ) );
      assertThat( store.getData()[777], is( 2 ) );

      store.reset();
      link_block = store.allocLinks();
      assertThat( link_block, is( 5 ) );
      store.allocNodes( 0, link_block );
      assertThat( store.getData()[link_block], is( 8 ) );
      assertThat( store.getData()[link_block + 1], is( 13 ) );
      assertThat( store.getData()[link_block + 2], is( 18 ) );
      assertThat( store.getData()[8], is( 0 ) );
      assertThat( store.getData()[13], is( 0 ) );
      assertThat( store.getData()[18], is( 0 ) );
      assertThat( store.getData().length, is( 1536 ) );
   }

   @Test
   public void testNodes()
   {
      TreeStore store = new TreeStore( TestNode.PROPERTY_COUNT, FANOUT );
      TestNode root = new TestNode( store );
      TreeCursor<TestNode> cursor = new TreeCursor<TestNode>( root );
      assertNode( cursor, true, true, Board.INVALID_POINT, Colors.BLACK, -1 );
      assertFalse( cursor.ascend() );
      TestPolicy policy = new TestPolicy();
      assertFalse( cursor.descend( policy ) );

      root.expand();
      assertNode( cursor, true, false, Board.INVALID_POINT, Colors.BLACK, -1 );
      assertTrue( cursor.descend( policy ) );
      cursor.getNode().setTestProperty( 123 );
      assertNode( cursor, false, true, Board.INVALID_POINT, Colors.WHITE, 123 );
      assertFalse( cursor.descend( policy ) );
      assertTrue( cursor.ascend() );
      policy.selectChild( 1 );
      assertTrue( cursor.descend( policy ) );
      assertNode( cursor, false, true, Board.INVALID_POINT, Colors.WHITE, -1 );
      assertFalse( cursor.descend( policy ) );

      TreeCursor<TestNode> cursor2 = new TreeCursor<TestNode>( cursor );
      assertNode( cursor2, false, true, Board.INVALID_POINT, Colors.WHITE, -1 );
      assertTrue( cursor2.ascend() );
      assertNode( cursor2, true, false, Board.INVALID_POINT, Colors.BLACK, -1 );
      assertNode( cursor, false, true, Board.INVALID_POINT, Colors.WHITE, -1 );
   }

   private void assertNode( TreeCursor<TestNode> cursor, boolean is_root, boolean is_leaf, short point, Colors player, int value )
   {
      if (is_root)
         assertTrue( cursor.isRoot() );
      else
         assertFalse( cursor.isRoot() );
      if (is_leaf)
         assertTrue( cursor.isLeaf() );
      else
         assertFalse( cursor.isLeaf() );
      TestNode node = cursor.getNode();
      assertThat( node.getPoint(), is( point ) );
      assertThat( node.getPlayer(), is( player ) );
      assertThat( node.getTestProperty(), is( value ) );
   }

   private static final int FANOUT = 3;
}
