//
// GamePropertiesTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.record;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.math.BigDecimal;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Test;

import com.dakengo.libs.base.Constants;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.record.GameProperties;
import com.dakengo.libs.record.Handicaps;
import com.dakengo.libs.record.PropertyException;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.tests.utils.TestPropVisitor;

/**
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class GamePropertiesTests
{
   public static void assertDefaults( GameProperties props )
   {
      for (int prop_id = 0; prop_id < PropertyID.COMMENT; prop_id++)
         assertThat( "property unexpectedly set: " + PropertyID.toString( prop_id ), props.has( prop_id ), is( false ) ); //$NON-NLS-1$

      assertThat( props.getAnnotatorName().isEmpty(), is( true ) );
      assertThat( props.getApplicationName().isEmpty(), is( true ) );
      assertThat( props.getApplicationVersion().isEmpty(), is( true ) );
      assertThat( props.getBackgroundInfo().isEmpty(), is( true ) );
      assertThat( props.getCharset(), is( Constants.UTF_8 ) );
      assertThat( props.getCopyright().isEmpty(), is( true ) );
      assertThat( props.getDate().isEmpty(), is( true ) );
      assertThat( props.getEventName().isEmpty(), is( true ) );
      assertThat( props.getGameName().isEmpty(), is( true ) );
      assertThat( props.getHandicap(), is( GameProperties.DEFAULT_HANDICAP ) );
      assertThat( props.getKomi(), is( Game.DEFAULT_KOMI ) );
      assertThat( props.getOpeningName().isEmpty(), is( true ) );
      assertThat( props.getOvertimeSystem().isEmpty(), is( true ) );
      assertThat( props.getPlaceName().isEmpty(), is( true ) );
      assertThat( props.getPlayerName( Colors.BLACK ).isEmpty(), is( true ) );
      assertThat( props.getPlayerName( Colors.WHITE ).isEmpty(), is( true ) );
      assertThat( props.getPlayerRank( Colors.BLACK ).isEmpty(), is( true ) );
      assertThat( props.getPlayerRank( Colors.WHITE ).isEmpty(), is( true ) );
      assertThat( props.getRenderingStyle(), is( GameProperties.DEFAULT_RENDERING_STYLE ) );
      assertThat( props.getResult(), is( GameProperties.DEFAULT_RESULT ) );
      assertThat( props.getRound(), is( GameProperties.DEFAULT_ROUND ) );
      assertThat( props.getRuleSystem().isEmpty(), is( true ) );
      assertThat( props.getSize(), is( Board.DEFAULT_SIZE ) );
      assertThat( props.getSourceName().isEmpty(), is( true ) );
      assertThat( props.getTeamName( Colors.BLACK ).isEmpty(), is( true ) );
      assertThat( props.getTeamName( Colors.WHITE ).isEmpty(), is( true ) );
      assertThat( props.getTimeLimit(), is( GameProperties.DEFAULT_TIME_LIMIT ) );
      assertThat( props.getUserName().isEmpty(), is( true ) );
   }

   @Test
   public void testValidProperties()
   {
      for (int prop_id = 0; prop_id < PropertyID.COMMENT; prop_id++)
         assertThat( "standard property: " + PropertyID.toString( prop_id ), GameProperties.isStandard( prop_id ), is( true ) ); //$NON-NLS-1$

      GameProperties props = new GameProperties();
      assertDefaults( props );

      props.setApplicationName( "testing" ); //$NON-NLS-1$
      props.setKomi( "-5.5" );   // make sure we can set reverse komi //$NON-NLS-1$
      props.setOvertimeSystem( "25 moves / 10 mins" ); //$NON-NLS-1$
      props.setRuleSystem( "NZ" ); //$NON-NLS-1$
      props.setSize( 13 );
      props.setTimeLimit( 300 );

      assertThat( props.getApplicationName(), is( "testing" ) ); //$NON-NLS-1$
      assertThat( "has APP_NAME", props.has( PropertyID.APP_NAME ), is( true ) ); //$NON-NLS-1$
      assertThat( props.getKomi(), is( new BigDecimal( "-5.5" ) ) ); //$NON-NLS-1$
      assertThat( "has KOMI", props.has( PropertyID.KOMI ), is( true ) ); //$NON-NLS-1$
      assertThat( props.getOvertimeSystem(), is( "25 moves / 10 mins" ) ); //$NON-NLS-1$
      assertThat( "has OVERTIME_SYSTEM", props.has( PropertyID.OVERTIME_SYSTEM ), is( true ) ); //$NON-NLS-1$
      assertThat( props.getRuleSystem(), is( "NZ" ) ); //$NON-NLS-1$
      assertThat( "has RULE_SYSTEM", props.has( PropertyID.RULE_SYSTEM ), is( true ) ); //$NON-NLS-1$
      assertThat( props.getSize(), is( 13 ) );
      assertThat( "has SIZE", props.has( PropertyID.SIZE ), is( true ) ); //$NON-NLS-1$
      assertThat( props.getTimeLimit(), is( 300 ) );
      assertThat( "has TIME_LIMIT", props.has( PropertyID.TIME_LIMIT ), is( true ) ); //$NON-NLS-1$

      props.setKomi( new BigDecimal( "8.5" ) ); //$NON-NLS-1$
      assertThat( props.getKomi(), is( new BigDecimal( "8.5" ) ) ); //$NON-NLS-1$

      TestPropVisitor visitor = new TestPropVisitor( false );
      props.visit( visitor );
      visitor.assertCounts( 0, 6 );
      visitor.assertValue( PropertyID.APP_NAME, 0, "testing" ); //$NON-NLS-1$
      visitor.assertValue( PropertyID.KOMI, 0, "8.5" ); //$NON-NLS-1$
      visitor.assertValue( PropertyID.OVERTIME_SYSTEM, 0, "25 moves / 10 mins" ); //$NON-NLS-1$
      visitor.assertValue( PropertyID.RULE_SYSTEM, 0, "NZ" ); //$NON-NLS-1$
      visitor.assertValue( PropertyID.SIZE, 0, "13" ); //$NON-NLS-1$
      visitor.assertValue( PropertyID.TIME_LIMIT, 0, "300" ); //$NON-NLS-1$

      GameProperties copy = new GameProperties( props );
      assertThat( copy.hashCode(), is( props.hashCode() ) );
      assertThat( copy, is( props ) );

      // try setting a value of valid type for each of the other properties
      //
      props.setAnnotatorName( "" ); //$NON-NLS-1$
      props.setApplicationVersion( "" ); //$NON-NLS-1$
      props.setBackgroundInfo( "" ); //$NON-NLS-1$
      props.setCharset( "" ); //$NON-NLS-1$
      props.setCopyright( "" ); //$NON-NLS-1$
      props.setDate( "" ); //$NON-NLS-1$
      props.setEventName( "" ); //$NON-NLS-1$
      props.setGameName( "" ); //$NON-NLS-1$
      props.setHandicap( ImmutablePair.of( Handicaps.FIXED, 2 ) );
      props.setOpeningName( "" ); //$NON-NLS-1$
      props.setPlaceName( "" ); //$NON-NLS-1$
      props.setPlayerName( Colors.BLACK, "" ); //$NON-NLS-1$
      props.setPlayerRank( Colors.WHITE, "" ); //$NON-NLS-1$
      props.setRenderingStyle( GameProperties.DEFAULT_RENDERING_STYLE );
      props.setResult( GameProperties.DEFAULT_RESULT );
      props.setRound( GameProperties.DEFAULT_ROUND );
      props.setRuleSystem( "" ); //$NON-NLS-1$
      props.setSize( 19 );
      props.setSourceName( "" ); //$NON-NLS-1$
      props.setTeamName( Colors.BLACK, "" ); //$NON-NLS-1$
      props.setTimeLimit( 1 );
      props.setUserName( "" ); //$NON-NLS-1$

      for (int prop_id = 0; prop_id < PropertyID.COMMENT; prop_id++)
         assertThat( "unset property: " + PropertyID.toString( prop_id ), props.has( prop_id ), is( true ) ); //$NON-NLS-1$

      props.clear();
      assertDefaults( props );
   }

   @Test
   public void testInvalidProps()
   {
      final GameProperties props = new GameProperties();

      // try setting out of range / illegal values
      //
      assertThrows(
         new Runnable() { @Override public void run() { props.setSize( 1000 ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.setTimeLimit( 0 ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.set( 1, ImmutablePair.of( "", "" ) ); } }, //$NON-NLS-1$ //$NON-NLS-2$
         PropertyException.class, null );

      // try setting values of an incorrect type for each supported property
      //
      int[] prop_ids = new int[]
      {
         PropertyID.ANNOTATOR_NAME, PropertyID.APP_NAME, PropertyID.APP_VERSION, PropertyID.BACKGROUND_INFO, PropertyID.CHARSET, PropertyID.COPYRIGHT,
         PropertyID.DATE, PropertyID.EVENT_NAME, PropertyID.GAME_NAME, PropertyID.OPENING_NAME, PropertyID.OVERTIME_SYSTEM, PropertyID.PLACE_NAME,
         PropertyID.RULE_SYSTEM, PropertyID.SOURCE_NAME, PropertyID.USER_NAME
      };
      int prop_count = prop_ids.length;
      for (int prop : prop_ids)
      {
         final int tmp = prop;
         assertThrows(
            new Runnable() { @Override public void run() { props.set( tmp, 1000 ); } },
            PropertyException.class, null );
      }
      prop_ids = new int[]
      {
         PropertyID.HANDICAP, PropertyID.KOMI, PropertyID.PLAYER_NAME, PropertyID.PLAYER_RANK, PropertyID.RENDERING_STYLE, PropertyID.RESULT,
         PropertyID.ROUND, PropertyID.SIZE, PropertyID.TEAM_NAME, PropertyID.TIME_LIMIT
      };
      prop_count += prop_ids.length;
      for (int prop : prop_ids)
      {
         final int tmp = prop;
         assertThrows(
            new Runnable() { @Override public void run() { props.set( tmp, 'X' ); } },
            PropertyException.class, null );
      }
      assertThat( "count of tested game properties", prop_count, is( PropertyID.COMMENT ) ); // COMMENT is first position property ID //$NON-NLS-1$

      // try setting position properties on the game properties object
      //
      for (int prop_id = PropertyID.COMMENT; prop_id <= PropertyID.MAX_STD_ID; prop_id++)
      {
         final int tmp = prop_id;
         assertThrows(
            new Runnable() { @Override public void run() { props.set( tmp, 'X' ); } },
            PropertyException.class, null );
      }

      assertDefaults( props );
   }

   @Test
   @SuppressWarnings( "unchecked" )
   public void testNonStdProps()
   {
      GameProperties props = new GameProperties();

      props.set( 1000, ImmutablePair.of( "testprop", "some value" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      assertThat( "non-standard prop 1000", props.has( 1000 ), is( true ) ); //$NON-NLS-1$
      ImmutablePair<String,String> value = (ImmutablePair<String,String>) props.get( 1000 );
      assertThat( value.getLeft(), is( "testprop" ) ); //$NON-NLS-1$
      assertThat( value.getRight(), is( "some value" ) ); //$NON-NLS-1$
   }

   @Test
   public void testEquals()
   {
      GameProperties props1 = new GameProperties();
      props1.setApplicationName( "foo!" ); //$NON-NLS-1$

      GameProperties props2 = new GameProperties();
      props2.setUserName( "bar" ); //$NON-NLS-1$
 
      GameProperties props3 = new GameProperties();

      GameProperties props4 = new GameProperties();
      props4.setUserName( "bar" ); //$NON-NLS-1$

      GameProperties props5 = new GameProperties();
      props5.setUserName( "yuck" ); //$NON-NLS-1$

      assertThat( props1, is( props1 ) );
      assertThat( props1.hashCode(), is( props1.hashCode() ) );
      assertThat( props2, is( props4 ) );
      assertThat( props2.hashCode(), is( props4.hashCode() ) );
      assertThat( props1, is( not( props2 ) ) );
      assertThat( props1, is( not( props3 ) ) );
      assertThat( props2, is( not( props5 ) ) );
   }
}
