//
// PositionPropertiesTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.record;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.math.BigDecimal;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Test;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.record.Connectors;
import com.dakengo.libs.record.DiagramSettings;
import com.dakengo.libs.record.Grades;
import com.dakengo.libs.record.GradeTypes;
import com.dakengo.libs.record.Notations;
import com.dakengo.libs.record.PositionProperties;
import com.dakengo.libs.record.PrintSettings;
import com.dakengo.libs.record.PropertyException;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.tests.utils.TestPropVisitor;

public class PositionPropertiesTests
{
   public static void assertDefaults( PositionProperties props )
   {
      for (int prop_id = PropertyID.COMMENT; prop_id <= PropertyID.MAX_STD_ID; prop_id++)
         assertThat( "property unexpectedly set: " + PropertyID.toString( prop_id ), props.has( prop_id ), is( false ) );
   }

   @Test
   public void testValidProperties()
   {
      Board.setSize( 19 );
      for (int prop_id = PropertyID.COMMENT; prop_id <= PropertyID.MAX_STD_ID; prop_id++)
         assertThat( "standard property: " + PropertyID.toString( prop_id ), PositionProperties.isStandard( prop_id ), is( true ) );

      PositionProperties props = new PositionProperties( true );
      assertDefaults( props );
      assertThat( "is root", props.isRoot(), is( true ) );

      props.addComment( "this is a test" );
      props.addGrade( Colors.BLACK, GradeTypes.POSITION, Grades.GOOD );
      props.addMove( Colors.BLACK, Convert.stringToPoint( "c13" ) );
      props.addNote( 3, "note label", "note value" );

      assertThat( "has COMMENT", props.has( PropertyID.COMMENT ), is( true ) ); 
      assertThat( "has GRADE", props.has( PropertyID.GRADE ), is( true ) ); 
      assertThat( "has MOVE", props.has( PropertyID.MOVE ), is( true ) ); 
      assertThat( "has NOTE", props.has( PropertyID.NOTE ), is( true ) );

      TestPropVisitor visitor = new TestPropVisitor( false );
      props.visit( visitor, true );
      visitor.assertCounts( 0, 4 );
      visitor.assertValue( PropertyID.COMMENT, 0, "this is a test" );
      visitor.assertValue( PropertyID.GRADE, 0, "(BLACK,(POSITION,GOOD))" );
      visitor.assertValue( PropertyID.MOVE, 0, "BLACK:c13" );
      visitor.assertValue( PropertyID.NOTE, 0, "(3,(note label,note value))" );

      visitor.clear();
      props.addTurn( Colors.WHITE );
      props.addTerritory( Colors.BLACK, Convert.stringsToPoints( new String[] { "b5", "d14", "s1" } ) );
      props.visit( visitor, true );
      visitor.assertCounts( 0, 6 );
      visitor.assertValue( PropertyID.TERRITORY, 0, "BLACK:b5,d14,s1" );
      visitor.assertValue( PropertyID.TURN, 0, "WHITE" );

      visitor.clear();
      props.addStones( Colors.WHITE, Convert.stringsToPoints( new String[] { "r9", "g8" } ) );
      props.visit( visitor, true );
      visitor.assertCounts( 0, 7 );
      visitor.assertValue( PropertyID.STONES, 0, "WHITE:g8,r9" );

      PositionProperties copy = new PositionProperties( props );
      assertThat( copy.hashCode(), is( props.hashCode() ) );
      assertThat( copy, is( props ) );

      // try adding a value of valid type for each of the other properties
      //
      props.addConnector( Connectors.LINE, new HashSet<ImmutablePair<Short,Short>>() );
      props.addDiagramSettings( "test message", new DiagramSettings() );
      props.addIllegalMove( Colors.WHITE, (short) 30 );
      HashSet<ImmutablePair<Short,String>> ann_pnts = new HashSet<ImmutablePair<Short,String>>();
      ann_pnts.add( ImmutablePair.of( (short) 1, "xyz" ) );
      props.addLabel( ann_pnts );
      props.addMark( Notations.DIMMED, Convert.stringsToPoints( new String[] { "b2" } ) );
      props.addPrintSettings( new PrintSettings() );
      props.addRemainingMoves( Colors.WHITE, 0 );
      props.addRemainingTime( Colors.BLACK, BigDecimal.ZERO );
      props.addVisible( Convert.stringsToPoints( new String[] { "a1" } ) );
      props.addBranch( new PositionProperties( false ) );
   }

   @Test
   public void testIllegalValues()
   {
      final PositionProperties props = new PositionProperties( false );

      final Set<ImmutablePair<Short,Short>> pnt_pairs = new HashSet<ImmutablePair<Short,Short>>();
      pnt_pairs.add( ImmutablePair.of( (short) -1, (short) 43 ) );
      assertThrows(
         new Runnable() { @Override public void run() { props.addConnector( Connectors.LINE, pnt_pairs ); } },
         PropertyException.class, null );
      pnt_pairs.clear();
      pnt_pairs.add( ImmutablePair.of( (short) 0, (short) 10000 ) ); 
      assertThrows(
         new Runnable() { @Override public void run() { props.addConnector( Connectors.LINE, pnt_pairs ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addGrade( Colors.EMPTY, GradeTypes.MOVE, Grades.BAD ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addIllegalMove( Colors.EMPTY, (short) 0 ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addIllegalMove( Colors.BLACK, (short) 10000 ); } },
         PropertyException.class, null );
      final Set<ImmutablePair<Short,String>> label_pairs = new HashSet<ImmutablePair<Short,String>>();
      label_pairs.add( ImmutablePair.of( (short) 10000, "" ) );
      assertThrows(
         new Runnable() { @Override public void run() { props.addLabel( label_pairs ); } },
         PropertyException.class, null );
      final Set<Short> pnts = new HashSet<Short>();
      assertThrows(
         new Runnable() { @Override public void run() { props.addMark( Notations.DIMMED, pnts ); } },
         PropertyException.class, null );
      pnts.add( (short) 1000 );
      assertThrows(
         new Runnable() { @Override public void run() { props.addMark( Notations.DIMMED, pnts ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addRemainingMoves( Colors.DONT_CARE, 1 ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addRemainingMoves( Colors.WHITE, -1 ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addRemainingTime( Colors.DONT_CARE, BigDecimal.ONE ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addRemainingTime( Colors.BLACK, new BigDecimal( "-1" ) ); } },
         PropertyException.class, null );
      pnts.clear();
      assertThrows(
         new Runnable() { @Override public void run() { props.addStones( Colors.EMPTY, pnts ); } },
         PropertyException.class, null );
      pnts.add( (short) 999 );
      assertThrows(
         new Runnable() { @Override public void run() { props.addStones( Colors.BLACK, pnts ); } },
         PropertyException.class, null );
      pnts.clear();
      assertThrows(
         new Runnable() { @Override public void run() { props.addTerritory( Colors.EMPTY, pnts ); } },
         PropertyException.class, null );
      pnts.add( (short) 999 );
      assertThrows(
         new Runnable() { @Override public void run() { props.addTerritory( Colors.BLACK, pnts ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addTurn( Colors.EMPTY ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.addVisible( pnts ); } },
         PropertyException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { props.add( 1, ImmutablePair.of( "", "" ) ); } },
         PropertyException.class, null );
   }

   @Test
   public void testIncorrectTypes()
   {
      final PositionProperties props = new PositionProperties( false );

      assertThrows(
         new Runnable() { @Override public void run() { props.add( PropertyID.COMMENT, 'a' ); } },
         PropertyException.class, null );
      for (int prop_id = PropertyID.COMMENT + 1; prop_id <= PropertyID.MAX_STD_ID; prop_id++)
      {
         final int id = prop_id;
         assertThrows(
            new Runnable() { @Override public void run() { props.add( id, "" ); } },
            PropertyException.class, null );
      }
   }

   @Test
   public void testAddGameProps()
   {
      final PositionProperties props = new PositionProperties( false );
      int[] prop_ids = new int[]
      {
         PropertyID.ANNOTATOR_NAME, PropertyID.APP_NAME, PropertyID.APP_VERSION, PropertyID.BACKGROUND_INFO, PropertyID.CHARSET,
         PropertyID.COPYRIGHT, PropertyID.DATE, PropertyID.EVENT_NAME, PropertyID.GAME_NAME, PropertyID.OPENING_NAME,
         PropertyID.OVERTIME_SYSTEM, PropertyID.PLACE_NAME, PropertyID.RULE_SYSTEM, PropertyID.SOURCE_NAME, PropertyID.USER_NAME
      };
      for (int prop_id : prop_ids)
      {
         final int id = prop_id;
         assertThrows(
            new Runnable() { @Override public void run() { props.add( id, "" ); } },
            PropertyException.class, null );
      }

      prop_ids = new int[]
      {
         PropertyID.HANDICAP, PropertyID.KOMI, PropertyID.PLAYER_NAME, PropertyID.PLAYER_RANK, PropertyID.RENDERING_STYLE,
         PropertyID.RESULT, PropertyID.ROUND, PropertyID.SIZE, PropertyID.TEAM_NAME, PropertyID.TIME_LIMIT, PropertyID.USER_NAME
      };
      for (int prop_id : prop_ids)
      {
         final int id = prop_id;
         assertThrows(
            new Runnable() { @Override public void run() { props.add( id, 1 ); } },
            PropertyException.class, null );
      }
   }

   @Test
   public void testNonStandard()
   {
      PositionProperties props = new PositionProperties( false );

      props.add( 1000, ImmutablePair.of( "testprop", "some value" ) );
      assertThat( "non-standard prop 1000", props.has( 1000 ), is( true ) );
      TestPropVisitor visitor = new TestPropVisitor( false );
      props.visit( visitor, true );
      visitor.assertValue( 1000, 0, "testprop:some value" );
   }

   @Test
   public void testBranches()
   {
      PositionProperties branch1 = new PositionProperties( false );
      branch1.addComment( "branch 1" );

      PositionProperties branch2 = new PositionProperties( false );
      branch2.addComment( "branch 2" );

      PositionProperties root = new PositionProperties( true );
      root.addBranch( branch1 );
      root.addComment( "root <a>" );
      root.addBranch( branch2 );
      root.addComment( "root <b>" );

      TestPropVisitor visitor = new TestPropVisitor( false );
      root.visit( visitor, false );
      visitor.assertCounts( 0, 2 );
      visitor.assertValues( PropertyID.COMMENT, new String[] { "root <a>", "root <b>" } );
      visitor.clear();
      root.visit( visitor, true );
      visitor.assertCounts( 0, 2 );
      visitor.assertValues( PropertyID.COMMENT, new String[] { "root <a>", "root <b>" } );

      visitor = new TestPropVisitor( true );
      root.visit( visitor, true );
      visitor.assertCounts( 1, 1 );
      visitor.assertValues( PropertyID.COMMENT, new String[] { "branch 1" } );
      visitor.clear();
      root.visit( visitor, false );
      visitor.assertCounts( 2, 4 );
      visitor.assertValues( PropertyID.COMMENT, new String[] { "branch 1", "root <a>", "branch 2", "root <b>" } );
   }

   @Test
   public void testEquals()
   {
      PositionProperties props1 = new PositionProperties( false );
      props1.addComment( "foo!" );

      PositionProperties props2 = new PositionProperties( false );
      props2.addMove( Colors.BLACK, (short) 0 );

      PositionProperties props3 = new PositionProperties( false );

      PositionProperties props4 = new PositionProperties( true );
      props4.addComment( "foo!" );

      PositionProperties props5 = new PositionProperties( false );
      props5.addMove( Colors.BLACK, (short) 0 );

      PositionProperties props6 = new PositionProperties( false );
      props6.addMove( Colors.WHITE, (short) 1 );

      assertThat( props1, is( props1 ) );
      assertThat( props1.hashCode(), is( props1.hashCode() ) );
      assertThat( props2, is( props5 ) );
      assertThat( props2.hashCode(), is( props5.hashCode() ) );
      assertThat( props1, is( not( props2 ) ) );
      assertThat( props2, is( not( props3 ) ) );
      assertThat( props1, is( not( props4 ) ) );
      assertThat( props2, is( not( props6 ) ) );
   }
}
