//
// SGFConvertTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.record;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.math.BigDecimal;
import java.util.Calendar;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Result;
import com.dakengo.libs.record.DiagramSettings;
import com.dakengo.libs.record.Emphasis;
import com.dakengo.libs.record.PrintSettings;
import com.dakengo.libs.record.RenderingStyle;
import com.dakengo.libs.record.Round;
import com.dakengo.plugins.record.sgf.Convert;
import com.dakengo.tests.utils.CollectionUtils;

public class SGFConvertTests
{
   public SGFConvertTests()
   {
   }

   @Before
   public void setUp()
   {
      Board.setSize( 5 );
   }

   @Test
   public void testPoints()
   {
      Board.setSize( 19 );

      String[] sgf_text = new String[] { "bb", "Rj", "cr", "Sa", "ss" };
      short[] points = new short[] { (short) 36, (short) 332, (short) 39, (short) 360, (short) 342 };
      assertThat( sgf_text.length, is( points.length ) );
      for (int i = 0; i < sgf_text.length; i++)
      {
         String sgf = sgf_text[i];
         short actual = Convert.stringToPoint( sgf, -1 );
         assertThat( sgf, actual, is( points[i] ) );
         sgf = sgf.toLowerCase();
         assertThat( Convert.pointToString( actual ), is( sgf ) );
      }

      String[] bad_text = new String[]
      {
         "a1", "12", "a1a", "1a", "a0", "a999", "z1", "z100" 
      };
      for (String text : bad_text)
      {
         final String str = text;
         assertThrows(
            String.format( "'%s'", text ),
            new Runnable() { @Override public void run() { Convert.stringToPoint( str, -1 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testAnnotatedPoints()
   {
      ImmutablePair<Short,String> pair = ImmutablePair.of( (short) 10, "xyz" );
      assertThat( Convert.annotatedPointToString( pair ), is( "ce:xyz" ) );

      pair = ImmutablePair.of( (short) 0, "[]():" );
      assertThat( Convert.annotatedPointToString( pair ), is( "ae:\\[\\]\\(\\)\\:" ) );

      pair = Convert.stringToAnnotatedPoint( "ce:xyz", 0 );
      assertThat( pair.getLeft(), is( (short) 10 ) );
      assertThat( pair.getRight(), is( "xyz" ) );

      pair = Convert.stringToAnnotatedPoint( "aa:xy:z", 0 );
      assertThat( pair.getLeft(), is( (short) 4 ) );
      assertThat( pair.getRight(), is( "xy:z" ) );

      pair = Convert.stringToAnnotatedPoint( "aa:xy\\:z", 0 );
      assertThat( pair.getLeft(), is( (short) 4 ) );
      assertThat( pair.getRight(), is( "xy\\:z" ) );

      assertThrows(
         new Runnable() { @Override public void run() { Convert.stringToAnnotatedPoint( "mm:xyz", 0 ); } },
         IllegalArgumentException.class, null );
      String[] bad_text = new String[]
      {
         "abc", ":xyz", "aa:", "xyz:aa", " ", ""
      };
      for (String text : bad_text)
      {
         final String tmp = text;
         assertThrows(
            String.format( "'%s'", text ),
            new Runnable() { @Override public void run() { Convert.stringToAnnotatedPoint( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testPointLists()
   {
      Set<Short> points = new HashSet<Short>();
      points.addAll( CollectionUtils.arrayToSet( new Short[] { (short) 1, (short) 2, (short) 7, (short) 8, (short) 20, (short) 22 } ) );
      String list_str = "ad][ac][bc][bb][ee][ec";
      assertThat( Convert.pointsToString( new TreeSet<Short>( points ) ), is( list_str ) );
      assertThat( Convert.stringToPoints( list_str, 0 ), is( points ) );
      points.clear();
      list_str = "ee][ad:ac][bc:bb][ec";
      points.addAll( CollectionUtils.arrayToSet( new Short[] { (short) 20, (short) 1, (short) 2, (short) 7, (short) 8, (short) 22 } ) );
      assertThat( Convert.stringToPoints( list_str, 0 ), is( points ) );

      points.clear();
      points.add( (short) 3 );
      assertThat( Convert.pointsToString( points ), is( "ab" ) );
      assertThat( Convert.stringToPoints( "ab", 0 ), is( points ) );

      points.clear();
      points.add( Board.INVALID_POINT );
      assertThat( Convert.pointsToString( points ), is( "" ) );
      assertThat( Convert.stringToPoints( "tt", 0 ), is( points ) );

      assertThat( Convert.stringToPoints( "", 0 ).size(), is( 0 ) );

      assertThrows(
         "'nm'",
         new Runnable() { @Override public void run() { Convert.stringToPoints( "nm", 0 ); } },
         IllegalArgumentException.class, null );
      String[] bad_str = new String[]
      {
         " ", "  ", "   ", "a", "aac", " a ", "xy", "ab;cd", "][", "ab][", "][ab"
      };
      for (String text : bad_str)
      {
         final String tmp = text;
         assertThrows(
            String.format( "'%s'", text ),
            new Runnable() { @Override public void run() { Convert.stringToPoints( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testAnnotatedPointLists()
   {
      Set<ImmutablePair<Short,String>> points = new HashSet<ImmutablePair<Short,String>>();
      points.add( ImmutablePair.of( (short) 10, "xyx" ) );
      points.add( ImmutablePair.of( (short) 11, "abc" ) );
      points.add( ImmutablePair.of( (short) 12, "def" ) );
      String text = "ce:xyx][cd:abc][cc:def";
      assertThat( Convert.annotatedPointsToString( new TreeSet<ImmutablePair<Short,String>>( points ) ), is( text ) );
      assertThat( Convert.stringToAnnotatedPoints( text, 0 ), is( points ) );

      points = Convert.stringToAnnotatedPoints( "aa:a]][ab:[][ac:x", 0 );
      Set<ImmutablePair<Short,String>> expected = new HashSet<ImmutablePair<Short,String>>();
      expected.add( ImmutablePair.of( (short) 4, "a]" ) );
      expected.add( ImmutablePair.of( (short) 3, "[" ) );
      expected.add( ImmutablePair.of( (short) 2, "x" ) );
      assertThat( points, is( expected ) );

      points.clear();
      points.add( ImmutablePair.of( (short) 1, "]" ) );
      assertThat( Convert.annotatedPointsToString( points ), is( "ad:\\]" ) );
      assertThat( Convert.stringToAnnotatedPoints( "ad:]", 0 ), is( points ) );

      assertThrows(
         new Runnable() { @Override public void run() { Convert.stringToAnnotatedPoints( "gn:abc", 0 ); } },
         IllegalArgumentException.class, null );
      String[] bad_text = new String[]
      {
         " ", "aa", "aa:", ":#$%", "  ", " aa:qqq",
         "][", "aa:xxx][", "zy:abc",
         "][bb:yyy", "aa:xxx][][bb:yyy"
      };
      for (String text2 : bad_text)
      {
         final String tmp = text2;
         assertThrows(
            String.format( "'%s'", text2 ),
            new Runnable() { @Override public void run() { Convert.stringToAnnotatedPoints( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testPointPairs()
   {
      assertThat( Convert.pointPairToString( ImmutablePair.of( (short) 0, (short) 24 ) ), is( "ae:ea" ) );
      assertThat( Convert.stringToPointPair( "aa:cc", 0 ), is( ImmutablePair.of( (short) 4, (short) 12 ) ) );
      assertThrows(
         new Runnable() { @Override public void run() { Convert.stringToPointPair( "gn:ng", 0 ); } },
         IllegalArgumentException.class, null );
      String[] bad_text = new String[]
      {
         "xy:yx", "a1:b2", "aa", " ", ""
      };
      for (String text : bad_text)
      {
         final String tmp = text;
         assertThrows(
            String.format( "'%s'", text ),
            new Runnable() { @Override public void run() { Convert.stringToPointPair( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testPointPairLists()
   {
      Set<ImmutablePair<Short,Short>> points = new HashSet<ImmutablePair<Short,Short>>();
      points.add( ImmutablePair.of( (short) 4, (short) 12 ) );
      points.add( ImmutablePair.of( (short) 0, (short) 24 ) );
      assertThat( Convert.pointPairsToString( points ), is( "aa:cc][ae:ea" ) );
      assertThat( Convert.stringToPointPairs( "aa:cc][ae:ea", 0 ), is( points ) );

      points.clear();
      points.add( ImmutablePair.of( (short) 0, (short) 24 ) );
      assertThat( Convert.pointPairsToString( points ), is( "ae:ea" ) );
      assertThat( Convert.stringToPointPairs( "ae:ea", 0 ), is( points ) );

      assertThat( Convert.stringToPointPairs( "", 0 ).size(), is( 0 ) );

      String[] bad_text = new String[]
      {
         " ", "  ", "   ", "a", " a ", "xy", "nm", "ab;cd", "a:c", "][", "ab:cd][", "][ab:cd"
      };
      for (String text : bad_text)
      {
         final String tmp = text;
         assertThrows(
            String.format( "'%s'", text ),
            new Runnable() { @Override public void run() { Convert.stringToPointPairs( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testColors()
   {
      assertThat( Convert.colorToChar( Colors.BLACK ), is( 'B' ) );
      assertThat( Convert.colorToChar( Colors.WHITE ), is( 'W' ) );
      assertThat( Convert.colorToChar( Colors.EMPTY ), is( 'E' ) );
      assertThat( Convert.colorToChar( Colors.DONT_CARE ), is( '?' ) );

      assertThat( Convert.charToColor( 'b', 0 ), is( Colors.BLACK ) );
      assertThat( Convert.charToColor( 'B', 0 ), is( Colors.BLACK ) );
      assertThat( Convert.charToColor( 'w', 0 ), is( Colors.WHITE ) );
      assertThat( Convert.charToColor( 'W', 0 ), is( Colors.WHITE ) );
      assertThat( Convert.charToColor( 'e', 0 ), is( Colors.EMPTY ) );
      assertThat( Convert.charToColor( 'E', 0 ), is( Colors.EMPTY ) );

      assertThrows(
         new Runnable() { @Override public void run() { Convert.charToColor( 'f', 0 ); } },
         ParseException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { Convert.charToColor( '?', 0 ); } },
         ParseException.class, null );
   }

   @Test
   public void testDates()
   {
      Calendar cal = Calendar.getInstance();
      cal.clear();
      cal.set( 2012, 2, 14 );
      Date date = cal.getTime();
      assertThat( Convert.dateToString( date ), is( "2012-03-14" ) );
      String[] date_str = new String[]
      {
         "3/14/2012", /*"14/3/2012",*/ "2012-03-14", /*"14-3-2012",*/ "Mar 14, 2012",
         "March 14 2012", "Mar-14-2012", "March-14-2012", "2012/03/14", "14 March 2012",
         "14 Mar 2012", "14-March-2012", "14-Mar-2012"
      };
      for (String text : date_str)
         assertThat( text, Convert.stringToDate( text, 0 ), is( date ) );
   }

   @Test
   public void testDecimals()
   {
      BigDecimal d1 = new BigDecimal( "7.0" );
      BigDecimal d2 = new BigDecimal( "0.4" );
      BigDecimal d3 = new BigDecimal( "1.2" );
      BigDecimal d4 = new BigDecimal( "-4.52" );
      assertThat( Convert.decimalToString( d1 ), is( "7.0" ) );
      assertThat( Convert.decimalToString( d2 ), is( "0.4" ) );
      assertThat( Convert.decimalToString( d3 ), is( "1.2" ) );
      assertThat( Convert.decimalToString( d4 ), is( "-4.52" ) );

      assertThat( Convert.stringToDecimal( "7.0", 0 ), is( d1 ) );
      assertThat( Convert.stringToDecimal( "0.4", 0 ), is( d2 ) );
      assertThat( Convert.stringToDecimal( "1.2", 0 ), is( d3 ) );
      assertThat( Convert.stringToDecimal( "-4.52", 0 ), is( d4 ) );
   }

   @Test
   public void testDiagramSettings()
   {
      DiagramSettings settings = new DiagramSettings( false, false, true, false, false, true );
      ImmutablePair<String,DiagramSettings> value = ImmutablePair.of( "foobar", settings );
      assertThat( Convert.diagramSettingsToString( value ), is( "32775:foobar" ) );
      settings.enableAppDefaults( true );
      assertThat( Convert.diagramSettingsToString( value ), is( "7:foobar" ) );

      value = Convert.stringToDiagramSettings( "514:xyz", 0 );
      assertThat( value.getLeft(), is( "xyz" ) );
      settings = new DiagramSettings( true, true, false, true, false, true );
      assertThat( value.getRight(), is( settings ) );

      assertThrows(
         new Runnable() { @Override public void run() { Convert.stringToDiagramSettings( "172", 0 ); } },
         ParseException.class, null );
   }

   @Test
   public void testEmphasis()
   {
      assertThat( Convert.emphasisToChar( Emphasis.STRONG ), is( '1' ) );
      assertThat( Convert.emphasisToChar( Emphasis.VERY_STRONG ), is( '2' ) );
      assertThrows(
         new Runnable() { @Override public void run() { Convert.emphasisToChar( Emphasis.NONE ); } },
         IllegalArgumentException.class, null );

      assertThat( Convert.charToEmphasis( '1', 0 ), is( Emphasis.STRONG ) );
      assertThat( Convert.charToEmphasis( '2', 0 ), is( Emphasis.VERY_STRONG ) );
      assertThrows(
         new Runnable() { @Override public void run() { Convert.charToEmphasis( 'a', 0 ); } },
         ParseException.class, null );
   }

   @Test
   public void testPrintSettings()
   {
      PrintSettings settings = new PrintSettings();
      settings.setMoveNumbers( PrintSettings.MoveNumbers.DISABLE );
      assertThat( Convert.printSettingsToChar( settings ), is( '0' ) );
      settings.setMoveNumbers( PrintSettings.MoveNumbers.AS_IS );
      assertThat( Convert.printSettingsToChar( settings ), is( '1' ) );
      settings.setMoveNumbers( PrintSettings.MoveNumbers.MODULO );
      assertThat( Convert.printSettingsToChar( settings ), is( '2' ) );
      settings.setMoveNumbers( PrintSettings.MoveNumbers.DISABLE );
      settings.setMoveNumberModulo( 100 );
      assertThat( Convert.printSettingsToChar( settings ), is( '2' ) );

      assertThrows(
         new Runnable() { @Override public void run()
         {
            PrintSettings tmp = new PrintSettings();
            tmp.setMoveNumberModulo( 10 );
            Convert.printSettingsToChar( tmp ); }
         },
         IllegalArgumentException.class, null );

      settings = Convert.charToPrintSettings( '0', 0 );
      assertThat( settings.getMoveNumbers(), is( PrintSettings.MoveNumbers.DISABLE ) );
      settings = Convert.charToPrintSettings( '1', 0 );
      assertThat( settings.getMoveNumbers(), is( PrintSettings.MoveNumbers.AS_IS ) );
      settings = Convert.charToPrintSettings( '2', 0 );
      assertThat( settings.getMoveNumbers(), is( PrintSettings.MoveNumbers.MODULO ) );
      assertThat( settings.getMoveNumberModulo(), is( 100 ) );

      char[] bad_text = new char[]
      {
         '\0', '4', 'a', '!'
      };
      for (char ch : bad_text)
      {
         final char tmp = ch;
         assertThrows(
            new Runnable() { @Override public void run() { Convert.charToPrintSettings( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testRenderingStyles()
   {
      RenderingStyle style1 = new RenderingStyle( false, false );
      RenderingStyle style2 = new RenderingStyle( true, false );
      RenderingStyle style3 = new RenderingStyle( false, true );
      RenderingStyle style4 = new RenderingStyle( true, true );
      assertThat( Convert.renderingStyleToChar( style1 ), is( '2' ) );
      assertThat( Convert.renderingStyleToChar( style2 ), is( '3' ) );
      assertThat( Convert.renderingStyleToChar( style3 ), is( '0' ) );
      assertThat( Convert.renderingStyleToChar( style4 ), is( '1' ) );

      assertThat( Convert.charToRenderingStyle( '2', 0 ), is( style1 ) );
      assertThat( Convert.charToRenderingStyle( '3', 0 ), is( style2 ) );
      assertThat( Convert.charToRenderingStyle( '0', 0 ), is( style3 ) );
      assertThat( Convert.charToRenderingStyle( '1', 0 ), is( style4 ) );

      char[] bad_chars = new char[]
      {
         '\0', ' ', '5'
      };
      for (char ch : bad_chars)
      {
         final char tmp = ch;
         assertThrows(
            String.format( "'%c'", ch ),
            new Runnable() { @Override public void run() { Convert.charToRenderingStyle( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testResults()
   {
      Result result = new Result( Colors.BLACK, Result.Status.WON, "0.5" );
      String text = result.toString();
      assertThat( text, is( "B+0.5" ) );
      Result r2 = Convert.stringToResult( text, 0 );
      assertThat( r2, is( result ) );

      result = new Result( Colors.WHITE, Result.Status.RESIGNED );
      text = result.toString();
      assertThat( text, is( "W+Resign" ) );
      r2 = Convert.stringToResult( text, 0 );
      assertThat( r2, is( result ) );

      result = new Result( Colors.BLACK, Result.Status.FORFEIT );
      text = result.toString();
      assertThat( text, is( "B+Forfeit" ) );
      r2 = Convert.stringToResult( text, 0 );
      assertThat( r2, is( result ) );

      result = new Result( Colors.WHITE, Result.Status.LOST_ON_TIME );
      text = result.toString();
      assertThat( text, is( "W+Time" ) );
      r2 = Convert.stringToResult( text, 0 );
      assertThat( r2, is( result ) );

      result = new Result( Colors.DONT_CARE, Result.Status.DRAW );
      text = result.toString();
      assertThat( text, is( "Draw" ) );
      r2 = Convert.stringToResult( text, 0 );
      assertThat( r2, is( result ) );

      assertThrows(
         new Runnable() { @Override public void run() { new Result( Colors.BLACK, Result.Status.DRAW ); } },
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { new Result( Colors.EMPTY, Result.Status.FORFEIT ); } },
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { new Result( Colors.DONT_CARE, Result.Status.LOST_ON_TIME ); } },
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { new Result( Colors.DONT_CARE, Result.Status.RESIGNED ); } },
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { new Result( Colors.BLACK, Result.Status.WON ); } },
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { new Result( Colors.EMPTY, Result.Status.WON, "1.5" ); } },
         IllegalArgumentException.class, null );
      assertThrows(
         new Runnable() { @Override public void run() { new Result( Colors.WHITE, Result.Status.WON, "-3.5" ); } },
         IllegalArgumentException.class, null );
      String[] bad_text = new String[]
      {
         "", "sdf", "W F", "B-R", "S+0.4", "W+", "+T"
      };
      for (String str : bad_text)
      {
         final String tmp = str;
         assertThrows(
            String.format( "'%s'", str ),
            new Runnable() { @Override public void run() { Convert.stringToResult( tmp, 0 ); } },
            ParseException.class, null );
      }
   }

   @Test
   public void testRounds()
   {
      Round round = new Round( 10, Round.Type.LEAGUE, "" );
      String text = Convert.roundToString( round );
      assertThat( text, is( "10 (LG)" ) );
      Round round2 = Convert.stringToRound( text, 0 );
      assertThat( round2, is( round ) );

      round = new Round( 1, Round.Type.FINAL, "" );
      text = Convert.roundToString( round );
      assertThat( text, is( "1 (FN)" ) );
      round2 = Convert.stringToRound( text, 0 );
      assertThat( round2, is( round ) );

      round = new Round( 4, Round.Type.PLAYOFF, "" );
      text = Convert.roundToString( round );
      assertThat( text, is( "4 (PO)" ) );
      round2 = Convert.stringToRound( text, 0 );
      assertThat( round2, is( round ) );

      round = new Round( 2, Round.Type.UNDEFINED, "" );
      text = Convert.roundToString( round );
      assertThat( text, is( "2" ) );
      round2 = Convert.stringToRound( text, 0 );
      assertThat( round2, is( round ) );

      round = new Round( 1, Round.Type.PRELIM, "" );
      text = Convert.roundToString( round );
      assertThat( text, is( "1 (PR)" ) );
      round2 = Convert.stringToRound( text, 0 );
      assertThat( round2, is( round ) );

      round = new Round( 1, Round.Type.GAME, "" );
      text = Convert.roundToString( round );
      assertThat( text, is( "1 (GM)" ) );
      round2 = Convert.stringToRound( text, 0 );
      assertThat( round2, is( round ) );

      round = new Round( 4, Round.Type.PLAYOFF, "Quarter-finals" );
      text = Convert.roundToString( round );
      assertThat( text, is( "4 (PO) Quarter-finals" ) );
      round2 = Convert.stringToRound( text, 0 );
      assertThat( round2, is( round ) );

      // TODO: consider more in-depth parsing to extract type
      // and number from the following cases

      round = Convert.stringToRound( "Game 6", 0 );
      text = Convert.roundToString( round );
      assertThat( "Game 6", is( text ) );

      round = Convert.stringToRound( "Game (7)", 0 );
      text = Convert.roundToString( round );
      assertThat( "Game (7)", is( text ) );

      round = Convert.stringToRound( "3rd Preliminary", 0 );
      text = Convert.roundToString( round );
      assertThat( "3rd Preliminary", is( text ) );

      round = Convert.stringToRound( "( FN )", 0 );
      text = Convert.roundToString( round );
      assertThat( "( FN )", is( text ) );  // because all of the text went into the extra field

      round = Convert.stringToRound( "(SD)", 0 );
      text = Convert.roundToString( round );
      assertThat( "(SD)", is( text ) );

      round = Convert.stringToRound( "Final", 0 );
      text = Convert.roundToString( round );
      assertThat( "Final", is( text ) );
   }

   @Test
   public void testTextPairs()
   {
      ImmutablePair<String,String> pair = Convert.stringToTextPair( "abc:def", 0 );
      assertThat( pair.getLeft(), is( "abc" ) );
      assertThat( pair.getRight(), is( "def" ) );

      pair = Convert.stringToTextPair( "[:]:()", 0 );
      assertThat( pair.getLeft(), is( "[" ) );
      assertThat( pair.getRight(), is( "]:()" ) );

      pair = Convert.stringToTextPair( "123:456.4", 0 );
      assertThat( pair.getLeft(), is( "123" ) );
      assertThat( pair.getRight(), is( "456.4" ) );

      pair = Convert.stringToTextPair( "", 0 );
      assertThat( pair.getLeft().length(), is( 0 ) );
      assertThat( pair.getRight().length(), is( 0 ) );

      pair = Convert.stringToTextPair( ":", 0 );
      assertThat( pair.getLeft().length(), is( 0 ) );
      assertThat( pair.getRight().length(), is( 0 ) );

      pair = Convert.stringToTextPair( "::", 0 );
      assertThat( pair.getLeft().length(), is( 0 ) );
      assertThat( pair.getRight(), is( ":" ) );

      pair = Convert.stringToTextPair( "b:", 0 );
      assertThat( pair.getLeft(), is( "b" ) );
      assertThat( pair.getRight().length(), is( 0 ) );

      pair = Convert.stringToTextPair( ":c", 0 );
      assertThat( pair.getLeft().length(), is( 0 ) );
      assertThat( pair.getRight(), is( "c" ) );

      pair = Convert.stringToTextPair( "a:b:c", 0 );
      assertThat( pair.getLeft(), is( "a" ) );
      assertThat( pair.getRight(), is( "b:c" ) );

      String[] bad_text = new String[]
      {
         " ", "a"
      };
      for (String text : bad_text)
      {
         final String tmp = text;
         assertThrows(
            String.format( "'%s'", text ),
            new Runnable() { @Override public void run() { Convert.stringToTextPair( tmp, 0 ); } },
            ParseException.class, null );
      }

      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "", "" ) ), is( ":" ) );
      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "", "ab" ) ), is( ":ab" ) );
      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "d", "" ) ), is( "d:" ) );
      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "12", "34" ) ), is( "12:34" ) );
      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "12:", "34" ) ), is( "12\\::34" ) );
      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "12", "3:4" ) ), is( "12:3\\:4" ) );
      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "[12", "3]4" ) ), is( "\\[12:3\\]4" ) );
      assertThat( Convert.textPairToString( new ImmutablePair<String,String>( "()12", "3" ) ), is( "\\(\\)12:3" ) );
   }

   @Test
   public void testTransformText()
   {
      assertThat( Convert.transformText( "", false ), is( "" ) );
      assertThat( Convert.transformText( "abc", false ), is( "abc" ) );
      String orig = "a\nb]\t[)(:";
      String eol = System.getProperty( "line.separator" );
      assertThat( Convert.transformText( orig, false ), is( String.format( "a%sb\\] \\[\\)\\(\\:", eol ) ) );
      assertThat( Convert.transformText( orig, true ), is( "a b\\] \\[\\)\\(\\:" ) );
   }
}
