//
// SGFRecordTests.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.record;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import static com.dakengo.tests.utils.AssertUtils.*;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.math.BigDecimal;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.junit.Before;
import org.junit.Test;

import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.Result;
import com.dakengo.libs.game.Signature;
import com.dakengo.libs.record.GameProperties;
import com.dakengo.libs.record.Handicaps;
import com.dakengo.libs.record.PositionProperties;
import com.dakengo.libs.record.PropertyException;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.libs.record.Record;
import com.dakengo.libs.record.RecordReader;
import com.dakengo.libs.record.SignatureBuilder;
import com.dakengo.tests.utils.BaseRecordTests;
import com.dakengo.tests.utils.TestPropVisitor;
import com.dakengo.tests.utils.TestRunnable;

public class SGFRecordTests extends BaseRecordTests
{
   public SGFRecordTests() throws IOException
   {
   }

   @Test
   public void testInvalidRecords() throws IOException
   {
      runExceptionCases( ParseException.class, new String[] { "(", "(;AP[xxx]", "(;C;[doc 1](;AB[aa][bb]))", "(;C((;AB[aa][bb]))", "(C[abcdefg])" } );
      runExceptionCases( ParseException.class, new String[] { "(;GM[2])", "(;GM[])", "(;GM[abc])" } );
      runExceptionCases( ParseException.class, new String[] { "(;B[1])", "(;W[B])", "(;B[a19])" } );
      runExceptionCases( ParseException.class, new String[] { "(;B[aa]W[bb])", "(;B[aa]B[bb])", "(;B[aa]C[foo]W[bb])" } );
      runExceptionCases( PropertyException.class, new String[] { "(;W[aa]B[bb])" } );
      runExceptionCases( PropertyException.class, new String[]
      {
         "(;AB[aa][aa])", "(;AE[aa][aa])", "(;AW[aa][aa])", "(;CR[aa][aa])", "(;DD[aa][aa])", "(;MA[aa][aa])",
         "(;SL[aa][aa])", "(;SQ[aa][aa])", "(;TR[aa][aa])", "(;TB[aa][aa])", "(;TW[aa][aa])",
         "(;AB[][aa])",   "(;AE[][aa])",   "(;AW[][aa])",   "(;CR[][aa])",   "(;DD[][aa])",   "(;MA[][aa])",
         "(;SL[][aa])",   "(;SQ[][aa])",   "(;TR[][aa])",   "(;TB[][aa])",   "(;TW[][aa])",
/*
         "(;AB[aa][])",   "(;AE[aa][])",   "(;AW[aa][])",   "(;CR[aa][])",   "(;DD[aa][])",   "(;MA[aa][])",
         "(;SL[aa][])",   "(;SQ[aa][])",   "(;TR[aa][])",   "(;TB[aa][])",   "(;TW[aa][])",
*/
         "(;AB[][])",     "(;AE[][])",     "(;AW[][])",     "(;CR[][])",     "(;DD[][])",     "(;MA[][])",
         "(;SL[][])",     "(;SQ[][])",     "(;TR[][])",     "(;TB[][])",     "(;TW[][])",
         "(;LB[][aa:x])", "(;LB[aa:x][aa:y])"
      } );
      runExceptionCases( ParseException.class, new String[]
      {
         "(;LB[1])",
         "(;LB[ab])",
         "(;LB[:aa])",
         "(;LB[1:aa])",
/*
         "(;LB[aa:ab][])",
*/
         "(;LB[aa:ab][1])",
         "(;LB[aa:ab][ab])",
         "(;LB[aa:ab][:aa])",
         "(;LB[aa:ab][1:aa])"
      } );
      runExceptionCases( ParseException.class, new String[]
      {
         "(;AR[1])",            "(;LN[1])",
         "(;AR[ab])",           "(;LN[ab])",
         "(;AR[:aa])",          "(;LN[:aa])",
         "(;AR[aa:])",          "(;LN[aa:])",
         "(;AR[aa:1])",         "(;LN[aa:1])",
         "(;AR[1:aa])",         "(;LN[1:aa])",
         "(;AR[aa:aa])",        "(;LN[aa:aa])",
         "(;AR[aa:ab][1])",     "(;LN[aa:ab][1])",
         "(;AR[aa:ab][ab])",    "(;LN[aa:ab][ab])",
         "(;AR[aa:ab][:aa])",   "(;LN[aa:ab][:aa])",
         "(;AR[aa:ab][aa:])",   "(;LN[aa:ab][aa:])",
         "(;AR[aa:ab][1:aa])",  "(;LN[aa:ab][1:aa])",
         "(;AR[aa:ab][aa:1])",  "(;LN[aa:ab][aa:1])",
         "(;AR[aa:ab][aa:aa])", "(;LN[aa:ab][aa:aa])"
      } );
      runExceptionCases( ParseException.class, new String[] { "(;PL[])", "(;PL[Z])", "(;PL[1])", "(;PL[E])" } );
      runExceptionCases( ParseException.class, new String[] { "(;MN[ab])", "(;V[abc])", "(;ST[abc])", "(;ST[4])", "(;GM[1]FF[9])" } );

      runMessageCases( 1, 0, 2, new String[]
      {
         "(;PB[foo]PW[bar]DM[3])", "(;PB[foo]PW[bar]GB[0])", "(;PB[foo]PW[bar]GW[3])", "(;PB[foo]PW[bar]UC[3])",
         "(;PB[foo]PW[bar]BM[0])", "(;PB[foo]PW[bar]TE[3])", "(;KO[1]PB[b]PW[w])",     "(;IT[xxx]PB[b]PW[w])",
         "(;PB[a]PB[b]PW[c])"
      } );
      runMessageCases( 1, 0, 3, new String[]
      {
         "(;AP[xxx]AP[yyy]PB[a]PW[c])"
      } );

      // TODO: add cases for FG and VW
   }

   @Test
   public void testInfoProps() throws IOException
   {
      m_reader.setInput( new StringReader( "(;SZ[7]AP[some app:1.2]RU[AGA]TM[360]PW[white]PB[black])" ) );
      Record rec = m_reader.read();
      assertParseMessageCount( 0 );
      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getSize(), is( 7 ) );
      assertThat( info_props.getApplicationName(), is( "some app" ) );
      assertThat( info_props.getApplicationVersion(), is( "1.2" ) );
      assertThat( info_props.getRuleSystem(), is( "AGA" ) );
      assertThat( info_props.getTimeLimit(), is( 360 ) );
      assertThat( info_props.getPlayerName( Colors.BLACK ), is( "black" ) );
      assertThat( info_props.getPlayerName( Colors.WHITE ), is( "white" ) );
      assertNull( m_reader.read() );

      m_reader.setInput( new StringReader( "(;GM[1]PB[some player]C[Root Node];KM[5.5]C[child node])" ) );
      rec = m_reader.read();
      assertParseMessageCount( 1 );  // one warning due to missing PW property
      info_props = rec.getGameProps();
      assertThat( info_props.getKomi(), is( new BigDecimal( "5.5" ) ) );
      assertNull( m_reader.read() );

      rec.getMainBranch().visit( m_prop_visitor, false );
      m_prop_visitor.assertCounts( 0, 4 ); // two node properties + two comment properties
      m_prop_visitor.assertValues( PropertyID.COMMENT, new String[] { "Root Node", "child node" } );
      m_prop_visitor.clear();
      info_props.visit( m_prop_visitor );
      m_prop_visitor.assertCounts( 0, 3 ); // GM property isn't stored, but KM and PB are, plus m_reader adds PW
      m_prop_visitor.assertValues( PropertyID.KOMI, new String[] { "5.5" } );

      assertNull( m_reader.read() );
   }

   @Test
   public void testSetupProps() throws IOException
   {
      m_reader.setInput( new StringReader( "(;SZ[5];AB[aa][ab][ac]AW[ba][bb][bc];TB[ca][cb][cc]TW[da][db][dc];AE[ea][eb])" ) );
      Record rec = m_reader.read();
      assertParseMessageCount( 2 ); // two warnings due to missing PB/PW properties
      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getSize(), is( 5 ) );

      rec.getMainBranch().visit( m_prop_visitor, false );
      m_prop_visitor.assertCounts( 0, 9 );
      m_prop_visitor.assertValues( PropertyID.NOTE, new String[] { "(0,(NODE,))", "(1,(NODE,))", "(2,(NODE,))", "(3,(NODE,))" } );
      m_prop_visitor.assertValues( PropertyID.STONES, new String[] { "BLACK:a3,a4,a5", "WHITE:b3,b4,b5", "EMPTY:e4,e5" } );
      m_prop_visitor.assertValues( PropertyID.TERRITORY, new String[] { "BLACK:c3,c4,c5", "WHITE:d3,d4,d5" } );

      m_prop_visitor.clear();
      info_props.visit( m_prop_visitor );
      m_prop_visitor.assertCounts( 0, 4 );  // reader added three props: missing player names and handicap

      assertNull( m_reader.read() );
   }

   @Test
   public void testHandicapViaMoves() throws IOException
   {
      Board.setSize( 9 );
      Board board0 = new Board();
      Set<Short> black_pnts = Convert.stringsToPoints( "a1", "a2", "a3" );
      board0.setColor( black_pnts, Colors.BLACK );
      Board board1 = new Board();
      board1.setColor( black_pnts, Colors.BLACK );
      board1.setColor( Convert.stringsToPoints( "a4" ), Colors.WHITE );

      m_reader.setInput( new StringReader( "(;FF[4]SZ[9]HA[3];B[ai];B[ah];B[ag];W[af])" ) );
      Record rec = m_reader.read();
      assertParseMessageCount( 2 ); // two warnings due to missing PB/PW properties
      assertNull( m_reader.read() );

      m_collector.collect( rec, Integer.MAX_VALUE );
      Iterator<ImmutablePair<Board,Integer>> iter = m_collector.getBoards().iterator();
      ImmutablePair<Board,Integer> pair = iter.next();
      assertThat( "handicap stones", pair.getLeft(), is( board0 ) );
      assertThat( "turn 0", pair.getRight(), is( 0 ) );
      pair = iter.next();
      assertThat( "final board", pair.getLeft(), is( board1 ) );
      assertThat( "turn 1", pair.getRight(), is( 1 ) );
      ImmutablePair<Board,Integer> last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 3 ) );
      assertThat( m_collector.getHandicapBoard(), is( board0 ) );

      m_collector.collect( rec, 0 );
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      assertThat( "handicap stones", pair.getLeft(), is( board0 ) );
      assertThat( "turn 0", pair.getRight(), is( 0 ) );
      last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 3 ) );
      assertThat( m_collector.getHandicapBoard(), is( board0 ) );

      m_collector.collect( rec, -1 );
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      assertThat( "final board", pair.getLeft(), is( board1 ) );
      assertThat( "turn 1", pair.getRight(), is( 1 ) );
      last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 3 ) );
      assertThat( m_collector.getHandicapBoard(), is( board0 ) );
   }

   @Test
   public void testNonStdProps() throws IOException
   {
      m_reader.setInput( new StringReader( "(;AZ[some value]ZX[other value])" ) );
      Record rec = m_reader.read();
      assertParseMessageCount( 4 );
      rec.getGameProps().visit( m_prop_visitor );
      m_prop_visitor.assertCounts( 0, 4 );
      m_prop_visitor.assertValue( PropertyID.MAX_STD_ID + 1, 0, "AZ:some value" );
      m_prop_visitor.assertValue( PropertyID.MAX_STD_ID + 2, 0, "ZX:other value" );
   }

   @Test
   public void testTrivial() throws IOException
   {
      String[] test_cases = new String[] { "()", "(;)", "(;;)" };
      for (String text : test_cases)
      {
         m_reader.setInput( new StringReader( text ) );
         assertNotNull( text, m_reader.read() );
         assertParseMessageCount( 2 );
         assertNull( text, m_reader.read() );
      }
   }

   @Test
   public void testFile_simple() throws IOException
   {
      Record rec = load( "simple.sgf", 6 );

      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getApplicationName(), is( "Explorer" ) );
      assertThat( info_props.getApplicationVersion(), is( "0" ) );
      assertThat( info_props.getSize(), is( 19 ) );

      rec.visit( m_prop_visitor, false );
      m_prop_visitor.assertCounts( 1, 18 );  // neither FF nor GM are stored by SGFReader, 13 SGF props + 2 added by reader for missing PB/PW + 3 NOTE props for nodes
      m_prop_visitor.assertValue( PropertyID.GRADE, 0, "(DONT_CARE,(MOVE,VERY_BAD))" );
      m_prop_visitor.assertValue( PropertyID.MARK, 0, "CIRCLE:" + sgfToStd( "ph" ) );
      m_prop_visitor.assertValue( PropertyID.MOVE, 0, "WHITE:" + sgfToStd( "qh" ) );
      m_prop_visitor.assertValues( PropertyID.NOTE, "(0,(NODE,))", "(0,(NAME,Blunder))", "(1,(NODE,))", "(2,(NODE,))" );
      StringBuilder buffer = new StringBuilder();
      buffer.append( "BLACK:" );
      buffer.append( sgfListToStd( "kb][mc][oc][pd][df][qf][qg][pi][ok][qk][op][qp][lq" ) );
      String black_str = buffer.toString();
      buffer = new StringBuilder();
      buffer.append( "WHITE:" );
      buffer.append( sgfListToStd( "fc][dd][id][kd][pg][ch][cj][dl][co][gp][dq][iq" ) );
      m_prop_visitor.assertValues( PropertyID.STONES, black_str, buffer.toString() );

      Set<Short> pnts = Convert.stringsToPoints( "d14", "l18", "m3", "n17", "p4", "p9", "p17", "q11", "q16", "r4", "r9", "r13", "r14" );
      Board board0 = new Board();
      board0.setColor( pnts, Colors.BLACK );
      Board board1 = new Board( board0 );
      pnts = Convert.stringsToPoints( "c5", "c10", "c12", "d3", "d8", "d16", "f17", "g4", "j3", "j16", "l16", "q13" );
      board1.setColor( pnts, Colors.WHITE );
      Board board2 = new Board( board1 );
      board2.setColor( Convert.stringToPoint( "r12" ), Colors.WHITE );

      m_collector.collect( rec, Integer.MAX_VALUE );
      Iterator<ImmutablePair<Board,Integer>> iter = m_collector.getBoards().iterator();
      ImmutablePair<Board,Integer> pair = iter.next();
      assertThat( "black setup stones", pair.getLeft(), is( board0 ) );
      assertThat( "turn 0", pair.getRight(), is( 0 ) );
      pair = iter.next();
      assertThat( "plus white setup stones", pair.getLeft(), is( board1 ) );
      assertThat( "still turn 0", pair.getRight(), is( 0 ) );
      pair = iter.next();
      assertThat( "white move", pair.getLeft(), is( board2 ) );
      assertThat( "turn 1", pair.getRight(), is( 1 ) );
      ImmutablePair<Board,Integer> last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 0 ) );
      assertNull( m_collector.getHandicapBoard() );

      m_collector.collect( rec, 0 );
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      assertThat( "black setup stones", pair.getLeft(), is( board0 ) );
      assertThat( "turn 0", pair.getRight(), is( 0 ) );
      pair = iter.next();
      assertThat( "plus white setup stones", pair.getLeft(), is( board1 ) );
      assertThat( "still turn 0", pair.getRight(), is( 0 ) );
      last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 0 ) );
      assertNull( m_collector.getHandicapBoard() );

      m_collector.collect( rec, -1 );
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      assertThat( "final board", pair.getLeft(), is( board2 ) );
      assertThat( "turn 1", pair.getRight(), is( 1 ) );
      last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 0 ) );
      assertNull( m_collector.getHandicapBoard() );
   }

   @Test
   public void testFile_white_setup_first() throws IOException
   {
      Record rec = load( "white_setup_first.sgf", 2 );

      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getCharset(), is( "UTF-8" ) );
      assertThat( info_props.getApplicationName(), is( "GoGui" ) );
      assertThat( info_props.getApplicationVersion(), is( "1.4.2" ) );
      assertThat( info_props.getSize(), is( 9 ) );
      assertThat( info_props.getKomi(), is( BigDecimal.ZERO ) );
      assertThat( info_props.getDate(), is( "2012-02-26" ) );

      Board board0 = new Board();
      board0.setColor( Convert.stringsToPoints( "a1", "a2" ), Colors.WHITE );
      Board board1 = new Board( board0 );
      board1.setColor( Convert.stringsToPoints( "a3", "a4" ), Colors.BLACK );
      Board board2 = new Board( board1 );
      board2.setColor( Convert.stringToPoint( "a9" ), Colors.WHITE );

      m_collector.collect( rec, Integer.MAX_VALUE );
      Iterator<ImmutablePair<Board,Integer>> iter = m_collector.getBoards().iterator();
      ImmutablePair<Board,Integer> pair = iter.next();
      assertThat( "white setup stones", pair.getLeft(), is( board0 ) );
      assertThat( "turn 0", pair.getRight(), is( 0 ) );
      pair = iter.next();
      assertThat( "plus black setup stones", pair.getLeft(), is( board1 ) );
      assertThat( "still turn 0", pair.getRight(), is( 0 ) );
      pair = iter.next();
      assertThat( "white move", pair.getLeft(), is( board2 ) );
      assertThat( "turn 1", pair.getRight(), is( 1 ) );
      ImmutablePair<Board,Integer> last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 0 ) );
      assertNull( m_collector.getHandicapBoard() );

      m_collector.collect( rec, 0 );
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      assertThat( "white setup stones", pair.getLeft(), is( board0 ) );
      assertThat( "turn 0", pair.getRight(), is( 0 ) );
      pair = iter.next();
      assertThat( "plus black setup stones", pair.getLeft(), is( board1 ) );
      assertThat( "still turn 0", pair.getRight(), is( 0 ) );
      last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 0 ) );
      assertNull( m_collector.getHandicapBoard() );

      m_collector.collect( rec, -1 );
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      assertThat( "final board", pair.getLeft(), is( board2 ) );
      assertThat( "turn 1", pair.getRight(), is( 1 ) );
      last = m_collector.getLast();
      assertThat( last.getLeft(), is( pair.getLeft() ) );
      assertThat( last.getRight(), is( pair.getRight() ) );
      assertFalse( iter.hasNext() );
      assertThat( m_collector.getHandicapStoneCount(), is( 0 ) );
      assertNull( m_collector.getHandicapBoard() );
   }

   @SuppressWarnings( "unchecked" )
   @Test
   public void testFile_wing_yuhiko_gnugo() throws IOException
   {
      Record rec = load( "wing-yuhiko-gnugo.sgf", 5 );

      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getGameName(), is( "yuhiko(W) - robot1(B) on WING" ) );
      assertThat( info_props.getSize(), is( 19 ) );
      assertThat( info_props.getDate(), is( "2002-05-01" ) );
      assertThat( info_props.getTimeLimit(), is( 300 ) );
      assertThat( info_props.getHandicap(), is( ImmutablePair.of( Handicaps.UNKNOWN, 2 ) ) );
      StringWriter buffer = new StringWriter();
      BufferedWriter bw = new BufferedWriter( buffer );
      bw.newLine();
      bw.append( "  This game was played on WING" );
      bw.newLine();
      bw.append( "  Permission to reproduce this game is given." );
      bw.close();
      ImmutablePair<String,String> value = (ImmutablePair<String,String>) info_props.get( PropertyID.MAX_STD_ID + 1 );
      assertThat( value.getLeft(), is( "COPYRIGHT" ) );
      assertThat( value.getRight(), is( buffer.toString() ) );
      value = (ImmutablePair<String,String>) info_props.get( PropertyID.MAX_STD_ID + 2 );
      assertThat( value.getLeft(), is( "NW" ) );
      assertThat( value.getRight(), is( "1756 (-9)" ) );
      value = (ImmutablePair<String,String>) info_props.get( PropertyID.MAX_STD_ID + 3 );
      assertThat( value.getLeft(), is( "NB" ) );
      assertThat( value.getRight(), is( "1600 (+14)" ) );

      rec.getMainBranch().visit( m_prop_visitor, true );
      m_prop_visitor.assertValue( PropertyID.MOVE, 5, "BLACK:" + sgfToStd( "qn" ) );
      m_prop_visitor.assertValue( PropertyID.MOVE, 12, "WHITE:" + sgfToStd( "df" ) );
      m_prop_visitor.assertValue( PropertyID.MOVE, 281, "BLACK:pass" );
      m_prop_visitor.assertValue( PropertyID.MOVE, 282, "WHITE:pass" );
      m_prop_visitor.assertValue( PropertyID.MOVE, 283, "BLACK:pass" );
      buffer = new StringWriter();
      bw = new BufferedWriter( buffer );
      bw.newLine();
      bw.append( " robot1 dead @ A18" );
      bw.newLine();
      bw.append( " robot1 dead @ R17" );
      bw.newLine();
      bw.append( " robot1 dead @ L16" );
      bw.newLine();
      bw.append( " robot1 dead @ S3" );
      bw.newLine();
      bw.append( " robot1 dead @ F1" );
      bw.newLine();
      bw.close();
      m_prop_visitor.assertValue( PropertyID.COMMENT, 0, buffer.toString() );
   }

   @Test
   public void testFile_ff4_ex() throws IOException
   {
      List<Record> records = load( "ff4_ex.sgf", 2, -1 );
      assertThat( records.size(), is( 2 ) );

      Record rec = records.get( 0 );
      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getApplicationName(), is( "Primiview" ) );
      assertThat( info_props.getApplicationVersion(), is( "3.1" ) );
      assertThat( info_props.getSize(), is( 19 ) );
      assertThat( info_props.getGameName(), is( "Gametree 1: properties" ) );
      assertThat( info_props.getUserName(), is( "Arno Hollosi" ) );

      rec.visit( m_prop_visitor, false );

      m_prop_visitor.assertValue( PropertyID.GRADE, 3, "(DONT_CARE,(POSITION,EVEN))" );
      m_prop_visitor.assertValue( PropertyID.MOVE, 4, "BLACK:" + sgfToStd( "pj" ) );
      m_prop_visitor.assertValue( PropertyID.COMMENT, 4, "Marked as \"Even position\"" );
      m_prop_visitor.assertValue( PropertyID.MOVE, 5, "WHITE:" + sgfToStd( "ci" ) );
      
      m_prop_visitor.assertValue( PropertyID.MOVE, 11, "WHITE:pass" );
      m_prop_visitor.assertValue( PropertyID.MOVE, 12, "BLACK:pass" );

      // Note: this SGF file is in Unix eol format.
      //
      StringBuilder buffer = new StringBuilder();
      buffer.append( "There are hard linebreaks & soft linebreaks.\n" );
      buffer.append( "Soft linebreaks are linebreaks preceeded by '\\' like this one >ok<. Hard line breaks are all other linebreaks.\n" );
      buffer.append( "Soft linebreaks are converted to >nothing<, i.e. removed.\n" );
      buffer.append( "\n" );
      buffer.append( "Note that linebreaks are coded differently on different systems.\n" );
      buffer.append( "\n" );
      buffer.append( "Examples (>ok< shouldn't be split):\n" );
      buffer.append( "\n" );
      buffer.append( "linebreak 1 \"\\n\": >ok<\n" );
      buffer.append( "linebreak 2 \"\\n\\r\": >o\rk<\n" );  // TODO: the intent of this case might have been that LFCR is a valid eol, but I don't agree
      buffer.append( "linebreak 3 \"\\r\\n\": >ok<\n" );
      buffer.append( "linebreak 4 \"\\r\": >ok<" );
      m_prop_visitor.assertValue( PropertyID.COMMENT, 20, buffer.toString() );

      m_prop_visitor.assertValue( PropertyID.STONES, 0, "BLACK:" + sgfListToStd( "dd][de][df][dg][do:gq" ) );
      m_prop_visitor.assertValue( PropertyID.STONES, 1, "WHITE:" + sgfListToStd( "jd][je][jf][jg][kn:lq][pn:pq" ) );

      m_prop_visitor.clear();
      rec.visit( m_prop_visitor, true );
      m_prop_visitor.assertValue( PropertyID.MOVE, 10, "BLACK:" + sgfToStd( "jj" ) );
      m_prop_visitor.assertValue( PropertyID.MOVE, 9, "WHITE:" + sgfToStd( "de" ) );
      m_prop_visitor.assertValue( PropertyID.MOVE, 8, "BLACK:" + sgfToStd( "gd" ) );
      m_prop_visitor.assertValue( PropertyID.MOVE, 7, "WHITE:" + sgfToStd( "jp" ) );

      Board board = new Board();
      m_collector.collect( rec, Integer.MAX_VALUE );
      Iterator<ImmutablePair<Board,Integer>> iter = m_collector.getBoards().iterator();
      board.setColor( com.dakengo.plugins.record.sgf.Convert.stringToPoint( "pd", 0 ), Colors.BLACK );
      ImmutablePair<Board,Integer> pair = iter.next();
      assertThat( "B[pd]", pair.getLeft(), is( board ) );
      assertThat( pair.getRight(), is( 1 ) );
      board.setColor( com.dakengo.plugins.record.sgf.Convert.stringToPoint( "dp", 0 ), Colors.WHITE );
      pair = iter.next();
      assertThat( "W[dp]", pair.getLeft(), is( board ) );
      assertThat( pair.getRight(), is( 2 ) );
      board.setColor( com.dakengo.plugins.record.sgf.Convert.stringToPoint( "pp", 0 ), Colors.BLACK );
      pair = iter.next();
      assertThat( "B[pp]", pair.getLeft(), is( board ) );
      assertThat( pair.getRight(), is( 3 ) );
      board.setColor( com.dakengo.plugins.record.sgf.Convert.stringToPoint( "dc", 0 ), Colors.WHITE );
      pair = iter.next();
      assertThat( "W[dc]", pair.getLeft(), is( board ) );
      assertThat( pair.getRight(), is( 4 ) );

      rec = records.get( 1 );
      info_props = rec.getGameProps();
      assertThat( info_props.getApplicationName(), is( "Primiview" ) );
      assertThat( info_props.getApplicationVersion(), is( "3.1" ) );
      assertThat( info_props.getSize(), is( 19 ) );

      // FIXME: further validation of the 2nd game record deferred until we
      // support multiple games stored in a single game tree.
      //
/*
      assertThat( info_props.getPlayerName( Colors.WHITE ), is( "W. Hite" ) );
      assertThat( info_props.getPlayerName( Colors.BLACK ), is( "B. Lack" ) );
 */
   }

   @Test
   public void testFile_jinen_pachi30s_2() throws IOException
   {
      Record rec = load( "jinen-pachi30s-2.sgf", 0 );

      Board board = new Board();
      board.setColor( Convert.stringsToPoints( "d4", "d16", "q16" ), Colors.BLACK );

      m_collector.setKomi( rec.getGameProps().getKomi() );
      m_collector.collect( rec, 0 );
      assertThat( m_collector.getGame().getKomi(), is( m_collector.getKomi() ) );
      assertThat( m_collector.getHandicapBoard(), is( board ) );
      assertThat( m_collector.getHandicapStoneCount(), is( 3 ) );
      Iterator<ImmutablePair<Board,Integer>> iter = m_collector.getBoards().iterator();
      ImmutablePair<Board,Integer> pair = iter.next();
      assertThat( pair.getLeft(), is( board ) );
      assertFalse( iter.hasNext() );

      m_collector.collect( rec, 1 );
      assertThat( m_collector.getHandicapBoard(), is( board ) );
      assertThat( m_collector.getHandicapStoneCount(), is( 3 ) );
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      assertThat( pair.getLeft(), is( board ) );
      Board next_board = new Board( board );
      next_board.setColor( Convert.stringToPoint( "r4", 0 ), Colors.WHITE );
      pair = iter.next();
      assertThat( pair.getLeft(), is( next_board ) );
      assertFalse( iter.hasNext() );

      m_collector.collect( rec, Integer.MAX_VALUE );
      assertThat( m_collector.getHandicapBoard(), is( board ) );
      assertThat( m_collector.getHandicapStoneCount(), is( 3 ) );
      int count = 0;
      iter = m_collector.getBoards().iterator();
      pair = iter.next();
      count++;
      assertThat( pair.getLeft(), is( board ) );
      next_board = new Board( board );
      next_board.setColor( Convert.stringToPoint( "r4", 0 ), Colors.WHITE );
      pair = iter.next();
      count++;
      assertThat( pair.getLeft(), is( next_board ) );
      while (iter.hasNext())
      {
         iter.next();
         count++;
      }
      assertThat( count, is( 254 ) );

      board = m_collector.getLast().getLeft();
      Signature sig1 = buildExpectedSig( new String[] { "n8", "l6", "h9", "b16", "g16", "q18" }, Colors.BLACK, board );
      Signature sig2 = SignatureBuilder.compute( rec );
      assertThat( sig2, is( sig1 ) );
      assertThat( sig2.getTurn(), is( Signature.TURN_NUMBERS[Signature.TURN_NUMBERS.length - 1] ) );
   }

   @Test
   public void testFile_kage_5_stone_1() throws IOException
   {
      Record rec = load( "kage-5-stone-1.sgf", 0 );

      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getSize(), is( 19 ) );
      assertThat( info_props.getPlayerName( Colors.BLACK ), is( "Kageyama Toshiro" ) );
      assertThat( info_props.getPlayerName( Colors.WHITE ), is( "Yokoyama Koichi" ) );
      assertThat( info_props.getKomi(), is( BigDecimal.ZERO ) );
      assertThat( info_props.getGameName(), is( "KAGE-5-STONE-1" ) );
      assertThat( info_props.getEventName(), is( "Kage's Handicap Go, 5 stone game 1" ) );

      Board board = new Board();
      board.setColor( Convert.stringsToPoints( "d4", "d16", "q16", "q4", "k10" ), Colors.BLACK );

      m_collector.setKomi( rec.getGameProps().getKomi() );
      m_collector.collect( rec, Integer.MAX_VALUE );
      assertThat( m_collector.getGame().getKomi(), is( m_collector.getKomi() ) );
      assertThat( m_collector.getHandicapBoard(), is( board ) );
      assertThat( m_collector.getHandicapStoneCount(), is( 5 ) );

      board = m_collector.getLast().getLeft();
      Signature sig1 = buildExpectedSig( new String[] { "r5", "q12", "r13", "f3", "h6", "m6" }, Colors.DONT_CARE, board );  // was BLACK but r5 gets captured early in the game
      Signature sig2 = SignatureBuilder.compute( rec );
      assertThat( sig2, is( sig1 ) );
      assertThat( sig2.getTurn(), is( Signature.TURN_NUMBERS[Signature.TURN_NUMBERS.length - 1] ) );
      assertThat( com.dakengo.plugins.record.sgf.Convert.signatureToString( sig2 ), is( "qoqghnphfqln" ) );
      assertThat( com.dakengo.plugins.record.sgf.Convert.stringToSignature( "qoqghnphfqln", 0 ), is( sig2 ) );
   }

   @Test
   public void testFile_1970_rin_kajiwara() throws IOException
   {
      Record rec = load( "1970-rin-kajiwara.sgf", 4 );
      m_collector.setKomi( rec.getGameProps().getKomi() );
      m_collector.collect( rec, Integer.MAX_VALUE );
      Board board = m_collector.getLast().getLeft();
      Signature sig1 = buildExpectedSig( new String[] { "r3", "s4", "p8", "j4" }, Colors.WHITE, board );
      Signature sig2 = SignatureBuilder.compute( rec );
      assertThat( sig2, is( sig1 ) );
      assertThat( sig2.getTurn(), is( Signature.TURN_NUMBERS[3] ) );
      assertThat( com.dakengo.plugins.record.sgf.Convert.signatureToString( sig2 ), is( "qqol??rpip??" ) );
      assertThat( com.dakengo.plugins.record.sgf.Convert.stringToSignature( "qqol??rpip??", 0 ), is( sig2 ) );
   }

   @Test
   public void testFile_8623_emerus_SimonN_Eloriot() throws IOException
   {
      Record rec = load( "8623-emerus-SimonN-Eloriot.sgf", 0 );

      GameProperties info_props = rec.getGameProps();
      assertThat( info_props.getHandicap(), is( ImmutablePair.of( Handicaps.UNKNOWN, 2 ) ) );
      assertThat( info_props.getOvertimeSystem(), is( "5x30 byo-yomi" ) );
      assertThat( info_props.getPlaceName(), is( "The KGS Go Server at http://www.gokgs.com/" ) );
      Result result = new Result( Colors.BLACK, Result.Status.WON, new BigDecimal( "5.5" ) );
      assertThat( info_props.getResult(), is( result ) );

      rec.visit( m_prop_visitor, false );
      m_prop_visitor.assertValue( PropertyID.MARK, 0, "TRIANGLE:" + sgfToStd( "mq" ) );
      m_prop_visitor.assertValue( PropertyID.MARK, 1, "TRIANGLE:" + sgfToStd( "fl" ) );

      Board board = new Board();
      board.setColor( Convert.stringsToPoints( "d4", "q16" ), Colors.BLACK );

      m_collector.setKomi( info_props.getKomi() );
      m_collector.collect( rec, Integer.MAX_VALUE );
      assertThat( m_collector.getHandicapStoneCount(), is( 2 ) );
      assertThat( m_collector.getHandicapBoard(), is( board ) );
   }

   // "incident258" refers to a GNU Go issue, not a Dakengo issue
   @Test
   public void testFile_incident258() throws IOException
   {
      Record rec = load( "incident258.sgf", 1 );
      assertThat( rec.getGameProps().getKomi(), is( new BigDecimal( "-4.5" ) ) );
   }

   @Test
   public void testFile_yi203() throws IOException
   {
      Record rec = load( "yi203.sgf", 0 );

      rec.visit( m_prop_visitor, false );
      m_prop_visitor.assertValue( PropertyID.EVENT_NAME, 0, "Event: Preliminary round of Ing Cup." );
      m_prop_visitor.assertValue( PropertyID.MOVE, 4, "BLACK:d5" );
      m_prop_visitor.assertValue( PropertyID.MOVE, 5, "WHITE:c5" );

      m_collector.setKomi( rec.getGameProps().getKomi() );
      m_collector.collect( rec, Integer.MAX_VALUE );
      assertThat( m_collector.getGame().getKomi(), is( m_collector.getKomi() ) );
      assertNull( m_collector.getHandicapBoard() );
      assertThat( m_collector.getHandicapStoneCount(), is( 0 ) );

      Board board = m_collector.getLast().getLeft();
      Signature sig1 = buildExpectedSig( new String[] { "s17", "b14", "r6", "p5", "o9", "k8" }, Colors.WHITE, board );
      Signature sig2 = SignatureBuilder.compute( rec );
      assertThat( sig2, is( sig1 ) );
      assertThat( sig2.getTurn(), is( Signature.TURN_NUMBERS[Signature.TURN_NUMBERS.length - 1] ) );
      assertThat( com.dakengo.plugins.record.sgf.Convert.signatureToString( sig2 ), is( "rcqnnkbfoojl" ) );
      assertThat( com.dakengo.plugins.record.sgf.Convert.stringToSignature( "rcqnnkbfoojl", 0 ), is( sig2 ) );
   }

   @Test
   public void testFromScratch() throws IOException
   {
      Board.setSize( 19 );
      Record rec = new Record();
      GameProperties info_props = rec.getGameProps();
      info_props.setApplicationName( "unittests" );
      info_props.setSize( 19 );
      info_props.setKomi( new BigDecimal( "6.5" ) );
      info_props.setPlayerName( Colors.BLACK, "black player" );
      info_props.setPlayerName( Colors.WHITE, "white player" );
      PositionProperties main_branch = rec.getMainBranch();
      String test_text1 = "test ][)(;";
      main_branch.addComment( test_text1 );
      main_branch.addMove( Colors.BLACK, (short) 10 );

      PositionProperties branch = new PositionProperties( false );
      branch.addMove( Colors.WHITE, (short) 11 );
      String test_text2 = "abcdef";
      branch.addComment( test_text2 );
      branch.addNote( 1, "NODE", "" );
      branch.addMove( Colors.BLACK, (short) 12 );
      branch.addNote( 2, "NODE", "" );
      branch.addMove( Colors.WHITE, (short) 13 );
      branch.addNote( 3, "NODE", "" );
      branch.addMove( Colors.BLACK, (short) 14 );
      branch.addNote( 4, "NODE", "" );
      branch.addMove( Colors.WHITE, Board.INVALID_POINT );
      main_branch.addBranch( branch );

      Record rec2 = null;
      File sgf_file = File.createTempFile( "unittests", ".sgf" );
      sgf_file.deleteOnExit();
      try (FileWriter output = new FileWriter( sgf_file ))
      {
         m_writer.setOutput( new BufferedWriter( output ) );
         m_writer.write( rec );
      }
      try (FileReader input = new FileReader( sgf_file ))
      {
         m_reader.setInput( input );
         rec2 = m_reader.read();
         assertNotNull( rec2 );
         assertNull( m_reader.read() );
      }

      rec2.visit( m_prop_visitor, true );
      m_prop_visitor.assertValue( PropertyID.APP_NAME, 0, "unittests" );
      m_prop_visitor.assertValue( PropertyID.SIZE, 0, "19" );
      m_prop_visitor.assertValue( PropertyID.KOMI, 0, "6.5" );
      m_prop_visitor.assertValues( PropertyID.PLAYER_NAME, "BLACK:black player", "WHITE:white player" );
      m_prop_visitor.assertValues( PropertyID.COMMENT, test_text1, test_text2 );
   }

   @Before
   public void setUp()
   {
      initReader( "sgf" );
      initWriter( "sgf" );
      m_prop_visitor.clear();
   }

   private <T> void runExceptionCases( Class<T> ex_class, String[] test_cases )
   {
      final RecordReader reader = m_reader; 
      for (String text : test_cases)
      {
         reader.setInput( new StringReader( text ) );
         assertThrows(
            String.format( "'%s'", text ),
            new TestRunnable() { @Override public void runTest() throws Exception { reader.read(); } },
            ex_class, null );
      }
   }

   private void runMessageCases( int msg_count, int branch_count, int prop_count, String[] test_cases ) throws IOException
   {
      TestPropVisitor visitor = new TestPropVisitor( false );
      for (String text : test_cases)
      {
         visitor.clear();
         String msg = String.format( "input = '%s'", text );
         m_reader.setInput( new StringReader( text ) );
         Record rec = m_reader.read();
         rec.visit( visitor, false );
         assertThat( msg, m_reader.getMessages().size(), is( msg_count ) );
         visitor.assertCounts( branch_count, prop_count );
      }
   }

   private String sgfListToStd( String sgf_list )
   {
      return Convert.pointsToString( new TreeSet<Short>( com.dakengo.plugins.record.sgf.Convert.stringToPoints( sgf_list, 0 ) ), "," );
   }

   private String sgfToStd( String sgf_pnt )
   {
      return Convert.pointToString( com.dakengo.plugins.record.sgf.Convert.stringToPoint( sgf_pnt, 0 ) );
   }

   private Signature buildExpectedSig( String[] pnt_str, Colors first_color, Board last_board )
   {
      Signature sig = new Signature();
      Colors color = first_color;
      for (int i = 0; i < pnt_str.length; i++)
      {
         Short pnt = Convert.stringToPoint( pnt_str[i], 0 );
         if (first_color == Colors.BLACK || first_color == Colors.WHITE)
         {
            assertThat( String.format( "element %d point %s", i, pnt_str[i] ), last_board.getColor( pnt ), is( color ) );
            color = Game.otherPlayer( color );
         }
         sig.addPoint( pnt );
      }
      return sig;
   }
}
