//
// AssertUtils.java
//
// Copyright (c) 2012-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.utils;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

/**
 * This class provides assertion utility methods for implementing unit tests.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public abstract class AssertUtils
{
   /**
    * This method is a substitute for the existing JUnit exception assertion facilities
    * which allows multiple test cases to be contained within a single test method.
    * 
    * @param action     supplies the code sequence that should throw an exception
    * @param ex_class   expected exception type
    * @param ex_msg     expected exception message, pass null if the exception message
    *                   does not need to be asserted 
    */
   public static void assertThrows( Runnable action, Class<?> ex_class, String ex_msg )
   {
      assertThrows( "", action, ex_class, ex_msg ); //$NON-NLS-1$
   }

   /**
    * This method is a substitute for the existing JUnit exception assertion facilities
    * which allows multiple test cases to be contained within a single test method.
    * 
    * @param details    text that further describes the test case
    * @param action     supplies the code sequence that should throw an exception
    * @param ex_class   expected exception type
    * @param ex_msg     expected exception message, pass null if the exception message
    *                   does not need to be asserted 
    */
   public static void assertThrows( String details, Runnable action, Class<?> ex_class, String ex_msg )
   {
      boolean thrown = false;
      String colon = (details != null && !details.isEmpty()) ? ": " : ""; //$NON-NLS-1$ //$NON-NLS-2$
      String err_msg = String.format( "%s%sexpected exception type not thrown", details, colon );
      String ex_classname = ex_class.getName();
      String actual_classname = ""; //$NON-NLS-1$
      String actual_msg = ""; //$NON-NLS-1$
      try
      {
         action.run();
      }
      catch (ForwardedException ex)
      {
         actual_classname = ex.getCause().getClass().getName();
         actual_msg = ex.getCause().getMessage();
         thrown = true;
      }
      catch (Throwable t)
      {
         actual_classname = t.getClass().getName();
         actual_msg = t.getMessage();
         thrown = true;
      }
      if (thrown)
      {
         assertThat( err_msg, actual_classname, is( ex_classname ) );
         err_msg = String.format( "%s%sexception details", details, colon ); 
         if (ex_msg != null)
            assertThat( err_msg, actual_msg, is( ex_msg ) );
      }
      err_msg = String.format( "%s%sexception type %s was not thrown", details, colon, ex_classname );
      assertThat( err_msg, thrown, is( true ) );
   }
}
