//
// TestPropVisitor.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.tests.utils;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.application.TextUtils;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.record.Notations;
import com.dakengo.libs.record.PositionProperties;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.libs.record.PropertyVisitor;

public class TestPropVisitor implements PropertyVisitor
{
   public TestPropVisitor( boolean descend )
   {
      m_descend = descend;
   }

   public void assertCounts( int branches, int props )
   {
      assertThat( "branches", m_branch_count, is( branches ) );
      int count = 0;
      StringBuilder buffer = new StringBuilder();
      for (int prop_id : m_props.keySet())
      {
         List<String> values = m_props.get( prop_id );
         if (buffer.length() > 0)
            buffer.append( "," );
         buffer.append( "(" );
         buffer.append( prop_id );
         buffer.append( ":" );
         buffer.append( TextUtils.collectionToString( values, ",", false, null ) );
         buffer.append( ")" );
         count += values.size();
      }
      assertThat( "actual properties = " + buffer.toString(), count, is( props ) );
   }

   public void assertValue( int prop_id, int index, String expected )
   {
      String msg = "property ID: " + prop_id;
      List<String> actual = null;
      assertThat( msg, m_props.containsKey( prop_id ), is( true ) );
      actual = m_props.get( prop_id );
      assertThat( msg, actual.size(), is( greaterThan( index ) ) );
      assertThat( msg, actual.get( index ), is( expected ) );
   }

   public void assertValues( int prop_id, String... expected )
   {
      assertThat( "property ID: " + prop_id, m_props.containsKey( prop_id ), is( true ) );
      List<String> actual = m_props.get( prop_id );
      String msg = "property ID: " + prop_id;
      assertThat( msg, actual.size(), is( expected.length ) );
      int i = 0;
      for (String str : actual)
         assertThat( msg, str, is( expected[i++] ) );
   }

   public void assertValues( int prop_id, Collection<String> expected )
   {
      assertThat( "property ID: " + prop_id, m_props.containsKey( prop_id ), is( true ) );
      List<String> actual = m_props.get( prop_id );
      assertThat( "property ID: " + prop_id, actual, is( expected ) );
   }

   public void clear()
   {
      m_branch_count = 0;
      m_props.clear();
   }

   public int getBranchCount()
   {
      return m_branch_count;
   }

   public int getPropertyCount()
   {
      return m_props.size();
   }

   @Override
   public boolean startBranch( PositionProperties branch )
   {
      if (m_descend)
         m_branch_count++;
      return m_descend;
   }

   @Override
   public void finishBranch( PositionProperties branch )
   {
      // empty
   }

   @Override
   @SuppressWarnings( "unchecked" )
   public void accept( int prop_id, Object value )
   {
      String text = "";
      if (prop_id > PropertyID.MAX_STD_ID)
      {
         ImmutablePair<String,String> pair = (ImmutablePair<String,String>) value;
         text = pair.getLeft() + ":" + pair.getRight();
      }
      else
      {
         switch (prop_id)
         {
            case PropertyID.MARK:
               ImmutablePair<Notations,Set<Short>> mark_value = (ImmutablePair<Notations,Set<Short>>) value;
               text = String.format( "%s:%s", mark_value.getLeft(), Convert.pointsToString( new TreeSet<Short>( mark_value.getRight() ), "," ) );
               break;
            case PropertyID.MOVE:
               ImmutablePair<Colors,Short> move_value = (ImmutablePair<Colors,Short>) value;
               text = String.format( "%s:%s", move_value.getLeft(), Convert.pointToString( move_value.getRight() ) );
               break;
            case PropertyID.PLAYER_NAME:
               ImmutablePair<Colors,String> name_value = (ImmutablePair<Colors,String>) value;
               text = String.format( "%s:%s", name_value.getLeft(), name_value.getRight() );
               break;
            case PropertyID.STONES:
            case PropertyID.TERRITORY:
               ImmutablePair<Colors,Set<Short>> terr_value = (ImmutablePair<Colors,Set<Short>>) value;
               text = String.format( "%s:%s", terr_value.getLeft(), Convert.pointsToString( new TreeSet<Short>( terr_value.getRight() ), "," ) );
               break;
            default:
               text = value.toString();
               break;
         }
      }
      List<String> values = null;
      if (!m_props.containsKey( prop_id ))
      {
         values = new ArrayList<String>();
         m_props.put( prop_id, values );
      }
      else
      {
         values = m_props.get( prop_id );
      }
      values.add( text );
   }

   private boolean                   m_descend      = false;
   private int                       m_branch_count = 0;
   private Map<Integer,List<String>> m_props        = new HashMap<Integer,List<String>>();
}
