//
// Program.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.apps.bot;

import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.util.Collection;

import net.xeoh.plugins.base.PluginManager;

import com.dakengo.libs.application.CommandLineParser;
import com.dakengo.libs.application.ConfigReader;
import com.dakengo.libs.application.Configuration;
import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.application.GtpEngine;
import com.dakengo.libs.application.GtpRecordStore;
import com.dakengo.libs.application.GtpReplyHandler;
import com.dakengo.libs.application.PluginUtils;
import com.dakengo.libs.record.RecordWriter;

/**
 * An implementation of a bot program which connects computer players
 * with the outside world via standard protocols such as GTP.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class Program
{
   public static final String GTP_CMD      = "gtp"; //$NON-NLS-1$

   public static final char   CONFIG_OPT   = 'c';
   public static final char   OPPONENT_OPT = 'o';
   public static final char   PLAYER_OPT   = 'p';

   /**
    * @param args
    */
   public static void main( String[] args )
   {
      int exit_code = 1;
      try
      {
         Program prog = new Program();
         exit_code = prog.run( args );
      }
      catch (Throwable t)
      {
/*
         System.err.println( "dkbot: " + t.getMessage() );
*/
         t.printStackTrace();
         System.err.println();
      }
      System.exit( exit_code );
   }

   private Program()
   {
   }

   private int run( String[] args ) throws IOException
   {
      CommandLineParser cmdline = parseArgs( args );
      if (cmdline == null)
         return 1;
      Configuration config = loadConfig( cmdline );
      if (config == null)
         return 1;
      if (cmdline.optionSet( OPPONENT_OPT ))
         config.setOpponentName( cmdline.getOptionValue( OPPONENT_OPT ) );

      PluginManager plugins = PluginUtils.createManager( config );
      String player_type = cmdline.getOptionValue( PLAYER_OPT );
      final String game_id = "dkbot"; //$NON-NLS-1$
      GameContext context = new GameContext( config, plugins, player_type, game_id );

      RecordWriter writer = plugins.getPlugin( RecordWriter.class, PluginUtils.getRecordCapabilities( Configuration.DEFAULT_RECORD_TYPE, false ) );
      if (writer == null)
      {
         System.err.println( Resources.formatString( "msg.7014", Configuration.DEFAULT_RECORD_TYPE ) ); //$NON-NLS-1$
         return 1;
      }
      GtpReplyHandler handler = new GtpOutputHandler();
      GtpRecordStore store = new FileStore( config, writer );

      int exit_status = 1;
      switch (cmdline.getCommand())
      {
         case GTP_CMD:
            GtpEngine engine = new GtpEngine( plugins, handler, store );
            engine.loop( new InputStreamReader( System.in ), context );
            break;
         default:
            printUsage();
            break;
      }
      return exit_status;
   }

   private CommandLineParser parseArgs( String[] args )
   {
      if (args.length == 0)
      {
         printVersion();
         printUsage();
         return null;
      }
      CommandLineParser cmdline = new CommandLineParser();
      cmdline.addCommand( GTP_CMD );
      cmdline.addOption( CONFIG_OPT, true );
      cmdline.addOption( OPPONENT_OPT, true );
      cmdline.addOption( PLAYER_OPT, true );
      cmdline.process( args );
      Collection<String> msgs = cmdline.getMessages();
      for (String msg : msgs)
         System.err.println( "dkbot: " + msg ); //$NON-NLS-1$
      if (!msgs.isEmpty())
      {
         System.err.println();
         return null;
      }
      if (!cmdline.getCommand().equals( GTP_CMD ) || !cmdline.getRemainingArgs().isEmpty() || !cmdline.optionSet( PLAYER_OPT ))
      {
         printUsage();
         return null;
      }
      return cmdline;
   }

   private Configuration loadConfig( CommandLineParser cmdline )
   {
      if (!cmdline.optionSet( CONFIG_OPT ))
         return new Configuration();
      Configuration config = null;
      try (FileReader reader = new FileReader( cmdline.getOptionValue( CONFIG_OPT )))
      {
         config = ConfigReader.read( reader );
      }
      catch (IOException ex)
      {
         System.err.println( Resources.formatString( "msg.7001", ex.getMessage() ) ); //$NON-NLS-1$
         return null;
      }
      return config;
   }

   private static void printVersion()
   {
      System.err.println( Resources.getString( "msg.7002" ) );            //$NON-NLS-1$
      System.err.println( Resources.getString( "msg.7003" ) );            //$NON-NLS-1$
      System.err.println();
      System.err.println( Resources.getString( "msg.7004" ) );            //$NON-NLS-1$
      System.err.println( "http://www.apache.org/licenses/LICENSE-2.0" ); //$NON-NLS-1$
      System.err.println();
   }

   private void printUsage()
   {
      String[] lines = new String[]
      {
         Resources.getString( "msg.7005" ), //$NON-NLS-1$
         "",                                //$NON-NLS-1$
         Resources.getString( "msg.7006" ), //$NON-NLS-1$
         "",                                //$NON-NLS-1$
         Resources.getString( "msg.7007" ), //$NON-NLS-1$
         "",                                //$NON-NLS-1$
         Resources.getString( "msg.7008" ), //$NON-NLS-1$
         "",                                //$NON-NLS-1$
         Resources.getString( "msg.7009" ), //$NON-NLS-1$
         Resources.getString( "msg.7015" ), //$NON-NLS-1$
         Resources.getString( "msg.7010" )  //$NON-NLS-1$
      };
      for (String line : lines)
         System.err.println( line );
      System.err.println();
   }
}
