//
// BaseGtpCommandHandler.java
//
// Copyright (c) 2012-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.io.IOException;

import net.xeoh.plugins.base.annotations.Capabilities;
import net.xeoh.plugins.base.annotations.events.Init;
import net.xeoh.plugins.base.annotations.injections.InjectPlugin;
import net.xeoh.plugins.informationbroker.InformationBroker;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

import com.dakengo.libs.game.Board;

/**
 * A base class for GTP command plugins.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public abstract class BaseGtpCommandHandler implements GtpCommandHandler
{
   public boolean shouldSave()
   {
      return m_should_save;
   }

   @Override
   public void handleCommand( GtpCommand cmd, GameContext context, GtpReplyHandler callback )
   {
      try
      {
         doCommand( cmd, context, callback );
      }
      catch (AssertionError e)
      {
         throw e;
      }
      catch (Throwable t)
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2037" ) ); //$NON-NLS-1$
         logThrowable( t );
      }
   }

   /**
    * Publishes the name of this command.
    * 
    * @see net.xeoh.plugins.informationbroker.InformationBroker
    */
   @Init
   public void init()
   {
      PluginUtils.publishCommand( m_broker, m_name );
   }

   /**
    * Returns the capabilities for this plugin.
    */
   @Capabilities
   public String[] getCapabilities()
   {
      return new String[] { Constants.GTP_CAP_PREFIX + m_name };
   }

   protected abstract void doCommand( GtpCommand cmd, GameContext context, GtpReplyHandler callback );

   protected BaseGtpCommandHandler( String name, boolean should_save )
   {
      m_name = name;
      m_should_save = should_save;
   }

   protected void setPlayGame( GtpCommand cmd, GameContext context, GtpReplyHandler callback )
   {
      context.setPlayGame();
      try
      {
         context.loadLayers( Board.getSize() );
      }
      catch (IOException ex)
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2037" ) ); //$NON-NLS-1$
         logThrowable( ex );
         return;
      }
   }

   protected void logDebug( String msg )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      String str = Resources.getString( "msg.2035" ); //$NON-NLS-1$
      log.debug( String.format( "%s %s: %s", m_name, str, msg ) ); //$NON-NLS-1$
   }

   protected void logError( String msg )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      String str = Resources.getString( "msg.2035" ); //$NON-NLS-1$
      log.error( String.format( "%s %s: %s", m_name, str, msg ) ); //$NON-NLS-1$
   }

   protected void logInfo( String msg )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      log.info( msg );
   }

   protected void logWarning( String msg )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      String str = Resources.getString( "msg.2035" ); //$NON-NLS-1$
      log.warn( String.format( "%s %s: %s", m_name, str, msg ) ); //$NON-NLS-1$
   }

   protected void logThrowable( Throwable t )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      String str = Resources.getString( "msg.2042" ); //$NON-NLS-1$
      log.error( String.format( "%s: %s", str, m_name ), t ); //$NON-NLS-1$
   }

   protected void reportInvalidNumArgs( GtpCommand cmd, GtpReplyHandler callback )
   {
      callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2044" ) ); //$NON-NLS-1$
      logError( Resources.getString( "msg.2040" ) ); //$NON-NLS-1$
   }

   @InjectPlugin
   public InformationBroker m_broker;  // jspf injection requires public access specifier

   private String  m_name        = ""; //$NON-NLS-1$
   private boolean m_should_save = false;
}
