//
// CommandLineParser.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * A class for parsing a program's command-line arguments. A command-line may be
 * composed of the following elements:<p>
 * <ul>
 * <li>a command string; must be the first argument if specified
 * <li>zero or more options, each comprised of a leading dash and a character,
 * such as {@code -a}
 * <li>zero or more options, each comprised of a leading dash and a character followed
 * immediately by a space and the option's value, such as {@code -a foobar}
 * <li>anything else on the command-line, including arguments that look like options
 * but which are not declared in the parser's configuration
 * </ul>
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class CommandLineParser
{
   /**
    * Returns the command argument, or an empty string if none was specified.
    */
   public String getCommand()
   {
      return m_command;
   }

   /**
    * Returns the list of messages produced while the command-line arguments
    * were being parsed.
    */
   public Collection<String> getMessages()
   {
      return m_messages;
   }

   /**
    * Remaining command-line arguments that did not match any command or option.
    */
   public Collection<String> getRemainingArgs()
   {
      return m_remaining;
   }

   /**
    * Adds a string to be recognized as a command. The specified string should
    * not contain embedded whitespace.
    * 
    * @param cmd  the command to be added
    */
   public void addCommand( String cmd )
   {
      cmd = cmd.trim();
      if (cmd.length() == 0)
         throw new IllegalArgumentException( Resources.getString( "msg.2027" ) ); //$NON-NLS-1$
      m_commands.add( cmd );
   }

   /**
    * Adds a character identifying a program option. Set needs_value to true
    * if the option must have a subsequent argument which provides the option's
    * value.
    * 
    * @param opt_char     the option char to add
    * @param needs_value  true if the option requires an argument
    */
   public void addOption( char opt_char, boolean needs_value )
   {
      if (!Character.isLetterOrDigit( opt_char ))
         throw new IllegalArgumentException( Resources.getString( "msg.2029" ) ); //$NON-NLS-1$
      m_option_types.put( opt_char, needs_value );
   }

   /**
    * Traverse the argument list, attempting to match command and option strings,
    * saving any unmatched arguments. Warning messages generated during argument
    * parsing are collected for later retrieval.
    * 
    * @param args  the argument list
    */
   public void process( String[] args )
   {
      m_command = ""; //$NON-NLS-1$
      m_remaining.clear();
      m_messages.clear();
      m_option_values.clear();

      char opt_char = '\0';
      boolean opt_needs_val = false;
      for (int i = 0; i < args.length; i++)
      {
         String arg = args[i];
         if (i == 0 && m_commands.contains( arg ))
         {
            m_command = arg;
            continue;
         }
         if (arg.length() == 2 && arg.charAt( 0 ) == '-')
         {
            if (opt_needs_val)
            {
               addValueMessage( opt_char );
               opt_char = '\0';
               opt_needs_val = false;
            }
            Boolean bool_val = m_option_types.get( arg.charAt( 1 ) );
            if (bool_val != null)
            {
               opt_needs_val = bool_val.booleanValue();
               opt_char = arg.charAt( 1 );
               if (!opt_needs_val)
                  m_option_values.put( opt_char, "" ); //$NON-NLS-1$
            }
            else
            {
               m_remaining.add( arg );
            }
         }
         else if (opt_needs_val && arg.charAt( 0 ) != '-')
         {
            m_option_values.put( opt_char, arg );
            opt_char = '\0';
            opt_needs_val = false;
         }
         else
         {
            if (opt_needs_val)
               addValueMessage( opt_char );
            m_remaining.add( arg );
            opt_char = '\0';
            opt_needs_val = false;
         }
      }
      if (opt_needs_val)
         addValueMessage( opt_char );
   }

   /**
    * Does the specified option require the next argument to be a non-option
    * value?
    * 
    * @param opt  the option to test for value requirement
    */
   public boolean optionRequiresValue( char opt )
   {
      Boolean flag = m_option_types.get( opt );
      if (flag == null)
         throw new IllegalArgumentException( Resources.formatString( "msg.2001", opt ) ); //$NON-NLS-1$
      return flag.booleanValue();
   }

   /**
    * Was the specified option found on the command-line?
    * 
    * @param opt  the option to test if set
    */
   public boolean optionSet( char opt )
   {
      return m_option_values.containsKey( opt );
   }

   /**
    * Returns the option's command-line value.
    * 
    * @param opt  the option to retrieve
    */
   public String getOptionValue( char opt )
   {
      String value = m_option_values.get( opt );
      if (value == null)
         throw new IllegalArgumentException( Resources.formatString( "msg.2002", opt ) ); //$NON-NLS-1$
      return value;
   }

   private void addValueMessage( char opt )
   {
      m_messages.add( Resources.formatString( "msg.2003", opt ) ); //$NON-NLS-1$
   }

   private Map<Character,Boolean> m_option_types  = new HashMap<Character,Boolean>();
   private Map<Character,String>  m_option_values = new HashMap<Character,String>();
   private Set<String>            m_commands      = new HashSet<String>();
   private List<String>           m_messages      = new ArrayList<String>();
   private List<String>           m_remaining     = new ArrayList<String>();
   private String                 m_command       = ""; //$NON-NLS-1$
}
