//
// ConfigReader.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.io.IOException;
import java.io.Reader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.google.gson.JsonElement;
import com.google.gson.JsonParseException;
import com.google.gson.JsonParser;
import com.google.gson.stream.JsonReader;

import com.dakengo.libs.base.ParseException;

/**
 * Call {@link #read} to deserialize a {@link Configuration} object from
 * an external source.<p>
 * 
 * See {@code etc/sample-config.json} for documentation of the configuration
 * file format, including all supported properties and variables
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see ConfigVariableValueProvider
 */
public abstract class ConfigReader
{
   /**
    * Parses the input to produce a new Configuration object.
    * 
    * @param input
    * 
    * @throws com.dakengo.libs.base.ParseException
    */
   public static Configuration read( Reader input ) throws IOException
   {
      Configuration config = new Configuration();
      Set<String> names = new HashSet<String>();
      try
      {
         JsonReader jreader = new JsonReader( input );
         jreader.setLenient( true ); // so we can include comments in config files
         jreader.beginObject();
         while (jreader.hasNext())
         {
            String prop_name = jreader.nextName();
            switch (prop_name)
            {
               case Configuration.DATE_FORMAT_VAR:
                  config.setDateFormat( jreader.nextString() );
                  break;
               case Configuration.ENGINE_NAME_VAR:
                  config.setEngineName( jreader.nextString() );
                  break;
               case Configuration.ENGINE_VERSION_VAR:
                  config.setEngineVersion( jreader.nextString() );
                  break;
               case Configuration.FILENAME_FORMAT_VAR:
                  config.setFilenameFormat( jreader.nextString() );
                  break;
               case Configuration.KO_RULE_VAR:
                  config.setKoRule( jreader.nextString() );
                  break;
               case Configuration.KOMI_VAR:
                  config.setKomi( jreader.nextString() );
                  break;
               case Configuration.LAYERS_FILE_VAR:
                  config.setLayersFile( jreader.nextString() );
                  break;
               case Configuration.OUTPUT_DIR_VAR:
                  config.setOutputDirectory( jreader.nextString() );
                  break;
               case PLAYER_DATA_KEYWORD:
                  readPlayerData( jreader, config );
                  break;
               case PLUGIN_URIS_KEYWORD:
                  jreader.beginArray();
                  while (jreader.hasNext())
                     config.addPluginURI( jreader.nextString() );
                  jreader.endArray();
                  break;
               case Configuration.SCORING_VAR:
                  config.setScoring( jreader.nextString() );
                  break;
               case Configuration.SUICIDE_VAR:
                  config.setAllowSuicide( jreader.nextString() );
                  break;
               default:
                  throw new ParseException( Resources.getString( "msg.2006" ), prop_name ); //$NON-NLS-1$
            }
            if (names.contains( prop_name ))
               throw new ParseException( Resources.getString( "msg.2007" ), prop_name ); //$NON-NLS-1$
            names.add( prop_name );
         }
         jreader.endObject();
      }
      catch (IllegalStateException ex)
      {
         throw new ParseException( ex.getMessage(), ex );
      }
      catch (JsonParseException ex)
      {
         throw new ParseException( ex.getMessage(), ex );
      }

      return config;
   }

   private static void readPlayerData( JsonReader jreader, Configuration config ) throws IOException
   {
      String player_type = ""; //$NON-NLS-1$
      JsonParser parser = new JsonParser();
      Map<String,JsonElement> props = null;
      jreader.beginArray();
      while (jreader.hasNext())
      {
         jreader.beginObject();
         props = new HashMap<String,JsonElement>();
         while (jreader.hasNext())
         {
            String prop_name = jreader.nextName();
            JsonElement prop_value = parser.parse( jreader );
            if (prop_name.equals( PLAYER_TYPE_KEYWORD ))
               player_type = prop_value.getAsString();
            props.put( prop_name, prop_value );
         }
         jreader.endObject();
         if (player_type.isEmpty())
            throw new ParseException( Resources.getString( "msg.2008" ) ); //$NON-NLS-1$
         config.setPlayerData( player_type, props );
      }
      jreader.endArray();
   }

   private static final String PLAYER_DATA_KEYWORD = "player-data"; //$NON-NLS-1$
   private static final String PLAYER_TYPE_KEYWORD = "player-type"; //$NON-NLS-1$
   private static final String PLUGIN_URIS_KEYWORD = "plugin-uris"; //$NON-NLS-1$
}
