//
// Configuration.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.io.File;
import java.math.BigDecimal;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.gson.JsonElement;

import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.KoRules;

/**
 * Provides programmatic access to configuration data.<p>
 *
 * See {@code etc/sample-config.json} for documentation of the configuration
 * file format.
 *
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see ConfigReader
 * @see ConfigVariableValueProvider
 */
public class Configuration
{
   /**
    * The default date format for date values.
    * 
    * @see <a href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html">SimpleDateFormat</a>
    */
   public static final String DEFAULT_DATE_FORMAT     = "yyyyMMdd"; //$NON-NLS-1$

   /**
    * The default name of the Go engine.
    */
   public static final String DEFAULT_ENGINE_NAME     = "Dakengo"; //$NON-NLS-1$

   /**
    * The default format for constructing game record file names.
    */
   public static final String DEFAULT_FILENAME_FORMAT = "$date.$file-ext"; //$NON-NLS-1$

   /**
    * The default ko rule.
    */
   public static final KoRules DEFAULT_KO_RULE        = KoRules.POSITIONAL;

   /**
    * The default file provides named collections of points.
    */
   public static final String DEFAULT_LAYERS_FILE     = "$app-home/etc/layers.json"; //$NON-NLS-1$

   /**
    * The default name of the player opposing the computer player.
    */
   public static final String DEFAULT_OPPONENT_NAME   = Resources.getString( "msg.2009" ); //$NON-NLS-1$

   /**
    * The default plugin uri.
    */
   public static final String DEFAULT_PLUGIN_URI      = "file://$app-home/plugins"; //$NON-NLS-1$

   /**
    * Identifies the default format used for storing game records.
    */
   public static final String DEFAULT_RECORD_TYPE     = "sgf"; //$NON-NLS-1$

   /**
    * Identifies the default scoring system.
    */
   public static final String DEFAULT_SCORING_SYSTEM  = "area"; //$NON-NLS-1$

   /**
    * The character identifying variables to be expanded.
    */
   public static final char   VAR_PREFIX              = '$';

   /**
    * Replaces variables in an input string with their corresponding values. A variable starts with
    * the var_prefix character. This is done in a single pass; nested variables (e.g.,
    * $FOO expands to the string "$BAR") are not recursively expanded. Unrecognized variables are left
    * as-is. The var_prefix character serves as its own escape character.<p>
    * 
    * @param input        input string which may contain variables to be expanded
    * @param var_prefix   character prefix identifying a variable
    * @param providers    table of variable names (excluding the variable prefix character)
    *                     and corresponding providers
    */
   public static String expandVariables( String input, char var_prefix, Map<String,ConfigVariableValueProvider> providers )
   {
      if (var_prefix == '\0' || providers.isEmpty() || input.isEmpty())
         return input;
      StringBuilder output = new StringBuilder();
      StringBuilder var_buffer = null;
      boolean collecting = false;
      int index = 0;
      char ch = '\0';
      char last_ch = '\0';
      while (index < input.length())
      {
         last_ch = ch;
         ch = input.charAt( index++ );
         if (ch == var_prefix && last_ch == var_prefix)
         {
            collecting = false;
            ch = '\0';
            output.append( var_prefix );
            continue;
         }
         if (collecting)
         {
            if (ch == var_prefix)
            {
               output.append( var_prefix );
               output.append( var_buffer.toString() );
               var_buffer.setLength( 0 );
               continue;
            }
            var_buffer.append( ch );
            String name = var_buffer.toString();
            if (providers.containsKey( name ))
            {
               output.append( providers.get( name ).getValue() );
               collecting = false;
            }
            continue;
         }
         if (ch == var_prefix)
         {
            collecting = true;
            if (var_buffer == null)
               var_buffer = new StringBuilder();
            else
               var_buffer.setLength( 0 );
            continue;
         }
         output.append( ch );
      }
      if (collecting && var_buffer.length() > 0)
      {
         output.append( var_prefix );
         output.append( var_buffer.toString() );
      }
      return output.toString();
   }

   /**
    * A helper method for evenly allocating workloads across threads
    * such that N games will be run on a maximum of P threads. The size
    * of the returned enumerable specifies how many worker threads should
    * actually be created. Each element specifies how many games will be
    * run on the corresponding thread.
    *
    * @param games    total number of games to be played
    * @param threads  number of threads available
    */
   public static Collection<Integer> computeInstanceCounts( int games, int threads )
   {
      if (games <= 0)
         throw new IllegalArgumentException( Resources.getString( "msg.2010" ) ); //$NON-NLS-1$
      if (threads <= 0)
         throw new IllegalArgumentException( Resources.getString( "msg.2011" ) ); //$NON-NLS-1$
      Collection<Integer> counts = new ArrayList<Integer>();
      int instances = games / threads;
      int remainder = games % threads;
      if (games < threads)
         threads = games;
      for (int i = 0; i < threads; i++)
      {
         int count = instances;
         if (remainder > 0)
         {
            count++;
            remainder--;
         }
         counts.add( count );
      }
      return counts;
   }

   public Configuration()
   {
      m_engine_version = Configuration.class.getPackage().getImplementationVersion();
      if (m_engine_version == null)
         m_engine_version = ""; //$NON-NLS-1$
      m_providers = new HashMap<String,ConfigVariableValueProvider>();
      m_providers.put( APP_HOME_VAR,        m_app_home_value );
      m_providers.put( DATE_VAR,            new DateValueProvider( this ) );
      m_providers.put( DATE_FORMAT_VAR,     new ConfigPropertyValueProvider( this, DATE_FORMAT_VAR ) );
      m_providers.put( ENGINE_NAME_VAR,     new ConfigPropertyValueProvider( this, ENGINE_NAME_VAR ) );
      m_providers.put( ENGINE_VERSION_VAR,  new ConfigPropertyValueProvider( this, ENGINE_VERSION_VAR ) );
      m_providers.put( FILE_EXT_VAR,        m_file_ext_value );
      m_providers.put( FILENAME_FORMAT_VAR, new ConfigPropertyValueProvider( this, FILENAME_FORMAT_VAR ) );
      m_providers.put( KO_RULE_VAR,         new ConfigPropertyValueProvider( this, KO_RULE_VAR ) );
      m_providers.put( KOMI_VAR,            new ConfigPropertyValueProvider( this, KOMI_VAR ) );
      m_providers.put( LAYERS_FILE_VAR,     new ConfigPropertyValueProvider( this, LAYERS_FILE_VAR ) );
      m_providers.put( OUTPUT_DIR_VAR,      new ConfigPropertyValueProvider( this, OUTPUT_DIR_VAR ) );
      m_providers.put( SCORING_VAR,         new ConfigPropertyValueProvider( this, SCORING_VAR ) );
      m_providers.put( SUICIDE_VAR,         new ConfigPropertyValueProvider( this, SUICIDE_VAR ) );
      m_providers.put( SIZE_VAR,            m_size_value );
      m_providers.put( TEMP_VAR,            m_temp_dir_value );
      m_providers.put( USER_HOME_VAR,       m_user_home_value );
      m_providers.put( USER_NAME_VAR,       m_user_name_value );
   }

   /**
    * Initialize a configuration by making a deep copy of another configuration.
    *
    * @param other
    */
   public Configuration( Configuration other )
   {
      m_providers = new HashMap<String,ConfigVariableValueProvider>( other.m_providers );
      m_engine_name = other.m_engine_name;
      m_engine_version = other.m_engine_version;
      m_ko_rule = other.m_ko_rule;
      m_komi = other.m_komi;
      m_layers_file = other.m_layers_file;
      m_output_dir = other.m_output_dir;
      m_player_data = new HashMap<String,Map<String,JsonElement>>( other.m_player_data );
      m_plugin_uris = new ArrayList<String>( other.m_plugin_uris );
      m_scoring = other.m_scoring;
      m_suicide = other.m_suicide;
   }

   /**
    * Returns the date format.
    * 
    * @see #DEFAULT_DATE_FORMAT
    */
   public SimpleDateFormat getDateFormat()
   {
      return new SimpleDateFormat( m_date_fmt );
   }

   /**
    * Sets the date format.
    * 
    * @param fmt
    * 
    * @see <a href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html"><code>SimpleDateFormat</code></a>
    */
   public void setDateFormat( String fmt )
   {
      if (fmt.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2012" ) ); //$NON-NLS-1$
      m_date_fmt = fmt;
   }

   /**
    * Returns the name of the engine to be returned to users
    * via user interfaces or log messages.
    * 
    * @see #DEFAULT_ENGINE_NAME
    */
   public String getEngineName()
   {
      return expandVariables( m_engine_name, VAR_PREFIX, m_providers );
   }

   /**
    * Sets the name of the engine.
    *
    * @param name
    */
   public void setEngineName( String name )
   {
      if (name.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2013" ) ); //$NON-NLS-1$
      m_engine_name = name;
   }

   /**
    * Returns the version of the engine to be returned to users
    * via user interfaces or log messages. The default value is
    * the Dakengo build version string as specified via the
    * {@code version.properties} file.
    */
   public String getEngineVersion()
   {
      return expandVariables( m_engine_version, VAR_PREFIX, m_providers );
   }

   /**
    * Sets the version of the engine. This method allows an empty
    * string to be supplied for the version.
    *
    * @param version
    */
   public void setEngineVersion( String version )
   {
      m_engine_version = version;
   }

   /**
    * Returns the format for constructing game record file names.
    * 
    * @see #getOutputFile
    * @see #setFilenameFormat
    */
   public String getFilenameFormat()
   {
      return m_filename_fmt;
   }

   /**
    * Sets the format for constructing game record file names.
    * 
    * @see #getFilenameFormat
    * @see #getOutputFile
    */
   public void setFilenameFormat( String fmt )
   {
      if (fmt.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2014" ) ); //$NON-NLS-1$
      m_filename_fmt = fmt;
   }

   /**
    * Returns the komi for this configuration.
    * 
    * @see com.dakengo.libs.game.Game#DEFAULT_KOMI
    */
   public BigDecimal getKomi()
   {
      return m_komi;
   }

   /**
    * Sets the komi for this configuration.
    *
    * @param komi
    */
   public void setKomi( BigDecimal komi )
   {
      m_komi = komi;
   }

   /**
    * Sets the komi for this configuration.
    *
    * @param komi
    */
   public void setKomi( String komi )
   {
      try
      {
         m_komi = new BigDecimal( komi );
      }
      catch (NumberFormatException unused)
      {
         throw new ParseException( Resources.getString( "msg.2015" ), komi ); //$NON-NLS-1$
      }
   }

   /**
    * Returns the ko rule for this configuration.
    * 
    * @see #DEFAULT_KO_RULE
    */
   public KoRules getKoRule()
   {
      return m_ko_rule;
   }

   /**
    * Sets the ko rule for this configuration.
    *
    * @param rule
    */
   public void setKoRule( KoRules rule )
   {
      m_ko_rule = rule;
   }

   /**
    * Sets the ko rule for this configuration.
    *
    * @param rule
    */
   public void setKoRule( String rule )
   {
      try
      {
         m_ko_rule = Enum.valueOf( KoRules.class, rule.toUpperCase() );
      }
      catch (IllegalArgumentException unused)
      {
         throw new ParseException( Resources.getString( "msg.2016" ), rule ); //$NON-NLS-1$
      }
   }

   /**
    * Returns the path to a file containing layer definitions.
    *
    * @see #DEFAULT_LAYERS_FILE
    * @see com.dakengo.libs.game.Layers
    */
   public File getLayersFile()
   {
      return new File( expandVariables( m_layers_file, VAR_PREFIX, m_providers ) );
   }

   /**
    * Sets the path to a file containing layer definitions.
    *
    * @param path
    *
    * @see com.dakengo.libs.game.Layers
    */
   public void setLayersFile( File path )
   {
      m_layers_file = path.getAbsolutePath();
   }

   /**
    * Sets the path to a file containing layer definitions.
    *
    * @param path
    *
    * @see com.dakengo.libs.game.Layers
    */
   public void setLayersFile( String path )
   {
      m_layers_file = path;
   }

   /**
    * Returns the name of the opponent of the selected computer player.
    * 
    * @see #DEFAULT_OPPONENT_NAME
    */
   public String getOpponentName()
   {
      return expandVariables( m_opponent_name, VAR_PREFIX, m_providers );
   }

   /**
    * Sets the name of the opponent of the selected computer player.
    * @param name
    */
   public void setOpponentName( String name )
   {
      if (name.isEmpty())
         m_opponent_name = DEFAULT_OPPONENT_NAME;
      else
         m_opponent_name = name;
   }

   /**
    * Returns the game record output directory for this configuration.
    * The default is the current directory.
    * 
    * @see java.lang.System#getProperty
    */
   public File getOutputDirectory()
   {
      return new File( expandVariables( m_output_dir, VAR_PREFIX, m_providers ) );
   }

   /**
    * Sets the game record output directory for this configuration.
    *
    * @param path
    */
   public void setOutputDirectory( File path )
   {
      m_output_dir = path.getAbsolutePath();
   }

   /**
    * Sets the game record output directory for this configuration.
    *
    * @param path
    */
   public void setOutputDirectory( String path )
   {
      m_output_dir = path;
   }

   /**
    * Returns the current file name for storing game records. The returned
    * file path includes the directory returned by {@link #getOutputDirectory}.
    *
    * @see #getFilenameFormat
    * @see #setFilenameFormat
    * @see #DEFAULT_FILENAME_FORMAT
    */
   public File getOutputFile()
   {
      return new File( getOutputDirectory(), expandVariables( m_filename_fmt, VAR_PREFIX, m_providers ) );
   }

   /**
    * Returns an object representing any configuration data that
    * was supplied for the player. Note that configuration data
    * is not required, and therefore the returned map may be
    * empty.
    *
    * @param player_type  the computer player type name
    */
   public Map<String,JsonElement> getPlayerData( String player_type )
   {
      if (player_type.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2017" ) ); //$NON-NLS-1$
      if (!m_player_data.containsKey( player_type ))
         return new HashMap<String,JsonElement>();
      return m_player_data.get( player_type );
   }

   /**
    * Sets (replaces any existing) configuration data for the specified computer player.
    *
    * @param player_type  the computer player type name
    * @param data         the player's configuration data, {@code null} or
    *                     {@link com.google.gson.JsonNull} are also allowed
    */
   public void setPlayerData( String player_type, Map<String,JsonElement> data )
   {
      if (player_type.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2017" ) ); //$NON-NLS-1$
      m_player_data.put( player_type, data );
   }

   /**
    * Returns a list of URI objects representing the plugin URIs specified
    * in the configuration file.
    * 
    * @see #DEFAULT_PLUGIN_URI
    */
   public List<URI> getPluginURIs() throws URISyntaxException
   {
      List<URI> plugin_uris = new ArrayList<URI>();
      if (m_plugin_uris.isEmpty())
      {
         plugin_uris.add( createURI( DEFAULT_PLUGIN_URI ) );
      }
      else
      {
         for (String path : m_plugin_uris)
            plugin_uris.add( createURI( path ) );
      }
      return plugin_uris;
   }

   /**
    * Adds a new plugin directory URI to this configuration.
    *
    * @param uri
    */
   public void addPluginURI( String uri )
   {
      m_plugin_uris.add( uri );
   }

   /**
    * Returns the scoring system for this configuration.
    * 
    * @see #DEFAULT_SCORING_SYSTEM
    */
   public String getScoring()
   {
      return m_scoring;
   }

   /**
    * Sets the scoring system for this configuration.
    *
    * @param name  must be a name published by one of the loaded scoring plugins
    */
   public void setScoring( String name )
   {
      if (name.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2018" ) ); //$NON-NLS-1$
      if (!BaseScorer.getSystems().contains( name ))
         throw new ParseException( Resources.getString( "msg.2019" ), name ); //$NON-NLS-1$
      m_scoring = name;
   }

   /**
    * Returns true if this configuration allows suicide moves. The default
    * is false.
    */
   public boolean allowSuicide()
   {
      return m_suicide;
   }

   /**
    * Specifies whether this configuration allows suicide moves.
    *
    * @param flag
    */
   public void setAllowSuicide( boolean flag )
   {
      m_suicide = flag;
   }

   /**
    * Specifies whether this configuration allows suicide moves.
    *
    * @param flag
    */
   public void setAllowSuicide( String flag )
   {
      flag = flag.toLowerCase();
      switch (flag)
      {
         case "yes": m_suicide = true;  break; //$NON-NLS-1$
         case "no":  m_suicide = false; break; //$NON-NLS-1$
         default:    throw new ParseException( Resources.getString( "msg.2020" ), flag ); //$NON-NLS-1$
      }
   }

   private URI createURI( String path ) throws URISyntaxException
   {
      String expanded_path = expandVariables( path, VAR_PREFIX, m_providers ).replace( "\\", "/" ); //$NON-NLS-1$ //$NON-NLS-2$
      Matcher matcher = m_scheme_pattern.matcher( expanded_path );
      if (matcher.matches() && matcher.start( 1 ) >= 0)
         expanded_path = "file:///" + expanded_path.substring( matcher.start( 1 ) ); //$NON-NLS-1$
      return new URI( expanded_path );
   }

   static final String APP_HOME_VAR        = "app-home";        //$NON-NLS-1$
   static final String DATE_VAR            = "date";            //$NON-NLS-1$
   static final String DATE_FORMAT_VAR     = "date-format";     //$NON-NLS-1$
   static final String ENGINE_NAME_VAR     = "engine-name";     //$NON-NLS-1$
   static final String ENGINE_VERSION_VAR  = "engine-version";  //$NON-NLS-1$
   static final String FILE_EXT_VAR        = "file-ext";        //$NON-NLS-1$
   static final String FILENAME_FORMAT_VAR = "filename-format"; //$NON-NLS-1$
   static final String KO_RULE_VAR         = "ko-rule";         //$NON-NLS-1$
   static final String KOMI_VAR            = "komi";            //$NON-NLS-1$
   static final String LAYERS_FILE_VAR     = "layers-file";     //$NON-NLS-1$
   static final String OUTPUT_DIR_VAR      = "output-dir";      //$NON-NLS-1$
   static final String SCORING_VAR         = "scoring";         //$NON-NLS-1$
   static final String SIZE_VAR            = "size";            //$NON-NLS-1$
   static final String SUICIDE_VAR         = "suicide";         //$NON-NLS-1$
   static final String TEMP_VAR            = "temp";            //$NON-NLS-1$
   static final String USER_HOME_VAR       = "user-home";       //$NON-NLS-1$
   static final String USER_NAME_VAR       = "user-name";       //$NON-NLS-1$

   private static final ConfigVariableValueProvider m_app_home_value  = new SystemPropertyValueProvider( Constants.APP_HOME_PROP );
   private static final ConfigVariableValueProvider m_file_ext_value  = new ConfigVariableValueProvider()
   {
      @Override
      public String getValue() { return "sgf"; } //$NON-NLS-1$
   };
   private static final ConfigVariableValueProvider m_size_value      = new BoardSizeValueProvider();
   private static final ConfigVariableValueProvider m_temp_dir_value  = new SystemPropertyValueProvider( "java.io.tmpdir" ); //$NON-NLS-1$
   private static final ConfigVariableValueProvider m_user_home_value = new SystemPropertyValueProvider( "user.home" );      //$NON-NLS-1$
   private static final ConfigVariableValueProvider m_user_name_value = new SystemPropertyValueProvider( "user.name" );      //$NON-NLS-1$

   private String                                  m_date_fmt       = DEFAULT_DATE_FORMAT;
   private Map<String,ConfigVariableValueProvider> m_providers      = new HashMap<String,ConfigVariableValueProvider>();
   private String                                  m_engine_name    = DEFAULT_ENGINE_NAME;
   private String                                  m_engine_version = "";                                               //$NON-NLS-1$
   private String                                  m_filename_fmt   = DEFAULT_FILENAME_FORMAT;
   private String                                  m_opponent_name  = DEFAULT_OPPONENT_NAME;
   private KoRules                                 m_ko_rule        = DEFAULT_KO_RULE;
   private BigDecimal                              m_komi           = Game.DEFAULT_KOMI;
   private String                                  m_layers_file    = "$app-home/etc/layers.json";                      //$NON-NLS-1$
   private String                                  m_output_dir     = System.getProperty( "user.dir", "." );            //$NON-NLS-1$ //$NON-NLS-2$
   private Map<String,Map<String,JsonElement>>     m_player_data    = new HashMap<String,Map<String,JsonElement>>();
   private List<String>                            m_plugin_uris    = new ArrayList<String>();
   private String                                  m_scoring        = DEFAULT_SCORING_SYSTEM;
   private boolean                                 m_suicide        = false;
   private Pattern                                 m_scheme_pattern = Pattern.compile( "^file:/{1,2}([a-zA-Z]:).*" );   //$NON-NLS-1$
}
