//
// GameContext.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.ThreadContext;
import net.xeoh.plugins.base.PluginManager;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.Layers;
import com.dakengo.libs.game.Scorer;
import com.dakengo.libs.player.Player;
import com.dakengo.libs.record.PositionProperties;
import com.dakengo.libs.record.Record;

/**
 * This class provides context for a game engine.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class GameContext
{
   /**
    * Initializes a new context object.
    * 
    * @param config       configuration data
    * @param plugins      plugin manager
    * @param player_type  identifies the desired player plugin
    * @param game_id      application-defined game identifier
    */
   public GameContext( Configuration config, PluginManager plugins, String player_type, String game_id )
   {
      if (player_type.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2017" ) ); //$NON-NLS-1$
      m_config = config;
      m_play_game = new Game( config.getKoRule(), config.allowSuicide(), config.getKomi() );
      m_play_game_id = game_id;
      m_player_type = player_type;
      m_plugins = plugins;
      m_game = m_play_game;
      m_game_id = m_play_game_id;
      ThreadContext.put( Constants.GAME_ID_MDC_PROP, game_id );
   }

   /**
    * Returns the black player's name, which will be the computer player if the first
    * GTP command received by the engine was {@code genmove}, otherwise it will be the opponent.
    * 
    * @see #getWhiteName
    * @see #setOpponentName
    * @see #setPlayerName
    */
   public String getBlackName()
   {
      return m_black_name;
   }

   /**
    * Returns the black player's name, which will be the computer player if the first
    * GTP command received by the engine was {@code play}, otherwise it will be the opponent.
    * 
    * @see #getBlackName
    * @see #setOpponentName
    * @see #setPlayerName
    */
   public String getWhiteName()
   {
      return m_white_name;
   }

   /**
    * This method binds the name of the opposing player to the
    * specified color. This is used for game records written out
    * by the GTP engine.
    * 
    * @param color
    */
   public void setOpponentName( Colors color )
   {
      Game.checkColor( color );
      if (color == Colors.BLACK)
      {
         if (m_black_name.isEmpty())
            m_black_name = m_config.getOpponentName();
      }
      else
      {
         if (m_white_name.isEmpty())
            m_white_name = m_config.getOpponentName();
      }
   }

   /**
    * This method binds the name of the computer player to the
    * specified color. This is used for game records written out
    * by the GTP engine.
    * 
    * @param color
    */
   public void setPlayerName( Colors color )
   {
      Game.checkColor( color );
      Player player = getPlayer();
      if (color == Colors.BLACK)
      {
         if (m_black_name.isEmpty())
            m_black_name = player.getName( m_config.getEngineName(), m_config.getEngineVersion() );
      }
      else
      {
         if (m_white_name.isEmpty())
            m_white_name = player.getName( m_config.getEngineName(), m_config.getEngineVersion() );
      }
   }

   /**
    * Returns the name of the engine to be returned to users via user interfaces or log messages.
    * 
    * @see Configuration#getEngineName
    */
   public String getEngineName()
   {
      return m_config.getEngineName();
   }

   /**
    * Returns the version of the engine to be returned to users via user interfaces or log messages.
    * 
    * @see Configuration#getEngineVersion
    */
   public String getEngineVersion()
   {
      return m_config.getEngineVersion();
   }

   /**
    * Returns the configuration's date format.
    * 
    * @see Configuration#getDateFormat
    */
   public SimpleDateFormat getDateFormat()
   {
      return m_config.getDateFormat();
   }

   /**
    * Returns the current game. The game object representing an in-progress
    * game is kept separate from any game object initialized from an external
    * source, such as by the {@code loadsgf} GTP command.
    */
   public Game getGame()
   {
      return m_game;
   }

   /**
    * Substitutes a game initialized from an external source in place of the
    * in-progress game object. The application is responsible for defining how
    * to identify games. For example, in Dakengo the {@code loadsgf} command
    * specifies the path of the SGF file as the game ID.
    * 
    * @param game
    * @param game_id
    */
   public void setGame( Game game, String game_id )
   {
      m_game = game;
      m_game_id = game_id;
      Map<String,String> ctx_map = ThreadContext.getContext();
      ctx_map.put( Constants.GAME_ID_MDC_PROP, game_id );
   }

   /**
    * Returns a game identifier.
    */
   public String getGameID()
   {
      return m_game_id;
   }

   /**
    * Returns a map containing the most recently loaded layers.
    * 
    * @see com.dakengo.libs.game.Layers
    */
   public Map<String,Set<Short>> getLayers()
   {
      return m_layers;
   }

   /**
    * A convenience method for obtaining the {@link Player} plugin selected by the application.
    * 
    * @return The computer player.
    */
   public Player getPlayer()
   {
      Player player = m_plugins.getPlugin( Player.class, PluginUtils.getPlayerCapability( m_player_type ) );
      if (player == null)
         throw new PluginException( Resources.getString( "msg.2052" ), m_player_type );
      if (!m_player_init)
      {
         player.init( this, m_config.getPlayerData( m_player_type ) );
         m_player_init = true;
      }
      return player;
   }

   /**
    * Provides access to the plugin manager.
    * 
    * @return The plugin manager.
    */
   public PluginManager getPlugins()
   {
      return m_plugins;
   }

   /**
    * Returns the game record that is kept up-to-date as GTP commands are executed.
    */
   public Record getRecord()
   {
      return m_record;
   }

   /**
    * A convenience method for obtaining the {@link com.dakengo.libs.game.Scorer} plugin selected by the application.
    * 
    * @return The scoring object.
    */
   public Scorer getScorer()
   {
      String scoring = m_config.getScoring();
      Scorer scorer = m_plugins.getPlugin( Scorer.class, PluginUtils.getScoringCapability( scoring ) );
      if (scorer == null)
         throw new PluginException( Resources.formatString( "msg.2053" ), scoring ); //$NON-NLS-1$
      return scorer;
   }

   /**
    * Loads into memory the layers corresponding to the specified board size.
    * Implicitly sets the current board size to the requested size.
    * 
    * @param size
    * 
    * @see com.dakengo.libs.game.Layers
    */
   public void loadLayers( int size ) throws IOException
   {
      if (!Board.isValidSize( size ))
         throw new IllegalArgumentException( Resources.formatString( "msg.2022", size ) ); //$NON-NLS-1$
      if (size == m_layer_size)
      {
         Board.setSize( size );
         return;
      }
      try (BufferedReader br = new BufferedReader( new FileReader( m_config.getLayersFile() ) ))
      {
         m_layers = Layers.load( br, size );
         m_layer_size = size;
      }
   }

   /**
    * Call this method at the end of each game to clean up intermediate
    * state.
    */
   public void reset()
   {
      m_black_name = ""; //$NON-NLS-1$
      m_white_name = ""; //$NON-NLS-1$
      m_play_size = Board.getSize();
      m_play_game.reset();
      m_game = m_play_game;
      m_record = new Record( new PositionProperties( true ) );
      try
      {
         getPlayer().reset( m_game_id );
      }
      catch (Throwable t)
      {
         Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
         log.error( Resources.getString( "msg.2050" ), t ); //$NON-NLS-1$
      }
   }

   /**
    * Sets this context to use the game object for playing
    * a real game, as opposed to the game object used for
    * GTP commands such as {@code loadsgf}.
    */
   public void setPlayGame()
   {
      Board.setSize( m_play_size );
      m_game = m_play_game;
      m_game_id = m_play_game_id;
   }

   private String                 m_black_name   = "";            // black player name to use when writing out a game record //$NON-NLS-1$
   private String                 m_white_name   = "";            // white player name to use when writing out a game record //$NON-NLS-1$
   private Configuration          m_config;
   private Game                   m_game;                         // may be set to m_play_game, or a game created by loadsgf
   private String                 m_game_id;                      // for m_play_game, this is a UUID in string form; for a game loaded by loadsgf, this is the path to the SGF file
   private Map<String,Set<Short>> m_layers       = new HashMap<String,Set<Short>>();
   private int                    m_layer_size   = 0;
   private Game                   m_play_game;                    // game to use when playing a real game (as opposed to loadsgf)
   private String                 m_play_game_id = "";            // game ID corresponding to m_play_game                    //$NON-NLS-1$
   private int                    m_play_size    = Board.DEFAULT_SIZE;
   private boolean                m_player_init  = false;
   private String                 m_player_type;
   private PluginManager          m_plugins;
   private Record                 m_record       = new Record( new PositionProperties( true ) );
}
