//
// GtpCommand.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import net.xeoh.plugins.informationbroker.InformationItem;

import com.dakengo.libs.base.HashUtils;
import com.dakengo.libs.base.ParseException;

/**
 * This class represents a command and its associated parameters
 * that can be executed by {@link GtpEngine}.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 *
 * @see GtpCommandHandler
 * @see GameContext
 * @see GtpReplyHandler
 * @see <a href="http://www.lysator.liu.se/~gunnar/gtp/gtp2-spec-draft2/gtp2-spec.html">Go Text Protocol</a>
 */
public class GtpCommand
{
   /**
    * A plugin broker information item for publishing or subscribing
    * to the names of commands that a GTP subset implements.
    * 
    * @see net.xeoh.plugins.informationbroker.InformationBroker
    */
   public class NamesInfoItem implements InformationItem<Set<String>> {}

   /** Identifies a command that was not accompanied by an ID number. */
   public static final long NO_ID = -1;

   /**
    * Parse a GTP command string.
    * 
    * @param text
    */
   public static GtpCommand parse( String text )
   {
      String ex_msg = Resources.getString( "msg.2021" ); //$NON-NLS-1$
      if (text.isEmpty())
         throw new ParseException( ex_msg );

      GtpCommand cmd = new GtpCommand();
      String[] parts = text.split( " " ); //$NON-NLS-1$
      if (parts.length == 0)
         throw new ParseException( ex_msg );
      int index = 0;
      while (index < parts.length && parts[index].isEmpty())
         index++;
      if (index >= parts.length)
         throw new ParseException( ex_msg );

      try
      {
         cmd.m_id = Long.parseLong( parts[index] );
         index++;
      }
      catch (NumberFormatException ex)
      {
         // first field might not have been the ID, therefore not an integer
      }
      while (index < parts.length && parts[index].isEmpty())
         index++;
      if (index >= parts.length)
         throw new ParseException( ex_msg );
      cmd.m_name = parts[index++].toLowerCase();

      while (index < parts.length)
      {
         if (!parts[index].isEmpty())
            cmd.m_args.add( parts[index] );
         index++;
      }

      return cmd;
   }

   /**
    * Return the list of arguments that accompanied this command, if any.
    */
   public List<String> getArguments()
   {
      return m_args;
   }

   /**
    * Return this command's ID number. This method will return {@link #NO_ID}
    * if no ID number was specified when the command was read.
    */
   public long getID()
   {
      return m_id;
   }

   /**
    * Returns this command's name, which identifies the action to be executed
    * by the GTP engine.
    */
   public String getName()
   {
      return m_name;
   }

   @Override
   public boolean equals( Object obj )
   {
      if (!(obj instanceof GtpCommand))
         return false;
      if (obj == this)
         return true;
      GtpCommand other = (GtpCommand) obj;
      return m_id == other.m_id && m_name.equals( other.m_name ) && m_args.equals( other.m_args );
   }

   @Override
   public int hashCode()
   {
      int args_hash = HashUtils.compute( m_args );
      return HashUtils.compute( m_name, args_hash, m_id );
   }

   @Override
   public String toString()
   {
      StringBuilder buffer = new StringBuilder();
      buffer.append( String.format( "%d,%s", m_id, m_name ) ); //$NON-NLS-1$
      for (String arg : m_args)
         buffer.append( String.format( ",%s", arg ) ); //$NON-NLS-1$
      return buffer.toString();
   }

   private GtpCommand()
   {
   }

   private List<String> m_args = new ArrayList<String>();
   private String       m_name = ""; //$NON-NLS-1$
   private long         m_id   = NO_ID; 
}
