//
// GtpEngine.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Date;
import java.text.DateFormat;

import net.xeoh.plugins.base.PluginManager;
import net.xeoh.plugins.informationbroker.InformationBroker;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.record.GameProperties;
import com.dakengo.libs.record.Record;

/**
 * This class is the main point of integration for the Go Text Protocol.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 *
 * @see GtpCommand
 * @see GtpCommandHandler
 * @see GameContext
 * @see GtpReplyHandler
 * @see <a href="http://www.lysator.liu.se/~gunnar/gtp/gtp2-spec-draft2/gtp2-spec.html">Go Text Protocol</a>
 */
public class GtpEngine
{
   /**
    * The version of the Go Text Protocol implemented by Dakengo.
    */
   public static final String PROTOCOL_VERSION       = "2"; //$NON-NLS-1$

   /**
    * Initializes a new GTP engine object.
    * 
    * @param plugin_mgr   provides the engine's computer player
    * @param handler      destination for GTP command results
    * @param store        destination for completed game records
    */
   public GtpEngine( PluginManager plugin_mgr, GtpReplyHandler handler, GtpRecordStore store )
   {
      m_plugin_mgr = plugin_mgr;
      m_handler = handler;
      m_store = store;
      InformationBroker broker = m_plugin_mgr.getPlugin( InformationBroker.class );
      PluginUtils.publishCommand( broker, "quit" ); //$NON-NLS-1$
   }

   /**
    * Returns the plugin manager for this engine.
    */
   public PluginManager getPluginManager()
   {
      return m_plugin_mgr;
   }

   /**
    * Read GTP commands from an input stream and execute them. Results are
    * returned via the engine's reply handler.
    * 
    * @param input
    * @param context
    */
   public void loop( Reader input, GameContext context ) throws IOException
   {
      // TODO: log GTP and player startup messages, version info, and config
      //
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      boolean keep_going = true;
      BufferedReader reader = new BufferedReader( input );
      try
      {
         do
         {
            String buffer = reader.readLine();
            if (buffer != null)
            {
               buffer = buffer.trim();
               if (buffer.isEmpty() || buffer.charAt( 0 ) == '#')
                  continue;
               log.debug( Resources.formatString( "msg.2043", buffer ) ); //$NON-NLS-1$
               keep_going = execute( GtpCommand.parse( buffer ), context );
            }
            else
            {
               log.debug( Resources.getString( "msg.2039" ) ); //$NON-NLS-1$
               keep_going = false;
            }
         } while (keep_going);
      }
      catch (Throwable t)
      {
         log.error( Resources.getString( "msg.2048" ), t ); //$NON-NLS-1$
         throw t;
      }
   }

   /**
    * Execute an individual GTP command. Results are
    * returned via the engine's reply handler. Returns true
    * if the application should continue processing commands.
    * 
    * @param cmd
    * @param context
    */
   public boolean execute( GtpCommand cmd, GameContext context )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      try
      {
         log.debug( Resources.formatString( "msg.2034", cmd ) ); //$NON-NLS-1$
         if (cmd.getName().equals( "quit" )) //$NON-NLS-1$
         {
            m_handler.handleReply( false, cmd.getID(), "" ); //$NON-NLS-1$
            saveGame( context );
            context.reset();
            return false;
         }
         GtpCommandHandler cmd_handler = m_plugin_mgr.getPlugin( GtpCommandHandler.class, PluginUtils.getGtpCapability( cmd.getName() ) );
         if (cmd_handler == null)
         {
            String str = Resources.getString( "msg.2046" ); //$NON-NLS-1$
            m_handler.handleReply( true, cmd.getID(), str );
            log.debug( str );
            return true;
         }
         if (cmd_handler.shouldSave())
            saveGame( context );
         cmd_handler.handleCommand( cmd, context, m_handler );
      }
      catch (AssertionError e)
      {
         throw e;
      }
      catch (Throwable t)
      {
         m_handler.handleReply( true, cmd.getID(), Resources.getString( "msg.2037" ) ); //$NON-NLS-1$
         log.error( Resources.getString( "msg.2049" ), t ); //$NON-NLS-1$
      }
      return true;
   }

   private void saveGame( GameContext context )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );
      Game game = context.getGame();
      if (game.getTurnNumber() <= 1)
         return;
      String black_name = context.getBlackName();
      if (black_name.isEmpty())
      {
         log.warn( Resources.formatString( "msg.2036", Resources.getString( "msg.2032" ) ) ); //$NON-NLS-1$ //$NON-NLS-2$
         return;
      }
      String white_name = context.getWhiteName();
      if (white_name.isEmpty())
      {
         log.warn( Resources.formatString( "msg.2036", Resources.getString( "msg.2047" ) ) ); //$NON-NLS-1$ //$NON-NLS-2$
         return;
      }
      try
      {
         Record rec = context.getRecord();
         GameProperties props = rec.getGameProps();
         props.setApplicationName( context.getEngineName() );
         props.setApplicationVersion( context.getEngineVersion() );
         props.setCharset( "ISO-8859-1" ); //$NON-NLS-1$
         props.setSize( Board.getSize() );
         props.setKomi( game.getKomi() );
         props.setDate( m_date_fmt.format( new Date() ) );
         props.setResult( game.getResult( context.getScorer() ) );
         props.setPlayerName( Colors.BLACK, black_name );
         props.setPlayerName( Colors.WHITE, white_name );
         m_store.save( rec );
      }
      catch (Throwable t)
      {
         log.error( Resources.getString( "msg.2049" ), t ); //$NON-NLS-1$
      }
   }

   private static final DateFormat m_date_fmt = DateFormat.getDateInstance();

   private PluginManager   m_plugin_mgr;
   private GtpReplyHandler m_handler;
   private GtpRecordStore  m_store;
}
