//
// PluginUtils.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashSet;
import java.util.Set;

import net.xeoh.plugins.base.PluginManager;
import net.xeoh.plugins.base.impl.PluginManagerFactory;
import net.xeoh.plugins.base.options.getplugin.OptionCapabilities;
import net.xeoh.plugins.base.util.JSPFProperties;
import net.xeoh.plugins.informationbroker.InformationBroker;
import net.xeoh.plugins.informationbroker.util.InformationBrokerUtil;

/**
 * Helper methods for instantiating plugins.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see <a href="https://code.google.com/p/jspf/">JSPF</a>
 */
public abstract class PluginUtils
{
   /**
    * Create and return a new plugin manager. Note that the JSPF manual recommends having only
    * one plugin manager per application, therefore this method should only be called once.
    * 
    * @param config
    */
   public static PluginManager createManager( Configuration config )
   {
      JSPFProperties props = new JSPFProperties();
      props.setProperty( PluginManager.class, "cache.enabled", "false" );  // manual says false is the default, but I prefer being explicit //$NON-NLS-1$ //$NON-NLS-2$
      props.setProperty( PluginManager.class, "logging.level", "OFF" );                                                                     //$NON-NLS-1$ //$NON-NLS-2$
      PluginManager pm = PluginManagerFactory.createPluginManager( props );
      try
      {
         for (URI uri : config.getPluginURIs())
            pm.addPluginsFrom( uri );
      }
      catch (URISyntaxException ex)
      {
         // log.warn( ex.getMessage() );
      }
      return pm;
   }

   /**
    * Returns an option capabilities object suitable for requesting a GTP command plugin.
    * 
    * @param cmd_name  GTP command name
    *
    * @see GtpCommandHandler
    */
   public static OptionCapabilities getGtpCapability( String cmd_name )
   {
      if (cmd_name.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2024" ) ); //$NON-NLS-1$
      return new OptionCapabilities( Constants.GTP_CAP_PREFIX + cmd_name );
   }

   /**
    * Returns an option capabilities object suitable for requesting a player plugin.
    * 
    * @param player_type
    * 
    * @see com.dakengo.libs.player.Player
    */
   public static OptionCapabilities getPlayerCapability( String player_type )
   {
      if (player_type.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2017" ) ); //$NON-NLS-1$
      return new OptionCapabilities( Constants.PLAYER_CAP_PREFIX + player_type );
   }

   /**
    * Returns an option capabilities object suitable for requesting a record plugin.
    * 
    * @param file_ext  file extension of desired record type, with no leading '.'
    * @param reader    true if the desired plugin is a reader, false for a writer
    * 
    * @see com.dakengo.libs.record.RecordReader
    * @see com.dakengo.libs.record.RecordWriter
    */
   public static OptionCapabilities getRecordCapabilities( String file_ext, boolean reader )
   {
      if (file_ext.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2025" ) ); //$NON-NLS-1$
      String type_cap = Constants.FORMAT_CAP_PREFIX + file_ext;
      String mode_suffix = (reader ? Constants.READER_CAP_SUFFIX : Constants.WRITER_CAP_SUFFIX);
      String readwrite_cap = Constants.MODE_CAP_PREFIX + mode_suffix;
      return new OptionCapabilities( type_cap, readwrite_cap );
   }

   /**
    * Returns an option capabilities object suitable for requesting a scoring plugin.
    * 
    * @param method
    */
   public static OptionCapabilities getScoringCapability( String method )
   {
      if (method.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2018" ) ); //$NON-NLS-1$
      return new OptionCapabilities( Constants.SCORING_CAP_PREFIX + method );
   }

   /**
    * Each GTP command plugin calls this method to add its command name to the
    * global list of commands.
    * 
    * @param broker
    * @param name
    */
   @SuppressWarnings("unchecked")    // I'd rather use @SuppressWarnings("varargs")
   public static void publishCommand( InformationBroker broker, String name )
   {
      if (name.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2024" ) ); //$NON-NLS-1$
      InformationBrokerUtil util = new InformationBrokerUtil( broker );
      Set<String> names = util.get( GtpCommand.NamesInfoItem.class, new HashSet<String>() );
      names.add( name );
      broker.publish( GtpCommand.NamesInfoItem.class, names );
   }

   /**
    * Each scoring plugin calls this method to add its name to the global list of
    * scoring systems.
    * 
    * @param broker
    * @param name
    */
   @SuppressWarnings("unchecked")    // I'd rather use @SuppressWarnings("varargs")
   public static void publishScoringSystem( InformationBroker broker, String name )
   {
      if (name.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.2018" ) ); //$NON-NLS-1$
      InformationBrokerUtil util = new InformationBrokerUtil( broker );
      Set<String> names = util.get( BaseScorer.NamesInfoItem.class, new HashSet<String>() );
      names.add( name );
      broker.publish( BaseScorer.NamesInfoItem.class, names );
   }
}
