//
// TextUtils.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.application;

import java.util.Collection;
import java.util.Comparator;
import java.util.SortedSet;
import java.util.TreeSet;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;

/**
 * Text utility methods.
 * 
 * <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public abstract class TextUtils
{
   /**
    * Produces a pretty-printed string representation of a board.
    * 
    * @param board
    * 
    * @return  A pretty-printed string representation of a board.
    */
   public static String print( Board board )
   {
      String eol = System.getProperty( "line.separator" ); //$NON-NLS-1$
      StringBuilder buffer = new StringBuilder();
      int size = Board.getSize();
      for (int i = 0; i < size; i++)
      {
         int row = size - i - 1;
         buffer.append( Integer.toString( row + 1 ) );
         buffer.append( (row + 1) > 9 ? "  " : "   " );
         for (int col = 0; col < size; col++)
         {
            Colors color = board.getColor( (short) (col * size + row) );
            switch (color)
            {
               case BLACK:
                  buffer.append( "X " );
                  break;
               case DONT_CARE:
                  buffer.append( "? " );
                  break;
               case EMPTY:
                  buffer.append( ". " );
                  break;
               case WHITE:
                  buffer.append( "O " );
                  break;
               default:
                  throw new IllegalStateException( Resources.formatString( "msg.2051", color ) ); //$NON-NLS-1$
            }
         }
         buffer.append( eol );
      }
      buffer.append( eol );
      buffer.append( "    " ); //$NON-NLS-1$
      buffer.append( "A B C D E F G H J K L M N O P Q R S T U V W X Y Z".substring( 0, size * 2 - 1 ) ); //$NON-NLS-1$
      return buffer.toString();
   }

   /**
    * Produces a string representation of the specified collection, relying
    * on the {@code toString} method of the template type argument to render each
    * element.
    * 
    * @param ary
    * @param separator
    * 
    * @return  A string representation of the specified collection.
    */
   public static <T> String arrayToString( T[] ary, String separator )
   {
      if (ary.length == 0)
         return Resources.getString( "msg.1" ); //$NON-NLS-1$
      StringBuilder buffer = new StringBuilder();
      boolean first = true;
      for (int i = 0; i < ary.length; i++)
      {
         if (!first)
            buffer.append( separator );
         buffer.append( ary[i].toString() );
         first = false;
      }
      return buffer.toString();
   }

   /**
    * Produces a string representation of the specified collection, relying
    * on the {@code toString} method of the template type argument to render each
    * element.
    * 
    * @param coll
    * @param separator
    * @param sort
    * @param cmp
    * 
    * @return  A string representation of the specified collection.
    */
   public static <T> String collectionToString( Collection<T> coll, String separator, boolean sort, Comparator<T> cmp )
   {
      if (coll.isEmpty())
         return Resources.getString( "msg.2" ); //$NON-NLS-1$
      if (sort)
      {
         SortedSet<T> tmp = null;
         if (cmp != null)
            tmp = new TreeSet<T>( cmp );
         else
            tmp = new TreeSet<T>();
         for (T elem : coll)
            tmp.add( elem );
         coll = tmp;
      }
      StringBuilder buffer = new StringBuilder();
      boolean first = true;
      for (T elem : coll)
      {
         if (!first)
            buffer.append( separator );
         buffer.append( elem.toString() );
         first = false;
      }
      return buffer.toString();
   }
}
