//
// HashUtils.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.base;

import java.util.Collection;

/**
 * Helper methods for implementing hashCode(). There are many variations of
 * this kind of helper class floating around. This one was ported
 * to Java from a StackOverflow answer provided by user 'nightcoder'.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 *
 * @see <a href="http://stackoverflow.com/a/2575444">What is the best algorithm for an overridden System.Object.GetHashCode?</a>
 */
public abstract class HashUtils
{
   public static <T1, T2> int compute( T1 arg1, T2 arg2 )
   {
      return PRIME * arg1.hashCode() + arg2.hashCode();
   }

   public static <T1, T2, T3> int compute( T1 arg1, T2 arg2, T3 arg3 )
   {
      int hash = arg1.hashCode();
      hash = PRIME * hash + arg2.hashCode();
      return PRIME * hash + arg3.hashCode();
   }

   public static <T1, T2, T3, T4> int compute( T1 arg1, T2 arg2, T3 arg3, T4 arg4 )
   {
      int hash = arg1.hashCode();
      hash = PRIME * hash + arg2.hashCode();
      hash = PRIME * hash + arg3.hashCode();
      return PRIME * hash + arg4.hashCode();
   }

   public static <T1, T2, T3, T4, T5> int compute( T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5 )
   {
      int hash = arg1.hashCode();
      hash = PRIME * hash + arg2.hashCode();
      hash = PRIME * hash + arg3.hashCode();
      hash = PRIME * hash + arg4.hashCode();
      return PRIME * hash + arg5.hashCode();
   }

   public static <T1, T2, T3, T4, T5, T6> int compute( T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6 )
   {
      int hash = arg1.hashCode();
      hash = PRIME * hash + arg2.hashCode();
      hash = PRIME * hash + arg3.hashCode();
      hash = PRIME * hash + arg4.hashCode();
      hash = PRIME * hash + arg5.hashCode();
      return PRIME * hash + arg6.hashCode();
   }

   /**
    * Computes a hash code for an array when the order of items matters.
    */
   public static int compute( boolean[] ary )
   {
      int hash = PRIME2;
      for (boolean item : ary)
         hash = PRIME * hash + (item ? 1 : 0);
      return hash;
   }

   /**
    * Computes a hash code for an array when the order of items matters.
    */
   public static int compute( char[] ary )
   {
      int hash = PRIME2;
      for (char item : ary)
         hash = PRIME * hash + (int) item;
      return hash;
   }

   /**
    * Computes a hash code for an array when the order of items matters.
    */
   public static int compute( short[] ary )
   {
      int hash = PRIME2;
      for (short item : ary)
         hash = PRIME * hash + (int) item;
      return hash;
   }

   /**
    * Computes a hash code for an array when the order of items matters.
    */
   public static int compute( int[] ary )
   {
      int hash = PRIME2;
      for (int item : ary)
         hash = PRIME * hash + item;
      return hash;
   }

   /**
    * Computes a hash code for an array when the order of items matters.
    */
   public static int compute( long[] ary )
   {
      int hash = PRIME2;
      for (long item : ary)
         hash = PRIME * hash + (int) (item & 0xFFFFFFFFL);
      return hash;
   }

   /**
    * Computes a hash code for an array when the order of items matters.
    */
   public static int compute( float[] ary )
   {
      int hash = PRIME2;
      for (float item : ary)
         hash = PRIME * hash + (int) Math.floor( item );
      return hash;
   }

   public static int compute( double[] ary )
   {
      int hash = PRIME2;
      for (double item : ary)
         hash = PRIME * hash + (int) Math.floor( item );
      return hash;
   }

   /**
    * Computes a hash code for an array when the order of items matters.
    */
   public static <T> int compute( T[] ary )
   {
      int hash = PRIME2;
      for (T item : ary)
         hash = PRIME * hash + item.hashCode();
      return hash;
   }

   /**
    * Computes a hash code for a collection when the order of items matters.
    */
   public static <T> int compute( Collection<T> coll )
   {
      int hash = PRIME2;
      for (T item : coll)
         hash = PRIME * hash + item.hashCode();
      return hash;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    */
   public static int computeAnyOrder( boolean[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (boolean item : ary)
      {
         hash += (item ? 1 : 0);
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    */
   public static int computeAnyOrder( char[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (char item : ary)
      {
         hash += (int) item;
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    */
   public static int computeAnyOrder( short[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (short item : ary)
      {
         hash += (int) item;
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    */
   public static int computeAnyOrder( int[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (int item : ary)
      {
         hash += item;
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    */
   public static int computeAnyOrder( long[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (long item : ary)
      {
         hash += (int) (item & 0xFFFFFFFFL);
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    */
   public static int computeAnyOrder( float[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (float item : ary)
      {
         hash += (int) Math.floor( item );
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    */
   public static int computeAnyOrder( double[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (double item : ary)
      {
         hash += (int) Math.floor( item );
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    * E.g., {1, 2, 3} and {3, 2, 1} get the same hash code.
    */
   public static <T> int computeAnyOrder( T[] ary )
   {
      int hash = PRIME2;
      int count = 0;
      for (T item : ary)
      {
         hash += item.hashCode();
         count++;
      }
      return PRIME * hash + count;
   }

   /**
    * Computes a hash code for a collection when the order of items doesn't matter.
    * E.g., {1, 2, 3} and {3, 2, 1} get the same hash code.
    */
   public static <T> int computeAnyOrder( Collection<T> coll )
   {
      int hash = PRIME2;
      int count = 0;
      for (T item : coll)
      {
         hash += item.hashCode();
         count++;
      }
      return PRIME * hash + count;
   }

   private static final int PRIME  = 29;
   private static final int PRIME2 = 37;
}
