//
// Lexer.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.base;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.apache.commons.lang3.tuple.ImmutablePair;

/**
 * This class converts a sequence of characters into a sequence of tokens which
 * can then be fed into a parser.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class Lexer
{
   /** Represents an empty character sequence. */
   public static final int NULL_TOKEN_ID       = Integer.MAX_VALUE;
   /** Represents a sequence of spaces, tabs, and/or form-feeds. */
   public static final int WS_TOKEN_ID         = Integer.MAX_VALUE - 1;
   /** Represents CRLF and/or just LF. Considered a line terminator, not a separator. */
   public static final int EOL_TOKEN_ID        = Integer.MAX_VALUE - 2;
   /** Represents a character sequence having no particular interpretation. */
   public static final int TEXT_TOKEN_ID       = Integer.MAX_VALUE - 3;
   /** Represents an integer value, optionally preceded by a negative sign. */ 
   public static final int INTEGER_TOKEN_ID    = Integer.MAX_VALUE - 4;
   /** Represents a decimal value, optionally with a separator and/or a negative sign. */
   public static final int DECIMAL_TOKEN_ID    = Integer.MAX_VALUE - 5;

   /**
    * Initializes a new lexer object.
    * 
    * @param input    input stream
    * @param tokens   tokens to be recognized
    * @param escapes  escape characters, can be null or empty
    */
   public Lexer( Reader input, Collection<ImmutablePair<Integer,String>> tokens, String escapes )
   {
      m_input = new BufferedReader( input );
      init( tokens, escapes );
   }

   /**
    * Initializes a new lexer object with a specified size for the lexer's internal buffer.
    * 
    * @param input        input stream
    * @param tokens       tokens to be recognized
    * @param escapes      escape characters, can be null or empty
    * @param buffer_size  size of the input buffer to initialize  
    */
   public Lexer( Reader input, Collection<ImmutablePair<Integer,String>> tokens, String escapes, int buffer_size )
   {
      m_input = new BufferedReader( input, buffer_size );
      init( tokens, escapes );
   }

   /**
    * Iterate over tokens from the input stream until the first non-whitespace
    * token is encountered.
    */
   public void consumeWhitespace() throws IOException
   {
      ImmutablePair<Integer,String> token = getNext();
      int token_id = token.getLeft();
      while ((token_id == WS_TOKEN_ID || token_id == EOL_TOKEN_ID) && hasMore())
      {
         token = getNext();
         token_id = token.getLeft();
      }
      if (token_id != NULL_TOKEN_ID && token_id != WS_TOKEN_ID && token_id != EOL_TOKEN_ID)
         pushBack( token );
   }

   /**
    * Enables or disables recognition of the specified token.
    * 
    * @param token_id  the ID of the token to enable or disable
    * @param flag      pass true to enable, false to disable
    */
   public void enable( int token_id, boolean flag )
   {
      if (token_id == TEXT_TOKEN_ID)
         throw new IllegalArgumentException( Resources.getString( "msg.5" ) ); //$NON-NLS-1$
      if (flag)
         m_disabled.remove( token_id );
      else
         m_disabled.add( token_id );
   }

   /**
    * Enables or disables recognition of the specified tokens.
    * 
    * @param token_ids  the IDs of the tokens to enable or disable
    * @param flag       pass true to enable, false to disable
    */
   public void enable( int[] token_ids, boolean flag )
   {
      for (int id : token_ids)
         enable( id, flag );
   }

   /**
    * Enables or disables recognition of the specified tokens.
    * 
    * @param token_ids  the IDs of the tokens to enable or disable
    * @param flag       pass true to enable, false to disable
    */
   public void enable( Collection<Integer> token_ids, boolean flag )
   {
      for (int id : token_ids)
         enable( id, flag );
   }

   /**
    * Returns the current line number. Line numbers start at 1.
    */
   public int getLineNumber()
   {
      return m_linenum;
   }

   /**
    * Returns the next available token.
    */
   public ImmutablePair<Integer,String> getNext() throws IOException
   {
      if (!m_stack.isEmpty())
         return m_stack.pop();
      boolean is_escape = false;
      boolean was_escape = false;
      char last_char = '\0';
      while (m_stack.isEmpty() && hasMore())
      {
         is_escape = isEscape( m_curr_char );
         if (was_escape || !is_escape)
            processCurrentChar( last_char, was_escape );
         last_char = m_curr_char;
         was_escape = (is_escape && !was_escape) || (m_curr_char == '\r' && was_escape);
         m_curr_char = '\0';
      }
      if (m_buffer.length() > 0)
      {
         String token_text = m_buffer.toString();
         m_buffer.delete( 0, token_text.length() );
         return ImmutablePair.of( maybeRecognizeNumber( token_text ), token_text );
      }
      else if (!m_stack.isEmpty())
      {
         return m_stack.pop();
      }
      else
      {
         return ImmutablePair.of( NULL_TOKEN_ID, "" ); //$NON-NLS-1$
      }
   }

   /**
    * Returns true if this lexer has more tokens available.
    */
   public boolean hasMore() throws IOException
   {
      if (m_stack.size() > 0 || m_curr_char != '\0')
         return true;
      int ch = m_input.read();
      if (ch == -1)
         return false;
      m_curr_char = (char) ch;
      return true;
   }

   /**
    * Returns true if this lexer is configured to recognize the specified token.
    * 
    * @param token_id  the token ID to be tested
    */
   public boolean isEnabled( int token_id )
   {
      return token_id == TEXT_TOKEN_ID ||
             (!m_disabled.contains( token_id ) &&
              (token_id == WS_TOKEN_ID         ||
               token_id == EOL_TOKEN_ID        ||
               token_id == INTEGER_TOKEN_ID    ||
               token_id == DECIMAL_TOKEN_ID    ||
               m_table.containsValue( token_id )));
   }

   /**
    * Returns true if the specified character is one of this lexer's configured escape
    * characters.
    * 
    * @param ch  the character to test
    */
   public boolean isEscape( char ch )
   {
      return m_escapes.indexOf( ch ) > -1;
   }

   /**
    * Returns true if the specified character is whitespace. The set of characters
    * considered as whitespace by this method is a subset of those recognized by
    * Character.isWhitespace(), in that '\r' and '\n' are considered a separate
    * character class.
    * 
    * @param ch  the character to test
    */
   public boolean isWhitespace( char ch )
   {
      return Character.isWhitespace( ch ) && ch != '\n' && ch != '\r';
   }

   /**
    * Pushes the specified token back onto the lexer, such that it will be
    * returned in a subsequent call to getNext() as if it had come from the
    * input stream. Multiple tokens can be pushed, and are returned in LIFO
    * order.
    * 
    * @param token  the token to be pushed
    */
   public void pushBack( ImmutablePair<Integer,String> token )
   {
      m_stack.push( token );
   }

   /**
    * Sets the I/O reader for this lexer. The specified reader is wrapper in
    * a BufferReader, so the caller does not have to.
    *   
    * @param input
    */
   public void setInput( Reader input )
   {
      m_input = new BufferedReader( input );
      m_stack.clear();
      m_linenum = 1;
      m_inc_line = false;
      m_curr_char = '\0';
   }

   private void init( Collection<ImmutablePair<Integer,String>> tokens, String escapes )
   {
      for (ImmutablePair<Integer,String> token : tokens)
      {
         String text = token.getRight();
         m_table.put( text, token.getLeft() );
         if (text.length() == 1)
         {
            if (isWhitespace( text.charAt( 0 ) ))
               m_disabled.add( WS_TOKEN_ID );
            else if (text.charAt( 0 ) == '\r' || text.charAt( 0 ) == '\n')
               m_disabled.add( EOL_TOKEN_ID );
         }
      }
      if (escapes != null)
         m_escapes = escapes;
   }

   private int maybeRecognizeNumber( String text )
   {
      // try to recognize a decimal or integer value
      //
      if (isEnabled( INTEGER_TOKEN_ID ))
      {
         try
         {
            Long.parseLong( text );
            return INTEGER_TOKEN_ID;
         }
         catch( NumberFormatException ex )
         {
         }
      }
      if (isEnabled( DECIMAL_TOKEN_ID ))
      {
         try
         {
            new BigDecimal( text );
            return DECIMAL_TOKEN_ID;
         }
         catch( NumberFormatException ex )
         {
         }
      }
      
      // nope, call it a text token
      //
      return TEXT_TOKEN_ID;
   }

   private void processCurrentChar( char last_char, boolean was_escape )
   {
      if (m_curr_char == '\n')
      {
         if (m_inc_line)
            m_linenum++;
         m_inc_line = true;
         if (isEnabled( EOL_TOKEN_ID ) && !was_escape)
         {
            if (last_char == '\r')
            {
               int len = m_buffer.length();
               m_buffer.deleteCharAt( len - 1 );
               pushBack( ImmutablePair.of( EOL_TOKEN_ID, "\r\n" ) ); //$NON-NLS-1$
            }
            else
            {
               pushBack( ImmutablePair.of( EOL_TOKEN_ID, "\n" ) ); //$NON-NLS-1$
            }
            return;
         }
      }
      else if (isWhitespace( m_curr_char ) && !was_escape)
      {
         if (isEnabled( WS_TOKEN_ID ))
         {
            pushBack( ImmutablePair.of( WS_TOKEN_ID, Character.toString( m_curr_char ) ) );
            return;
         }
      }
      else if (m_inc_line)
      {
         m_linenum++;
         m_inc_line = false;
      }

      if ((m_curr_char != '\r' && m_curr_char != '\n') || ((m_curr_char == '\r' || m_curr_char == '\n') && !was_escape))
         m_buffer.append( m_curr_char );
      if (was_escape)
         return;

      String text = m_buffer.toString();
      int text_len = text.length();
      for (Map.Entry<String,Integer> entry : m_table.entrySet())
      {
         String key = entry.getKey();
         int id = entry.getValue();
         int pos = text.lastIndexOf( key, text_len - 1 );
         if (pos != -1 && isEnabled( id ) && pos + key.length() == text_len)
         {
            // token string must be a suffix
            pushBack( ImmutablePair.of( id, key ) );
            m_buffer.delete( pos, text_len );
            break;
         }
      }
   }

   private StringBuilder                        m_buffer    = new StringBuilder();
   private char                                 m_curr_char = '\0';
   private Set<Integer>                         m_disabled  = new HashSet<Integer>();
   private String                               m_escapes   = ""; //$NON-NLS-1$
   private boolean                              m_inc_line  = false;
   private Reader                               m_input     = new StringReader( "" ); //$NON-NLS-1$
   private int                                  m_linenum   = 1;
   private Stack<ImmutablePair<Integer,String>> m_stack     = new Stack<ImmutablePair<Integer,String>>();
   private Map<String,Integer>                  m_table     = new HashMap<String,Integer>();
}
