//
// ParseException.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.base;

import java.io.IOException;
import java.io.ObjectOutputStream;

/**
 * An exception class representing a parsing error.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
@SuppressWarnings( "serial" )
public class ParseException extends RuntimeException
{
   public ParseException( Throwable cause )
   {
      super( cause );
   }

   public ParseException( String details )
   {
      super( details );
      m_details = details;
      m_linenum = 0;
   }

   public ParseException( String details, Throwable cause )
   {
      super( details, cause );
      m_details = details;
      m_linenum = 0;
   }

   public ParseException( int linenum, String details )
   {
      super( createMessage( "", linenum, details, "" ) ); //$NON-NLS-1$ //$NON-NLS-2$
      m_details = details;
      m_linenum = linenum;
   }

   public ParseException( int linenum, String details, Throwable cause )
   {
      super( createMessage( "", linenum, details, "" ), cause ); //$NON-NLS-1$ //$NON-NLS-2$
      m_details = details;
      m_linenum = linenum;
   }

   public ParseException( String path, int linenum, String details )
   {
      super( createMessage( path, linenum, details, "" ) ); //$NON-NLS-1$
      m_details = details;
      m_linenum = linenum;
      m_path    = path;
   }

   public ParseException( String path, int linenum, String details, Throwable cause )
   {
      super( createMessage( path, linenum, details, "" ), cause ); //$NON-NLS-1$
      m_details = details;
      m_linenum = linenum;
      m_path    = path;
   }

   public ParseException( String details, String token )
   {
      super( createMessage( "", -1, details, token ) ); //$NON-NLS-1$
      m_details = details;
      m_token   = token;
   }

   public ParseException( String details, String token, Throwable cause )
   {
      super( createMessage( "", -1, details, token ), cause ); //$NON-NLS-1$
      m_details = details;
      m_token   = token;
   }

   public ParseException( String path, String details, String token )
   {
      super( createMessage( path, -1, details, token ) );
      m_details = details;
      m_token   = token;
      m_path    = path;
   }

   public ParseException( String path, String details, String token, Throwable cause )
   {
      super( createMessage( path, -1, details, token ), cause );
      m_details = details;
      m_token   = token;
      m_path    = path;
   }

   public ParseException( int linenum, String details, String token )
   {
      super( createMessage( "", linenum, details, token ) ); //$NON-NLS-1$
      m_details = details;
      m_token   = token;
      m_linenum = linenum;
   }

   public ParseException( int linenum, String details, String token, Throwable cause )
   {
      super( createMessage( "", linenum, details, token ), cause ); //$NON-NLS-1$
      m_details = details;
      m_token   = token;
      m_linenum = linenum;
   }

   public ParseException( String path, int linenum, String details, String token )
   {
      super( createMessage( path, linenum, details, token ) );
      m_details = details;
      m_path    = path;
      m_token   = token;
      m_linenum = linenum;
   }

   public ParseException( String path, int linenum, String details, String token, Throwable cause )
   {
      super( createMessage( path, linenum, details, token ), cause );
      m_details = details;
      m_path    = path;
      m_token   = token;
      m_linenum = linenum;
   }

   public String getDetails()
   {
      return m_details;
   }

   public int getLineNumber()
   {
      return m_linenum;
   }

   public String getPath()
   {
      return m_path;
   }

   public String getToken()
   {
      return m_token;
   }

   private static String createMessage( String path, int linenum, String details, String token )
   {
      StringBuilder buffer = new StringBuilder();
      if (!path.isEmpty())
      {
         buffer.append( path );
         buffer.append( ": " ); //$NON-NLS-1$
      }
      if (linenum > 0)
      {
         buffer.append( Resources.getString( "msg.3" ) ); //$NON-NLS-1$
         buffer.append( " " );  //$NON-NLS-1$
         buffer.append( linenum );
         buffer.append( ": " ); //$NON-NLS-1$
      }
      buffer.append( details );
      if (!token.isEmpty())
      {
         buffer.append( ": '" ); //$NON-NLS-1$
         buffer.append( token );
         buffer.append( "'" ); //$NON-NLS-1$
      }
      return buffer.toString();
   }

   private void writeObject( ObjectOutputStream output ) throws IOException
   {
      throw new UnsupportedOperationException( Resources.getString( "msg.4" ) ); //$NON-NLS-1$
   }

   private String m_details = ""; //$NON-NLS-1$
   private String m_path    = ""; //$NON-NLS-1$
   private String m_token   = ""; //$NON-NLS-1$
   private int    m_linenum = -1;
}
