//
// Board.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

/**
 * Represents a Go board. Points on the board are identified using
 * short integer values in the range {@code 0 <= point < total}.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class Board
{
   /**
    * The default board size used when game records or GTP engines do
    * not specify board size explicitly.
    */
   public static final int   DEFAULT_SIZE  = 19;

   /**
    * A sentinel value that does not represent any valid point.
    * One of its uses is to represent a pass.
    */
   public static final short INVALID_POINT = -1;

   /**
    * Returns true if two points are neighbors of each other given
    * the current board size, false otherwise.
    * 
    * @param point1
    * @param point2
    */
   public static boolean areNeighbors( short point1, short point2 )
   {
      checkPoint( point1 );
      checkPoint( point2 );
      if (point1 == point2)
         return false;
      int[][] ary = getNeighborArray();
      for (int i = 0; i < 4; i++)
      {
         if (ary[i][point1] == point2)
            return true;
      }
      return false;
   }

   /**
    * Verifies that the specified point is valid for the current board size.
    * Otherwise, throws IllegalArgumentException. Board.INVALID_POINT fails
    * this test.
    * 
    * @param point
    */
   public static void checkPoint( short point )
   {
      if (point >= 0 && point < getTotal())
         return;
      throw new IllegalArgumentException( Resources.formatString( "msg.1004", point ) ); //$NON-NLS-1$
   }

   /**
    * Returns a new point that is the transformation of the specified point
    * as a result of a vertical flip of the board.
    * 
    * @param point
    */
   public static int flipPoint( int point )
   {
      int offset = (point / m_size) * m_size;
      point -= offset;
      point = (short) ((m_size / 2 - point) * 2 + point);
      return (short) (point + offset);
   }

   /**
    * Classifies the type of point.
    * 
    * @param point
    */
   public static Locations getLocation( short point )
   {
      checkPoint( point );
      int ll = 0;
      int ul = m_size - 1;
      int lr = m_total - m_size;
      int ur = m_total - 1;
      int r = point % m_size;
      if (point == ll || point == ul || point == lr || point == ur)
         return Locations.CORNER;
      else if ((point > ll && point < ul) || (point > lr && point < ur))
         return Locations.EDGE;
      else if (r == ll || r == ul) // here ll and ul represent the top and bottom edges, respectively
         return Locations.EDGE;
      return Locations.INTERIOR;
   }

   /**
    * Updates the supplied array with the location of each point, where each such
    * point value serves as an index into the output array. The array must be
    * at least {@link #getTotal} elements in size.
    * 
    * @param points     collection of points to be classified
    * @param locations  output array
    * 
    * @see #getLocation
    */
   public static void getLocations( short[] points, Locations[] locations )
   {
      for (int i = 0; i < points.length; i++)
      {
         short pnt = points[i];
         locations[pnt] = getLocation( pnt );
      }
   }

   /**
    * Updates the supplied array with the location of each point, where each such
    * point value serves as an index into the output array. The array must be
    * at least {@link #getTotal} elements in size.
    * 
    * @param points     collection of points to be classified
    * @param locations  output array
    * 
    * @see #getLocation
    */
   public static void getLocations( Iterable<Short> points, Locations[] locations )
   {
      for (short pnt : points)
         locations[pnt] = getLocation( pnt );
   }

   /**
    * Returns the size of all boards (e.g., for a 19x19 board, this
    * method returns the value 19). This is a global value that affects
    * the creation of new {@link Board} objects and the translation of
    * string representations of points to point values.
    */
   public static int getSize()
   {
      return m_size;
   }

   /**
    * Sets the size of all boards (e.g., a value of 19 represents a
    * 19x19 board). This is a global value that affects the creation
    * of new {@link Board} objects and the translation of string
    * representations of points to point values.
    */
   public static void setSize( int size )
   {
      if (!isValidSize( size ))
         throw new IllegalArgumentException( Resources.formatString( "msg.1008", size ) ); //$NON-NLS-1$
      m_size = size;
      m_total = (short) (m_size * m_size);
      initRotationArray( getRotationArray() );
      initNeighborArray( getNeighborArray() );
      initDiagonalArray( getDiagonalArray() );
   }

   /**
    * Returns the total number of points on a board.
    */
   public static int getTotal()
   {
      return m_total;
   }

   /**
    * Returns true if the specified size is supported by Dakengo. A board
    * size must be an odd value in the range 3 <= size <= 19.
    */
   public static boolean isValidSize( int size )
   {
      return size >= 3 && size <= 19 && (size & 1) == 1;
   }

   /**
    * Initializes a new, empty board.
    */
   public Board()
   {
      m_data = new int[m_total];
      clear();
   }

   /**
    * Creates a new board that is a copy of the original.
    * 
    * @param other
    */
   public Board( Board other )
   {
      m_data = new int[m_total];
      copyFrom( other );
   }

   /**
    * Copy the specified board into this board.
    * 
    * @param other
    */
   public void copyFrom( Board other )
   {
      System.arraycopy( other.m_data, 0, m_data, 0, m_total );
      m_black_pnts = new PointLinkedList( m_data, other.m_black_pnts );
      m_empty_pnts = new PointLinkedList( m_data, other.m_empty_pnts );
      m_white_pnts = new PointLinkedList( m_data, other.m_white_pnts );
      m_hash = other.m_hash;
   }

   /**
    * Set all points to Color.EMPTY.
    */
   public void clear()
   {
      m_hash = 0L;
      m_black_pnts = new PointLinkedList( m_data, PointLinkedList.BLACK_VALUE );
      m_white_pnts = new PointLinkedList( m_data, PointLinkedList.WHITE_VALUE );
      m_empty_pnts = new PointLinkedList( m_data, PointLinkedList.EMPTY_VALUE );
      m_empty_pnts.reserve();
   }

   /**
    * Returns the color of the specified point, or Colors.EMPTY
    * if there is no stone on the point.
    * 
    * @param point
    */
   public Colors getColor( short point )
   {
      checkPoint( point );
      int value = m_data[point] & PointLinkedList.COLOR_MASK;
      if (value == PointLinkedList.BLACK_VALUE)
         return Colors.BLACK;
      if (value == PointLinkedList.WHITE_VALUE)
         return Colors.WHITE;
      return Colors.EMPTY;
   }

   /**
    * Sets the color of the specified point.
    * 
    * @param point
    * @param color
    */
   public void setColor( short point, Colors color )
   {
      checkPoint( point );
      m_hash = 0L;
      setColorInternal( point, color );
   }

   /**
    * Sets a collection of points to the specified color.
    * 
    * @param points
    * @param color
    */
   public void setColor( short[] points, Colors color )
   {
      m_hash = 0L;
      for (int i = 0; i < points.length; i++)
         setColorInternal( points[i], color );
   }

   /**
    * Sets a collection of points to the specified color.
    * 
    * @param points
    * @param color
    */
   public void setColor( Iterable<Short> points, Colors color )
   {
      m_hash = 0L;
      for (short pnt : points)
         setColorInternal( pnt, color );
   }

   /**
    * Returns the number of points occupied by the specified color.
    * 
    * @param color
    */
   public int getCount( Colors color )
   {
      switch (color)
      {
         case BLACK: return m_black_pnts.size();
         case WHITE: return m_white_pnts.size();
         case EMPTY: return m_empty_pnts.size();
         default:    return 0;
      }
   }

   /**
    * Returns the full Zobrist hash for this board. The hash is computed
    * the first time it's needed, and then cached. An empty board will
    * return a hash value of 0.
    */
   public long getZobrist()
   {
      if (m_hash != 0L)
         return m_hash;
      for (int i = 0; i < m_total; i++)
      {
         int color = m_data[i] & PointLinkedList.COLOR_MASK;
         if (color == PointLinkedList.BLACK_VALUE)
            m_hash ^= Constants.BLACK_POS_VALUES[i];
         else if (color == PointLinkedList.WHITE_VALUE)
            m_hash ^= Constants.WHITE_POS_VALUES[i];
      }
      return m_hash;
   }

   /**
    * Apply reflection and/or rotation transformations to produce a new
    * board. The rotate parameter accepts an integer specifying
    * the number of clockwise (if positive) or counter-clockwise
    * (if negative) 90-degree rotations. Setting the flip parameter
    * to true causes the board to be flipped vertically.
    */
   public Board transform( int rotate, boolean flip )
   {
      // A 3x3 example of transforms:
      //
      //  X . O
      //  . O X   2,0,1,0,2,0,0,1,2
      //  O . .
      //
      //  O . X
      //  . O .   0,0,2,1,2,0,2,0,1
      //  . X O
      //
      //  . . O
      //  X O .   2,1,0,0,2,0,1,0,2
      //  O . X
      //
      //  O X .
      //  . O .   1,0,2,0,2,1,2,0,0
      //  X . O
      //
      //  O . .
      //  . O X   1,0,2,0,2,0,2,1,0  (inverted)
      //  X . O

      rotate = rotate % 4;
      if (!flip && rotate == 0)
         return new Board( this );

      Board board = new Board();
      if (flip)
      {
         for (int orig_pnt = m_black_pnts.getHead(); orig_pnt >= 0; orig_pnt = m_black_pnts.getNext( orig_pnt ))
         {
            int new_pnt = flipPoint( orig_pnt );
            board.m_empty_pnts.remove( new_pnt );
            board.m_black_pnts.insert( new_pnt );
         }
         for (int orig_pnt = m_white_pnts.getHead(); orig_pnt >= 0; orig_pnt = m_white_pnts.getNext( orig_pnt ))
         {
            int new_pnt = flipPoint( orig_pnt );
            board.m_empty_pnts.remove( new_pnt );
            board.m_white_pnts.insert( new_pnt );
         }
      }
      else
      {
         board.copyFrom( this );
      }

      if (rotate == 0)
         return board;
      switch( rotate )
      {
         case -1: rotate = 3; break;
         case -2: rotate = 2; break;
         case -3: rotate = 1; break;
         default:             break;
      }
      rotate--;
      int[][] ary = getRotationArray();
      int[] new_data = new int[m_total];
      PointLinkedList new_black_pnts = new PointLinkedList( new_data, PointLinkedList.BLACK_VALUE );
      PointLinkedList new_white_pnts = new PointLinkedList( new_data, PointLinkedList.WHITE_VALUE );
      PointLinkedList new_empty_pnts = new PointLinkedList( new_data, PointLinkedList.EMPTY_VALUE );
      new_empty_pnts.reserve();
      for (int orig_pnt = board.m_black_pnts.getHead(); orig_pnt >= 0; orig_pnt = board.m_black_pnts.getNext( orig_pnt ))
      {
         int new_pnt = ary[rotate][orig_pnt];
         new_empty_pnts.remove( new_pnt );
         new_black_pnts.insert( new_pnt );
      }
      for (int orig_pnt = board.m_white_pnts.getHead(); orig_pnt >= 0; orig_pnt = board.m_white_pnts.getNext( orig_pnt ))
      {
         int new_pnt = ary[rotate][orig_pnt];
         new_empty_pnts.remove( new_pnt );
         new_white_pnts.insert( new_pnt );
      }
      board.m_data       = new_data;
      board.m_black_pnts = new_black_pnts;
      board.m_empty_pnts = new_empty_pnts;
      board.m_white_pnts = new_white_pnts;

      return board;
   }

   /**
    * Calls {@code friendly_visitor} for each neighbor of the specified point that has the same color. Calls
    * {@code lib_visitor} for each empty point adjacent to a friendly point.
    * 
    * @param point              a point in a chain to be traversed
    * @param friendly_visitor   visitor for points in the chain, can be null
    * @param lib_visitor        visitor for liberties, can be null
    */
   public void visitChain( short point, PointVisitor friendly_visitor, PointVisitor lib_visitor )
   {
      checkPoint( point );
      for (int i = 0; i < m_visited.length; i++)
         m_visited[i] = false;
      int[][] ary = getNeighborArray();
      int pnt_color = m_data[point] & PointLinkedList.COLOR_MASK;

      int index = 0;
      m_chain_pnts[index] = point;
      while (index >= 0)
      {
         point = m_chain_pnts[index--];
         m_visited[point] = true;
         for (int i = 0; i < 4; i++)
         {
            int nbr = ary[i][point];
            if (nbr == INVALID_POINT || m_visited[nbr])
               continue;
            m_visited[nbr] = true;
            int nbr_color = m_data[nbr] & PointLinkedList.COLOR_MASK;
            if (nbr_color == pnt_color)
            {
               m_chain_pnts[++index] = (short) nbr;
               if (friendly_visitor != null && !friendly_visitor.accept( this, (short) nbr ))
                  return;
            }
            else if (nbr_color == PointLinkedList.EMPTY_VALUE)
            {
               if (lib_visitor != null && !lib_visitor.accept( this, (short) nbr ))
                  return;
            }
         }
      }
   }

   /**
    * Calls the visitor for each diagonal of the specified point.
    * 
    * @param point
    * @param visitor
    */
   public void visitDiagonals( short point, PointVisitor visitor )
   {
      checkPoint( point );
      int[][] ary = getDiagonalArray();
      for (int i = 0; i < 4; i++)
      {
         int diag = ary[i][point];
         if (diag != INVALID_POINT && !visitor.accept( this, (short) diag ))
            return;
      }
   }

   /**
    * Calls the visitor for each neighbor of the specified point.
    * 
    * @param point
    * @param visitor
    */
   public void visitNeighbors( short point, PointVisitor visitor )
   {
      checkPoint( point );
      int[][] ary = getNeighborArray();
      for (int i = 0; i < 4; i++)
      {
         int nbr = ary[i][point];
         if (nbr != INVALID_POINT && !visitor.accept( this, (short) nbr ))
            return;
      }
   }

   /**
    * Calls the visitor for each point occupied by a stone
    * of the specified color. Pass Colors.EMPTY to get the empty points. Pass
    * Colors.DONT_CARE to get all points.
    * 
    * @param color
    * @param visitor
    */
   public void visitPoints( Colors color, PointVisitor visitor )
   {
      if (color == Colors.DONT_CARE)
      {
         for (short pnt = 0; pnt < m_total && visitor.accept( this, pnt ); pnt++) {}
         return;
      }
      PointLinkedList list = null;
      switch (color)
      {
         case BLACK: list = m_black_pnts; break;
         case EMPTY: list = m_empty_pnts; break;
         case WHITE: list = m_white_pnts; break;
         default:                         break;
      }
      for (int pnt = list.getHead(); pnt >= 0 && visitor.accept( this, (short) pnt); pnt = list.getNext( pnt )) {}
   }

   /**
    * Calls the visitor for each of the specified points.
    * 
    * @param points
    * @param visitor
    */
   public void visitPoints( short[] points, PointVisitor visitor )
   {
      for (int i = 0; i < points.length; i++)
      {
         short pnt = points[i];
         checkPoint( pnt );
         if (!visitor.accept( this, pnt ))
            return;
      }
   }

   /**
    * Calls the visitor for each of the specified points.
    * 
    * @param points
    * @param visitor
    */
   public void visitPoints( Iterable<Short> points, PointVisitor visitor )
   {
      for (short pnt : points)
      {
         checkPoint( pnt );
         if (!visitor.accept( this, pnt ))
            return;
      }
   }

   /**
    * Truncates the Zobrist hash. Use the getZobrist() method when the full
    * Zorbrist hash is required.
    */
   @Override
   public final int hashCode()
   {
      return (int) getZobrist();
   }

   @Override
   public final boolean equals( Object obj )
   {
      if (!(obj instanceof Board))
         return false;
      if (this == obj)
         return true;
      Board other = (Board) obj;
      if (m_hash != 0 && other.m_hash != 0 && m_hash != other.m_hash)
         return false;
      return m_black_pnts.equals( other.m_black_pnts ) && m_white_pnts.equals( other.m_white_pnts );
   }

   @Override
   public final String toString()
   {
      return String.format( "%s,%s,%X", m_black_pnts.toString(), m_white_pnts.toString(), m_hash ); //$NON-NLS-1$
   }

   private static int[][] getRotationArray()
   {
      int[][] ary = null;
      switch (m_size)
      {
         case 3:   ary = SIZE3_INDEXES;  break;
         case 5:   ary = SIZE5_INDEXES;  break;
         case 7:   ary = SIZE7_INDEXES;  break;
         case 9:   ary = SIZE9_INDEXES;  break;
         case 11:  ary = SIZE11_INDEXES; break;
         case 13:  ary = SIZE13_INDEXES; break;
         case 15:  ary = SIZE15_INDEXES; break;
         case 17:  ary = SIZE17_INDEXES; break;
         case 19:  ary = SIZE19_INDEXES; break;
         default:
            throw new IllegalArgumentException( Resources.formatString( "msg.1020", m_size ) ); //$NON-NLS-1$
      }
      return ary;
   }

   private static void initRotationArray( int[][] ary )
   {
      if (ary[0] != null)
         return;
      ary[0] = new int[m_total];
      ary[1] = new int[m_total];
      ary[2] = new int[m_total];
      for (int col = 0; col < m_size; col++)
      {
         for (int row = 0; row < m_size; row++)
         {
            int dest = col * m_size + row;
            ary[0][dest] = row * m_size + m_size - col - 1;
            ary[1][dest] = (m_size - col - 1) * m_size + m_size - row - 1;
            ary[2][dest] = (m_size - row - 1) * m_size + col;
         }
      }
   }

   private static int[][] getNeighborArray()
   {
      int[][] ary = null;
      switch( m_size )
      {
         case 3:   ary = SIZE3_NBRS;  break;
         case 5:   ary = SIZE5_NBRS;  break;
         case 7:   ary = SIZE7_NBRS;  break;
         case 9:   ary = SIZE9_NBRS;  break;
         case 11:  ary = SIZE11_NBRS; break;
         case 13:  ary = SIZE13_NBRS; break;
         case 15:  ary = SIZE15_NBRS; break;
         case 17:  ary = SIZE17_NBRS; break;
         case 19:  ary = SIZE19_NBRS; break;
         default:
            throw new IllegalArgumentException( Resources.formatString( "msg.1020", m_size ) ); //$NON-NLS-1$
      }
      return ary;
   }

   private static void initNeighborArray( int[][] ary )
   {
      if (ary[0] != null)
         return;
      ary[0] = new int[m_total];
      ary[1] = new int[m_total];
      ary[2] = new int[m_total];
      ary[3] = new int[m_total];
      ary[0][0] = m_total + 1;
      for (int pnt = 0; pnt < m_total; pnt++)
      {
         int col = pnt / m_size;

         // south of point
         //
         int nbr = pnt - 1;
         if (nbr >= 0 && nbr / m_size == col)
            ary[Directions.SOUTH.ordinal()][pnt] = nbr;
         else
            ary[Directions.SOUTH.ordinal()][pnt] = Board.INVALID_POINT;

         // north of point
         //
         nbr = pnt + 1;
         if (nbr < m_total && nbr / m_size == col)
            ary[Directions.NORTH.ordinal()][pnt] = nbr;
         else
            ary[Directions.NORTH.ordinal()][pnt] = Board.INVALID_POINT;

         // west of point
         //
         nbr = pnt - m_size;
         if (nbr >= 0)
            ary[Directions.WEST.ordinal()][pnt] = nbr;
         else
            ary[Directions.WEST.ordinal()][pnt] = Board.INVALID_POINT;

         // east of point
         //
         nbr = pnt + m_size;
         if (nbr < m_total)
            ary[Directions.EAST.ordinal()][pnt] = nbr;
         else
            ary[Directions.EAST.ordinal()][pnt] = Board.INVALID_POINT;
      }
   }

   private static int[][] getDiagonalArray()
   {
      int[][] ary = null;
      switch( m_size )
      {
         case 3:   ary = SIZE3_DIAGS;  break;
         case 5:   ary = SIZE5_DIAGS;  break;
         case 7:   ary = SIZE7_DIAGS;  break;
         case 9:   ary = SIZE9_DIAGS;  break;
         case 11:  ary = SIZE11_DIAGS; break;
         case 13:  ary = SIZE13_DIAGS; break;
         case 15:  ary = SIZE15_DIAGS; break;
         case 17:  ary = SIZE17_DIAGS; break;
         case 19:  ary = SIZE19_DIAGS; break;
         default:
            throw new IllegalArgumentException( Resources.formatString( "msg.1020", m_size ) ); //$NON-NLS-1$
      }
      return ary;
   }

   private static void initDiagonalArray( int[][] ary )
   {
      if (ary[0] != null)
         return;
      ary[0] = new int[m_total];
      ary[1] = new int[m_total];
      ary[2] = new int[m_total];
      ary[3] = new int[m_total];
      ary[0][0] = m_total + 1;
      for (int pnt = 0; pnt < m_total; pnt++)
      {
         int col = pnt / m_size;

         // south-east of point
         //
         int index = Directions.SOUTH_EAST.ordinal() - 4;
         int diag = pnt + m_size - 1;
         if (diag < m_total && diag / m_size == col + 1)
            ary[index][pnt] = diag;
         else
            ary[index][pnt] = Board.INVALID_POINT;

         // south-west of point
         //
         index = Directions.SOUTH_WEST.ordinal() - 4;
         diag = pnt - m_size - 1;
         if (diag >= 0 && diag / m_size == col - 1)
            ary[index][pnt] = diag;
         else
            ary[index][pnt] = Board.INVALID_POINT;

         // north-east of point
         //
         index = Directions.NORTH_EAST.ordinal() - 4;
         diag = pnt + m_size + 1;
         if (diag < m_total && diag / m_size == col + 1)
            ary[index][pnt] = diag;
         else
            ary[index][pnt] = Board.INVALID_POINT;

         // north-west of point
         //
         index = Directions.NORTH_WEST.ordinal() - 4;
         diag = pnt - m_size + 1;
         if (diag >= 0 && diag / m_size == col - 1)
            ary[index][pnt] = diag;
         else
            ary[index][pnt] = Board.INVALID_POINT;
      }
   }

   private void setColorInternal( short point, Colors color )
   {
      PointLinkedList src_list;
      switch (m_data[point] & PointLinkedList.COLOR_MASK)
      {
         case PointLinkedList.BLACK_VALUE: src_list = m_black_pnts; break;
         case PointLinkedList.WHITE_VALUE: src_list = m_white_pnts; break;
         default:                          src_list = m_empty_pnts; break;
      }

      PointLinkedList dest_list;
      switch (color)
      {
         case BLACK:
            dest_list = m_black_pnts;
            break;
         case EMPTY:
            dest_list = m_empty_pnts;
            break;
         case WHITE:
            dest_list = m_white_pnts;
            break;
         default:
            throw new IllegalArgumentException( Resources.formatString( "msg.1019", color ) ); //$NON-NLS-1$
      }

      src_list.remove( point );
      dest_list.insert( point );
   }

   static
   {
      m_size = DEFAULT_SIZE;
      m_total = DEFAULT_SIZE * DEFAULT_SIZE;
      SIZE3_INDEXES  = new int[3][];
      SIZE5_INDEXES  = new int[3][];
      SIZE7_INDEXES  = new int[3][];
      SIZE9_INDEXES  = new int[3][];
      SIZE11_INDEXES = new int[3][];
      SIZE13_INDEXES = new int[3][];
      SIZE15_INDEXES = new int[3][];
      SIZE17_INDEXES = new int[3][];
      SIZE19_INDEXES = new int[3][];
      SIZE3_NBRS     = new int[4][];
      SIZE5_NBRS     = new int[4][];
      SIZE7_NBRS     = new int[4][];
      SIZE9_NBRS     = new int[4][];
      SIZE11_NBRS    = new int[4][];
      SIZE13_NBRS    = new int[4][];
      SIZE15_NBRS    = new int[4][];
      SIZE17_NBRS    = new int[4][];
      SIZE19_NBRS    = new int[4][];
      SIZE3_DIAGS    = new int[4][];
      SIZE5_DIAGS    = new int[4][];
      SIZE7_DIAGS    = new int[4][];
      SIZE9_DIAGS    = new int[4][];
      SIZE11_DIAGS   = new int[4][];
      SIZE13_DIAGS   = new int[4][];
      SIZE15_DIAGS   = new int[4][];
      SIZE17_DIAGS   = new int[4][];
      SIZE19_DIAGS   = new int[4][];
      initRotationArray( getRotationArray() );
      initNeighborArray( getNeighborArray() );
      initDiagonalArray( getDiagonalArray() );
   }

   private static final int[][] SIZE3_INDEXES;
   private static final int[][] SIZE5_INDEXES;
   private static final int[][] SIZE7_INDEXES;
   private static final int[][] SIZE9_INDEXES;
   private static final int[][] SIZE11_INDEXES;
   private static final int[][] SIZE13_INDEXES;
   private static final int[][] SIZE15_INDEXES;
   private static final int[][] SIZE17_INDEXES;
   private static final int[][] SIZE19_INDEXES;

   private static final int[][] SIZE3_NBRS;
   private static final int[][] SIZE5_NBRS;
   private static final int[][] SIZE7_NBRS;
   private static final int[][] SIZE9_NBRS;
   private static final int[][] SIZE11_NBRS;
   private static final int[][] SIZE13_NBRS;
   private static final int[][] SIZE15_NBRS;
   private static final int[][] SIZE17_NBRS;
   private static final int[][] SIZE19_NBRS;

   private static final int[][] SIZE3_DIAGS;
   private static final int[][] SIZE5_DIAGS;
   private static final int[][] SIZE7_DIAGS;
   private static final int[][] SIZE9_DIAGS;
   private static final int[][] SIZE11_DIAGS;
   private static final int[][] SIZE13_DIAGS;
   private static final int[][] SIZE15_DIAGS;
   private static final int[][] SIZE17_DIAGS;
   private static final int[][] SIZE19_DIAGS;

   private static int m_size;
   private static int m_total;

   private int             m_data[];
   private PointLinkedList m_black_pnts;
   private PointLinkedList m_white_pnts;
   private PointLinkedList m_empty_pnts;
   private long            m_hash        = 0L;

   // for temporary use during visit***() methods; these do
   // not need to be copied
   private short[]         m_chain_pnts  = new short[Board.DEFAULT_SIZE * Board.DEFAULT_SIZE];
   private boolean[]       m_visited     = new boolean[Board.DEFAULT_SIZE * Board.DEFAULT_SIZE];
}
