//
// Chain.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import com.dakengo.libs.base.HashUtils;

/**
 * "A chain is a group of stones that are all connected along the lines of the board."<p>
 *
 * "There is a strict connection between two stones if there is a path between them
 * so that all elements of the path have the same color."
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 *
 * @see <a href="http://senseis.xmp.net/?Chain">Chain</a>
 * @see <a href="http://senseis.xmp.net/?StrictConnection">StrictConnection</a>
 */
public class Chain
{
   /**
    * Initializes an empty chain.
    */
   public Chain()
   {
   }

   /**
    * Initializes a new chain object by inspecting the board at the specified point
    * and collecting all of stones of the same color connected to the point. Also
    * collects the immediate liberties of the chain, if any.
    *
    * @param board
    * @param point
    */
   public Chain( Board board, short point )
   {
      board.visitChain( point, m_pnt_collector, m_lib_collector );
      m_points.add( point );
   }

   /**
    * Returns true if the specified point is strictly connected to this
    * chain, or if the point is already in this chain. This method does
    * not consider whether the specified point is occupied by the
    * correct color.
    * 
    * @param point
    */
   public boolean connects( short point )
   {
      Board.checkPoint( point );
      return m_points.contains( point ) || m_libs.contains( point );
   }

   /**
    * Returns true if any pair of points, where one point comes from this chain and
    * the other point from the specified chain, are strictly connected. This method
    * also returns true if the two chains are overlapping. This method does not
    * consider whether the two chains are the same color.
    *
    * @param other
    */
   public boolean connects( Chain other )
   {
      for (short pnt : m_points)
      {
         if (other.m_points.contains( pnt ) || other.m_libs.contains( pnt ))
            return true;
      }
      return false;
   }

   /**
    * Returns true if the specified point is a member of this chain.
    *
    * @param point
    */
   public boolean contains( short point )
   {
      Board.checkPoint( point );
      return m_points.contains( point );
   }

   /**
    * Returns one of this chain's points as an identifier for the chain.
    * Returns Board.INVALID_POINT if this chain is empty.
    */
   public short getAnchor()
   {
      if (m_points.isEmpty())
         return Board.INVALID_POINT;
      return m_points.last();
   }

   /**
    * Returns the number of points comprising this chain.
    */
   public int getCount()
   {
      return m_points.size();
   }

   /**
    * Returns the points representing all direct liberties of this chain.
    */
   public Set<Short> getLiberties()
   {
      return m_libs;
   }

   /**
    * Returns the points comprising this chain.
    */
   public Set<Short> getPoints()
   {
      return m_points;
   }

   @Override
   public boolean equals( Object obj )
   {
      if (!(obj instanceof Chain))
         return false;
      if (this == obj)
         return true;
      Chain other = (Chain) obj;
      return m_points.equals( other.m_points );
   }

   @Override
   public int hashCode()
   {
      return HashUtils.computeAnyOrder( m_points ); 
   }

   @Override
   public String toString()
   {
      return String.format( "%s:%s", Convert.pointsToString( m_points, "," ), Convert.pointsToString( m_libs, "," ) ); //$NON-NLS-1$
   }

   private PointVisitor m_pnt_collector = new PointVisitor()
   {
      @Override
      public boolean accept( Board board, short point )
      {
         m_points.add( point );
         return true;
      }
   };

   private PointVisitor m_lib_collector = new PointVisitor()
   {
      @Override
      public boolean accept( Board board, short point )
      {
         m_libs.add( point );
         return true;
      }
   };

   private Set<Short>       m_libs   = new HashSet<Short>();
   private SortedSet<Short> m_points = new TreeSet<Short>();
}
