//
// ChainFinder.java
//
// Copyright (c) 2012-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.collections.map.MultiValueMap;

/**
 * A utility class for identifying candidate eyespace.
 *
 * @see <a href="http://senseis.xmp.net/?Eyespace">Eyespace</a>
 */
public class ChainFinder
{
   private class ChainCollector implements PointVisitor
   {
      public ChainCollector( Map<Short,Chain> chains, HashSet<Short> history )
      {
         m_chains = chains;
         m_history = history;
      }

      @Override
      public boolean accept( Board board, short point )
      {
         if (m_history.contains( point ))
            return true;
         Chain chain = new Chain( board, point );
         for (short pnt : chain.getPoints())
            m_history.add( pnt );
         m_chains.put( chain.getAnchor(), chain );
         return true;
      }

      private Map<Short,Chain> m_chains;
      private HashSet<Short>   m_history;
   }

   public ChainFinder()
   {
   }

   /**
    * Find all chains on the board in the specified color. Optionally
    * find corresponding candidate eyespaces.
    *
    * @param board
    * @param color
    * @param plus_eyespace
    */
   public void find( Board board, Colors color, boolean plus_eyespace )
   {
      Game.checkColor( color );
      m_chains.clear();
      m_spaces.clear();

      HashSet<Short> history = new HashSet<Short>();
      ChainCollector chain_coll = new ChainCollector( m_chains, history );
      board.visitPoints( color, chain_coll );

      if (plus_eyespace)
      {
         for (Map.Entry<Short,Chain> entry : m_chains.entrySet())
         {
            Chain chain = entry.getValue();
            if (chain.getLiberties().isEmpty())
               continue;
            findEyeSpace( board, chain ); 
         }
      }
   }

   /**
    * Find candidate eyespaces on the board for the specified chain.
    *
    * @param board
    * @param chain
    */
   public void find( Board board, Chain chain )
   {
      m_chains.clear();
      m_spaces.clear();
      m_chains.put( chain.getAnchor(), chain );
      if (chain.getLiberties().isEmpty())
         return;
      findEyeSpace( board, chain );
   }

   /**
    * Returns a map in which each key is a chain anchor point and each
    * value is the corresponding chain object.
    */
   public Map<Short,Chain> getChains()
   {
      return m_chains;
   }

   /**
    * Return the multi-map of eyespaces, where each key is a chain anchor
    * point, associated with a collection of point lists being the eyespaces
    * for that chain.
    */
   public MultiValueMap getSpaces()
   {
      return m_spaces;
   }

   @SuppressWarnings( "unchecked" )
   private void findEyeSpace( Board board, Chain chain )
   {
      short anchor = chain.getAnchor();
      if (m_spaces.containsKey( anchor ))
         return;
      HashSet<Short> libs = new HashSet<Short>( chain.getLiberties() );
      int max_count = 0;
      final HashSet<Short> tmp = new HashSet<Short>();
      PointVisitor lib_collector = new PointVisitor() {
         @Override
         public boolean accept( Board b, short p )
         {
            tmp.add( p );
            return true;
         }
      };
      while (libs.size() > 0)
      {
         tmp.clear();
         short next_pnt = libs.iterator().next();
         tmp.add( next_pnt );
         board.visitChain( next_pnt, lib_collector, null );
         if (tmp.size() > max_count)
            max_count = tmp.size();
         m_spaces.put( anchor, new HashSet<Short>( tmp ) );
         libs.removeAll( tmp );
      }
      if (m_spaces.getCollection( anchor ).size() <= 1)
      {
         m_spaces.remove( anchor );
         return;
      }
      boolean keep_going = true;
      while (keep_going)
      {
         keep_going = false;
         Iterator<?> iter = m_spaces.iterator( anchor );
         while (iter.hasNext())
         {
            HashSet<Short> pnts = (HashSet<Short>) iter.next();
            if (pnts.size() >= max_count)
            {
               m_spaces.remove( anchor, pnts );
               keep_going = true;
               break;
            }
         }
      }
   }

   private Map<Short,Chain> m_chains = new HashMap<Short,Chain>();
   private MultiValueMap    m_spaces = new MultiValueMap();
}
