//
// Convert.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.dakengo.libs.base.ParseException;

/**
 * Helper methods to convert primitive data types into and out of
 * standard string representations.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public abstract class Convert
{
   public static Colors charToColor( char ch )
   {
      return charToColor( ch, 0 );
   }

   public static Colors charToColor( char ch, int linenum )
   {
      switch( ch )
      {
         case 'B': case 'b': return Colors.BLACK;
         case '?':           return Colors.DONT_CARE;
         case '.':           return Colors.EMPTY;
         case 'W': case 'w': return Colors.WHITE;
         default:            throwInvalidColor( linenum, Character.toString( ch ) );
      }
      return Colors.DONT_CARE;
   }

   public static char colorToChar( Colors color )
   {
      switch( color )
      {
         case BLACK:
            return 'B';
         case DONT_CARE:
            return '?';
         case EMPTY:
            return '.';
         case WHITE:
            return 'W';
         default:
            return '*';
      }
   }

   public static Colors stringToColor( String text )
   {
      return stringToColor( text, 0 );
   }

   public static Colors stringToColor( String text, int linenum )
   {
      try
      {
         if (text.length() == 1)
            return charToColor( text.charAt( 0 ), linenum );
         else
            return Colors.valueOf( text.toUpperCase() );
      }
      catch (Throwable t)
      {
         throwInvalidColor( linenum, text );
      }
      return Colors.DONT_CARE;
   }

   public static String colorToString( Colors color )
   {
      switch (color)
      {
         case BLACK:
         case EMPTY:
         case WHITE:
            return color.toString().toLowerCase();
         case DONT_CARE:
            return "?";  //$NON-NLS-1$
         default:
            return Resources.formatString( "msg.1019", color ); //$NON-NLS-1$
      }
   }

   public static short stringToPoint( String text )
   {
      return stringToPoint( text, 0 );
   }

   public static short stringToPoint( String text, int linenum )
   {
      int len = text.length();
      if (len < 2 || len > 3)
         throwInvalidPointString( linenum, text );
      String copy = text.toLowerCase();
      char letter = copy.charAt( 0 );
      int col = 0;
      if (letter >= 'i')
         col = letter - 'a' - 1;
      else
         col = letter - 'a';
      if (col < 0 || col >= Board.getSize())
         throwInvalidPointString( linenum, text );
      int row = 0;
      try
      {
         row = Integer.parseInt( copy.substring( 1 ) );
      }
      catch( NumberFormatException ex )
      {
         throwInvalidPointString( linenum, text );
      }
      if (row <= 0 || row > Board.getSize())
         throwInvalidPointString( linenum, text );
      return (short) (col * Board.getSize() + row - 1);
   }

   public static String pointToString( short point )
   {
      if (point == Board.INVALID_POINT)
         return "pass"; //$NON-NLS-1$
      Board.checkPoint( point );
      int col = point / Board.getSize();
      int row = point % Board.getSize();
      String text = ""; //$NON-NLS-1$
      if (col >= 8)
         text = Character.toString( (char) ('a' + col + 1) );
      else
         text = Character.toString( (char) ('a' + col) );
      text += Integer.toString( row + 1 );
      return text;
   }

   /**
    * Parses the string representation of a collection of points.
    * 
    * @param text
    * @param separator
    * 
    * @return a collection of points.
    */
   public static Set<Short> stringToPoints( String text, String separator )
   {
      return stringToPoints( text, separator, 0 );
   }

   /**
    * Parses the string representation of a collection of points.
    * 
    * @param text
    * @param separator
    * 
    * @return a collection of points.
    */
   public static Set<Short> stringToPoints( String text, String separator, int linenum )
   {
      if (separator.isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.1026" ) ); //$NON-NLS-1$
      Set<Short> points = new HashSet<Short>();
      String[] tmp = text.split( separator );
      for (String pnt_str : tmp)
      {
         if (!pnt_str.isEmpty())
            points.add( stringToPoint( pnt_str, linenum ) );
      }
      return points;
   }

   /**
    * Produces the string representation of a collection of points.
    *
    * @param points     the points to print
    * @param separator  character sequence to use to separate list elements
    */
   public static String pointsToString( short[] points, String separator )
   {
      StringBuilder buffer = new StringBuilder();
      boolean first = true;
      for (int i = 0; i < points.length; i++)
      {
         if (first)
            first = false;
         else
            buffer.append( separator );
         buffer.append( pointToString( points[i] ) );
      }
      return buffer.toString();
   }

   /**
    * Produces the string representation of a collection of points.
    *
    * @param points     the points to print
    * @param separator  character sequence to use to separate list elements
    */
   public static String pointsToString( Iterable<Short> points, String separator )
   {
      StringBuilder buffer = new StringBuilder();
      boolean first = true;
      for (short pnt : points)
      {
         if (first)
            first = false;
         else
            buffer.append( separator );
         buffer.append( pointToString( pnt ) );
      }
      return buffer.toString();
   }

   public static Set<Short> stringsToPoints( String... strings )
   {
      Set<Short> points = new HashSet<Short>();
      for (String text : strings)
         points.add( stringToPoint( text ) );
      return points;
   }

   public static Set<Short> stringsToPoints( Iterable<String> strings )
   {
      Set<Short> points = new HashSet<Short>();
      for (String text : strings)
         points.add( stringToPoint( text ) );
      return points;
   }

   public static List<String> pointsToStrings( short[] points )
   {
      List<String> strings = new ArrayList<String>();
      for (int i = 0; i < points.length; i++)
         strings.add( pointToString( points[i] ) );
      return strings;
   }

   public static List<String> pointsToStrings( Iterable<Short> points )
   {
      List<String> strings = new ArrayList<String>();
      for (short pnt : points)
         strings.add( pointToString( pnt ) );
      return strings;
   }

   private static void throwInvalidColor( int linenum, String token )
   {
      throw new ParseException( linenum, Resources.getString( "msg.1003" ), token ); //$NON-NLS-1$
   }

   private static void throwInvalidPointString( int linenum, String token )
   {
      throw new ParseException( linenum, Resources.getString( "msg.1005" ), token ); //$NON-NLS-1$
   }
}
