//
// Game.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.math.BigDecimal;
import java.util.Iterator;
import java.util.Set;

import org.apache.commons.collections.map.MultiValueMap;
import org.apache.commons.lang3.tuple.ImmutablePair;

/**
 * This class encapsulates the state of a Go game, including setup,
 * the current board, and the current turn.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class Game
{
   public static final BigDecimal DEFAULT_KOMI       = new BigDecimal( "6.5" ); //$NON-NLS-1$
   public static final int        MIN_HANDICAP_COUNT = 2;
   public static final int        MAX_HANDICAP_COUNT = 9;

   /**
    * Verifies that the specified color is a player color.
    * 
    * @param color
    * 
    * @throws IllegalArgumentException  if the specified color is not either {@link Colors#BLACK}
    *                                   or {@link Colors#WHITE}
    */
   public static void checkColor( Colors color )
   {
      if (color != Colors.BLACK && color != Colors.WHITE)
         throw new IllegalArgumentException( Resources.formatString( "msg.1019", color ) ); //$NON-NLS-1$
   }

   /**
    * Returns true if the specified color is {@link Colors#BLACK} or (@link Colors#WHITE}.
    * 
    * @param color
    */
   public static boolean isPlayerColor( Colors color )
   {
      return color == Colors.BLACK || color == Colors.WHITE;
   }

   /**
    * Returns true if the specified count is a valid handicap.
    * 
    * @param count
    */
   public static boolean isValidHandicap( int count )
   {
      return count >= MIN_HANDICAP_COUNT && count <= MAX_HANDICAP_COUNT;
   }

   /**
    * For black or white, returns the opposite player color. Any other
    * color is returned unchanged.
    * 
    * @param color
    */
   public static Colors otherPlayer( Colors color )
   {
      if (color == Colors.BLACK)
         return Colors.WHITE;
      if (color == Colors.WHITE)
         return Colors.BLACK;
      return color;
   }

   /**
    * @param rule            ko rule
    * @param allow_suicide   true if suicide moves are allowed
    * @param komi            komi value
    */
   public Game( KoRules rule, boolean allow_suicide, BigDecimal komi )
   {
      if (rule == KoRules.SITUATIONAL)
         throw new IllegalArgumentException( "situational ko not yet supported" ); //$NON-NLS-1$
      m_rule = rule;
      m_allow_suicide = allow_suicide;
      m_komi = komi;
   }

   /**
    * Copy constructor.
    * 
    * @param other
    */
   public Game( Game other )
   {
      m_board          = new Board( other.m_board );
      m_rule           = other.m_rule;
      m_player         = other.m_player;
      m_turn_num       = other.m_turn_num;
      m_pass_count     = other.m_pass_count;
      m_last_point     = other.m_last_point;
      m_last_type      = other.m_last_type;
      m_komi           = other.m_komi;
      m_allow_suicide  = other.m_allow_suicide;
      m_pris_finder    = new PrisonerFinder();
      m_sig            = new Signature( other.m_sig );
      m_ko_table       = new MultiValueMap();
      m_ko_table.putAll( other.m_ko_table );
   }

   /**
    * Are players allowed to make suicide moves?
    */
   public boolean allowSuicide()
   {
      return m_allow_suicide;
   }

   /**
    * Checks whether the specified board represents a ko position as determined by
    * the game's ko rule.
    * 
    * @param board
    * 
    * @return {@code true} if the board represents a ko position; {@code false} otherwise.
    */
   @SuppressWarnings( "unchecked" )
   public boolean isKo( Board board )
   {
      if (m_rule == KoRules.BASIC || m_rule == KoRules.POSITIONAL) 
      {
         long zhash = board.getZobrist();
         if (m_ko_table.containsKey( zhash ))
         {
            Iterator<?> iter = m_ko_table.getCollection( zhash ).iterator();
            while (iter.hasNext())
            {
               ImmutablePair<Board,Integer> entry = (ImmutablePair<Board,Integer>) iter.next();
               if ((m_rule == KoRules.POSITIONAL || entry.getRight() == m_turn_num - 2) && entry.getLeft().equals( board ))
                  return true;
            }
         }
      }
      else if (m_rule != KoRules.NONE)
      {
         throw new IllegalArgumentException( Resources.getString( "msg.1023" ) ); //$NON-NLS-1$
      }
      return false;
   }

   /**
    * Analyzes the move for the specified {@code board} and returns one of the
    * following possible move types:<p>
    * 
    * <ul>
    * <li>{@link MoveTypes#KO}</li>
    * <li>{@link MoveTypes#LEGAL}</li>
    * <li>{@link MoveTypes#OCCUPIED}</li>
    * <li>{@link MoveTypes#PASS}</li>
    * <li>{@link MoveTypes#SUICIDE}</li>
    * </ul><p>
    * 
    * If {@code prisoners} is non-null, this method adds any stones captured
    * by the move.<p>
    * 
    * {@code board} will be updated with the result of the move, unless
    * the resulting move type is not {@link MoveTypes#LEGAL}, in which
    * case {@code board} will be left in its original state.<p>
    *
    * TODO: this method does not implement {@link KoRules#SITUATIONAL}
    *
    * @param board      board in which to evaluate the move, modified by this method
    * @param color      the player who is making a move
    * @param point      the move point
    * @param prisoners  a container for prisoners to be returned; null is allowed
    * 
    * @see #isKo
    * @see #play
    */
   public MoveTypes classify( Board board, Colors color, short point, Set<Short> prisoners )
   {
      if (prisoners != null)
         prisoners.clear();
      if (point == Board.INVALID_POINT)
         return MoveTypes.PASS;
      if (board.getColor( point ) != Colors.EMPTY)
         return MoveTypes.OCCUPIED;

      // compute prisoners and liberties for this move
      //
      m_pris_finder.find( board, point, color );

      // check for suicide move
      //
      if (!m_allow_suicide && !m_pris_finder.hasLiberty())
         return MoveTypes.SUICIDE;

      // update board and check ko status
      //
      board.setColor( point, color );
      Set<Short> new_prisoners = m_pris_finder.getPrisoners();
      board.setColor( new_prisoners, Colors.EMPTY );
      if (isKo( board ))
      {
         board.setColor( point, Colors.EMPTY );
         board.setColor( new_prisoners, Game.otherPlayer( color ) );
         return MoveTypes.KO;
      }
      if (prisoners != null)
         prisoners.addAll( new_prisoners );

      return MoveTypes.LEGAL;
   }

   /**
    * Updates this game's state and returns the type of move.
    * The move is validated against the current suicide and ko settings
    * of the game.
    *
    * @param point      the move point
    * @param prisoners  a container for prisoners to be returned; null is allowed
    * 
    * @see #classify
    * @see #getLastMoveType
    * @see #getLastPoint
    * @see #isKo
    */
   public MoveTypes play( short point, Set<Short> prisoners )
   {
      if (m_pass_count >= 2)
         throw new GameException( Resources.getString( "msg.1013" ) ); //$NON-NLS-1$

      if (point == Board.INVALID_POINT)
      {
         if (Signature.isTurnNeeded( m_turn_num - 1 ))
            m_sig.addPoint( point );
         m_pass_count++;
         m_turn_num++;
         m_player = otherPlayer( m_player );
         m_last_type = MoveTypes.PASS;
         return m_last_type;
      }
      if (m_pass_count > 0)
         m_pass_count--;

      // TODO: think harder about the exceptions being thrown here;
      // define a subclass specifically for ko? for all of the cases?
      //
      m_last_type = classify( m_board, m_player, point, prisoners );
      switch( m_last_type )
      {
         case OCCUPIED:
            throw new GameException( Resources.getString( "msg.1014" ), point ); //$NON-NLS-1$
         case SUICIDE:
            if (!m_allow_suicide)
               throw new GameException( Resources.getString( "msg.1016" ), point ); //$NON-NLS-1$
            break;
         case KO:
            throw new GameException( Resources.formatString( "msg.1009", m_rule ), m_player, point ); //$NON-NLS-1$
         default:
            break;  // PASS handled above
      }

      // update m_ko_table
      //
      if (m_rule != KoRules.NONE)
      {
         ImmutablePair<Board,Integer> entry = ImmutablePair.of( new Board( m_board ), m_turn_num );
         m_ko_table.put( m_board.getZobrist(), entry );
      }

      // TODO: how to communicate atari status back to user?
      // not just move type, but the specific group?
      //

      m_last_point = point;

      if (Signature.isTurnNeeded( m_turn_num - 1 ))
         m_sig.addPoint( point );
      m_turn_num++;
      m_player = otherPlayer( m_player );
      return m_last_type;
   }

   /**
    * Returns this game's board.
    */
   public Board getBoard()
   {
      return m_board;
   }

   /**
    * Returns the game's komi value, which is white's compensation
    * for black's first move advantage.
    */
   public BigDecimal getKomi()
   {
      return m_komi;
   }

   /**
    * Sets the game's komi value, which is white's compensation
    * for black's first move advantage. Reverse komi
    * (negative komi) may be specified as well.
    * 
    * @param komi
    */
   public void setKomi( BigDecimal komi )
   {
      m_komi = komi;
   }

   /**
    * Returns the ko rule in force for this game.
    */
   public KoRules getKoRule()
   {
      return m_rule;
   }

   /**
    * Returns the type of the last move played by either player.
    */
   public MoveTypes getLastMoveType()
   {
      return m_last_type;
   }

   /**
    * Returns the last point played. Returns {@link Board#INVALID_POINT} to indicate a
    * game in which no moves have yet been played; such a return value
    * does not indicate a pass.
    */
   public short getLastPoint()
   {
      return m_last_point;
   }

   /**
    * Returns the color of the player whose turn it is.
    */
   public Colors getPlayer()
   {
      return m_player;
   }

   /**
    * Use this method to override the normal alternation of black and white
    * by specifying which player goes next.
    * 
    * @param color
    */
   public void setPlayer( Colors color )
   {
      checkColor( color );
      m_player = color;
   }

   /**
    * Returns this game's signature.
    */
   public Signature getSignature()
   {
      return m_sig;
   }

   /**
    * Returns the current turn number.
    */
   public int getTurnNumber()
   {
      return m_turn_num;
   }

   /**
    * Computes the game's status (win or draw) and final score. Other game
    * statuses such as {@link Result.Status#FORFEIT}, {@link Result.Status#LOST_ON_TIME},
    * or {@link Result.Status#RESIGNED} can be determined by the application without
    * calculating any score.
    */
   public Result getResult( Scorer scorer )
   {
      Colors color = Colors.DONT_CARE;
      Result.Status status = Result.Status.WON;
      ImmutablePair<Integer,Integer> scores = scorer.computeScores( m_board, null );
      BigDecimal b_score = new BigDecimal( scores.getLeft() );
      BigDecimal w_score = new BigDecimal( scores.getRight() ).add( m_komi );
      BigDecimal score = BigDecimal.ZERO;
      if (b_score.compareTo( w_score ) > 0)
      {
         color = Colors.BLACK;
         score = b_score.subtract( w_score );
      }
      else if (w_score.compareTo( b_score ) > 0)
      {
         color = Colors.WHITE;
         score = w_score.subtract( b_score );
      }
      else
      {
         color = Colors.DONT_CARE;
         status = Result.Status.DRAW;
      }
      return new Result( color, status, score );
   }

   /**
    * Returns true if this game has completed, such that no further
    * moves are required.
    */
   public boolean isOver()
   {
      return m_pass_count >= 2;
   }

   /**
    * Places the stones indicated by the specified layer on the board; does
    * not check whether the board is already in play. This method can be
    * used to place handicap stones, and in the case of a fixed handicap,
    * it is the caller's responsibility to specify the correct points and to
    * specify black as the color.
    * 
    * @param color
    * @param points
    */
   public void placeStones( Colors color, short[] points )
   {
      if (color == Colors.DONT_CARE)
         throw new IllegalArgumentException( Resources.getString( "msg.1003" ) ); //$NON-NLS-1$
      m_board.setColor( points, color );
      m_ko_table.clear();
      m_ko_table.put( m_board.getZobrist(), ImmutablePair.of( new Board( m_board ), 1 ) );
   }

   /**
    * Places the stones indicated by the specified layer on the board; does
    * not check whether the board is already in play. This method can be
    * used to place handicap stones, and in the case of a fixed handicap,
    * it is the caller's responsibility to specify the correct points and to
    * specify black as the color.
    * 
    * @param color
    * @param points
    */
   public void placeStones( Colors color, Iterable<Short> points )
   {
      if (color == Colors.DONT_CARE)
         throw new IllegalArgumentException( Resources.getString( "msg.1003" ) ); //$NON-NLS-1$
      m_board.setColor( points, color );
      m_ko_table.clear();
      m_ko_table.put( m_board.getZobrist(), ImmutablePair.of( new Board( m_board ), 1 ) );
   }

   /**
    * Restores the state of this game to its initial conditions.
    */
   public void reset()
   {
      m_board = new Board();
      m_sig = new Signature();
      m_player = Colors.BLACK;
      m_turn_num = 1;
      m_pass_count = 0;
      m_ko_table.clear();
      m_last_point = Board.INVALID_POINT;
      m_last_type = MoveTypes.OUT_OF_TURN;
   }

   /**
    * Indicates that the current player is resigning the game.
    */
   public void resign()
   {
      m_pass_count++;
      m_player = otherPlayer( m_player );
   }

   private Board          m_board          = new Board();
   private KoRules        m_rule           = KoRules.BASIC;
   private Colors         m_player         = Colors.BLACK;
   private int            m_turn_num       = 1;
   private int            m_pass_count     = 0;
   private short          m_last_point     = Board.INVALID_POINT;
   private MoveTypes      m_last_type      = MoveTypes.OUT_OF_TURN;
   private BigDecimal     m_komi           = DEFAULT_KOMI;
   private boolean        m_allow_suicide  = false;
   private PrisonerFinder m_pris_finder    = new PrisonerFinder();
   private MultiValueMap  m_ko_table       = new MultiValueMap();  // key = Zobrist hash, value = ImmutablePair<Board,Integer>
   private Signature      m_sig            = new Signature();
}
