//
// Layers.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.io.IOException;
import java.io.Reader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.google.gson.JsonParseException;
import com.google.gson.stream.JsonReader;
import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.base.ParseException;

/**
 * A layer is a set of points representing an important portion of the board. This
 * class provides helper methods for creating layers or loading them from an
 * external file.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public abstract class Layers
{
   /**
    * Create a new rectangular region, where ll identifies the lower-left corner
    * and ur identifies the upper-right corner.
    * 
    * @param ll
    * @param ur
    */
   public static Set<Short> create( short ll, short ur )
   {
      Set<Short> points = new HashSet<Short>();
      expandRect( ll, ur, points );
      return points;
   }

   /**
    * Create a new rectangular region, where rect specifies two opposite corners.
    * 
    * @param rect
    */
   public static Set<Short> create( ImmutablePair<Short,Short> rect )
   {
      Set<Short> points = new HashSet<Short>();
      expandRect( rect, points );
      return points;
   }

   /**
    * Create a new region that is the union of a collection of rectangles. Each
    * pair specifies two opposite corners of a rectangle.
    * 
    * @param rects
    */
   public static Set<Short> create( Iterable<ImmutablePair<Short,Short>> rects )
   {
      Set<Short> points = new HashSet<Short>();
      for (ImmutablePair<Short,Short> rect : rects)
         expandRect( rect, points );
      return points;
   }

   /**
    * A helper method for adding points to a collection based on two board coordinates
    * representing the opposite sides of a rectangle. Returns the number of points
    * added.
    * 
    * @param rect
    * @param dest
    */
   public static int expandRect( ImmutablePair<Short,Short> rect, Set<Short> dest )
   {
      return expandRect( rect.getLeft(), rect.getRight(), dest );
   }

   /**
    * A helper method for adding points to a collection based on two board coordinates
    * representing the opposite sides of a rectangle. Returns the number of points
    * added.
    * 
    * @param ll
    * @param ur
    * @param dest
    */
   public static int expandRect( short ll, short ur, Set<Short> dest )
   {
      int size = Board.getSize();
      int delta_y = ur % size - ll % size;
      int start = Board.INVALID_POINT;
      int end = Board.INVALID_POINT;
      int pnt = Board.INVALID_POINT;
      int count = 0;
      if (ll > ur)
      {
         start = ur;
         end = ll;
         delta_y = -delta_y;
      }
      else
      {
         start = ll;
         end = ur;
      }
      if (delta_y < 0)
      {
         delta_y = -delta_y;
         start -= delta_y;   // flips the rect coordinates aronud
         end += delta_y;
      }
      pnt = start;
      while (pnt <= end)
      {
         dest.add( (short) pnt );
         count++;
         pnt++;
         if (pnt - start > delta_y)
         {
            start += size;
            pnt = start;
         }
      }
      return count;
   }

   /**
    * Parses JSON to produce a mapping of layer names to sets of layer points. See
    * {@code etc/layers.json} for an example. Implicitly sets the current board size
    * equal to the value of {@code requested_size}.
    * 
    * @param input             input reader
    * @param requested_size    which set of layers to load
    * 
    * @see Board#setSize
    */
   public static Map<String,Set<Short>> load( Reader input, int requested_size ) throws IOException
   {
      int orig_size = Board.getSize();
      if (!Board.isValidSize( requested_size ))
         throw new IllegalArgumentException( Resources.formatString( "msg.1008", requested_size ) ); //$NON-NLS-1$
      Map<String,Set<Short>> layers = new HashMap<String,Set<Short>>();
      try
      {
         JsonReader jreader = new JsonReader( input );
         jreader.setLenient( true ); // so we can include comments in config files
         jreader.beginArray();
         int size = 0;
         Map<String,Set<Short>> new_layers = null;
         while (jreader.hasNext())
         {
            jreader.beginObject();
            while (jreader.hasNext())
            {
               String prop_name = jreader.nextName();
               if (prop_name.equals( SIZE_KEYWORD ))
               {
                  size = jreader.nextInt();
                  Board.setSize( size );
                  if (size == requested_size && new_layers != null)
                  {
                     layers.putAll( new_layers );
                     new_layers = null;
                  }
               }
               else if (prop_name.equals( LAYERS_KEYWORD ))
               {
                  if (size == 0)
                     throw new ParseException( Resources.getString( "msg.1024" ) ); //$NON-NLS-1$
                  new_layers = readLayers( jreader );
                  if (size == requested_size)
                  {
                     for (String name : new_layers.keySet())
                        layers.put( name, new_layers.get( name ) );
                     new_layers = null;
                  }
                  size = 0;
               }
               else
               {
                  throwUnrecognizedKeyword( prop_name );
               }
            }
            jreader.endObject();
         }
         jreader.endArray();
      }
      catch (IllegalStateException ex)
      {
         requested_size = orig_size;
         throw new ParseException( ex.getMessage(), ex );
      }
      catch (NumberFormatException ex)
      {
         requested_size = orig_size;
         throw new ParseException( ex.getMessage(), ex );
      }
      catch (JsonParseException ex)
      {
         requested_size = orig_size;
         throw new ParseException( ex.getMessage(), ex );
      }
      finally
      {
         Board.setSize( requested_size );
      }
      return layers;
   }

   private static Map<String,Set<Short>> readLayers( JsonReader jreader ) throws IOException
   {
      Map<String,Set<Short>> layers = new HashMap<String,Set<Short>>();
      Set<String> names = null;
      Set<Short> points = null;
      jreader.beginArray();
      while (jreader.hasNext())
      {
         jreader.beginObject();
         while (jreader.hasNext())
         {
            String prop_name = jreader.nextName();
            switch (prop_name)
            {
               case INTERSECT_KEYWORD:
                  points = readIntersect( jreader, layers );
                  break;
               case NAMES_KEYWORD:
                  names = readNames( jreader );
                  break;
               case POINTS_KEYWORD:
                  points = readPoints( jreader );
                  break;
               case RECTANGLE_KEYWORD:
                  points = readRectangle( jreader );
                  break;
               case UNION_KEYWORD:
                  points = readUnion( jreader, layers );
                  break;
               default:
                  throwUnrecognizedKeyword( prop_name );
                  break;
            }
         }
         if (names == null || names.isEmpty())
            throw new ParseException( Resources.getString( "msg.1010" ) ); //$NON-NLS-1$
         if (points == null)
            throw new ParseException( Resources.getString( "msg.1011" ) ); //$NON-NLS-1$
         if (!points.isEmpty())
         {
            for (String name : names)
            {
               if (layers.containsKey( name ))
                  throw new ParseException( Resources.getString( "msg.1000" ), name ); //$NON-NLS-1$
               layers.put( name, points );
            }
         }
         jreader.endObject();
      }
      jreader.endArray();
      return layers;
   }

   private static Set<String> readNames( JsonReader jreader ) throws IOException
   {
      Set<String> names = new HashSet<String>();
      jreader.beginArray();
      while (jreader.hasNext())
         names.add( jreader.nextString() );
      jreader.endArray();
      return names;
   }

   private static Set<Short> readPoints( JsonReader jreader ) throws IOException
   {
      Set<Short> points = new HashSet<Short>();
      jreader.beginArray();
      while (jreader.hasNext())
         points.add( Convert.stringToPoint( jreader.nextString() ) );
      jreader.endArray();
      return points;
   }

   private static Set<Short> readRectangle( JsonReader jreader ) throws IOException
   {
      short pnt1 = Board.INVALID_POINT;
      short pnt2 = Board.INVALID_POINT;
      int count = 0;
      jreader.beginArray();
      while (jreader.hasNext())
      {
         if (count == 0)
            pnt1 = Convert.stringToPoint( jreader.nextString() );
         else if (count == 1)
            pnt2 = Convert.stringToPoint( jreader.nextString() );
         else if (count + 1 > 2)
            throw new ParseException( Resources.getString( "msg.1018" ) ); //$NON-NLS-1$
         count++;
      }
      jreader.endArray();
      if (count != 2)
         throw new ParseException( Resources.getString( "msg.1017" ) ); //$NON-NLS-1$
      return create( pnt1, pnt2 );
   }

   private static Set<Short> readIntersect( JsonReader jreader, Map<String,Set<Short>> layers ) throws IOException
   {
      Set<Short> points = null;
      jreader.beginArray();
      while (jreader.hasNext())
      {
         String text = jreader.nextString();
         try
         {
            short pnt = Convert.stringToPoint( text );
            boolean has_pnt = (points != null && points.contains( pnt ));
            if (has_pnt)
            {
               points.clear();
               points.add( pnt );
            }
            else if (points == null)
            {
               points = new HashSet<Short>();
               points.add( pnt );
            }
         }
         catch (ParseException ex)
         {
            if (!layers.containsKey( text ))
               throw new ParseException( Resources.formatString( "msg.1012", Resources.getString( "msg.1002" ) ), text ); //$NON-NLS-1$ //$NON-NLS-2$
            Set<Short> other = layers.get( text );
            if (points == null)
               points = new HashSet<Short>( other );
            else
               points.retainAll( other );
         }
      }
      jreader.endArray();
      if (points == null)
         throw new ParseException( Resources.formatString( "msg.1001", Resources.getString( "msg.1002" ) ) ); //$NON-NLS-1$ //$NON-NLS-2$
      return points;
   }

   private static Set<Short> readUnion( JsonReader jreader, Map<String,Set<Short>> layers ) throws IOException
   {
      Set<Short> points = new HashSet<Short>();
      jreader.beginArray();
      while (jreader.hasNext())
      {
         String text = jreader.nextString();
         try
         {
            short pnt = Convert.stringToPoint( text );
            points.add( pnt );
         }
         catch (ParseException ex)
         {
            if (!layers.containsKey( text ))
               throw new ParseException( Resources.formatString( "msg.1012", Resources.getString( "msg.1022" ) ), text ); //$NON-NLS-1$ //$NON-NLS-2$
            Set<Short> other = layers.get( text );
            points.addAll( other );
         }
      }
      jreader.endArray();
      if (points.isEmpty())
         throw new ParseException( Resources.formatString( "msg.1001", Resources.getString( "msg.1022" ) ) ); //$NON-NLS-1$ //$NON-NLS-2$
      return points;
   }

   private static void throwUnrecognizedKeyword( String token )
   {
      throw new ParseException( Resources.getString( "msg.1021" ), token ); //$NON-NLS-1$
   }

   private static final String INTERSECT_KEYWORD = "intersect"; //$NON-NLS-1$
   private static final String LAYERS_KEYWORD    = "layers";    //$NON-NLS-1$
   private static final String NAMES_KEYWORD     = "names";     //$NON-NLS-1$
   private static final String POINTS_KEYWORD    = "points";    //$NON-NLS-1$
   private static final String RECTANGLE_KEYWORD = "rectangle"; //$NON-NLS-1$
   private static final String SIZE_KEYWORD      = "size";      //$NON-NLS-1$
   private static final String UNION_KEYWORD     = "union";     //$NON-NLS-1$
}
