//
// PointLinkedList.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.util.HashSet;

/**
 * This class wraps a board data array, providing a linked list data structure
 * for a point color. This class does no error checking of its own.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class PointLinkedList
{
   public static final int COLOR_MASK  = 0x03000000;

   public static final int EMPTY_VALUE = 0x00000000;
   public static final int BLACK_VALUE = 0x02000000;
   public static final int WHITE_VALUE = 0x01000000;

   /**
    * Initializes a new linked list.
    * 
    * @param storage
    * @param color
    */
   public PointLinkedList( int[] storage, int color )
   {
      m_storage = storage;
      m_head = -1;
      m_tail = -1;
      m_color = color;
   }

   public PointLinkedList( int[] storage, PointLinkedList other )
   {
      m_storage = storage;
      m_head = other.m_head;
      m_tail = other.m_tail;
      m_color = other.m_color;
   }

   public final int getHead()
   {
      return m_head;
   }

   public final int getTail()
   {
      return m_tail;
   }

   public final int getPrev( int point )
   {
      int i = (m_storage[point] & PREV_INDEX_MASK) >> 12;
      if (i == NEXT_INDEX_MASK) // using NEXT_INDEX_MASK here just to check that the shifted value's bits are all set
         return -1;
      return i;
   }

   public final int getNext( int point )
   {
      int i = m_storage[point] & NEXT_INDEX_MASK;
      if (i == NEXT_INDEX_MASK)
         return -1;
      return i;
   }

   public final boolean isEmpty()
   {
      return m_head == -1;
   }

   public final int size()
   {
      if (m_head == -1)
         return 0;
      int size = 0;
      for (int i = m_head; i >= 0; i = getNext( i ))
         size++;
      return size;
   }

   /**
    * Inserts the specified point into this list and sets the corresponding
    * color. Should be called immediately after any call to {@link #remove}.
    * 
    * @param point
    */
   public final void insert( int point )
   {
      m_storage[point] &= ~COLOR_MASK;
      m_storage[point] |= m_color;
      if (m_head == -1)
      {
         m_head = point;
         m_tail = point;
         setBoth( point, -1, -1 );
         return;
      }
      setNext( m_tail, point );
      setBoth( point, m_tail, -1 );
      m_tail = point;
   }

   /**
    * Removes a point from the current list. {@link #insert} should be called
    * immediately after this method.
    * 
    * @param point
    */
   public final void remove( int point )
   {
      if (m_head == -1)
         return;
      if (m_head == point)
      {
         if (m_tail == point)
            m_tail = -1;
         m_head = getNext( m_head );
         if (m_head >= 0)
            setPrev( m_head, -1 );
         return;
      }
      if (m_tail == point)
      {
         m_tail = getPrev( m_tail );
         if (m_tail >= 0)
            setNext( m_tail, -1 );
         return;
      }
      int left = getPrev( point );
      int right = getNext( point );
      setNext( left, right );
      setPrev( right, left );
   }

   public final void clear()
   {
      m_head = -1;
      m_tail = -1;
   }

   public final void reserve()
   {
      m_head = 0;
      m_storage[m_head] = m_color;
      setBoth( m_head, -1, 1 );
      m_tail = m_storage.length - 1;
      m_storage[m_tail] = m_color;
      setBoth( m_tail, m_tail - 1, -1 );
      for (int i = 1; i < m_tail; i++)
      {
         m_storage[i] = m_color;
         if (i > m_head)
            setBoth( i, i - 1, i + 1 );
      }
   }

   @Override
   public final boolean equals( Object obj )
   {
      if (!(obj instanceof PointLinkedList))
         return false;
      if (this == obj)
         return true;
      PointLinkedList other = (PointLinkedList) obj;
      if (m_color != other.m_color)
         return false;
      HashSet<Integer> tmp = new HashSet<Integer>();
      for (int pnt = m_head; pnt >= 0; pnt = getNext( pnt ))
         tmp.add( pnt );
      int count = 0;
      for (int pnt = other.m_head; pnt >= 0; pnt = other.getNext( pnt ))
      {
         if (!tmp.contains( pnt ))
            return false;
         count++;
      }
      return tmp.size() == count;
   }

   @Override
   public final String toString()
   {
      StringBuilder buffer = new StringBuilder();
      for (int pnt = m_head; pnt >= 0; pnt = getNext( pnt ))
      {
         if (buffer.length() > 0)
            buffer.append( ' ' );
         buffer.append( Convert.pointToString( (short) pnt ) );
      }
      return buffer.toString();
   }

   private void setBoth( int point, int prev, int next )
   {
      if (prev < 0)
         prev = PREV_INDEX_MASK >> 12;
      if (next < 0)
         next = NEXT_INDEX_MASK;
      int value = m_storage[point];
      value &= ~(PREV_INDEX_MASK | NEXT_INDEX_MASK);
      value |= (prev << 12) | next;
      m_storage[point] = value;
   }

   private void setPrev( int point, int prev )
   {
      if (prev < 0)
         prev = PREV_INDEX_MASK >> 12;
      int value = m_storage[point];
      value &= ~PREV_INDEX_MASK;
      value |= prev << 12;
      m_storage[point] = value;
   }

   private void setNext( int point, int next )
   {
      if (next < 0)
         next = NEXT_INDEX_MASK;
      int value = m_storage[point];
      value &= ~NEXT_INDEX_MASK;
      value |= next;
      m_storage[point] = value;
   }

   private static final int PREV_INDEX_MASK = 0x00FFF000;
   private static final int NEXT_INDEX_MASK = 0x00000FFF;

   private int[] m_storage;  // owned by a Board object
   private int   m_head;
   private int   m_tail;
   private int   m_color;
}
