//
// PrisonerFinder.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.util.HashSet;
import java.util.Set;

/**
 * This class identifies prisoners captured by a given move.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see Board#visitChain
 */
public class PrisonerFinder
{
   /**
    * Default constructor.
    */
   public PrisonerFinder()
   {
   }

   /**
    * Identify enemy points that are captured if the specified color is played
    * at the specified point. It does not matter whether a stone is actually
    * present on the point.
    * 
    * @param board
    * @param point
    * @param color
    */
   public void find( Board board, short point, Colors color )
   {
      Game.checkColor( color );
      m_enemy_diags = 0;
      m_friendly_nbrs = 0;
      m_has_liberty = false;
      m_point = point;
      m_prisoners.clear();
      m_friendly = color;
      m_enemy = Game.otherPlayer( m_friendly );
      m_location = Board.getLocation( point );
      try
      {
         board.visitNeighbors( point, m_nbr_visitor );
         if (m_location == Locations.EDGE)
            m_friendly_nbrs++;
         else if (m_location == Locations.CORNER)
            m_friendly_nbrs += 2;
         m_has_liberty = m_has_liberty || !m_prisoners.isEmpty();
         if (m_has_liberty)
            board.visitDiagonals( point, m_diag_visitor );
      }
      finally
      {
         m_prisoner_group.clear();
         m_group_has_lib = false;
      }
   }

   /**
    * @return The location of the point passed to {@link #find}.
    */
   public final Locations getLocation()
   {
      return m_location;
   }

   /**
    * @return {@code true} if the point passed to {@link #find} has at least one liberty,
    *         either directly or indirectly; {@code false} otherwise.
    */
   public final boolean hasLiberty()
   {
      return m_has_liberty;
   }

   /**
    * An eye is an empty point meeting the following criteria:<p>
    * 
    * <ul>
    * <li>its neighbors are all of the friendly color
    * <li>its diagonals contain no more than 1 stone of the
    * opposite color, unless it's a border in which case no diagonal
    * enemies are allowed
    * </ul>
    *
    * @return {@code true} if the point passed to {@link #find} is an eye;
    *         {@code false} otherwise.
    */
   public final boolean isEye()
   {
      if (m_friendly_nbrs < 4)
         return false;
      return m_enemy_diags <= (m_location == Locations.INTERIOR ? 1 : 0);
   }

   /**
    * Returns the prisoners captured by the move specified in the call to find().
    */
   public Set<Short> getPrisoners()
   {
      return m_prisoners;
   }

   private PointVisitor m_diag_visitor = new PointVisitor()
   {
      @Override
      public final boolean accept( Board board, short point )
      {
         if (board.getColor( point ) == m_enemy)
            m_enemy_diags++;
         return true;
      }
   };

   private PointVisitor m_nbr_visitor = new PointVisitor()
   {
      @Override
      public final boolean accept( Board board, short point )
      {
         Colors color = board.getColor( point );
         if (color == Colors.EMPTY)
         {
            m_has_liberty = true;
            return true;
         }
         if (color == m_friendly)
         {
            m_friendly_nbrs++;
            if (!m_has_liberty)
            {
               m_group_has_lib = false;
               board.visitChain( point, null, m_lib_visitor );
               m_has_liberty = m_group_has_lib;
            }
            return true;
         }
         if (m_prisoners.contains( point ))
            return true;
         m_prisoner_group.clear();
         m_prisoner_group.add( point );
         m_group_has_lib = false;
         board.visitChain( point, m_chainpnt_visitor, m_lib_visitor );
         if (!m_group_has_lib)
            m_prisoners.addAll( m_prisoner_group );
         return true;
      }
   };

   private PointVisitor m_chainpnt_visitor = new PointVisitor()
   {
      @Override
      public final boolean accept( Board board, short point )
      {
         if (point == m_point)
            return true;
         m_prisoner_group.add( point );
         return true;
      }
   };

   private PointVisitor m_lib_visitor = new PointVisitor()
   {
      @Override
      public final boolean accept( Board board, short point )
      {
         if (point == m_point)
            return true;
         m_group_has_lib = true;
         return false;
      }
   };

   private Set<Short>       m_prisoners      = new HashSet<Short>();
   private Set<Short>       m_prisoner_group = new HashSet<Short>();
   private short            m_point          = Board.INVALID_POINT;
   private Colors           m_enemy          = Colors.DONT_CARE;
   private Colors           m_friendly       = Colors.DONT_CARE;
   private int              m_friendly_nbrs  = 0;
   private int              m_enemy_diags    = 0;
   private Locations        m_location       = Locations.INTERIOR;
   private boolean          m_has_liberty    = false;
   private boolean          m_group_has_lib  = false;
}
