//
// Result.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import java.math.BigDecimal;

import com.dakengo.libs.base.HashUtils;
import com.dakengo.libs.base.ParseException;

/**
 * This value class represents the conclusion of a game.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class Result
{
   public enum Status
   {
      DRAW, FORFEIT, LOST_ON_TIME, RESIGNED, WON
   }

   /**
    * @param player   must be Colors.DONT_CARE for a draw, or Colors.BLACK or Colors.WHITE
    *                 for any other result
    * @param status   must not be Status.WON
    */
   public Result( Colors player, Status status )
   {
      if (status == Status.DRAW && player != Colors.DONT_CARE)
         throw new IllegalArgumentException( Resources.getString( "msg.1003" ) ); //$NON-NLS-1$
      if (status != Status.DRAW)
         Game.checkColor( player );
      if (status == Status.WON)
         throw new IllegalArgumentException( Resources.getString( "msg.1006" ) ); //$NON-NLS-1$
      m_player = player;
      m_status = status;
   }

   /**
    * @param player  must be Colors.BLACK or Colors.WHITE
    * @param status  must be Status.WON
    * @param score
    */
   public Result( Colors player, Status status, BigDecimal score )
   {
      Game.checkColor( player );
      if (status != Status.WON)
         throw new IllegalArgumentException( Resources.getString( "msg.1006" ) ); //$NON-NLS-1$
      if (score.compareTo( BigDecimal.ZERO ) < 0)
         throw new IllegalArgumentException( Resources.getString( "msg.1007" ) ); //$NON-NLS-1$
      m_player = player;
      m_status = status;
      m_score = score;
   }

   /**
    * @param player  must be Colors.BLACK or Colors.WHITE
    * @param status  must be Status.WON
    * @param score
    */
   public Result( Colors player, Status status, String score )
   {
      Game.checkColor( player );
      if (status != Status.WON)
         throw new IllegalArgumentException( Resources.getString( "msg.1006" ) ); //$NON-NLS-1$
      try
      {
         m_score = new BigDecimal( score );
      }
      catch (NumberFormatException unused)
      {
         throw new ParseException( Resources.getString( "msg.1007" ), score ); //$NON-NLS-1$
      }
      if (m_score.compareTo( BigDecimal.ZERO ) < 0)
         throw new IllegalArgumentException( Resources.getString( "msg.1007" ) ); //$NON-NLS-1$
      m_player = player;
      m_status = status;
   }

   public Colors getPlayer()
   {
      return m_player;
   }

   public BigDecimal getScore()
   {
      return m_score;
   }

   public Status getStatus()
   {
      return m_status;
   }

   @Override
   public boolean equals( Object obj )
   {
      if (!(obj instanceof Result))
         return false;
      if (this == obj)
         return true;
      Result other = (Result) obj;
      return m_player == other.m_player && m_status == other.m_status && m_score.equals( other.m_score );
   }

   @Override
   public int hashCode()
   {
      return HashUtils.compute( m_player, m_status, m_score );
   }

   @Override
   public String toString()
   {
      StringBuilder buffer = new StringBuilder();
      if (m_player == Colors.BLACK)
         buffer.append( "B+" ); //$NON-NLS-1$
      else if (m_player == Colors.WHITE)
         buffer.append( "W+" ); //$NON-NLS-1$
      switch (m_status)
      {
         case DRAW:         buffer.append( "Draw" );    break; //$NON-NLS-1$
         case FORFEIT:      buffer.append( "Forfeit" ); break; //$NON-NLS-1$
         case LOST_ON_TIME: buffer.append( "Time" );    break; //$NON-NLS-1$
         case RESIGNED:     buffer.append( "Resign" );  break; //$NON-NLS-1$
         case WON:          buffer.append( m_score.toPlainString() ); break;
         default:           throw new IllegalArgumentException( Resources.formatString( "msg.1025", m_status ) ); //$NON-NLS-1$
      }
      return buffer.toString();
   }

   private Colors     m_player  = Colors.DONT_CARE;
   private Status     m_status  = Status.DRAW;
   private BigDecimal m_score   = BigDecimal.ZERO;
}
