//
// Scorer.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import net.xeoh.plugins.base.Plugin;
import org.apache.commons.lang3.tuple.ImmutablePair;

/**
 * Plugins implement this interface to compute player scores for a board position.
 * The implementation must be annotated with the {@code @PluginImplementation} class
 * annotation. The implementation must be thread-safe.<p>
 * 
 * Each scoring plugin must publish its name via {@link com.dakengo.libs.application.PluginUtils#publishScoringSystem}.<p>
 * 
 * Each plugin must implement a method that returns the plugin's capabilities,
 * identified via the {@code @Capabilities}
 * annotation. The required capability is a single string having the prefix
 * {@link com.dakengo.libs.application.Constants#SCORING_CAP_PREFIX}
 * and a suffix consisting of the scoring method.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see com.dakengo.libs.application.BaseScorer
 * @see <a href="http://data.xeoh.net/jspf/api/net/xeoh/plugins/base/annotations/Capabilities.html"><code>@Capabilities</code></a>
 * @see <a href="http://data.xeoh.net/jspf/api/net/xeoh/plugins/base/annotations/PluginImplementation.html"><code>@PluginImplementation</code></a>
 * @see <a href="http://senseis.xmp.net/?Scoring">Scoring</a>
 */
public interface Scorer extends Plugin
{
   /**
    * Compute both players' scores for the specified board. The left-hand
    * element of the returned pair is the black player's score, whereas the
    * right-hand element is white's score.
    * 
    * @param board  the board to be scored
    * @param available  filled with the open points collected by this scorer.
    *                   The remainder is set to {@link com.dakengo.libs.game.Board#INVALID_POINT};
    *                   Passing null is allowed.
    */
   ImmutablePair<Integer,Integer> computeScores( Board board, short[] available );
}
