//
// Signature.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.game;

import org.apache.commons.lang3.ArrayUtils;

/**
 * A game signature is a sequence of moves from a game that represent a statistically unique
 * identifier for that game. See http://www.andromeda.com/people/ddyer/go/signature-spec.html
 * for the inspiration for this class. Up to 6 moves will be sample from a record:
 * 19, 30, 39, 50, 59, and 70. If a game does not reach move 70, then the signature can still be
 * generated but will be incomplete. For purposes of this class, move numbers are zero-based.
 * In a handicap game, the initial setup is considered black's first move (which
 * would be move zero).
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class Signature
{
   public static final int[] TURN_NUMBERS  = new int[] { 19, 30, 39, 50, 59, 70 };

   public static final int[] ARRAY_INDICES = new int[] { 0, 3, 1, 4, 2, 5 };

   /**
    * Returns true if the specified turn number is needed for
    * computing a signature.
    * 
    * @param turn_num
    */
   public static boolean isTurnNeeded( int turn_num )
   {
      return ArrayUtils.contains( TURN_NUMBERS, turn_num );
   }

   public Signature()
   {
      m_index = 0;
      m_points = new short[TURN_NUMBERS.length];
      for (int i = 0; i < TURN_NUMBERS.length; i++)
         m_points[i] = Board.INVALID_POINT;
   }

   public Signature( Signature other )
   {
      m_index = other.m_index;
      m_points = new short[TURN_NUMBERS.length];
      System.arraycopy( other.m_points, 0, m_points, 0, TURN_NUMBERS.length );
   }

   /**
    * Adds the specified point to this signature.
    * 
    * @param point
    */
   public void addPoint( short point )
   {
      if (m_index >= TURN_NUMBERS.length)
         throw new IllegalArgumentException( Resources.getString( "msg.1015" ) ); //$NON-NLS-1$
      if (point < Board.INVALID_POINT || point >= Board.getTotal())
         throw new IllegalArgumentException( Resources.formatString( "msg.1004", point ) ); //$NON-NLS-1$
      m_points[ARRAY_INDICES[m_index++]] = point;
   }

   /**
    * Returns the points corresponding to the moves sampled from a game.
    * The returned array length may be <= 6. An empty array is returned
    * if the game had no moves.
    */
   public short[] getPoints()
   {
      return m_points;
   }

   /**
    * Returns the highest move number for which this signature was
    * computed. Returns 0 if the game had no moves.
    */
   public int getTurn()
   {
      if (m_index == 0)
         return 0;
      return TURN_NUMBERS[m_index - 1];
   }

   /**
    * Returns true if all 6 required moves were available in the game when
    * computing this signature.
    */
   public boolean isComplete()
   {
      return m_index == TURN_NUMBERS.length;
   }

   @Override
   public boolean equals( Object obj )
   {
      if (!(obj instanceof Signature))
         return false;
      if (obj == this)
         return true;
      Signature other = (Signature) obj;
      for (int i = 0; i < TURN_NUMBERS.length; i++)
      {
         if (m_points[i] != other.m_points[i])
            return false;
      }
      return true;
   }

   @Override
   public int hashCode()
   {
      int hash = 37;
      for (int i = 0; i < TURN_NUMBERS.length; i++)
         hash = 37 * hash + m_points[i];
      return hash;
   }

   @Override
   public String toString()
   {
      return Convert.pointsToString( m_points, "," ); //$NON-NLS-1$
   }

   private short[] m_points;
   private int     m_index;
}
