//
// BaseLegalMoveFilter.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.player;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.PointVisitor;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.PrisonerFinder;

/**
 * Base class for objects that choose candidate moves from amongst the overall
 * set of available points.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public abstract class BaseLegalMoveFilter
{
   /**
    * This method computes the overall set of available legal moves and passes
    * them to the {@link #sort} method for ordering. Skips moves that would be
    * suicide, but does not check for ko.
    * 
    * @param player
    *           color of player to move
    * @param board
    *           position to analyze
    * @param max_candidates
    *           limit on the number of candidate points to return
    *           
    * @return  A list of candidate moves, each move represented by a point and
    *          the board resulting from the move.
    */
   public List<ImmutablePair<Short, Board>> find( final Colors player, Board board, int max_candidates )
   {
      Game.checkColor( player );
      if (max_candidates <= 0)
         throw new IllegalArgumentException( Resources.getString( "msg.10001" ) ); //$NON-NLS-1$
      final List<ImmutablePair<Short, Board>> candidates = new ArrayList<ImmutablePair<Short, Board>>();
      board.visitPoints( Colors.EMPTY, new PointVisitor()
      {
         @Override
         public boolean accept( Board b, short p )
         {
            m_finder.find( b, p, player );
            if (m_finder.hasLiberty() && !m_finder.isEye())
            {
               Board new_board = new Board( b );
               new_board.setColor( p, player );
               new_board.setColor( m_finder.getPrisoners(), Colors.EMPTY );
               candidates.add( ImmutablePair.of( p, new_board ) );
            }
            return true;
         }
      } );
      sort( player, candidates );
      if (candidates.size() > max_candidates)
         return candidates.subList( 0, max_candidates );
      return candidates;
   }

   /**
    * Implement this method to sort the list of candidate moves according to
    * their priority as determined by the subclass.
    * 
    * @param player
    * @param candidates
    */
   protected abstract void sort( Colors player, List<ImmutablePair<Short, Board>> candidates );

   protected BaseLegalMoveFilter()
   {
   }

   private PrisonerFinder m_finder = new PrisonerFinder();
}
