//
// BaseTreeNode.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.player;

import com.dakengo.libs.base.Modifiable;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;

/**
 * This is a base class for game tree nodes. Every game tree node has:<p>
 * <ul>
 * <li>parent
 * <li>offset to child index array
 * <li>board point
 * </ul><p>
 * 
 * The color of a node is compute based on the node's level in the tree.
 * Subclasses may add additional properties.<p>
 * 
 * Node structure:<p>
 * {@code parent (4 bytes)}<p>
 * {@code link block (4 bytes)}<p>
 * {@code point (2 bytes + 2 bytes unused)}<p>
 * {@code subclass prop 1 (4 bytes)}<p>
 * {@code subclass prop 2 (4 bytes)}<p>
 * {@code ...}<p>
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see TreeCursor
 * @see TreePolicy
 */
public abstract class BaseTreeNode implements Modifiable<BaseTreeNode>
{
   public static final int PARENT_PROP_INDEX         = 0;
   public static final int LINKS_PROP_INDEX          = 1;
   public static final int POINT_PROP_INDEX          = 2;
   public static final int FIRST_SUBCLASS_PROP_INDEX = POINT_PROP_INDEX + 1;

   /**
    * @return  The underlying tree storage object.
    */
   public TreeStore getStore()
   {
      return m_store;
   }

   /**
    * @return The location of this node within the tree's storage.
    */
   public int getStoreIndex()
   {
      return m_index;
   }

   /**
    * @return The location of this node's parent, or {@code -1} if this is
    *         the tree's root.
    */
   public int getParentStoreIndex()
   {
      return getProperty( PARENT_PROP_INDEX );
   }

   /**
    * @return The player color corresponding to this node.
    */
   public Colors getPlayer()
   {
      int level = 0;
      int index = m_index;
      int[] data = m_store.getData();
      while (index > 0)
      {
         level++;
         index = data[index];
      }
      Colors color = m_store.getRootColor();
      return ((level & 1) == 0 ? color : Game.otherPlayer( color ));
   }

   /**
    * @return The board point identifying the position represented by this node.
    */
   public short getPoint()
   {
      return (short) (getProperty( POINT_PROP_INDEX ) & 0xFFFF); 
   }

   /**
    * Sets the board point identifying the position represented by this node.
    * 
    * @param point  new board point
    */
   public void setPoint( short point )
   {
      Board.checkPoint( point );
      int value = getProperty( POINT_PROP_INDEX );
      value &= 0xFFFF0000;
      value |= point;
      setProperty( POINT_PROP_INDEX, value );
   }

   /**
    * Initializes the children of this node, according to the fanout specified by the underlying tree
    * storage.
    * 
    * @see TreePolicy
    * @see TreeStore
    */
   public void expand()
   {
      int link_block = m_store.allocLinks();
      m_store.allocNodes( m_index, link_block );
      setProperty( LINKS_PROP_INDEX, link_block );
   }

   /**
    * @return {@code true} if this node has no children; otherwise {@code false}.
    */
   public boolean isLeaf()
   {
      return getProperty( LINKS_PROP_INDEX ) == -1;
   }

   /**
    * @return {@code true} if this node is the tree's root; otherwise {@code false}.
    */
   public boolean isRoot()
   {
      return getParentStoreIndex() < 0;
   }

   @Override
   public Object clone()
   {
      return cloneMe();
   }

   void setStoreIndex( int index )
   {
      m_index = index;
   }

   int getChildStoreIndex( int child_index )
   {
      int value = getProperty( LINKS_PROP_INDEX );
      if (value < 0)
         return -1;
      int[] data = m_store.getData();
      return data[value + child_index];
   }

   void setChildStoreIndex( int child_index, int store_index )
   {
      int value = getProperty( LINKS_PROP_INDEX );
      int[] data = m_store.getData();
      data[value + child_index] = store_index;
   }

   protected BaseTreeNode( TreeStore store )
   {
      m_index = 0;
      m_store = store;
   }

   protected BaseTreeNode( BaseTreeNode other )
   {
      m_index = other.m_index;
      m_store = other.m_store;
   }

   /**
    * @param prop_index  property index
    * 
    * @return  The value identified by the specified property index.
    */
   protected int getProperty( int prop_index )
   {
      return m_store.getData()[m_index + prop_index];
   }

   /**
    * Sets the value for the specified property.
    * 
    * @param prop_index  property index
    * @param value       property value
    */
   protected void setProperty( int prop_index, int value )
   {
      m_store.getData()[m_index + prop_index] = value;
   }

   /**
    * Updates the value of a property by an amount.
    * 
    * @param prop_index  property index
    * @param amount      amount that the property value should change
    */
   protected void updateProperty( int prop_index, int amount )
   {
      m_store.getData()[m_index + prop_index] += amount;
   }

   private int       m_index;
   private TreeStore m_store;
}
