//
// DefaultPlayout.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.player;

import java.util.Set;

import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.base.MersenneTwisterFast;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.PointVisitor;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.PrisonerFinder;
import com.dakengo.libs.game.Scorer;

/**
 * This is a portable implementation of random playouts, intended
 * to be used as a fallback wherever a faster, native implementation is
 * not available.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class DefaultPlayout implements Playout
{
   /**
    * @param prng    random number generator, assumed to be thread-local
    * @param scorer  scoring object, assumed to be thread-local
    */
   public DefaultPlayout( MersenneTwisterFast prng, Scorer scorer )
   {
      m_prng = prng;
      m_scorer = scorer;
   }

   @Override
   public double getKomi()
   {
      return m_komi;
   }

   @Override
   public void setKomi( double komi )
   {
      m_komi = komi;
   }

   @Override
   public ImmutablePair<Integer,Long> run( Board board, short test_pnt, Colors to_play, int count )
   {
      long start_millis = System.currentTimeMillis();

      if (count <= 0)
         throw new IllegalArgumentException( Resources.getString( "msg.10003" ) ); //$NON-NLS-1$
      Colors orig_color = board.getColor( test_pnt );
      if (orig_color != Colors.EMPTY)
         throw new IllegalArgumentException( Resources.formatString( "msg.10004", Convert.pointToString( test_pnt ) ) ); //$NON-NLS-1$

      m_last_black_pnt = Board.INVALID_POINT;
      m_copied_board.copyFrom( board );
      m_pnt_index = -1;
      if (!doMove( m_copied_board, test_pnt, to_play ))
         throw new IllegalStateException( Resources.formatString( "msg.10004", Convert.pointToString( test_pnt ) ) ); //$NON-NLS-1$

      m_point_count = 0;
      m_copied_board.visitPoints( Colors.EMPTY, m_copier );
      if (m_point_count == 0)
         throw new IllegalStateException( Resources.getString( "msg.10002" ) ); //$NON-NLS-1$
      else if (m_point_count == 1)
         return ImmutablePair.of( updateScore( m_copied_board, 0, to_play ), System.currentTimeMillis() - start_millis );

      int score = 0;
      for (int playout = 0; playout < count; playout++)
      {
         m_playout_board.copyFrom( m_copied_board );

         short pnt = Board.INVALID_POINT;
         m_pnt_index = 0;
         Colors color = to_play;
         do
         {
            color = Game.otherPlayer( color );
            pnt = Board.INVALID_POINT;
            while (pnt == Board.INVALID_POINT && m_pnt_index < m_point_count)
            {
               int swap_index = m_prng.nextInt( m_point_count - m_pnt_index ) + m_pnt_index;  // nextInt() returns an integer drawn uniformly from 0 to n-1
               pnt = m_playout_pnts[swap_index];
               m_playout_pnts[swap_index] = m_playout_pnts[m_pnt_index];
               m_playout_pnts[m_pnt_index] = pnt;
               m_pnt_index++;
               if (!doMove( m_playout_board, pnt, color ))
                  pnt = Board.INVALID_POINT;
               else if (color == Colors.BLACK)
                  m_last_black_pnt = pnt;
            }
         } while (pnt != Board.INVALID_POINT);

         score = updateScore( m_playout_board, score, to_play );
      }

      return ImmutablePair.of( score, System.currentTimeMillis() - start_millis );
   }

   private int updateScore( Board board, int curr_score, Colors to_play )
   {
      ImmutablePair<Integer,Integer> scores = m_scorer.computeScores( board, null );
      double diff = ((double) scores.getRight()) + m_komi - ((double) scores.getLeft());
      int delta = 0;
      if (diff > 0.0)
         delta = 1;  // win for WHITE
      else if (diff < 0.0)
         delta = -1; // win for BLACK
      if (to_play == Colors.BLACK)
         return curr_score - delta;
      else
         return curr_score + delta;
   }

   private boolean doMove( Board board, short pnt, Colors color )
   {
      if (color == Colors.BLACK && pnt == m_last_black_pnt)
         return false;
      m_pris_finder.find( board, pnt, color );
      if (!m_pris_finder.hasLiberty() || m_pris_finder.isEye())
         return false;
      board.setColor( pnt, color );
      Set<Short> prisoners = m_pris_finder.getPrisoners();
      if (prisoners.isEmpty())
         return true;
      int count = prisoners.size();
      board.setColor( prisoners, Colors.EMPTY );
      int dest_index = 0;
      m_pnt_index--;
      while (m_pnt_index > dest_index && count > 0)
      {
         short played_pnt = m_playout_pnts[m_pnt_index];
         if (!prisoners.contains( played_pnt ))
         {
            short tmp_pnt = m_playout_pnts[dest_index];
            m_playout_pnts[dest_index] = played_pnt;
            m_playout_pnts[m_pnt_index] = tmp_pnt;
            dest_index++;
         }
         else
         {
            m_pnt_index--;
            count--;
         }
      }
      return true;
   }

   private final PointVisitor m_copier = new PointVisitor()
   {
      @Override
      public boolean accept( Board b, short p )
      {
         m_playout_pnts[m_point_count++] = p;
         return true;
      }
   };

   private       MersenneTwisterFast m_prng;
   private       Scorer              m_scorer;
   private       int                 m_pnt_index      = -1;
   private       double              m_komi           = Game.DEFAULT_KOMI.doubleValue();
   private       short               m_last_black_pnt = Board.INVALID_POINT;
   private       int                 m_point_count    = 0;
   private final short[]             m_playout_pnts   = new short[Board.DEFAULT_SIZE * Board.DEFAULT_SIZE];
   private final PrisonerFinder      m_pris_finder    = new PrisonerFinder();
   private final Board               m_copied_board   = new Board();
   private final Board               m_playout_board  = new Board();
}
