//
// Player.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// TODO:
//   - add a method to be called during the opponent's turn so
//     that the player can ponder its next move
//

package com.dakengo.libs.player;

import java.util.Map;
import java.util.Set;

import com.google.gson.JsonElement;
import net.xeoh.plugins.base.Plugin;

import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.game.Game;

/**
 * Implement this interface to define a computer player. The implementation must be annotated with the
 * {@code @PluginImplementation} class annotation. The implementation must be thread-safe.<p>
 * 
 * Each plugin must implement a method that returns the plugin's capabilities,
 * identified via the {@code @Capabilities}
 * annotation. A single string is required for {@code Player}, having the prefix
 * {@link com.dakengo.libs.application.Constants#PLAYER_CAP_PREFIX}
 * and a suffix consisting of the player's abbreviated name.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see <a href="http://data.xeoh.net/jspf/api/net/xeoh/plugins/base/annotations/Capabilities.html"><code>@Capabilities</code></a>
 * @see <a href="http://data.xeoh.net/jspf/api/net/xeoh/plugins/base/annotations/PluginImplementation.html"><code>@PluginImplementation</code></a>
 */
public interface Player extends Plugin
{
   /**
    * Initialize the player with data obtained from a
    * {@link com.dakengo.libs.application.Configuration}. The player implementation
    * should expect this method to be called more than once. The {@code context} and
    * {@code data} parameters may change from one call to the next.
    * 
    * @param context  context data for the game that is about to start
    * @param data     player data obtained from a {@link com.dakengo.libs.application.Configuration}
    *                 object loaded for the game
    */
   void init( GameContext context, Map<String,JsonElement> data );

   /**
    * Return this player's next move.
    * 
    * @param game
    */
   short getMove( Game game );

   /**
    * Return the points selected by this player for
    * free handicap placement. 
    * 
    * @param count
    */
   Set<Short> getHandicap( int count );

   /**
    * Clean up the internal state of the player. Called at the conclusion of
    * each game.
    * 
    * @param game_id  identifier of the game that has ended
    */
   void reset( String game_id );

   /**
    * Returns the name of this player to be stored in a
    * game record.
    * 
    * @param engine_name     name of the engine hosting the computer player
    * @param engine_version  version of the engine hosting the computer player
    */
   String getName( String engine_name, String engine_version );
}
