//
// TreeCursor.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.player;

/**
 * Tree cursors provide a facility for navigating game trees.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class TreeCursor<T extends BaseTreeNode>
{
   /**
    * Default constructor.
    */
   public TreeCursor()
   {
   }

   /**
    * Initializes a new cursor to refer to the specified tree node.
    * 
    * @param node
    */
   public TreeCursor( T node )
   {
      m_node = node;
   }

   /**
    * Initializes a new cursor to refer to the same tree node as another cursor.
    * 
    * @param other
    */
   @SuppressWarnings("unchecked")
   public TreeCursor( TreeCursor<T> other )
   {
      if (other.m_node != null)
         m_node = (T) other.m_node.clone();
   }

   /**
    * Accesses the cursor's node.
    * 
    * @return This current node, or {@code null} if the cursor
    *         refers to no node.
    */
   public T getNode()
   {
      return m_node;
   }

   /**
    * Sets the cursor's node.
    * 
    * @param node
    */
   public void setNode( T node )
   {
      m_node = node;
   }

   /**
    * Distinguishes internal nodes from leaf nodes.
    * 
    * @return  {@code true} if the current node is not {@code null} and
    *          has no children, {@code false} otherwise.
    */
   public boolean isLeaf()
   {
      return m_node != null && m_node.isLeaf();
   }

   /**
    * Distinguishes the tree's root node from all other nodes in the tree.
    * 
    * @return  {@code true} if the current node is the root of a tree,
    *          {@code false} otherwise.
    */
   public boolean isRoot()
   {
      return m_node != null && m_node.isRoot();
   }

   /**
    * Moves this cursor to the node's parent.
    * 
    * @return  {@code true} if there was a parent node, or {@code false} if the node was
    *          {@code null} or was the root of a tree.
    */
   public boolean ascend()
   {
      if (m_node == null)
         return false;
      int parent = m_node.getParentStoreIndex();
      if (parent < 0)
         return false;
      m_node.setStoreIndex( parent );
      return true;
   }

   /**
    * Moves this cursor to one of the current node's children, based on
    * the specified tree policy.
    * 
    * @param policy
    * 
    * @return  {@code true} if the current node has at least one child and the policy
    *          selected it, or {@code false} if the node is a leaf or the policy selected
    *          none of its children.
    */
   public boolean descend( TreePolicy<T> policy )
   {
      if (m_node == null || m_node.isLeaf())
         return false;
      int index = policy.selectChild( this );
      if (index < 0)
         return false;
      m_node.setStoreIndex( m_node.getChildStoreIndex( index ) );
      return true;
   }

   @Override
   @SuppressWarnings("unchecked")
   public boolean equals( Object obj )
   {
      if (this == obj)
         return true;
      if (!(obj instanceof TreeCursor<?>))
         return false;
      TreeCursor<T> other = (TreeCursor<T>) obj;
      return m_node.equals( other.m_node );
   }

   @Override
   public int hashCode()
   {
      return m_node.hashCode();
   }

   @Override
   public String toString()
   {
      if (m_node == null)
         return "<null>";
      return m_node.toString();
   }

   private T m_node;
}
