//
// TreeStorage.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.player;

import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;

/**
 * This class wraps an integer array which serves as the storage medium for
 * a game tree. Applications define subclasses of {@link BaseTreeNode} to
 * represent positions in the tree.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class TreeStore
{
   /**
    * @param prop_count  number of subclass properties per node
    * @param fan_out     number of child links for each internal node
    */
   public TreeStore( int prop_count, int fan_out )
   {
      if (prop_count < 0)
         throw new IllegalArgumentException( Resources.getString( "msg.10007" ) ); //$NON-NLS-1$
      if (fan_out <= 0)
         throw new IllegalArgumentException( Resources.getString( "msg.10008" ) ); //$NON-NLS-1$
      m_prop_count = BaseTreeNode.FIRST_SUBCLASS_PROP_INDEX + prop_count;
      m_fan_out = fan_out;
      m_alloc_size = (m_prop_count + m_fan_out) * m_fan_out * GROWTH_FACTOR;
      m_data = new int[m_alloc_size];
      m_next_index = m_prop_count;
      initNode( 0 );
   }

   /**
    * @return The player color of the root node; all other node colors are determined by this.
    */
   public Colors getRootColor()
   {
      return m_root_color;
   }

   /**
    * Sets the player color of the root node; all other node colors are determined by this.
    * The specified color must be a player color.
    * 
    * @param color  player color
    * 
    * @see com.dakengo.libs.game.Game#checkColor
    */
   public void setRootColor( Colors color )
   {
      Game.checkColor( color );
      m_root_color = color;
   }

   /**
    * @return The number of child links for each internal node.
    */
   public int getFanOut()
   {
      return m_fan_out;
   }

   /**
    * @return The number of properties per node.
    */
   public int getPropertyCount()
   {
      return m_prop_count;
   }

   /**
    * Resets the tree such that there is only a root node.
    */
   public void reset()
   {
      initNode( 0 );
      m_next_index = m_prop_count;
   }

   /**
    * Allocates space within the tree's data storage for a number of nodes equal to
    * the tree's fanout.
    * 
    * @param parent       storage index of the new nodes' parent node
    * @param link_block   storage index of the link block that will contain the children's indexes
    * 
    * @see #allocLinks
    */
   public void allocNodes( int parent, int link_block )
   {
      if (m_next_index + (m_prop_count * m_fan_out) > m_data.length)
         grow();
      for (int i = 0; i < m_fan_out; i++)
      {
         initNode( m_next_index );
         m_data[m_next_index] = parent;
         m_data[link_block + i] = m_next_index;
         m_next_index += m_prop_count;
      }
   }

   /**
    * Allocates space within the tree's data storage for an array of storage indexes, to
    * contain child node indexes.
    * 
    * @return  A storage index.
    * 
    * @see #allocNodes
    */
   public int allocLinks()
   {
      if ((m_next_index + m_fan_out) > m_data.length)
         grow();
      int block_index = m_next_index;
      m_next_index += m_fan_out;
      return block_index;
   }

   /**
    * @return The underlying data pointer for the tree.
    */
   public int[] getData()
   {
      return m_data;
   }

   private void grow()
   {
      int[] new_data = new int[m_data.length + m_alloc_size];
      System.arraycopy( m_data, 0, new_data, 0, m_data.length );
      m_data = new_data;
   }

   private void initNode( int index )
   {
      for (int i = 0; i < m_prop_count; i++)
            m_data[index + i] = -1;
   }

   private int GROWTH_FACTOR = 32;

   private Colors m_root_color = Colors.BLACK;
   private int    m_fan_out;
   private int    m_prop_count;
   private int    m_next_index;
   private int    m_alloc_size;
   private int[]  m_data;
}
