//
// BoardCollector.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.record;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.KoRules;

/**
 * This class returns the board positions for turns in a game record.
 * The application may specify that the positions be rolled up into a
 * single board, thus saving memory usage. The application may also
 * request a Result for the final board position.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public class BoardCollector implements PropertyVisitor
{
   public BoardCollector()
   {
   }

   /**
    * Traverse the record's properties to gather up the boards representing
    * each position in the game. Call this method instead of calling
    * the Record or PositionProperties visit() methods directly.
    * 
    * @param rec
    * @param turn_num  Specifies which boards to collect, as follows:
    *                  <ul>
    *                  <li>0 means collect just the initial handicap board, if any
    *                  <li>negative means collect only the last board
    *                  <li>positive means collect all boards up to the indicated turn number
    *                  </ul>
    */
   public void collect( Record rec, int turn_num )
   {
      m_handicap_board = null;
      m_handicap_count = 0;
      m_black_stones.clear();
      m_game.reset();
      m_boards.clear();
      m_turn_num = turn_num;
      m_collect_called = true;
      try
      {
         rec.visit( this, true );
      }
      finally
      {
         m_collect_called = false;
      }
   }

   /**
    * Returns pairs representing the collected boards in the order in which
    * moves occurred. The lhs of each pair points to a board and rhs specifies
    * the corresponding turn number.
    */
   public Iterable<ImmutablePair<Board,Integer>> getBoards()
   {
      return m_boards;
   }

   /**
    * Returns the game object used while collecting boards.
    */
   public Game getGame()
   {
      return m_game;
   }

   /**
    * Returns a board representing the handicap stones for the game;
    * returns null if there was no handicap.
    */
   public Board getHandicapBoard()
   {
      return m_handicap_board;
   }

   /**
    * Returns the number of handicap stones.
    */
   public int getHandicapStoneCount()
   {
      return m_handicap_count;
   }

   /**
    * Returns the komi value used when calculating the final
    * board result.
    */
   public BigDecimal getKomi()
   {
      return m_game.getKomi();
   }

   /**
    * Sets the komi value used when calculating the final
    * board result.
    * 
    * @param komi
    */
   public void setKomi( BigDecimal komi )
   {
      m_game.setKomi( komi );
   }

   /**
    * Returns a pair containing the final board collected and
    * the corresponding turn number. Returns null if no boards
    * have been collected.
    */
   public ImmutablePair<Board,Integer> getLast()
   {
      if (m_boards.isEmpty())
         return null;
      return m_boards.get( m_boards.size() - 1 );
   }

   @Override
   public boolean startBranch( PositionProperties branch )
   {
      return !m_game.isOver();
   }

   @Override
   public void finishBranch( PositionProperties branch )
   {
      // empty
   }

   @SuppressWarnings( "unchecked" )
   @Override
   public void accept( int prop_id, Object value )
   {
      if (!m_collect_called)
         throw new IllegalStateException( Resources.getString( "msg.4000" ) ); //$NON-NLS-1$
      if (m_game.isOver())
         return;
      switch (prop_id)
      {
         case PropertyID.HANDICAP:
            if (m_handicap_count > 0)
               throw new IllegalStateException( Resources.getString( "msg.4001" ) ); //$NON-NLS-1$
            ImmutablePair<Handicaps,Integer> h_value = (ImmutablePair<Handicaps,Integer>) value;
            m_handicap_count = h_value.getRight();
            maybeInitHandicapBoard();
            break;
         case PropertyID.MOVE:
            if (m_turn_num == 0 || (m_turn_num > 0 && !m_boards.isEmpty() && m_boards.get( m_boards.size() - 1 ).getRight() >= m_turn_num))
               return;
            ImmutablePair<Colors,Short> move = (ImmutablePair<Colors,Short>) value;
            m_game.setPlayer( move.getLeft() );
            m_game.play( move.getRight(), null );
            addBoard( false );
            break;
         case PropertyID.STONES:
            ImmutablePair<Colors,Set<Short>> s_value = (ImmutablePair<Colors,Set<Short>>) value;
            if (m_black_stones.isEmpty() && s_value.getLeft() == Colors.BLACK)
               m_black_stones = new HashSet<Short>( s_value.getRight() );
            m_game.placeStones( s_value.getLeft(), s_value.getRight() );
            addBoard( true );
            maybeInitHandicapBoard();
            break;
      }
   }

   private void addBoard( boolean same_turn )
   {
      int turn = -1;
      if (m_boards.isEmpty())
      {
         turn = same_turn ? 0 : 1;
      }
      else
      {
         turn = m_boards.get( m_boards.size() - 1 ).getRight();
         if (!same_turn)
            turn++;
      }
      Board board = m_game.getBoard();
      if (m_turn_num >= 0)
         board = new Board( board );
      else if (!m_boards.isEmpty())
         m_boards.remove( m_boards.size() - 1 );
      m_boards.add( ImmutablePair.of( board, turn ) );
   }

   private void maybeInitHandicapBoard()
   {
      if (m_handicap_board != null || m_handicap_count == 0 || m_black_stones.isEmpty())
         return;
      if (m_handicap_count != m_black_stones.size())
         m_handicap_count = m_black_stones.size();
      m_handicap_board = new Board();
      m_handicap_board.setColor( m_black_stones, Colors.BLACK );
   }

   private List<ImmutablePair<Board,Integer>> m_boards          = new ArrayList<ImmutablePair<Board,Integer>>();
   private Game                               m_game            = new Game( KoRules.NONE, true, Game.DEFAULT_KOMI );
   private Board                              m_handicap_board;
   private int                                m_handicap_count;
   private int                                m_turn_num;
   private Set<Short>                         m_black_stones    = new HashSet<Short>();
   private boolean                            m_collect_called;
}
