//
// GameProperties.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// TODO:
//   - if practical, formalize the char encodings that can be set for
//     PropertyID.CHARSET
//

package com.dakengo.libs.record;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.base.Constants;
import com.dakengo.libs.base.HashUtils;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.Result;

/**
 * A container for properties that provide general information about
 * a game:<p>
 *
 * <ul>
 * <li>{@link PropertyID#ANNOTATOR_NAME}
 * <li>{@link PropertyID#APP_NAME}
 * <li>{@link PropertyID#APP_VERSION}
 * <li>{@link PropertyID#BACKGROUND_INFO}
 * <li>{@link PropertyID#CHARSET}
 * <li>{@link PropertyID#COPYRIGHT}
 * <li>{@link PropertyID#DATE}
 * <li>{@link PropertyID#EVENT_NAME}
 * <li>{@link PropertyID#GAME_NAME}
 * <li>{@link PropertyID#HANDICAP}
 * <li>{@link PropertyID#KOMI}
 * <li>{@link PropertyID#OPENING_NAME}
 * <li>{@link PropertyID#OVERTIME_SYSTEM}
 * <li>{@link PropertyID#PLACE_NAME}
 * <li>{@link PropertyID#PLAYER_NAME}
 * <li>{@link PropertyID#PLAYER_RANK}
 * <li>{@link PropertyID#RENDERING_STYLE}
 * <li>{@link PropertyID#RESULT}
 * <li>{@link PropertyID#ROUND}
 * <li>{@link PropertyID#RULE_SYSTEM}
 * <li>{@link PropertyID#SIZE}
 * <li>{@link PropertyID#SOURCE_NAME}
 * <li>{@link PropertyID#TEAM_NAME}
 * <li>{@link PropertyID#TIME_LIMIT}
 * <li>{@link PropertyID#USER_NAME}
 * </ul><p>
 * 
 * Non-standard properties (id > {@link PropertyID#MAX_STD_ID})
 * may also be set.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see PositionProperties
 * @see Record
 */
public class GameProperties
{
   public static final ImmutablePair<Handicaps,Integer> DEFAULT_HANDICAP        = ImmutablePair.of( Handicaps.NONE, 0 );
   public static final RenderingStyle                   DEFAULT_RENDERING_STYLE = new RenderingStyle();
   public static final Result                           DEFAULT_RESULT          = new Result( Colors.DONT_CARE, Result.Status.DRAW );
   public static final Round                            DEFAULT_ROUND           = new Round( 1, Round.Type.GAME, "" );  //$NON-NLS-1$
   public static final ImmutablePair<String,String>     DEFAULT_STRING_PAIR     = ImmutablePair.of( "", "" );           //$NON-NLS-1$ //$NON-NLS-2$
   public static final int                              DEFAULT_TIME_LIMIT      = Integer.MAX_VALUE;

   /**
    * Returns true if the specified ID is a standard game property ID.
    */
   public static boolean isStandard( int id )
   {
      return id >= PropertyID.ANNOTATOR_NAME && id <= PropertyID.USER_NAME;
   }

   public GameProperties()
   {
   }

   /**
    * Creates a deep copy of the specified properties object.
    * 
    * @param other
    */
   public GameProperties( GameProperties other )
   {
      m_nonplayer_props = new HashMap<Integer,Object>( other.m_nonplayer_props );
      m_black_props = new HashMap<Integer,String>( other.m_black_props );
      m_white_props = new HashMap<Integer,String>( other.m_white_props );
   }

   /**
    * Returns the name of the person who made annotations to the game record.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setAnnotatorName
    */
   public String getAnnotatorName()
   {
      return (String) get( PropertyID.ANNOTATOR_NAME );
   }

   /**
    * Sets the name of the person who made annotations to the game record.
    * 
    * @param name
    * 
    * @see #get
    * @see #getAnnotatorName
    * @see #has
    * @see #set
    */
   public void setAnnotatorName( String name )
   {
      set( PropertyID.ANNOTATOR_NAME, name );
   }

   /**
    * Returns the name of the application that constructed the game record.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setApplicationName
    */
   public String getApplicationName()
   {
      return (String) get( PropertyID.APP_NAME );
   }

   /**
    * Sets the name of the application that constructed the game record.
    * 
    * @param name
    * 
    * @see #get
    * @see #getApplicationName
    * @see #has
    * @see #set
    */
   public void setApplicationName( String name )
   {
      set( PropertyID.APP_NAME, name );
   }

   /**
    * Returns the version of the application that constructed the game record.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setApplicationVersion
    */
   public String getApplicationVersion()
   {
      return (String) get( PropertyID.APP_VERSION );
   }

   /**
    * Sets the version of the application that constructed the game record.
    * 
    * @param version
    * 
    * @see #get
    * @see #getApplicationVersion
    * @see #has
    * @see #set
    */
   public void setApplicationVersion( String version )
   {
      set( PropertyID.APP_VERSION, version );
   }

   /**
    * Returns extra information about, or a summary of, the game
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setBackgroundInfo
    */
   public String getBackgroundInfo()
   {
      return (String) get( PropertyID.BACKGROUND_INFO );
   }

   /**
    * Sets extra information about, or a summary of, the game.
    * 
    * @param info
    * 
    * @see #get
    * @see #getBackgroundInfo
    * @see #has
    * @see #set
    */
   public void setBackgroundInfo( String info )
   {
      set( PropertyID.BACKGROUND_INFO, info );
   }

   /**
    * Returns the character set in which the text of the game record is encoded.
    * By default, the record's charset is "utf-8".
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setCharset
    */
   public String getCharset()
   {
      return (String) get( PropertyID.CHARSET );
   }

   /**
    * Sets the character set in which the text of the game record is encoded.
    * 
    * @param charset
    * 
    * @see #get
    * @see #getCharset
    * @see #has
    * @see #set
    */
   public void setCharset( String charset )
   {
      set( PropertyID.CHARSET, charset );
   }

   /**
    * Returns the copyright text for the game record.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setCopyright
    */
   public String getCopyright()
   {
      return (String) get( PropertyID.COPYRIGHT );
   }

   /**
    * Sets the copyright text for the game record.
    * 
    * @param copyright
    * 
    * @see #get
    * @see #getCopyright
    * @see #has
    * @see #set
    */
   public void setCopyright( String copyright )
   {
      set( PropertyID.COPYRIGHT, copyright );
   }

   /**
    * Returns the date on which the game was played.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setDate
    */
   public String getDate()
   {
      return (String) get( PropertyID.DATE );
   }

   /**
    * Sets the date on which the game was played.
    * 
    * @param date
    * 
    * @see #get
    * @see #getDate
    * @see #has
    * @see #set
    */
   public void setDate( String date )
   {
      set( PropertyID.DATE, date );
   }

   /**
    * Returns the name of the event in which the game took place.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setEventName
    */
   public String getEventName()
   {
      return (String) get( PropertyID.EVENT_NAME );
   }

   /**
    * Sets the name of the event in which the game took place.
    * 
    * @param name
    * 
    * @see #get
    * @see #getEventName
    * @see #has
    * @see #set
    */
   public void setEventName( String name )
   {
      set( PropertyID.EVENT_NAME, name );
   }

   /**
    * Returns the name given to this particular game.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setGameName
    */
   public String getGameName()
   {
      return (String) get( PropertyID.GAME_NAME );
   }

   /**
    * Sets the name given to this particular game.
    * 
    * @param name
    * 
    * @see #get
    * @see #getGameName
    * @see #has
    * @see #set
    */
   public void setGameName( String name )
   {
      set( PropertyID.GAME_NAME, name );
   }

   /**
    * Returns the placement and number of handicap stones.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setHandicap
    */
   @SuppressWarnings( "unchecked" )
   public ImmutablePair<Handicaps,Integer> getHandicap()
   {
      return (ImmutablePair<Handicaps,Integer>) get( PropertyID.HANDICAP );
   }

   /**
    * Sets the placement and number of handicap stones.
    * 
    * @param value
    * 
    * @see #get
    * @see #getHandicap
    * @see #has
    * @see #set
    */
   public void setHandicap( ImmutablePair<Handicaps,Integer> value )
   {
      set( PropertyID.HANDICAP, value );
   }

   /**
    * Returns the komi value. By default, the komi is 6.5.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setKomi
    */
   public BigDecimal getKomi()
   {
      return (BigDecimal) get( PropertyID.KOMI );
   }

   /**
    * Sets the komi value.
    * 
    * @param komi
    * 
    * @see #get
    * @see #getKomi
    * @see #has
    * @see #set
    */
   public void setKomi( BigDecimal komi )
   {
      set( PropertyID.KOMI, komi );
   }

   /**
    * Sets the komi value.
    * 
    * @param komi
    * 
    * @see #get
    * @see #getKomi
    * @see #has
    * @see #set
    */
   public void setKomi( String komi )
   {
      try
      {
         set( PropertyID.KOMI, new BigDecimal( komi ) );
      }
      catch (NumberFormatException unused)
      {
         throw new ParseException( Resources.getString( "msg.4002" ), komi ); //$NON-NLS-1$
      }
   }

   /**
    * Returns the name of the opening that was played.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setOpeningName
    */
   public String getOpeningName()
   {
      return (String) get( PropertyID.OPENING_NAME );
   }

   /**
    * Sets the name of the opening that was played.
    * 
    * @param name
    * 
    * @see #get
    * @see #getOpeningName
    * @see #has
    * @see #set
    */
   public void setOpeningName( String name )
   {
      set( PropertyID.OPENING_NAME, name );
   }

   /**
    * Returns the overtime system in force for the game.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setOvertimeSystem
    */
   public String getOvertimeSystem()
   {
      return (String) get( PropertyID.OVERTIME_SYSTEM );
   }

   /**
    * Sets the overtime system in force for the game.
    * 
    * @param system
    * 
    * @see #get
    * @see #getOvertimeSystem
    * @see #has
    * @see #set
    */
   public void setOvertimeSystem( String system )
   {
      set( PropertyID.OVERTIME_SYSTEM, system );
   }

   /**
    * Returns the location where the game took place.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setPlaceName
    */
   public String getPlaceName()
   {
      return (String) get( PropertyID.PLACE_NAME );
   }

   /**
    * Sets the location where the game took place.
    * 
    * @param name
    * 
    * @see #get
    * @see #getPlaceName
    * @see #has
    * @see #set
    */
   public void setPlaceName( String name )
   {
      set( PropertyID.PLACE_NAME, name );
   }

   /**
    * Returns the name of a player.
    * 
    * @param color
    * 
    * @see #get
    * @see #getPlayerProperty
    * @see #has
    * @see #set
    * @see #setPlayerName
    */
   public String getPlayerName( Colors color )
   {
      return getPlayerProperty( PropertyID.PLAYER_NAME, color );
   }

   /**
    * Sets the name of a player.
    * 
    * @param color
    * @param name
    * 
    * @see #get
    * @see #getPlayerName
    * @see #getPlayerProperty
    * @see #has
    * @see #set
    */
   public void setPlayerName( Colors color, String name )
   {
      set( PropertyID.PLAYER_NAME, ImmutablePair.of( color, name ) );
   }

   /**
    * Returns the rank of a player.
    * 
    * @param color
    * 
    * @see #getPlayerProperty
    * @see #has
    * @see #set
    * @see #setPlayerRank
    */
   public String getPlayerRank( Colors color )
   {
      return getPlayerProperty( PropertyID.PLAYER_RANK, color );
   }

   /**
    * Sets the rank of a player.
    * 
    * @param color
    * @param name
    * 
    * @see #getPlayerRank
    * @see #getPlayerProperty
    * @see #has
    * @see #set
    */
   public void setPlayerRank( Colors color, String name )
   {
      set( PropertyID.PLAYER_RANK, ImmutablePair.of( color, name ) );
   }

   /**
    * Returns rendering hints. By default, the rendering style is to show markup
    * but hide siblings.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setRenderingStyle
    */
   public RenderingStyle getRenderingStyle()
   {
      return (RenderingStyle) get( PropertyID.RENDERING_STYLE );
   }

   /**
    * Sets rendering hints.
    * 
    * @param style
    * 
    * @see #get
    * @see #getRenderingStyle
    * @see #has
    * @see #set
    */
   public void setRenderingStyle( RenderingStyle style )
   {
      set( PropertyID.RENDERING_STYLE, style );
   }

   /**
    * Returns the game result. By default, the game result is a draw.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setResult
    */
   public Result getResult()
   {
      return (Result) get( PropertyID.RESULT );
   }

   /**
    * Sets the game result.
    * 
    * @param result
    * 
    * @see #get
    * @see #getResult
    * @see #has
    * @see #set
    */
   public void setResult( Result result )
   {
      set( PropertyID.RESULT, result );
   }

   /**
    * Returns the round in which this game was played. By default, the round
    * is game 1.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setRound
    */
   public Round getRound()
   {
      return (Round) get( PropertyID.ROUND );
   }

   /**
    * Sets the round in which this game was played.
    * 
    * @param round
    * 
    * @see #get
    * @see #getRound
    * @see #has
    * @see #set
    */
   public void setRound( Round round )
   {
      set( PropertyID.ROUND, round );
   }

   /**
    * Returns the rule system in force for the game.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setRuleSystem
    */
   public String getRuleSystem()
   {
      return (String) get( PropertyID.RULE_SYSTEM );
   }

   /**
    * Sets the rule system in force for the game.
    * 
    * @param system
    * 
    * @see #get
    * @see #getRuleSystem
    * @see #has
    * @see #set
    */
   public void setRuleSystem( String system )
   {
      set( PropertyID.RULE_SYSTEM, system );
   }

   /**
    * Returns the game board size. By default, the board size is
    * 19.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setSize
    */
   public int getSize()
   {
      return ((Integer) get( PropertyID.SIZE )).intValue();
   }

   /**
    * Sets the game board size.
    * 
    * @param size
    * 
    * @see #get
    * @see #getSize
    * @see #has
    * @see #set
    */
   public void setSize( int size )
   {
      set( PropertyID.SIZE, size );
   }

   /**
    * Returns the source from which the game record was obtained.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setSourceName
    */
   public String getSourceName()
   {
      return (String) get( PropertyID.SOURCE_NAME );
   }

   /**
    * Sets the source from which the game record was obtained.
    * 
    * @param name
    * 
    * @see #get
    * @see #getSourceName
    * @see #has
    * @see #set
    */
   public void setSourceName( String name )
   {
      set( PropertyID.SOURCE_NAME, name );
   }

   /**
    * Returns the name of a team.
    * 
    * @param color
    * 
    * @see #getPlayerProperty
    * @see #has
    * @see #set
    * @see #setTeamName
    */
   public String getTeamName( Colors color )
   {
      return getPlayerProperty( PropertyID.TEAM_NAME, color );
   }

   /**
    * Sets the name of a team.
    * 
    * @param color
    * @param name
    * 
    * @see #has
    * @see #getPlayerProperty
    * @see #getTeamName
    * @see #set
    */
   public void setTeamName( Colors color, String name )
   {
      set( PropertyID.TEAM_NAME, ImmutablePair.of( color, name ) );
   }

   /**
    * Returns the time limit for the game, in seconds. By default,
    * there is effectively no time limit.
    * 
    * @see #get
    * @see #has
    * @see #set
    * @see #setTimeLimit
    */
   public int getTimeLimit()
   {
      return ((Integer) get( PropertyID.TIME_LIMIT )).intValue();
   }

   /**
    * Sets the time limit for the game, in seconds.
    * 
    * @param limit
    * 
    * @see #has
    * @see #get
    * @see #getTimeLimit
    * @see #set
    */
   public void setTimeLimit( int limit )
   {
      set( PropertyID.TIME_LIMIT, limit );
   }

   /**
    * Returns the user name of the person who created the game record.
    * 
    * @see #has
    * @see #get
    * @see #set
    * @see #setUserName
    */
   public String getUserName()
   {
      return (String) get( PropertyID.USER_NAME );
   }

   /**
    * Sets the user name of the person who created the game record.
    * 
    * @param name
    * 
    * @see #has
    * @see #get
    * @see #getUserName
    * @see #set
    */
   public void setUserName( String name )
   {
      set( PropertyID.USER_NAME, name );
   }

   /**
    * A weakly-typed method of setting properties. Properties cannot be repeated within
    * a single GameProperties instance.
    *
    * @see #get
    * @see #getPlayerProperty
    * @see #has
    */
   @SuppressWarnings( "unchecked" )
   public void set( int prop_id, Object value )
   {
      if (value == null)
         throw new NullPointerException( Resources.formatString( "msg.4003", prop_id ) ); //$NON-NLS-1$
      ImmutablePair<Colors,String> pl_val = null;
      try
      {
         switch (prop_id)
         {
            case PropertyID.ANNOTATOR_NAME:
            case PropertyID.APP_NAME:
            case PropertyID.APP_VERSION:
            case PropertyID.BACKGROUND_INFO:
            case PropertyID.COPYRIGHT:
            case PropertyID.DATE:
            case PropertyID.EVENT_NAME:
            case PropertyID.GAME_NAME:
            case PropertyID.OPENING_NAME:
            case PropertyID.OVERTIME_SYSTEM:
            case PropertyID.PLACE_NAME:
            case PropertyID.RULE_SYSTEM:
            case PropertyID.SOURCE_NAME:
            case PropertyID.USER_NAME:
               m_nonplayer_props.put( prop_id, (String) value );
               break;
            case PropertyID.CHARSET:
               if (((String) value).isEmpty())
                  value = Constants.UTF_8;
               m_nonplayer_props.put( prop_id, value );
               break;
            case PropertyID.HANDICAP:
               ImmutablePair<Handicaps,Integer> h_value = (ImmutablePair<Handicaps,Integer>) value;
               if (!Game.isValidHandicap( h_value.getRight() ))
                  throw new PropertyException( prop_id, h_value.getRight().toString() );
               m_nonplayer_props.put( prop_id, value );
               break;
            case PropertyID.KOMI:
               m_nonplayer_props.put( prop_id, (BigDecimal) value );
               break;
            case PropertyID.RENDERING_STYLE:
               m_nonplayer_props.put( prop_id, (RenderingStyle) value );
               break;
            case PropertyID.RESULT:
               m_nonplayer_props.put( prop_id, (Result) value );
               break;
            case PropertyID.ROUND:
               m_nonplayer_props.put( prop_id, (Round) value );
               break;
            case PropertyID.PLAYER_NAME:
            case PropertyID.PLAYER_RANK:
            case PropertyID.TEAM_NAME:
               pl_val = (ImmutablePair<Colors,String>) value;
               Game.checkColor( pl_val.getLeft() );
               setPlayerProperty( prop_id, pl_val );
               break;
            case PropertyID.SIZE:
               if (!Board.isValidSize( ((Integer) value).intValue() ))
                  throw new PropertyException( prop_id, value.toString() );
               m_nonplayer_props.put( prop_id, value );
               break;
            case PropertyID.TIME_LIMIT:
               if (((Integer) value).intValue() <= 0)
                  throw new PropertyException( prop_id, value.toString() );
               m_nonplayer_props.put( prop_id, value );
               break;
            default:
               setNonStandard( prop_id, (ImmutablePair<String,String>) value );
               break;
         }
      }
      catch (ClassCastException ex)
      {
         throw new PropertyException( prop_id, ex );
      }
   }

   /**
    * Retrieves a property value. Provides a reasonable default if the property has not
    * been set. Throws a PropertyException if the property is not valid.
    *
    * @see #getPlayerProperty
    * @see #has
    * @see #set
    */
   public Object get( int id )
   {
      switch (id)
      {
         case PropertyID.PLAYER_NAME:
         case PropertyID.PLAYER_RANK:
         case PropertyID.TEAM_NAME:
            return new IllegalArgumentException( Resources.formatString( "msg.4004", id ) ); //$NON-NLS-1$

         case PropertyID.ANNOTATOR_NAME:
         case PropertyID.APP_NAME:
         case PropertyID.APP_VERSION:
         case PropertyID.BACKGROUND_INFO:
         case PropertyID.COPYRIGHT:
         case PropertyID.DATE:
         case PropertyID.EVENT_NAME:
         case PropertyID.GAME_NAME:
         case PropertyID.OPENING_NAME:
         case PropertyID.OVERTIME_SYSTEM:
         case PropertyID.PLACE_NAME:
         case PropertyID.RULE_SYSTEM:
         case PropertyID.SOURCE_NAME:
         case PropertyID.USER_NAME:
            return get( id, "" ); //$NON-NLS-1$

         case PropertyID.CHARSET:
            return get( id, Constants.UTF_8 );
         case PropertyID.KOMI:
            return get( id, Game.DEFAULT_KOMI );
         case PropertyID.RENDERING_STYLE:
            return get( id, DEFAULT_RENDERING_STYLE );
         case PropertyID.RESULT:
            return get( id, DEFAULT_RESULT );
         case PropertyID.ROUND:
            return get( id, DEFAULT_ROUND );
         case PropertyID.SIZE:
            return get( id, Board.DEFAULT_SIZE );
         case PropertyID.TIME_LIMIT:
            return get( id, DEFAULT_TIME_LIMIT );

         case PropertyID.HANDICAP:
            return get( id, DEFAULT_HANDICAP );

         default:
            return getNonStandard( id, DEFAULT_STRING_PAIR );
      }
   }

   /**
    * Returns a player-specific property value. If the specified property was
    * not set, this method returns an empty string.
    * 
    * @param id
    * @param color
    * 
    * @see #get
    * @see #has
    * @see #set
    */
   public String getPlayerProperty( int id, Colors color )
   {
      if (color == Colors.BLACK)
      {
         if (!m_black_props.containsKey( id ))
            return ""; //$NON-NLS-1$
         return m_black_props.get( id );
      }
      if (color == Colors.WHITE)
      {
         if (!m_white_props.containsKey( id ))
            return ""; //$NON-NLS-1$
         return m_white_props.get( id );
      }
      throw new IllegalArgumentException( Resources.formatString( "msg.4005", color ) ); //$NON-NLS-1$
   }

   /**
    * Returns true if the specified property has been set. Returns
    * false if the property has not been set, whether or not the property is
    * valid.
    *
    * @see #get
    * @see #getPlayerProperty
    * @see #set
    */
   public boolean has( int id )
   {
      return m_nonplayer_props.containsKey( id ) || m_black_props.containsKey( id ) || m_white_props.containsKey( id );
   }

   /**
    * Clear all properties.
    */
   public void clear()
   {
      m_nonplayer_props.clear();
      m_black_props.clear();
      m_white_props.clear();
   }

   /**
    * Calls the specified visitor for each property that has been set.
    */
   public void visit( PropertyVisitor visitor )
   {
      for (int id : m_nonplayer_props.keySet())
         visitor.accept( id, m_nonplayer_props.get( id ) );
      for (int id : m_black_props.keySet())
         visitor.accept( id, ImmutablePair.of( Colors.BLACK, m_black_props.get( id ) ) );
      for (int id : m_white_props.keySet())
         visitor.accept( id, ImmutablePair.of( Colors.WHITE, m_white_props.get( id ) ) );
   }

   @Override
   public boolean equals( Object obj )
   {
      if (!(obj instanceof GameProperties))
         return false;
      if (obj == this)
         return true;
      GameProperties other = (GameProperties) obj;
      if (m_nonplayer_props.size() != other.m_nonplayer_props.size() ||
          m_black_props.size() != other.m_black_props.size()         ||
          m_white_props.size() != other.m_white_props.size())
         return false;
      for (int id : m_nonplayer_props.keySet())
      {
         if (!other.m_nonplayer_props.containsKey( id ))
            return false;
         if (!other.m_nonplayer_props.get( id ).equals( m_nonplayer_props.get( id ) ))
            return false;
      }
      for (int id : m_black_props.keySet())
      {
         if (!other.m_black_props.containsKey( id ))
            return false;
         if (!other.m_black_props.get( id ).equals( m_black_props.get( id ) ))
            return false;
      }
      for (int id : m_white_props.keySet())
      {
         if (!other.m_white_props.containsKey( id ))
            return false;
         if (!other.m_white_props.get( id ).equals( m_white_props.get( id ) ))
            return false;
      }
      return true;
   }

   @Override
   public int hashCode()
   {
      return HashUtils.compute(
         HashUtils.computeAnyOrder( m_nonplayer_props.keySet() ),
         HashUtils.computeAnyOrder( m_black_props.keySet() ),
         HashUtils.computeAnyOrder( m_white_props.keySet() ) );
   }

   @SuppressWarnings( "unchecked" )
   private ImmutablePair<String,String> getNonStandard( int id, ImmutablePair<String,String> def_val )
   {
      if (id <= PropertyID.MAX_STD_ID)
         throw new PropertyException( id );
      if (!m_nonplayer_props.containsKey( id ))
         return def_val;
      return (ImmutablePair<String,String>) m_nonplayer_props.get( id );
   }

   private void setNonStandard( int id, ImmutablePair<String,String> value )
   {
      if (id < PropertyID.COMMENT)
         throw new PropertyException( id );
      m_nonplayer_props.put( id, value );
   }

   private void setPlayerProperty( int id, ImmutablePair<Colors,String> value )
   {
      switch( id )
      {
         case PropertyID.PLAYER_NAME:
         case PropertyID.PLAYER_RANK:
         case PropertyID.TEAM_NAME:
            if (value.getLeft() == Colors.BLACK)
               m_black_props.put( id, value.getRight() );
            else
               m_white_props.put( id, value.getRight() );
            break;
         default:
            throw new IllegalArgumentException( Resources.formatString( "msg.4006", id ) ); //$NON-NLS-1$
      }
   }

   private Object get( int id, Object def_val )
   {
      if (!m_nonplayer_props.containsKey( id ))
         return def_val;
      return m_nonplayer_props.get( id );
   }

   private Map<Integer,Object> m_nonplayer_props = new HashMap<Integer,Object>();
   private Map<Integer,String> m_black_props     = new HashMap<Integer,String>();
   private Map<Integer,String> m_white_props     = new HashMap<Integer,String>();
}
