//
// PositionProperties.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// TODO:
//   - if practical, formalize the char encodings that can be set for
//     the Charset property
//

package com.dakengo.libs.record;

import java.math.BigDecimal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.base.HashUtils;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;

/**
 * This class maintains a list of properties representing a line of play
 * in a game record:<p>
 *
 * <ul>
 * <li>{@link PropertyID#BRANCH}
 * <li>{@link PropertyID#COMMENT}
 * <li>{@link PropertyID#CONNECTOR}
 * <li>{@link PropertyID#DIAGRAM_SETTINGS}
 * <li>{@link PropertyID#GRADE}
 * <li>{@link PropertyID#ILLEGAL_MOVE}
 * <li>{@link PropertyID#LABEL}
 * <li>{@link PropertyID#MARK}
 * <li>{@link PropertyID#MOVE}
 * <li>{@link PropertyID#NOTE}
 * <li>{@link PropertyID#PRINT_SETTINGS}
 * <li>{@link PropertyID#REMAINING_MOVES}
 * <li>{@link PropertyID#REMAINING_TIME}
 * <li>{@link PropertyID#STONES}
 * <li>{@link PropertyID#TERRITORY}
 * <li>{@link PropertyID#TURN}
 * <li>{@link PropertyID#VISIBLE}
 * </ul><p>
 * 
 * Non-standard properties (id > {@link PropertyID#MAX_STD_ID})
 * may also be set.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see GameProperties
 * @see Record
 */
public class PositionProperties
{
   /**
    * Returns true if the specified ID is a standard position property ID.
    */
   public static boolean isStandard( int id )
   {
      return id >= PropertyID.COMMENT && id <= PropertyID.MAX_STD_ID;
   }

   /**
    * Initializes a new line of play.
    * 
    * @param is_root  pass true if this object will represent the initial (main) branch
    *                 of a game record
    */
   public PositionProperties( boolean is_root )
   {
      m_is_root = is_root;
   }

   /**
    * Creates a new line of play that is a deep copy of the specified
    * object.
    * 
    * @param other
    */
   @SuppressWarnings( "unchecked" )
   public PositionProperties( PositionProperties other )
   {
      m_is_root = other.m_is_root;
      for (ImmutablePair<Integer,?> entry : other.m_entries)
      {
         int prop_id = entry.getLeft();
         switch (prop_id)
         {
            case PropertyID.CONNECTOR:
               ImmutablePair<Connectors,Set<ImmutablePair<Short,Short>>> conn_val = (ImmutablePair<Connectors,Set<ImmutablePair<Short,Short>>>) entry.getRight();
               add( prop_id, ImmutablePair.of( conn_val.getLeft(), new HashSet<ImmutablePair<Short,Short>>( conn_val.getRight() ) ) );
               break;
            case PropertyID.DIAGRAM_SETTINGS:
               ImmutablePair<String,DiagramSettings> diag_val = (ImmutablePair<String,DiagramSettings>) entry.getRight();
               add( prop_id, ImmutablePair.of( prop_id, ImmutablePair.of( diag_val.getLeft(), new DiagramSettings( diag_val.getRight() ) ) ) );
               break;
            case PropertyID.LABEL:
               add( prop_id, new HashSet<ImmutablePair<Short,String>>( (Set<ImmutablePair<Short,String>>) entry.getRight() ) );
               break;
            case PropertyID.MARK:
               ImmutablePair<Notations,Set<Short>> mark_val = (ImmutablePair<Notations,Set<Short>>) entry.getRight();
               add( prop_id, ImmutablePair.of( mark_val.getLeft(), new HashSet<Short>( mark_val.getRight() ) ) );
               break;
            case PropertyID.PRINT_SETTINGS:
               add( prop_id, new PrintSettings( (PrintSettings) entry.getRight() ) );
               break;
            case PropertyID.STONES:
            case PropertyID.TERRITORY:
               ImmutablePair<Colors,Set<Short>> st_value = (ImmutablePair<Colors,Set<Short>>) entry.getRight();
               add( prop_id, ImmutablePair.of( st_value.getLeft(), new HashSet<Short>( st_value.getRight() ) ) );
               break;
            case PropertyID.VISIBLE:
               Set<Short> vis_value = (Set<Short>) entry.getRight();
               add( prop_id, new HashSet<Short>( vis_value ) );
               break;
            default:
               add( prop_id, entry.getRight() );
               break;
         }
      }
   }

   /**
    * Adds a comment.
    * 
    * @param text
    * 
    * @see #add
    * @see #has
    */
   public void addComment( String text )
   {
      add( PropertyID.COMMENT, text );
   }

   /**
    * Connects the first point to the second point of each pair.
    * 
    * @param connector
    * @param pnt_pairs
    * 
    * @see #add
    * @see #has
    */
   public void addConnector( Connectors connector, Set<ImmutablePair<Short,Short>> pnt_pairs )
   {
      add( PropertyID.CONNECTOR, ImmutablePair.of( connector, pnt_pairs ) );
   }

   /**
    * Configures the rendering of a diagram at the current location in the game record.
    * 
    * @param message
    * @param settings
    * 
    * @see #add
    * @see #has
    */
   public void addDiagramSettings( String message, DiagramSettings settings )
   {
      add( PropertyID.DIAGRAM_SETTINGS, ImmutablePair.of( message, new DiagramSettings( settings ) ) );
   }

   /**
    * Adds a notation for how good or bad the current game record element is.
    * 
    * @param color
    * @param type
    * @param grade
    * 
    * @see #add
    * @see #has
    */
   public void addGrade( Colors color, GradeTypes type, Grades grade )
   {
      add( PropertyID.GRADE, ImmutablePair.of( color, ImmutablePair.of( type, grade ) ) );
   }

   /**
    * Adds a move to be rendered (e.g., in a viewer) despite being illegal.
    * 
    * @param color
    * @param point
    * 
    * @see #add
    * @see #has
    */
   public void addIllegalMove( Colors color, short point )
   {
      add( PropertyID.ILLEGAL_MOVE, ImmutablePair.of( color, point ) );
   }

   /**
    * Adds points to be labeled with text.
    * 
    * @param labels
    * 
    * @see #add
    * @see #has
    */
   public void addLabel( Set<ImmutablePair<Short,String>> labels )
   {
      add( PropertyID.LABEL, labels );
   }

   /**
    * Adds an instruction to decorate the specified points.
    * 
    * @param notation
    * @param points
    * 
    * @see #add
    * @see #has
    */
   public void addMark( Notations notation, Set<Short> points )
   {
      add( PropertyID.MARK, ImmutablePair.of( notation, points ) );
   }

   /**
    * Adds a legal move.
    * 
    * @param color
    * @param point
    * 
    * @see #add
    * @see #has
    */
   public void addMove( Colors color, short point )
   {
      add( PropertyID.MOVE, ImmutablePair.of( color, point ) );
   }

   /**
    * Adds an arbitrary label/text pair along with a number.
    * 
    * @param number
    * @param label
    * @param text
    * 
    * @see #add
    * @see #has
    */
   public void addNote( int number, String label, String text )
   {
      add( PropertyID.NOTE, ImmutablePair.of( number, ImmutablePair.of( label, text ) ) );
   }

   /**
    * Configures printing at the current location in the game record.
    * 
    * @param settings
    * 
    * @see #add
    * @see #has
    */
   public void addPrintSettings( PrintSettings settings )
   {
      add( PropertyID.PRINT_SETTINGS, new PrintSettings( settings ) );
   }

   /**
    * Adds the number of remaining moves left in a byo-yomi period.
    * 
    * @param color
    * @param count
    * 
    * @see #add
    * @see #has
    */
   public void addRemainingMoves( Colors color, int count )
   {
      add( PropertyID.REMAINING_MOVES, ImmutablePair.of( color, count ) );
   }

   /**
    * Adds the remaining amount of time in seconds left for the
    * indicated player to play.
    * 
    * @param color
    * @param time
    * 
    * @see #add
    * @see #has
    */
   public void addRemainingTime( Colors color, BigDecimal time )
   {
      add( PropertyID.REMAINING_TIME, ImmutablePair.of( color, time ) );
   }

   /**
    * Adds a collection of points to be set for a player or cleared from the board.
    * 
    * @param color
    * @param points
    * 
    * @see #add
    * @see #has
    */
   public void addStones( Colors color, Set<Short> points )
   {
      add( PropertyID.STONES, ImmutablePair.of( color, points ) );
   }

   /**
    * Adds a collection of points representing territory for the indicated
    * player.
    * 
    * @param color
    * @param points
    * 
    * @see #add
    * @see #has
    */
   public void addTerritory( Colors color, Set<Short> points )
   {
      add( PropertyID.TERRITORY, ImmutablePair.of( color, points ) );
   }

   /**
    * Adds an indication of whose turn it is next.
    * 
    * @param color
    * 
    * @see #add
    * @see #has
    */
   public void addTurn( Colors color )
   {
      add( PropertyID.TURN, color );
   }

   /**
    * Limits the board points that are visible.
    * 
    * @param points
    * 
    * @see #add
    * @see #has
    */
   public void addVisible( Set<Short> points )
   {
      add( PropertyID.VISIBLE, points );
   }

   /**
    * Adds a branch to this line of play.
    * 
    * @param branch
    * 
    * @see #add
    * @see #has
    */
   public void addBranch( PositionProperties branch )
   {
      if (branch.isRoot())
         throw new PropertyException( PropertyID.BRANCH, Resources.getString( "msg.4007" ) ); //$NON-NLS-1$
      add( PropertyID.BRANCH, branch );
   }

   /**
    * Appends a new property.
    * 
    * @param prop_id
    * @param value
    * 
    * @see #addBranch
    * @see #addComment
    * @see #addConnector
    * @see #addDiagramSettings
    * @see #addGrade
    * @see #addIllegalMove
    * @see #addLabel
    * @see #addMark
    * @see #addMove
    * @see #addNote
    * @see #addPrintSettings
    * @see #addRemainingMoves
    * @see #addRemainingTime
    * @see #addStones
    * @see #addTerritory
    * @see #addTurn
    * @see #addVisible
    */
   @SuppressWarnings( "unchecked" )
   public void add( int prop_id, Object value )
   {
      if (value == null)
         throw new NullPointerException( Resources.formatString( "msg.4008", prop_id ) ); //$NON-NLS-1$
      try
      {
         switch (prop_id)
         {
            case PropertyID.COMMENT:
               m_entries.add( ImmutablePair.of( prop_id, (String) value ) );
               break;
            case PropertyID.CONNECTOR:
               ImmutablePair<Connectors,Set<ImmutablePair<Short,Short>>> conn_value = (ImmutablePair<Connectors,Set<ImmutablePair<Short,Short>>>) value;
               for (ImmutablePair<Short,Short> conn_pair : conn_value.getRight())
               {
                  Board.checkPoint( conn_pair.getLeft() );
                  Board.checkPoint( conn_pair.getRight() );
               }
               m_entries.add( ImmutablePair.of( prop_id, ImmutablePair.of( conn_value.getLeft(), new HashSet<ImmutablePair<Short,Short>>( conn_value.getRight() ) ) ) );
               break;
            case PropertyID.DIAGRAM_SETTINGS:
               ImmutablePair<String,DiagramSettings> diag_value = (ImmutablePair<String,DiagramSettings>) value;
               m_entries.add( ImmutablePair.of( prop_id, diag_value ) );
               break;
            case PropertyID.GRADE:
               ImmutablePair<Colors,ImmutablePair<GradeTypes,Grades>> grade_value = (ImmutablePair<Colors,ImmutablePair<GradeTypes,Grades>>) value;
               Colors gr_color = grade_value.getLeft();
               if (grade_value.getRight().getLeft() == GradeTypes.MOVE && gr_color != Colors.DONT_CARE)
                  throw new PropertyException( prop_id, gr_color.toString() );
               m_entries.add( ImmutablePair.of( prop_id, grade_value ) );
               break;
            case PropertyID.ILLEGAL_MOVE:
               ImmutablePair<Colors,Short> ill_move_value = (ImmutablePair<Colors,Short>) value;
               Game.checkColor( ill_move_value.getLeft() );
               Board.checkPoint( ill_move_value.getRight() );
               m_entries.add( ImmutablePair.of( prop_id, ill_move_value ) );
               break;
            case PropertyID.LABEL:
               Set<ImmutablePair<Short,String>> labels_value = (Set<ImmutablePair<Short,String>>) value;
               if (labels_value.isEmpty())
                  throw new PropertyException( Resources.getString( "msg.4009" ) ); //$NON-NLS-1$
               for (ImmutablePair<Short,String> label_pair : labels_value)
                  Board.checkPoint( label_pair.getLeft() );
               m_entries.add( ImmutablePair.of( prop_id, new HashSet<ImmutablePair<Short,String>>( labels_value ) ) );
               break;
            case PropertyID.MARK:
               ImmutablePair<Notations,Set<Short>> mark_value = (ImmutablePair<Notations,Set<Short>>) value;
               checkPoints( mark_value.getRight() );
               m_entries.add( ImmutablePair.of( prop_id, ImmutablePair.of( mark_value.getLeft(), new HashSet<Short>( mark_value.getRight() ) ) ) );
               break;
            case PropertyID.MOVE:
               ImmutablePair<Colors,Short> move_value = (ImmutablePair<Colors,Short>) value;
               Game.checkColor( move_value.getLeft() );
               if (move_value.getRight() != Board.INVALID_POINT)
                  Board.checkPoint( move_value.getRight() );
               m_entries.add( ImmutablePair.of( prop_id, (ImmutablePair<Colors,Short>) value ) );
               break;
            case PropertyID.NOTE:
               m_entries.add( ImmutablePair.of( prop_id, (ImmutablePair<Integer,ImmutablePair<String,String>>) value ) );
               break;
            case PropertyID.PRINT_SETTINGS:
               m_entries.add( ImmutablePair.of( prop_id, new PrintSettings( (PrintSettings) value ) ) );
               break;
            case PropertyID.REMAINING_MOVES:
               ImmutablePair<Colors,Integer> rem_move_value = (ImmutablePair<Colors,Integer>) value;
               Game.checkColor( rem_move_value.getLeft() );
               if (rem_move_value.getRight() < 0)
                  throw new PropertyException( prop_id, Integer.toString( rem_move_value.getRight() ) );
               m_entries.add( ImmutablePair.of( prop_id, rem_move_value ) );
               break;
            case PropertyID.REMAINING_TIME:
               ImmutablePair<Colors,BigDecimal> rem_time_value = (ImmutablePair<Colors,BigDecimal>) value;
               Game.checkColor( rem_time_value.getLeft() );
               if (rem_time_value.getRight().compareTo( BigDecimal.ZERO ) < 0)
                  throw new PropertyException( prop_id, rem_time_value.getRight().toString() );
               m_entries.add( ImmutablePair.of( prop_id, rem_time_value ) );
               break;
            case PropertyID.STONES:
               ImmutablePair<Colors,Set<Short>> st_value = (ImmutablePair<Colors,Set<Short>>) value;
               if (st_value.getLeft() == Colors.DONT_CARE)
                  throw new IllegalArgumentException( Resources.formatString( "msg.4010", Colors.DONT_CARE ) ); //$NON-NLS-1$
               checkPoints( st_value.getRight() );
               m_entries.add( ImmutablePair.of( prop_id, ImmutablePair.of( st_value.getLeft(), new HashSet<Short>( st_value.getRight() ) ) ) );
               break;
            case PropertyID.TERRITORY:
               ImmutablePair<Colors,Set<Short>> terr_value = (ImmutablePair<Colors,Set<Short>>) value;
               Game.checkColor( terr_value.getLeft() );
               checkPoints( terr_value.getRight() );
               m_entries.add( ImmutablePair.of( prop_id, ImmutablePair.of( terr_value.getLeft(), new HashSet<Short>( terr_value.getRight() ) ) ) );
               break;
            case PropertyID.TURN:
               Game.checkColor( (Colors) value );
               m_entries.add( ImmutablePair.of( prop_id, (Colors) value ) );
               break;
            case PropertyID.VISIBLE:
               Set<Short> vis_value = (Set<Short>) value;
               checkPoints( vis_value );
               m_entries.add( ImmutablePair.of( prop_id, new HashSet<Short>( vis_value ) ) );
               break;
            case PropertyID.BRANCH:
               m_entries.add( ImmutablePair.of( prop_id, new PositionProperties( (PositionProperties) value ) ) );
               break;
            default:
               addNonStandard( prop_id, (ImmutablePair<String,String>) value );
               break;
         }
      }
      catch (ClassCastException ex)
      {
         throw new PropertyException( prop_id, ex );
      }
      catch (IllegalArgumentException ex)
      {
         throw new PropertyException( prop_id, value.toString() );
      }
   }

   /**
    * Clears all properties.
    */
   public void clear()
   {
      m_entries.clear();
   }

   /**
    * Returns true if the specified property has been added.
    *
    * @see #add
    */
   public boolean has( int id )
   {
      for (ImmutablePair<Integer,?> entry : m_entries)
      {
         if (entry.getLeft() == id)
            return true;
      }
      return false;
   }

   /**
    * Returns true if this object represents the initial (main) branch of a game record.
    */
   public boolean isRoot()
   {
      return m_is_root;
   }

   /**
    * Calls the specified visitor for each property, descending branches
    * of the game record as needed.
    * 
    * @param visitor         the callback
    * @param mainline_only   if true, descend the first child branch that is
    *                        encountered and skip the remainder of the branch
    */
   public void visit( PropertyVisitor visitor, boolean mainline_only )
   {
      for (ImmutablePair<Integer,?> entry : m_entries)
      {
         if (entry.getLeft() == PropertyID.BRANCH)
         {
            PositionProperties child = (PositionProperties) entry.getRight();
            if (visitor.startBranch( child ))
            {
               child.visit( visitor, mainline_only );
               visitor.finishBranch( child );
               if (mainline_only)
                  return;
            }
         }
         else
         {
            visitor.accept( entry.getLeft(), entry.getRight() );
         }
      }
   }

   @Override
   public boolean equals( Object obj )
   {
      if (!(obj instanceof PositionProperties))
         return false;
      if (obj == this)
         return true;
      PositionProperties other = (PositionProperties) obj;
      if (m_is_root != other.m_is_root)
         return false;
      if (m_entries.size() != other.m_entries.size())
         return false;
      Iterator<ImmutablePair<Integer,?>> l_iter = m_entries.iterator();
      Iterator<ImmutablePair<Integer,?>> r_iter = other.m_entries.iterator();
      while (l_iter.hasNext() && r_iter.hasNext())
      {
         ImmutablePair<Integer,?> lhs = (ImmutablePair<Integer,?>) l_iter.next();
         ImmutablePair<Integer,?> rhs = (ImmutablePair<Integer,?>) r_iter.next();
         if (lhs.getLeft() != rhs.getLeft() || !lhs.getRight().equals( rhs.getRight() ))
            return false;
      }
      return true;
   }

   @Override
   public int hashCode()
   {
      return HashUtils.compute( m_is_root, HashUtils.compute( m_entries ) );
   }

   private void addNonStandard( int prop_id, ImmutablePair<String,String> value )
   {
      if (prop_id <= PropertyID.MAX_STD_ID)
         throw new PropertyException( prop_id );
      m_entries.add( ImmutablePair.of( prop_id, value ) );
   }

   private void checkPoints( Collection<Short> points )
   {
      if (points.isEmpty())
         throw new PropertyException( Resources.getString( "msg.4011" ) ); //$NON-NLS-1$
      for (short pnt : points)
         Board.checkPoint( pnt );
   }

   private List<ImmutablePair<Integer,?>> m_entries = new ArrayList<ImmutablePair<Integer,?>>();
   private boolean                        m_is_root = false;
}
