//
// PropertyID.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.record;

/**
 * Defines the properties that may be used in a game record (Record class). This is a class
 * full of constants instead of an enum because game records are allowed to use non-standard
 * properties. This class simply establishes constants for the pre-defined property IDs.<p>
 *
 * When adding a non-standard property (id > MAX_STD_ID), use a {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link String},{@link String}> to store
 * the property value, where {@code getLeft} is the key from the source record, and {@code getRight} is that key's value.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see GameProperties
 * @see PositionProperties
 */
public abstract class PropertyID
{
   // Game info properties
   //
   /** The name of the person who made annotations to the game record. Value type: {@link String} */
   public static final int ANNOTATOR_NAME    = 0;
   /** The application that constructed the game record. Value type: {@link String} */
   public static final int APP_NAME          = 1;
   /** The application version. Value type: {@link String} */
   public static final int APP_VERSION       = 2;
   /** Extra information or a summary of the game. Value type: {@link String} */
   public static final int BACKGROUND_INFO   = 3;
   /** The character set in which the text of the game record is encoded, by default {@code "utf-8"}. Value type: {@link String} */
   public static final int CHARSET           = 4;
   /** The copyright text for the game record. Value type: {@link String} */
   public static final int COPYRIGHT         = 5;
   /** The date on which the game was played. Value type: {@link String} */
   public static final int DATE              = 6;
   /** The name of the event in which the game took place. Value type: {@link String} */
   public static final int EVENT_NAME        = 7;
   /** The name given to this particular game. Value type: {@link String} */
   public static final int GAME_NAME         = 8;
   /** Placement and number of handicap stones. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link Handicaps},{@link Integer}> */
   public static final int HANDICAP          = 9;
   /** The komi value, by default 6.5. Value type: {@link java.math.BigDecimal} */
   public static final int KOMI              = 10;
   /** The name of the opening that was played. Value type: {@link String} */
   public static final int OPENING_NAME      = 11;
   /** The overtime system in force for the game. Value type: {@link String} */
   public static final int OVERTIME_SYSTEM   = 12;
   /** The location where the game took place. Value type: {@link String} */
   public static final int PLACE_NAME        = 13;
   /** The name of a player. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link String}> */
   public static final int PLAYER_NAME       = 14;
   /** The rank of a player. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link String}> */
   public static final int PLAYER_RANK       = 15;
   /** Rendering hints. Value type: {@link RenderingStyle} */
   public static final int RENDERING_STYLE   = 16;
   /** Game result. Value type: {@link com.dakengo.libs.game.Result} */
   public static final int RESULT            = 17;
   /** Game round. Value type: {@link Round} */
   public static final int ROUND             = 18;
   /** The rule system in force for the game. Value type: {@link String} */
   public static final int RULE_SYSTEM       = 19;
   /** The board size. Value type: {@link Integer} */
   public static final int SIZE              = 20;
   /** The source from which the game record was obtained. Value type: {@link String} */
   public static final int SOURCE_NAME       = 21;
   /** The name of a team. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link String}> */
   public static final int TEAM_NAME         = 22;
   /** The time limit for the game, in seconds. Value type: {@link Integer} */
   public static final int TIME_LIMIT        = 23;
   /** The user of the application. Value type: {@link String} */
   public static final int USER_NAME         = 24;

   // Position properties
   //
   /** Comment text. Value type: {@link String} */
   public static final int COMMENT           = 25;
   /** Connect the first point to the second of each pair. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link Connectors},{@link java.util.Set}<{@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link Short},{@link Short}>>> */
   public static final int CONNECTOR         = 26;
   /** Configure diagram rendering. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link String},{@link DiagramSettings}> */
   public static final int DIAGRAM_SETTINGS  = 27;
   /** How good or bad something is. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link GradeTypes},{@link Grades}>> */
   public static final int GRADE             = 28;
   /** A move to be executed (e.g., in a record viewer) even though it's illegal. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link Short}> */
   public static final int ILLEGAL_MOVE      = 29;
   /** Points to be annotated with text. Value type: {@link java.util.Set}<{@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link Short},{@link String}>> */
   public static final int LABEL             = 30;
   /** Mark the indicated points. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link Notations},{@link java.util.Set}<{@link Short}>> */
   public static final int MARK              = 31;
   /** A point representing a player move. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link Short}> */
   public static final int MOVE              = 32;
   /** Allows arbitrary text values to be added. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link Integer},{@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link String},{@link String}>> */
   public static final int NOTE              = 33;
   /** Configure print settings. Value type: {@link PrintSettings} */
   public static final int PRINT_SETTINGS    = 34;
   /** Number of moves left to play in a byo-yomi period. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link Integer}> */
   public static final int REMAINING_MOVES   = 35;
   /** Time left to play in seconds. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link java.math.BigDecimal}> */
   public static final int REMAINING_TIME    = 36;
   /** Stones to be set on (or cleared from) the board for a player. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link java.util.Set}<{@link Short}>> */
   public static final int STONES            = 37;
   /** Points representing territory for a player. Value type: {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link com.dakengo.libs.game.Colors},{@link java.util.Set}<{@link Short}>> */
   public static final int TERRITORY         = 38;
   /** Color identifying the player turn. Value type: {@link com.dakengo.libs.game.Colors} */
   public static final int TURN              = 39;
   /** Limit the board points that are visible. Value type: {@link java.util.Set}<{@link Short}> */
   public static final int VISIBLE           = 40;
   /** A sub-branch split off from the current branch. Value type: {@link PositionProperties} */
   public static final int BRANCH            = 41;

   /** The maximum standard property ID. The value type of properties with non-standard property IDs greater than this value is {@link org.apache.commons.lang3.tuple.ImmutablePair}<{@link String},{@link String}> */
   public static final int MAX_STD_ID        = BRANCH;

   /**
    * Converts a property name (e.g., "COMMENT") to a property ID (e.g., {@link #COMMENT}).
    * Returns {@link Integer#MAX_VALUE} for a string that does not match any property name.
    * 
    * @param prop_name
    */
   public static int fromString( String prop_name )
   {
      for (int i = 0; i < NAMES.length; i++)
      {
         if (prop_name.equals( NAMES[i] ))
            return i;
      }
      return Integer.MAX_VALUE;
   }

   /**
    * Returns a property name (e.g., "COMMENT") corresponding to a property ID (e.g., {@link #COMMENT}).
    * Negative ID values cause an {@link IllegalArgumentException}. Returns "NON_STANDARD" for ID values greater than
    * {@link #MAX_STD_ID}.
    * 
    * @param prop_id
    */
   public static String toString( int prop_id )
   {
      if (prop_id < 0)
         throw new IllegalArgumentException( Resources.formatString( "msg.4006", prop_id ) ); //$NON-NLS-1$
      if (prop_id > MAX_STD_ID)
         return "NON_STANDARD"; //$NON-NLS-1$
      return NAMES[prop_id];
   }

   private static final String[] NAMES = new String[]
   {
      "ANNOTATOR_NAME",   //$NON-NLS-1$
      "APP_NAME",         //$NON-NLS-1$
      "APP_VERSION",      //$NON-NLS-1$
      "BACKGROUND_INFO",  //$NON-NLS-1$
      "CHARSET",          //$NON-NLS-1$
      "COPYRIGHT",        //$NON-NLS-1$
      "DATE",             //$NON-NLS-1$
      "EVENT_NAME",       //$NON-NLS-1$
      "GAME_NAME",        //$NON-NLS-1$
      "HANDICAP",         //$NON-NLS-1$
      "KOMI",             //$NON-NLS-1$
      "OPENING_NAME",     //$NON-NLS-1$
      "OVERTIME_SYSTEM",  //$NON-NLS-1$
      "PLACE_NAME",       //$NON-NLS-1$
      "PLAYER_NAME",      //$NON-NLS-1$
      "PLAYER_RANK",      //$NON-NLS-1$
      "RENDERING_STYLE",  //$NON-NLS-1$
      "RESULT",           //$NON-NLS-1$
      "ROUND",            //$NON-NLS-1$
      "RULE_SYSTEM",      //$NON-NLS-1$
      "SIZE",             //$NON-NLS-1$
      "SOURCE_NAME",      //$NON-NLS-1$
      "TEAM_NAME",        //$NON-NLS-1$
      "TIME_LIMIT",       //$NON-NLS-1$
      "USER_NAME",        //$NON-NLS-1$

      "COMMENT",          //$NON-NLS-1$
      "CONNECTOR",        //$NON-NLS-1$
      "DIAGRAM_SETTINGS", //$NON-NLS-1$
      "GRADE",            //$NON-NLS-1$
      "ILLEGAL_MOVE",     //$NON-NLS-1$
      "LABEL",            //$NON-NLS-1$
      "MARK",             //$NON-NLS-1$
      "MOVE",             //$NON-NLS-1$
      "NOTE",             //$NON-NLS-1$
      "PRINT_SETTINGS",   //$NON-NLS-1$
      "REMAINING_MOVES",  //$NON-NLS-1$
      "REMAINING_TIME",   //$NON-NLS-1$
      "STONES",           //$NON-NLS-1$
      "TERRITORY",        //$NON-NLS-1$
      "TURN",             //$NON-NLS-1$
      "VISIBLE",          //$NON-NLS-1$
      "BRANCH"            //$NON-NLS-1$
   };

   private PropertyID()
   {
   }
}
