//
// Record.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.libs.record;

/**
 * This class records information about a game, which can then be
 * serialized to an external format such as SGF.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see GameProperties
 * @see PropertyID
 * @see PositionProperties
 * @see RecordReader
 * @see RecordWriter
 */
public class Record
{
   /**
    * Initializes a new game record.
    */
   public Record()
   {
      m_game_props = new GameProperties();
      m_main_branch = new PositionProperties( true );
   }

   /**
    * Initializes a new game record with an existing properties
    * object representing the main line of play.
    */
   public Record( PositionProperties branch )
   {
      if (branch == null)
         throw new NullPointerException( Resources.getString( "msg.4016" ) ); //$NON-NLS-1$
      if (!branch.isRoot())
         throw new IllegalArgumentException( Resources.getString( "msg.4017" ) ); //$NON-NLS-1$
      m_game_props = new GameProperties();
      m_main_branch = branch;
   }

   /**
    * Copy constructor.
    * 
    * @param orig
    */
   public Record( Record orig )
   {
      m_game_props = new GameProperties( orig.getGameProps() );
      m_main_branch = new PositionProperties( orig.getMainBranch() );
   }

   /**
    * Resets all properties to their defaults. Clears the main line of play
    * and removes any variation branches previously set.
    */
   public void clear()
   {
      m_game_props = new GameProperties();
      m_main_branch = new PositionProperties( true );
   }

   /**
    * Returns properties that describe the game.
    */
   public GameProperties getGameProps()
   {
      return m_game_props;
   }

   /**
    * Returns the properties representing the main line of
    * play. Variations may be added as child branches of the
    * main branch.
    */
   public PositionProperties getMainBranch()
   {
      return m_main_branch;
   }

   /**
    * Calls the visitor for each property stored in this game record,
    * with the option of traversing or skipping variations.
    *
    * @param visitor        the callback
    * @param mainline_only  if true, descend the first child branch that is
    *                       encountered and do not return to the remainder
    *                       of the root branch
    */
   public void visit( PropertyVisitor visitor, boolean mainline_only )
   {
      m_game_props.visit( visitor );
      if (!visitor.startBranch( m_main_branch ))
         return;
      m_main_branch.visit( visitor, mainline_only );
      visitor.finishBranch( m_main_branch );
   }

   private GameProperties     m_game_props;
   private PositionProperties m_main_branch;
}
