//
// FixedHandicapHandler.java
//
// Copyright (c) 2012-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.gtp.core;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import net.xeoh.plugins.base.annotations.PluginImplementation;

import com.dakengo.libs.application.BaseGtpCommandHandler;
import com.dakengo.libs.application.GtpCommand;
import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.application.GtpReplyHandler;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.record.PositionProperties;

/** 
 * Implements the {@code fixed_handicap} GTP command.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
@PluginImplementation
public class FixedHandicapHandler extends BaseGtpCommandHandler
{
   public FixedHandicapHandler()
   {
      super( "fixed_handicap", false ); //$NON-NLS-1$
   }

   @Override
   protected void doCommand( GtpCommand cmd, GameContext context, GtpReplyHandler callback )
   {
      List<String> args = cmd.getArguments();
      if (args.size() != 1)
      {
         reportInvalidNumArgs( cmd, callback );
         return;
      }
      int count = 0;
      try
      {
         count = Integer.parseInt( args.get( 0 ) );
      }
      catch (NumberFormatException ex)
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2044" ) ); //$NON-NLS-1$
         logError( Resources.getString( "msg.3004" ) ); //$NON-NLS-1$
         return;
      }
      if (!Game.isValidHandicap( count ))
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2041" ) ); //$NON-NLS-1$
         logError( Resources.getString( "msg.3004" ) ); //$NON-NLS-1$
         return;
      }
      setPlayGame( cmd, context, callback );
      Game game = context.getGame();
      if (game.getBoard().getCount( Colors.EMPTY ) != Board.getTotal())
      {
         String msg = Resources.getString( "msg.2033" ); //$NON-NLS-1$
         callback.handleReply( true, cmd.getID(), msg );
         logError( msg );
         return;
      }
      Map<String,Set<Short>> layers = context.getLayers();
      String key = "handicap-" + count; //$NON-NLS-1$
      if (!layers.containsKey( key ))
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2037" ) ); //$NON-NLS-1$
         logError( Resources.getString( "msg.3004" ) ); //$NON-NLS-1$
         return;
      }
      Set<Short> points = layers.get( key );
      game.placeStones( Colors.BLACK, points );
      game.setPlayer( Colors.WHITE );
      PositionProperties branch = context.getRecord().getMainBranch();
      branch.addStones( Colors.BLACK, points );
      callback.handleReply( false, cmd.getID(), Convert.pointsToString( new TreeSet<Short>( points ), " " ) ); //$NON-NLS-1$
      logInfo( Resources.formatString( "msg.3010", count ) ); //$NON-NLS-1$
   }
}
