//
// DkPlayoutHandler.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.gtp.regression;

import java.util.List;

import net.xeoh.plugins.base.annotations.PluginImplementation;
import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.application.BaseGtpCommandHandler;
import com.dakengo.libs.application.GtpCommand;
import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.application.GtpReplyHandler;
import com.dakengo.libs.base.MersenneTwisterFast;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Convert;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.player.DefaultPlayout;
import com.dakengo.libs.player.Playout;

/**
 * Implements the Dakengo-specific command {@code dk-playout}. This command takes three
 * arguments: the color to play, the point to play in standard notation, and the number
 * of playouts to be run. A temporary copy of the current game board is made, the point
 * is played on the copy, and then the specified number of playouts are run.<p>
 * 
 * The command returns two values: the playout score, where positive means more wins
 * than losses for the specified color, and the time in millis required to run the
 * playouts.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
@PluginImplementation
public class DkPlayoutHandler extends BaseGtpCommandHandler
{
   public DkPlayoutHandler()
   {
      super( "dk-playout", false ); //$NON-NLS-1$
   }

   @Override
   protected void doCommand( GtpCommand cmd, GameContext context, GtpReplyHandler callback )
   {
      List<String> args = cmd.getArguments();
      if (args.size() != 3)
      {
         reportInvalidNumArgs( cmd, callback );
         return;
      }
      Colors color = Colors.DONT_CARE;
      short pnt = Board.INVALID_POINT;
      int count = 0;
      try
      {
         color = Convert.stringToColor( args.get( 0 ) );
         if (!Game.isPlayerColor( color ))
         {
            callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2044" ) ); //$NON-NLS-1$
            logError( Resources.getString( "msg.11001" ) ); //$NON-NLS-1$
            return;
         }
         pnt = Convert.stringToPoint( args.get( 1 ) );
         count = Integer.parseInt( args.get( 2 ) );
         if (count <= 0)
         {
            String msg = Resources.getString( "msg.11003" ); //$NON-NLS-1$:
            callback.handleReply( true, cmd.getID(), msg );
            logError( msg );
            return;
         }
      }
      catch (ParseException ex)
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2044" ) ); //$NON-NLS-1$
         logError( ex.getMessage() );
         return;
      }
      catch (NumberFormatException ex)
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2044" ) ); //$NON-NLS-1$
         logError( Resources.getString( "msg.11002" ) ); //$NON-NLS-1$
         return;
      }
      Playout playout = new DefaultPlayout( new MersenneTwisterFast(), context.getScorer() );  // TODO: use a factory method to obtain playout
      ImmutablePair<Integer,Long> result = playout.run( context.getGame().getBoard(), pnt, color, count );
      callback.handleReply( false, cmd.getID(), String.format( "%d %d", result.getLeft(), result.getRight() ) ); //$NON-NLS-1$
   }
}
