//
// LoadSGFHandler.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.gtp.regression;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.List;

import net.xeoh.plugins.base.annotations.PluginImplementation;

import com.dakengo.libs.application.BaseGtpCommandHandler;
import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.application.GtpCommand;
import com.dakengo.libs.application.GtpReplyHandler;
import com.dakengo.libs.application.PluginUtils;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.record.BoardCollector;
import com.dakengo.libs.record.GameProperties;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.libs.record.Record;
import com.dakengo.libs.record.RecordReader;

/** 
 * Implements the {@code loadsgf} GTP command.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
@PluginImplementation
public class LoadSGFHandler extends BaseGtpCommandHandler
{
   public LoadSGFHandler()
   {
      super( "loadsgf", false ); //$NON-NLS-1$
   }

   @Override
   protected void doCommand( GtpCommand cmd, GameContext context, GtpReplyHandler callback )
   {
      List<String> args = cmd.getArguments();
      if (args.size() > 2)
      {
         reportInvalidNumArgs( cmd, callback );
         return;
      }
      int turn_num = Integer.MAX_VALUE;
      if (args.size() > 1)
      {
         try
         {
            turn_num = Integer.parseInt( args.get( 1 ) );
         }
         catch (NumberFormatException ex)
         {
            callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2044" ) ); //$NON-NLS-1$
            logError( Resources.getString( "msg.11005" ) ); //$NON-NLS-1$
            return;
         }
      }
      if (turn_num <= 0)
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2044" ) ); //$NON-NLS-1$
         logError( Resources.getString( "msg.11005" ) ); //$NON-NLS-1$
         return;
      }
      String file_path = args.get( 0 );
      String file_ext = "";
      int pos = file_path.lastIndexOf( '.' );
      if (pos > 0 && pos < file_path.length() - 1)
         file_ext = file_path.substring( pos + 1 );
      if (!file_ext.equalsIgnoreCase( FILE_EXT ))
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.11006" ) ); //$NON-NLS-1$
         logError( Resources.formatString( "msg.11007", file_ext ) ); //$NON-NLS-1$
         return;
      }
      RecordReader reader = context.getPlugins().getPlugin( RecordReader.class, PluginUtils.getRecordCapabilities( file_ext, true ) );
      if (reader == null)
      {
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.2037" ) ); //$NON-NLS-1$
         logError( Resources.formatString( "msg.11004", FILE_EXT ) ); //$NON-NLS-1$
         return;
      }
      reader.setInputPath( file_path );
      Record rec = null;
      GameProperties game_props = null;
      int orig_size = Board.getSize();
      try (BufferedReader br = new BufferedReader( new FileReader( file_path ) ))
      {
         reader.setInput( br );
         rec = reader.read();
         game_props = rec.getGameProps();
         context.loadLayers( game_props.getSize() );
      }
      catch (ParseException ex)
      {
         Board.setSize( orig_size );
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.11006" ) ); //$NON-NLS-1$
         logThrowable( ex );
         return;
      }
      catch (IOException ex)
      {
         Board.setSize( orig_size );
         callback.handleReply( true, cmd.getID(), Resources.getString( "msg.11006" ) ); //$NON-NLS-1$
         logThrowable( ex );
         return;
      }
      BigDecimal komi = Game.DEFAULT_KOMI;
      if (game_props.has( PropertyID.KOMI ))
         komi = game_props.getKomi();
      BoardCollector collector = new BoardCollector();
      collector.setKomi( komi );
      collector.collect( rec, turn_num - 1 );
      context.setGame( collector.getGame(), file_path );
      callback.handleReply( false, cmd.getID(), "" ); //$NON-NLS-1$
   }

   private static final String FILE_EXT = "sgf"; //$NON-NLS-1$
}
