//
// MonteCarloPlayer.java
//
// Copyright (c) 2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.player.sample;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.gson.JsonElement;
import net.xeoh.plugins.base.annotations.Capabilities;
import net.xeoh.plugins.base.annotations.PluginImplementation;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

import com.dakengo.libs.application.Constants;
import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.base.MersenneTwisterFast;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.MoveTypes;
import com.dakengo.libs.game.Scorer;
import com.dakengo.libs.player.BaseLegalMoveFilter;
import com.dakengo.libs.player.DefaultPlayout;
import com.dakengo.libs.player.Player;
import com.dakengo.libs.player.Playout;

/**
 * A sample computer player that runs random playouts for each available point
 * in order to choose its next move. The following configuration properties
 * are supported:<p>
 * 
 * <ul>
 * <li>{@code playouts} - playout count for each candidate point
 * </ul>
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
@PluginImplementation
public class MonteCarloPlayer implements Player
{
   @Override
   public void init( GameContext context, Map<String,JsonElement> data )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );

      m_scorer = context.getScorer();
      if (m_prng_tls.get() == null)
         m_prng_tls.set( new MersenneTwisterFast() );
      int value = 0;
      if (data.containsKey( PLAYOUT_COUNT_KEY ))
      {
         value = data.get( PLAYOUT_COUNT_KEY ).getAsInt();
         if (value <= 0)
            log.warn( Resources.getString( "msg.8006" ) ); //$NON-NLS-1$
         else
            m_playout_count = value;
      }
   }

   @Override
   public short getMove( Game game )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );

      // If could not find a legal move, then pass to initiate the end
      // game protocol.
      //
      Colors player = game.getPlayer();
      if (game.getLastMoveType() == MoveTypes.PASS && game.getTurnNumber() > 1)
      {
         log.debug( Resources.formatString( "msg.8001", player.toString().toLowerCase() ) ); //$NON-NLS-1$
         return Board.INVALID_POINT;
      }

      BaseLegalMoveFilter filter = new RandomMoveFilter( m_prng_tls.get() );
      Board board = game.getBoard();
      List<ImmutablePair<Short,Board>> candidates = filter.find( player, board, Board.getTotal() );
      if (candidates.isEmpty())
      {
         log.debug( Resources.formatString( "msg.8003", player.toString().toLowerCase() ) ); //$NON-NLS-1$
         return Board.INVALID_POINT;
      }

      int best_score = Integer.MIN_VALUE;
      short best_pnt = Board.INVALID_POINT;
      Playout playout = new DefaultPlayout( m_prng_tls.get(), m_scorer );  // TODO: use a factory method to obtain playout
      for (ImmutablePair<Short,Board> candidate : candidates)
      {
         short pnt = candidate.getLeft();
         if (game.isKo( candidate.getRight() ))
            continue;
         ImmutablePair<Integer,Long> result = playout.run( board, pnt, player, m_playout_count );
         if (result.getLeft() > best_score)
         {
            best_score = result.getLeft();
            best_pnt = pnt;
         }
      }
      return best_pnt;
   }

   @Override
   public Set<Short> getHandicap( int count )
   {
      // FIXME: return a random selection for the handicap
      return new HashSet<Short>();
   }

   @Override
   public void reset( String game_id )
   {
   }

   @Override
   public String getName( String engine_name, String engine_version )
   {
      return Resources.getString( "msg.8004" ); //$NON-NLS-1$
   }

   @Capabilities
   public String[] getCapabilities()
   {
      return CAPABILITIES;
   }

   private static final int    DEFAULT_PLAYOUT_COUNT   = 500;

   private static final String PLAYOUT_COUNT_KEY       = "playouts";

   private static ThreadLocal<MersenneTwisterFast> m_prng_tls = new ThreadLocal<MersenneTwisterFast>();

   private static String[] CAPABILITIES;

   static
   {
      CAPABILITIES = new String[1];
      CAPABILITIES[0] = Constants.PLAYER_CAP_PREFIX + "montecarlo"; //$NON-NLS-1$
   }

   private int    m_playout_count = DEFAULT_PLAYOUT_COUNT;
   private Scorer m_scorer;
}
