//
// RandomPlayer.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.player.sample;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.gson.JsonElement;
import net.xeoh.plugins.base.annotations.Capabilities;
import net.xeoh.plugins.base.annotations.PluginImplementation;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

import com.dakengo.libs.application.Constants;
import com.dakengo.libs.application.GameContext;
import com.dakengo.libs.base.MersenneTwisterFast;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.MoveTypes;
import com.dakengo.libs.player.BaseLegalMoveFilter;
import com.dakengo.libs.player.Player;

/**
 * A sample computer player that selects random legal moves.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
@PluginImplementation
public class RandomPlayer implements Player
{
   @Override
   public void init( GameContext context, Map<String, JsonElement> data )
   {
      if (m_prng_tls.get() == null)
         m_prng_tls.set( new MersenneTwisterFast() );
   }

   @Override
   public short getMove( Game game )
   {
      Logger log = LogManager.getLogger( Constants.LOGGER_NAME );

      // If no more empty points, or the other player passed,
      // then pass so as to end the game.
      //
      // If could not find a legal move, then pass to initiate the end
      // game protocol.
      //
      Colors player = game.getPlayer();
      if (game.getLastMoveType() == MoveTypes.PASS && game.getTurnNumber() > 1)
      {
         log.debug( Resources.formatString( "msg.8001", player.toString().toLowerCase() ) ); //$NON-NLS-1$
         return Board.INVALID_POINT;
      }
      Board board = game.getBoard();
      BaseLegalMoveFilter filter = new RandomMoveFilter( m_prng_tls.get() );
      List<ImmutablePair<Short,Board>> candidates = filter.find( player, board, 4 );
      if (candidates.isEmpty())
      {
         log.debug( Resources.formatString( "msg.8003", player.toString().toLowerCase() ) ); //$NON-NLS-1$
         return Board.INVALID_POINT;
      }
      for (ImmutablePair<Short,Board> candidate : candidates)
      {
         if (!game.isKo( candidate.getRight() ))
            return candidate.getLeft();
      }
      return Board.INVALID_POINT;
   }

   @Override
   public Set<Short> getHandicap( int count )
   {
      // FIXME: return a random selection for the handicap
      return new HashSet<Short>();
   }

   @Override
   public void reset( String game_id )
   {
   }

   @Override
   public String getName( String engine_name, String engine_version )
   {
      return Resources.getString( "msg.8002" ); //$NON-NLS-1$
   }

   @Capabilities
   public String[] getCapabilities()
   {
      return CAPABILITIES;
   }

   private static ThreadLocal<MersenneTwisterFast> m_prng_tls = new ThreadLocal<MersenneTwisterFast>();

   private static String[] CAPABILITIES;

   static
   {
      CAPABILITIES = new String[1];
      CAPABILITIES[0] = Constants.PLAYER_CAP_PREFIX + "random"; //$NON-NLS-1$
   }
}
