//
// Convert.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.record.sgf;

import java.io.IOException;
import java.io.StringReader;
import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.base.Lexer;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Layers;
import com.dakengo.libs.game.Result;
import com.dakengo.libs.game.Signature;
import com.dakengo.libs.record.DiagramSettings;
import com.dakengo.libs.record.Emphasis;
import com.dakengo.libs.record.PrintSettings;
import com.dakengo.libs.record.RenderingStyle;
import com.dakengo.libs.record.Round;

/**
 * Helper methods to convert primitive data types into and out of
 * SGF string representations.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 */
public abstract class Convert
{
   public interface TokenParser<T>
   {
      public void parse( String token, int linenum, Set<T> dest );
   };

   public static TokenParser<Short> m_pnt_parser = new TokenParser<Short>()
   {
      @Override
      public void parse( String token, int linenum, Set<Short> dest )
      {
         try
         {
            ImmutablePair<Short,Short> pair = stringToPointPair( token, linenum );
            Set<Short> tmp = Layers.create( pair );
            dest.addAll( tmp );
         }
         catch (ParseException ex)
         {
            dest.add( stringToPoint( token, linenum ) );
         }
      }
   };

   public static TokenParser<ImmutablePair<Short,Short>> m_pntpair_parser = new TokenParser<ImmutablePair<Short,Short>>()
   {
      @Override
      public void parse( String token, int linenum, Set<ImmutablePair<Short,Short>> dest )
      {
         dest.add( stringToPointPair( token, linenum ) );
      }
   };

   public static TokenParser<ImmutablePair<Short,String>> m_annpnt_parser = new TokenParser<ImmutablePair<Short,String>>()
   {
      @Override
      public void parse( String token, int linenum, Set<ImmutablePair<Short,String>> dest )
      {
         dest.add( stringToAnnotatedPoint( token, linenum ) );
      }
   };

   public static final int LIST_DIV_TOKEN_ID    = 1;
   public static final int LEFT_PAREN_TOKEN_ID  = 2;
   public static final int RIGHT_PAREN_TOKEN_ID = 3;

   public static final ImmutablePair<Integer,String> LEFT_PAREN  = new ImmutablePair<Integer,String>( LEFT_PAREN_TOKEN_ID, "(" );  //$NON-NLS-1$
   public static final ImmutablePair<Integer,String> RIGHT_PAREN = new ImmutablePair<Integer,String>( RIGHT_PAREN_TOKEN_ID, ")" ); //$NON-NLS-1$

   public static final int[] PAREN_TOKEN_IDS = new int[] { LEFT_PAREN_TOKEN_ID, RIGHT_PAREN_TOKEN_ID };

   public static final String[] DATE_PATTERNS = new String[]
   {
      "yyyy/MM/dd", "yyyy-MM-dd", "MM/dd/yyyy", "MM-dd-yyyy", "MMMMMMMMM-dd-yyyy", "MMMMMMMMMM dd yyyy", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
      "MMMMMMMMM dd, yyyy", "dd MMMMMMMMM yyyy", "dd-MMMMMMMMM-yy"                                       //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
   };

   public static String annotatedPointToString( ImmutablePair<Short,String> value )
   {
      String pnt_str = pointToString( value.getLeft() );
      String text = transformText( value.getRight(), true );
      return String.format( "%s:%s", pnt_str, text ); //$NON-NLS-1$
   }

   public static ImmutablePair<Short,String> stringToAnnotatedPoint( String text, int linenum )
   {
      int pos = text.indexOf( ':' );
      if (pos != 2 || text.length() <= 3)
         throw new ParseException( linenum, Resources.getString( "msg.5000" ), text ); //$NON-NLS-1$
      short pnt = stringToPoint( text.substring( 0, pos ), linenum );
      return ImmutablePair.of( pnt, text.substring( pos + 1 ) );
   }

   /**
    * @param points
    */
   public static String annotatedPointsToString( Set<ImmutablePair<Short,String>> points )
   {
      StringBuilder buffer = new StringBuilder();
      for (ImmutablePair<Short,String> pnt : points)
      {
         if (buffer.length() > 0)
            buffer.append( "][" ); //$NON-NLS-1$
         buffer.append( annotatedPointToString( pnt ) );
      }
      return buffer.toString();
   }

   public static Set<ImmutablePair<Short,String>> stringToAnnotatedPoints( String text, int linenum )
   {
      Set<ImmutablePair<Short,String>> points = new HashSet<ImmutablePair<Short,String>>();
      if (text.length() == 0)
         return points;
      parseList( text, linenum, m_annpnt_parser, points );
      return points;
   }

   public static Colors charToColor( char ch, int linenum )
   {
      switch (ch)
      {
         case 'B': case 'b': return Colors.BLACK;
         case 'E': case 'e': return Colors.EMPTY;
         case 'W': case 'w': return Colors.WHITE;
         default:            throw new ParseException( linenum, Resources.getString( "msg.5001" ), Character.toString( ch ) ); //$NON-NLS-1$
      }
   }

   public static char colorToChar( Colors color )
   {
      if (color.equals( Colors.DONT_CARE ))
         return '?';
      return color.toString().charAt( 0 );
   }

   public static String dateToString( Date date )
   {
      SimpleDateFormat fmt = new SimpleDateFormat( "yyyy-MM-dd" ); //$NON-NLS-1$
      return fmt.format( date );
   }

   public static Date stringToDate( String text, int linenum )
   {
      Calendar cal = Calendar.getInstance();
      SimpleDateFormat fmt = new SimpleDateFormat();
      fmt.setLenient( true );
      for (String pattern : DATE_PATTERNS)
      {
         try
         {
            fmt.applyPattern( pattern );
            Date date = fmt.parse( text );
            cal.setTime( date );
            if (cal.get( Calendar.YEAR ) >= 1583)
               return date;
         }
         catch (java.text.ParseException ex)
         {
         }
      }
      throw new ParseException( linenum, Resources.getString( "msg.5002" ), text ); //$NON-NLS-1$ 
   }

   public static String decimalToString( BigDecimal value )
   {
      return value.toPlainString();
   }

   public static BigDecimal stringToDecimal( String text, int linenum )
   {
      try
      {
         return new BigDecimal( text );
      }
      catch (NumberFormatException ex)
      {
         throw new ParseException( linenum, Resources.getString( "msg.5003" ), text ); //$NON-NLS-1$
      }
   }

   public static String diagramSettingsToString( ImmutablePair<String,DiagramSettings> value )
   {
      DiagramSettings settings = value.getRight();
      int num = 0;
      if (!settings.coordsEnabled())
         num += DIAG_COORDS;
      if (!settings.nameEnabled())
         num += DIAG_NAME;
      if (!settings.movesNotShownEnabled())
         num += DIAG_MOVES_NOT_SHOWN;
      if (!settings.removeCapturedEnabled())
         num += DIAG_REMOVE_CAPTURED;
      if (!settings.hoshiDotsEnabled())
         num += DIAG_HOSHI_DOTS;
      if (!settings.appDefaultsEnabled())
         num += DIAG_APP_DEFAULTS;
      return String.format( "%d:%s", num, value.getLeft() ); //$NON-NLS-1$
   }

   public static ImmutablePair<String,DiagramSettings> stringToDiagramSettings( String text, int linenum )
   {
      int pos = text.indexOf( ':' );
      if (pos < 0)
         throw new ParseException( linenum, Resources.getString( "msg.5004" ), text ); //$NON-NLS-1$
      int num = 0;
      try
      {
         num = Integer.parseInt( text.substring( 0, pos ) );
      }
      catch (NumberFormatException ex)
      {
         throw new ParseException( linenum, Resources.getString( "msg.5005" ), text ); //$NON-NLS-1$
      }
      DiagramSettings settings = new DiagramSettings();
      settings.enableAppDefaults( (num & DIAG_APP_DEFAULTS) == 0 );
      settings.enableHoshiDots( (num & DIAG_HOSHI_DOTS) == 0 );
      settings.enableRemoveCaptured( (num & DIAG_REMOVE_CAPTURED) == 0 );
      settings.enableMovesNotShown( (num & DIAG_MOVES_NOT_SHOWN) == 0 );
      settings.enableName( (num & DIAG_NAME) == 0 );
      settings.enableCoords( (num & DIAG_COORDS) == 0 );
      return ImmutablePair.of( text.substring( pos + 1 ), settings );
   }

   public static char emphasisToChar( Emphasis value )
   {
      if (value == Emphasis.STRONG)
         return '1';
      if (value == Emphasis.VERY_STRONG)
         return '2';
      throw new IllegalArgumentException( Resources.formatString( "msg.5006", value ) ); //$NON-NLS-1$
   }

   public static Emphasis charToEmphasis( char ch, int linenum )
   {
      if (ch == '1')
         return Emphasis.STRONG;
      if (ch == '2')
         return Emphasis.VERY_STRONG;
      throw new ParseException( linenum, Resources.getString( "msg.5007" ), Character.toString( ch ) ); //$NON-NLS-1$
   }

   public static String numberToString( long value )
   {
      return Long.toString( value );
   }

   public static long stringToNumber( String text, int linenum )
   {
      try
      {
         return Long.parseLong( text );
      }
      catch (NumberFormatException ex)
      {
         throw new ParseException( linenum, Resources.getString( "msg.5005" ), text ); //$NON-NLS-1$
      }
   }

   public static String pointToString( short point )
   {
      if (point < 0)
      {
         if (point == Board.INVALID_POINT)
            return ""; //$NON-NLS-1$
         throw new IllegalArgumentException( Resources.formatString( "msg.5008", point ) ); //$NON-NLS-1$
      }
      if (point >= Board.getTotal())
         throw new IllegalArgumentException( Resources.formatString( "msg.5009", point ) ); //$NON-NLS-1$
      int size = Board.getSize();
      int col = point / size;
      int row = point % size;
      String text = ""; //$NON-NLS-1$
      text += Character.toString( (char) ('a' + col) );
      text += Character.toString( (char) ('a' + size - row - 1) );
      return text;
   }

   public static short stringToPoint( String text, int linenum )
   {
      if (text.length() != 2)
         throwInvalidPointString( linenum, text );
      int size = Board.getSize();
      String copy = text.toLowerCase();
      char letter1 = copy.charAt( 0 );
      if (!Character.isLetter( letter1 ))
         throwInvalidPointString( linenum, text );
      char letter2 = copy.charAt( 1 );
      if (!Character.isLetter( letter2 ))
         throwInvalidPointString( linenum, text );
      if (letter1 >= 't' || letter2 >= 't')
      {
         if (letter1 == letter2)
            return Board.INVALID_POINT;
         throwInvalidPointString( linenum, text );
      }

      int col = letter1 - 'a';
      int row = letter2 - 'a';
      if (row >= size)
         throw new IllegalArgumentException( Resources.getString( "msg.5010" ) ); //$NON-NLS-1$
      int pnt = (col * size) + (size - row - 1);
      if (pnt >= Board.getTotal())
         throw new IllegalArgumentException( Resources.formatString( "msg.5009", pnt ) ); //$NON-NLS-1$

      return (short) pnt;
   }

   public static String pointsToString( Iterable<Short> points )
   {
      StringBuilder buffer = new StringBuilder();
      int i = 0;
      Iterator<Short> iter = points.iterator();
      while (iter.hasNext())
      {
         short pnt = iter.next();
         if (i > 0)
            buffer.append( "][" ); //$NON-NLS-1$
         buffer.append( pointToString( pnt ) );
         i++;
      }
      return buffer.toString();
   }

   public static Set<Short> stringToPoints( String text, int linenum )
   {
      Set<Short> points = new HashSet<Short>();
      if (text.isEmpty())
         return points;
      parseList( text, linenum, m_pnt_parser, points );
      return points;
   }

   public static String pointPairToString( ImmutablePair<Short,Short> pair )
   {
      StringBuilder buffer = new StringBuilder();
      buffer.append( pointToString( pair.getLeft() ) );
      buffer.append( ':' );
      buffer.append( pointToString( pair.getRight() ) );
      return buffer.toString();
   }

   public static ImmutablePair<Short,Short> stringToPointPair( String text, int linenum )
   {
      int pos = text.indexOf( ':' );
      if (pos <= 0 || pos == text.length() - 1 || text.lastIndexOf( ':' ) != pos)
         throw new ParseException( linenum, Resources.getString( "msg.5011" ), text ); //$NON-NLS-1$
      return ImmutablePair.of( stringToPoint( text.substring( 0, pos ), linenum ), stringToPoint( text.substring( pos + 1 ), linenum ) );
   }

   public static String pointPairsToString( Set<ImmutablePair<Short,Short>> pairs )
   {
      StringBuilder buffer = new StringBuilder();
      for (ImmutablePair<Short,Short> pair : pairs)
      {
         if (buffer.length() > 0)
            buffer.append( "][" ); //$NON-NLS-1$
         buffer.append( pointPairToString( pair ) );
      }
      return buffer.toString();
   }

   public static Set<ImmutablePair<Short,Short>> stringToPointPairs( String text, int linenum )
   {
      Set<ImmutablePair<Short,Short>> pairs = new HashSet<ImmutablePair<Short,Short>>();
      if (text.isEmpty())
         return pairs;
      parseList( text, linenum, m_pntpair_parser, pairs );
      return pairs;
   }

   public static char printSettingsToChar( PrintSettings settings )
   {
      switch (settings.getMoveNumbers())
      {
         case AS_IS:
            return '1';
         case DISABLE:
            return '0';
         case MODULO:
            int modulo = settings.getMoveNumberModulo(); 
            if (modulo != 100)
               throw new IllegalArgumentException( Resources.formatString( "msg.5012", modulo ) ); //$NON-NLS-1$
            return '2';
         default:
            throw new IllegalArgumentException( Resources.formatString( "msg.5013", settings.getMoveNumbers() ) ); //$NON-NLS-1$
      }
   }

   public static PrintSettings charToPrintSettings( char value, int linenum )
   {
      PrintSettings settings = new PrintSettings();
      switch (value)
      {
         case '0':
            settings.setMoveNumbers( PrintSettings.MoveNumbers.DISABLE );
            break;
         case '1':
            settings.setMoveNumbers( PrintSettings.MoveNumbers.AS_IS );
            break;
         case '2':
            settings.setMoveNumberModulo( 100 );
            break;
         default:
            throw new ParseException( linenum, Resources.getString( "msg.5014" ), Character.toString( value ) ); //$NON-NLS-1$
      }
      return settings;
   }

   public static char renderingStyleToChar( RenderingStyle style )
   {
      if (style.showSiblings())
      {
         if (style.showMarkup())
            return '1';
         else
            return '3';
      }
      else
      {
         if (style.showMarkup())
            return '0';
         else
            return '2';
      }
   }

   public static RenderingStyle charToRenderingStyle( char ch, int linenum )
   {
      switch (ch)
      {
         case '0': return new RenderingStyle( false, true );
         case '1': return new RenderingStyle( true, true );
         case '2': return new RenderingStyle( false, false );
         case '3': return new RenderingStyle( true, false );
         default:  throw new ParseException( linenum, Resources.getString( "msg.5015" ), Character.toString( ch ) ); //$NON-NLS-1$
      }
   }

   public static Result stringToResult( String text, int linenum )
   {
      if (text.length() < 3)
         throw new ParseException( linenum, Resources.getString( "msg.5017" ), text ); //$NON-NLS-1$
      String copy = text.toUpperCase();
      if (copy.equals( "DRAW" )) //$NON-NLS-1$
         return new Result( Colors.DONT_CARE, Result.Status.DRAW );
      String prefix = copy.substring( 0, 2 );
      String suffix = copy.substring( 2, copy.length() );
      Colors player = Colors.DONT_CARE;
      if (prefix.equals( "W+" )) //$NON-NLS-1$
         player = Colors.WHITE;
      else if (prefix.equals( "B+" )) //$NON-NLS-1$
         player = Colors.BLACK;
      else
         throw new ParseException( linenum, Resources.getString( "msg.5017" ), text ); //$NON-NLS-1$
      if (suffix.equals( "R" ) || suffix.equals( "RESIGN" ))                           //$NON-NLS-1$ //$NON-NLS-2$
         return new Result( player, Result.Status.RESIGNED );
      else if (suffix.equals( "F" ) || suffix.equals( "FORFEIT" ))                     //$NON-NLS-1$ //$NON-NLS-2$
         return new Result( player, Result.Status.FORFEIT );
      else if (suffix.equals( "T" ) || suffix.equals( "TIME" ))                        //$NON-NLS-1$ //$NON-NLS-2$
         return new Result( player, Result.Status.LOST_ON_TIME );
      else
         return new Result( player, Result.Status.WON, stringToDecimal( suffix, linenum ) );
   }

   public static String roundToString( Round round )
   {
      StringBuilder buffer = new StringBuilder();
      if (round.getNumber() > 0)
      {
         buffer.append( round.getNumber() );
         if (round.getType() != Round.Type.UNDEFINED)
            buffer.append( " " ); //$NON-NLS-1$
      }
      switch( round.getType() )
      {
         case FINAL:   buffer.append( "(FN)" ); break; //$NON-NLS-1$
         case GAME:    buffer.append( "(GM)" ); break; //$NON-NLS-1$
         case LEAGUE:  buffer.append( "(LG)" ); break; //$NON-NLS-1$
         case PLAYOFF: buffer.append( "(PO)" ); break; //$NON-NLS-1$
         case PRELIM:  buffer.append( "(PR)" ); break; //$NON-NLS-1$
         default:                               break;
      }
      if (!round.getExtra().isEmpty())
      {
         if (buffer.length() > 0)
            buffer.append( " " ); //$NON-NLS-1$
         buffer.append( round.getExtra() );
      }
      return buffer.toString();
   }

   public static Round stringToRound( String text, int linenum )
   {
      // This method has to gracefully deal with unstructured data. The strategy is
      // as follows:
      // 1) start with assumption that input will match the recommended
      //    format described in the SGF spec, and start parsing
      // 2) if the whole input matches the spec, then extra will be
      //    empty, because there was nothing left over
      // 3) if the initial round number is valid, but the remainder of
      //    input is not, then accept the round number, set the type
      //    to UNDEFINED, and put the rest in extra
      // 4) else put the entire input in extra
      //
      Collection<ImmutablePair<Integer,String>> tokens = new HashSet<ImmutablePair<Integer,String>>();
      tokens.add( LEFT_PAREN );
      tokens.add( RIGHT_PAREN );
      Lexer lexer = new Lexer( new StringReader( text ), tokens, "" ); //$NON-NLS-1$
      StringBuilder extra = new StringBuilder();
      int num = 0;
      Round.Type type = Round.Type.UNDEFINED;
      try
      {
         if (!lexer.hasMore())
            return new Round();
         ImmutablePair<Integer,String> token = lexer.getNext();
         if (token.getLeft() != Lexer.INTEGER_TOKEN_ID)
            return new Round( 0, Round.Type.UNDEFINED, text );
         num = (int) stringToNumber( token.getRight(), linenum );
         if (lexer.hasMore())
         {
            lexer.consumeWhitespace();
            if (lexer.hasMore())
            {
               token = lexer.getNext();
               if (token.getLeft() != LEFT_PAREN_TOKEN_ID)
                  return new Round( num, Round.Type.UNDEFINED, text );
               lexer.consumeWhitespace();
               if (!lexer.hasMore())
                  return new Round( num, Round.Type.UNDEFINED, text );
               token = lexer.getNext();
               if (token.getLeft() != Lexer.TEXT_TOKEN_ID)
                  return new Round( num, Round.Type.UNDEFINED, text );
               String abbrev = token.getRight().toUpperCase();
               lexer.consumeWhitespace();
               if (!lexer.hasMore())
                  return new Round( num, Round.Type.UNDEFINED, text );
               token = lexer.getNext();
               if (token.getLeft() != RIGHT_PAREN_TOKEN_ID)
                  return new Round( num, Round.Type.UNDEFINED, text );
               // FIXME: no idea whether the following abbreviations are right
               //
               if (abbrev.equals( "FN" )) //$NON-NLS-1$
                  type = Round.Type.FINAL;
               else if (abbrev.equals( "GM" )) //$NON-NLS-1$
                  type = Round.Type.GAME;
               else if (abbrev.equals( "LG" )) //$NON-NLS-1$
                  type = Round.Type.LEAGUE;
               else if (abbrev.equals( "PO" )) //$NON-NLS-1$
                  type = Round.Type.PLAYOFF;
               else if (abbrev.equals( "PR" )) //$NON-NLS-1$
                  type = Round.Type.PRELIM;
               lexer.consumeWhitespace();
               lexer.enable( PAREN_TOKEN_IDS, false );
               lexer.enable( Lexer.WS_TOKEN_ID, false );
               while (lexer.hasMore())
               {
                  token = lexer.getNext();
                  extra.append( token.getRight() );
               }
            }
         }
      }
      catch (IOException ex)
      {
         throw new ParseException( linenum, ex.getMessage(), text );
      }
      return new Round( num, type, extra.toString() );
   }

   /**
    * Returns a string of 12 characters in length. The first 6 characters represent
    * signature A and the remaining 6 characters represent signature B, as defined
    * by Dave Dyer's game record signature spec.
    * 
    * @see #stringToSignature
    * @see <a href="http://www.andromeda.com/people/ddyer/go/signature-spec.html">Signatures for Go game records</a>
    * 
    * @param sig
    */
   public static String signatureToString( Signature sig )
   {
      short[] sig_pnts = sig.getPoints();
      if (sig_pnts.length != Signature.TURN_NUMBERS.length)
         throw new IllegalStateException( Resources.getString( "msg.5018" ) ); //$NON-NLS-1$
      StringBuilder buffer = new StringBuilder();
      for (int i = 0; i < sig_pnts.length; i++)
      {
         short pnt = sig_pnts[i];
         if (pnt == Board.INVALID_POINT)
            buffer.append( "??" ); //$NON-NLS-1$
         else
            buffer.append( pointToString( pnt ) );
      }
      return buffer.toString();
   }

   /**
    * Instantiates a game signature based on a string of 12 characters as defined
    * by Dave Dyer's game record signature spec.
    * 
    * @see #signatureToString
    * @see <a href="http://www.andromeda.com/people/ddyer/go/signature-spec.html">Signatures for Go game records</a>
    * 
    * @param text
    * @param linenum
    */
   public static Signature stringToSignature( String text, int linenum )
   {
      Signature sig = new Signature();
      int count = Signature.TURN_NUMBERS.length;
      if (text.length() != count * 2)
         throw new ParseException( linenum, Resources.getString( "msg.5019" ), text ); //$NON-NLS-1$
      for (int i = 0; i < count; i++)
      {
         int j = Signature.ARRAY_INDICES[i] * 2;
         if (text.charAt( j ) == '?')
         {
            if (text.charAt( j + 1 ) != '?')
               throw new ParseException( linenum, Resources.getString( "msg.5019" ), text ); //$NON-NLS-1$
            sig.addPoint( Board.INVALID_POINT );
         }
         else
         {
            String tmp = text.substring( j, j + 2 );
            sig.addPoint( stringToPoint( tmp, linenum ) );
         }
      }
      return sig;
   }

   public static String textPairToString( ImmutablePair<String,String> pair )
   {
      return String.format( "%s:%s", transformText( pair.getLeft(), true ), transformText( pair.getRight(), true ) ); //$NON-NLS-1$
   }

   public static ImmutablePair<String,String> stringToTextPair( String text, int linenum )
   {
      if (text.isEmpty())
         return ImmutablePair.of( "", "" ); //$NON-NLS-1$ //$NON-NLS-2$
      int pos = text.indexOf( ':' );
      if (pos < 0)
         throw new ParseException( linenum, Resources.getString( "msg.5020" ), text ); //$NON-NLS-1$
      String lhs = ""; //$NON-NLS-1$
      String rhs = ""; //$NON-NLS-1$
      if (pos > 0)
         lhs = text.substring( 0, pos );
      if (pos < text.length() - 1)
         rhs = text.substring( pos + 1 );
      return ImmutablePair.of( lhs, rhs );
   }

   /**
    * Transforms the original string to simplified text as defined
    * by the SGF specification.<p>
    * 
    * <ul>
    * <li>whitespace other than unescaped eol characters (hard linebreaks) are
    * converted to spaces
    * <li>if eol_to_space is true, hard linebreaks are converted to spaces
    * <li>the following special characters
    * '[', ']', '(', ')', '\', and ':' are escaped by prefixing each
    * one with the '\' character
    * <li>explicitly escaped characters are emitted verbatim, except for
    * whitespace which is always converted to a space and eol characters
    * (soft linebreaks) which are always removed
    * </ul>
    * 
    * @see <a href="http://www.red-bean.com/sgf/sgf4.html#text">Section 3.2. Text</a>
    * @see <a href="http://www.red-bean.com/sgf/sgf4.html#simpletext">Section 3.3. SimpleText</a>
    * 
    * @param orig_text      original text
    * @param eol_to_space   if true, convert eol characters to spaces
    */
   public static String transformText( String orig_text, boolean eol_to_space )
   {
      StringBuilder buffer = new StringBuilder();
      boolean bslash = false;
      boolean eol = false;
      int len = orig_text.length();
      for (int i = 0; i < len; i++)
      {
         char ch = orig_text.charAt( i );
         if (ch == '\\')
         {
            if (bslash)
            {
               buffer.append( '\\' );
               bslash = false;
            }
            else
            {
               bslash = true;
            }
            eol = false;
         }
         else if (ch == '\n' || ch == '\r')
         {
            if (eol || bslash)
            {
               // skip it
            }
            else if (eol_to_space)
            {
               buffer.append( ' ' );
            }
            else
            {
               buffer.append( System.getProperty( "line.separator" ) ); //$NON-NLS-1$
            }
            eol = true;
            bslash = false;
         }
         else if (Character.isWhitespace( ch ))
         {
            buffer.append( ' ' );
            eol = false;
            bslash = false;
         }
         else
         {
            if (!bslash && (ch == '[' || ch == ']' || ch == '(' || ch == ')' || ch == ':'))
               buffer.append( '\\' );
            buffer.append( ch );
            eol = false;
            bslash = false;
         }
      }
      return buffer.toString();
   }

   private static <T> void parseList( String text, int linenum, TokenParser<T> parser, Set<T> dest )
   {
      Set<ImmutablePair<Integer,String>> tokens = new HashSet<ImmutablePair<Integer,String>>();
      tokens.add( ImmutablePair.of( LIST_DIV_TOKEN_ID, "][" ) ); //$NON-NLS-1$
      Lexer lexer = new Lexer( new StringReader( text ), tokens, "\\" ); //$NON-NLS-1$
      try
      {
         while (lexer.hasMore())
         {
            ImmutablePair<Integer,String> token = lexer.getNext();
            if (token.getLeft() != LIST_DIV_TOKEN_ID)
               parser.parse( token.getRight(), linenum, dest );
            else
               throw new ParseException( linenum, Resources.getString( "msg.5021" ), text ); //$NON-NLS-1$
            if (!lexer.hasMore())
               break;
            token = lexer.getNext();
            if (token.getLeft() != LIST_DIV_TOKEN_ID || !lexer.hasMore())
               throw new ParseException( linenum, Resources.getString( "msg.5021" ), text ); //$NON-NLS-1$
         }
      }
      catch (IOException ex)
      {
         throw new ParseException( linenum, ex.getMessage(), text );
      }
   }

   private static void throwInvalidPointString( int linenum, String token )
   {
      throw new ParseException( linenum, Resources.getString( "msg.5022" ), token ); //$NON-NLS-1$
   }

   private static final int DIAG_COORDS          = 0x00000001;
   private static final int DIAG_NAME            = 0x00000002;
   private static final int DIAG_MOVES_NOT_SHOWN = 0x00000004;
   private static final int DIAG_REMOVE_CAPTURED = 0x00000100;
   private static final int DIAG_HOSHI_DOTS      = 0x00000200;
   private static final int DIAG_APP_DEFAULTS    = 0x00008000;
}
