//
// SGFReader.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.record.sgf;

import java.io.IOException;
import java.io.Reader;
import java.math.BigDecimal;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.xeoh.plugins.base.annotations.Capabilities;
import net.xeoh.plugins.base.annotations.PluginImplementation;
import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.base.Lexer;
import com.dakengo.libs.base.ParseException;
import com.dakengo.libs.game.Board;
import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.Layers;
import com.dakengo.libs.record.Connectors;
import com.dakengo.libs.record.GameProperties;
import com.dakengo.libs.record.Grades;
import com.dakengo.libs.record.GradeTypes;
import com.dakengo.libs.record.Handicaps;
import com.dakengo.libs.record.Notations;
import com.dakengo.libs.record.PositionProperties;
import com.dakengo.libs.record.PrintSettings;
import com.dakengo.libs.record.PropertyException;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.libs.record.Record;
import com.dakengo.libs.record.RecordReader;
import com.dakengo.libs.record.Round;

/**
 * Reads game records in the SGF format.
 * Primarily targets FF[4] but makes allowances for FF[3] compatibility
 * whenever possible. Each node in the SGF tree is indicated using a NOTE property.
 * Further annotations for the node are also specified via NOTE properties. For
 * example:<p>
 * 
 * <table>
 * <tr><td>{@code ImmutablePair.of( 1, ImmutablePair.of( "NODE",    "" ) );}</td><td>a new node indicated by ';'</td></tr>
 * <tr><td>{@code ImmutablePair.of( 1, ImmutablePair.of( "ILLEGAL", "" );}</td><td>a "KO" property for the same node</td></tr>
 * <tr><td>{@code ImmutablePair.of( 1, ImmutablePair.of( "NAME",    "foo" );}</td><td>an "N" property for the same node</td></tr>
 * <tr><td>{@code ImmutablePair.of( 1, ImmutablePair.of( "NUMBER",  "12" );}</td><td>an "MN" property for the same node</td></tr>
 * <tr><td>{@code ImmutablePair.of( 1, ImmutablePair.of( "HOTSPOT", "1", );}</td><td>an "HO" property for the same node</td></tr>
 * <tr><td>{@code ImmutablePair.of( 1, ImmutablePair.of( "VALUE",   "W+65.5" );}</td><td>a "V" property for the same node</td></tr>
 * <tr><td>{@code ImmutablePair.of( 2, ImmutablePair.of( "NODE",    "" );}</td><td>the next node indicated by ';'</td></tr>
 * </table><p>
 * 
 * The "FG" and "VW" properties are represented in a record using a NOTE property and,
 * optionally, a subsequent property. The consumer of the record should look for a
 * subsequent {@link PropertyID#DIAGRAM_SETTINGS} or {@link PropertyID#VISIBLE} property on the
 * current branch. If found, the second property specifies further action to take.
 * If not found, the consumer should reset its state. For example, a "VISIBILITY"
 * note with no subsequent {@link PropertyID#VISIBLE} property means show the entire board.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see <a href="http://www.red-bean.com/sgf/">SGF File Format<a>
 */
@PluginImplementation
public class SGFReader implements RecordReader
{
   public SGFReader()
   {
      initStandardPropsTable();
   }

   @Override
   public void setInput( Reader reader )
   {
      if (m_lexer == null)
      {
         Set<ImmutablePair<Integer,String>> tokens = new HashSet<ImmutablePair<Integer,String>>();
         tokens.add( LEFT_PAREN );
         tokens.add( RIGHT_PAREN );
         tokens.add( LEFT_BRACKET );
         tokens.add( RIGHT_BRACKET );
         tokens.add( COLON );
         tokens.add( SEMI_COLON );
         m_lexer = new Lexer( reader, tokens, "\\" ); //$NON-NLS-1$
         m_lexer.enable( Lexer.INTEGER_TOKEN_ID, false );
         m_lexer.enable( Lexer.DECIMAL_TOKEN_ID, false );
      }
      else
      {
         m_lexer.setInput( reader );
      }
   }

   @Override
   public List<String> getMessages()
   {
      return new ArrayList<String>( m_messages );
   }

   @Override
   public String getInputPath()
   {
      return m_input_path;
   }

   @Override
   public void setInputPath( String path )
   {
      m_input_path = path;
   }

   @Override
   public Record read() throws IOException
   {
      m_black_moves.clear();
      m_black_stones = false;
      m_branch_stack.clear();
      m_messages.clear();
      m_next_id = PropertyID.MAX_STD_ID + 1;
      m_node_num = -1;
      m_nonstd_ids.clear();
      m_record = null;
      m_value_required = false;
      m_white_move = Board.INVALID_POINT;

      if (m_lexer == null)
         throw new IllegalStateException( Resources.getString( "msg.5023" ) ); //$NON-NLS-1$
      m_lexer.consumeWhitespace();
      if (!m_lexer.hasMore())
         return null;
      ImmutablePair<Integer,String> token = m_lexer.getNext();
      if (token.getLeft() != LEFT_PAREN_TOKEN_ID)
         throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5024" ), token.getRight() ); //$NON-NLS-1$
      readBranch();

      GameProperties props = m_record.getGameProps();
      if (props.getPlayerName( Colors.BLACK ).isEmpty())
      {
         m_messages.add( Resources.getString( "msg.5025" ) ); //$NON-NLS-1$
         props.set( PropertyID.PLAYER_NAME, ImmutablePair.of( Colors.BLACK, UNKNOWN_PLAYER ) );
      }
      if (props.getPlayerName( Colors.WHITE ).isEmpty())
      {
         m_messages.add( Resources.getString( "msg.5026" ) ); //$NON-NLS-1$
         props.set( PropertyID.PLAYER_NAME, ImmutablePair.of( Colors.WHITE, UNKNOWN_PLAYER ) );
      }

      return m_record;
   }

   /**
    * Returns an array of strings specifying the capabilities of this plugin.
    */
   @Capabilities
   public String[] getCapabilities()
   {
      return CAPABILITIES;
   }

   private interface PropertyReader
   {
      void read( String prop_name ) throws IOException;
   }

   /**
    * property: "AP"
    */
   private PropertyReader m_appprop_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         String[] pieces = text.split( ":" ); //$NON-NLS-1$
         GameProperties props = m_record.getGameProps();
         if (pieces.length == 1)
         {
            if (!SGFReader.this.isDuplicateInfoProp( PropertyID.APP_NAME, prop_name, text ))
               props.set( PropertyID.APP_NAME, pieces[0] );
         }
         else if (pieces.length == 2)
         {
            if (!SGFReader.this.isDuplicateInfoProp( PropertyID.APP_NAME, prop_name, text ))
               props.set( PropertyID.APP_NAME, pieces[0] );
            if (!SGFReader.this.isDuplicateInfoProp( PropertyID.APP_VERSION, prop_name, text ))
               props.set( PropertyID.APP_VERSION, pieces[1] );
         }
         else
         {
            throwUnexpectedValue( prop_name, text );
         }
      }
   };

   /**
    * property: "C"
    */
   private PropertyReader m_comment_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         m_curr_branch.add( PropertyID.COMMENT, readRawValue( prop_name ) );
      }
   };

   /**
    * property: "AR" or "LN"
    */
   private PropertyReader m_connector_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Connectors conn = Connectors.ARROW;
         if (prop_name.equals( "AR" ))      //$NON-NLS-1$
            conn = Connectors.ARROW;
         else if (prop_name.equals( "LN" )) //$NON-NLS-1$
            conn = Connectors.LINE;
         else
            throwUnexpectedProperty( prop_name );

         Set<ImmutablePair<Short,Short>> point_pairs = new HashSet<ImmutablePair<Short,Short>>();
         m_lexer.enable( COLON_TOKEN_ID, false );
         try
         {
            while (m_lexer.hasMore())
            {
               ImmutablePair<Integer,String> token = m_lexer.getNext();
               if (token.getLeft() == RIGHT_BRACKET_TOKEN_ID)
                  break;
               if (token.getLeft() != Lexer.TEXT_TOKEN_ID)
                  throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5027" ), token.getRight() ); //$NON-NLS-1$
               ImmutablePair<Short,Short> pair = null;
               try
               {
                  pair = Convert.stringToPointPair( token.getRight(), m_lexer.getLineNumber() );
               }
               catch (ParseException ex)
               {
                  rethrowWithInputPath( ex );
               }
               if (pair.getLeft().equals( pair.getRight() ))
                  throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5028" ), token.getRight() ); //$NON-NLS-1$
               point_pairs.add( pair );
               if (!m_lexer.hasMore())
                  throwUnexpectedEndOfInput();
               token = m_lexer.getNext();
               if (token.getLeft() != RIGHT_BRACKET_TOKEN_ID)
                  throwRightBracketExpected( token.getRight() );
               m_lexer.consumeWhitespace();
               if (!m_lexer.hasMore())
                  throwUnexpectedEndOfInput();
               token = m_lexer.getNext();
               while (token.getLeft() == Lexer.EOL_TOKEN_ID)
               {
                  if (!m_lexer.hasMore())
                     throwUnexpectedEndOfInput();
                  token = m_lexer.getNext();
               }
               if (token.getLeft() != LEFT_BRACKET_TOKEN_ID)
               {
                  m_lexer.pushBack( token );
                  break;
               }
            }
         }
         finally
         {
            m_lexer.enable( COLON_TOKEN_ID, true );
         }
         
         m_curr_branch.add( PropertyID.CONNECTOR, ImmutablePair.of( conn, point_pairs ) );
      }
   };

   /**
    * property: "FG"
    */
   private PropertyReader m_figure_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         m_curr_branch.add( PropertyID.NOTE, ImmutablePair.of( m_node_num, ImmutablePair.of( Constants.DIAGRAM_NOTE, "" ) ) ); //$NON-NLS-1$
         if (!text.isEmpty())
            m_curr_branch.add( PropertyID.DIAGRAM_SETTINGS, Convert.stringToDiagramSettings( text, m_lexer.getLineNumber() ) );
      }
   };

   /**
    * property: "FF"
    */
   private PropertyReader m_filefmt_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         int value = (int) readNumber( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         m_sgf_version = value;
         if (m_sgf_version < 1 || m_sgf_version > 4)
            throwUnexpectedValue( "FF", Integer.toString( m_sgf_version ) ); //$NON-NLS-1$
      }
   };

   /**
    * property: "AN", "CA", "CP", "DT", "EV", "GC", "GN", "ON", "OT", "PC", "RU", "SO", or "US"
    */
   private PropertyReader m_gameinfo_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         int prop_id = PropertyID.MAX_STD_ID;
         switch (prop_name)
         {
            case "AN": prop_id = PropertyID.ANNOTATOR_NAME;  break; //$NON-NLS-1$
            case "CA": prop_id = PropertyID.CHARSET;         break; //$NON-NLS-1$
            case "CP": prop_id = PropertyID.COPYRIGHT;       break; //$NON-NLS-1$
            case "DT": prop_id = PropertyID.DATE;            break; //$NON-NLS-1$
            case "EV": prop_id = PropertyID.EVENT_NAME;      break; //$NON-NLS-1$
            case "GC": prop_id = PropertyID.BACKGROUND_INFO; break; //$NON-NLS-1$
            case "GN": prop_id = PropertyID.GAME_NAME;       break; //$NON-NLS-1$
            case "ON": prop_id = PropertyID.OPENING_NAME;    break; //$NON-NLS-1$
            case "OT": prop_id = PropertyID.OVERTIME_SYSTEM; break; //$NON-NLS-1$
            case "PC": prop_id = PropertyID.PLACE_NAME;      break; //$NON-NLS-1$
            case "RU": prop_id = PropertyID.RULE_SYSTEM;     break; //$NON-NLS-1$
            case "SO": prop_id = PropertyID.SOURCE_NAME;     break; //$NON-NLS-1$
            case "US": prop_id = PropertyID.USER_NAME;       break; //$NON-NLS-1$
            default:   throwUnexpectedProperty( prop_name ); break;
         }

         // FIXME: a temporary sanity check until the code has better
         // support for character encodings
         if (prop_name.equals( "CA" ) && !text.equalsIgnoreCase( "utf-8" )) //$NON-NLS-1$ //$NON-NLS-2$
            m_messages.add( Resources.formatString( "line %d: unexpected charset: '%s'", m_lexer.getLineNumber(), text ) ); //$NON-NLS-1$

         if (!isDuplicateInfoProp( prop_id, prop_name, text ))
            m_record.getGameProps().set( prop_id, text );
      }
   };

   /**
    * property: "GM"
    */
   private PropertyReader m_gametype_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         long value = readNumber( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         if (value != 1L)
            throwUnexpectedValue( prop_name, Long.toString( value ) );
      }
   };

   /**
    * property: "HA"
    */
   private PropertyReader m_handicap_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         long value = readNumber( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         if (value == 0L)
         {
            logUnexpectedValue( prop_name, "0" ); //$NON-NLS-1$
            return;
         }
         if (value < 2L || value > 9L)
         {
            throwUnexpectedValue( prop_name, Long.toString( value ) );
         }
         if (!isDuplicateInfoProp( PropertyID.HANDICAP, prop_name, Long.toString( value ) ))
            m_record.getGameProps().set( PropertyID.HANDICAP, ImmutablePair.of( Handicaps.UNKNOWN, (int) value ) );
      }
   };

   /**
    * property: "KO"
    */
   private PropertyReader m_illegalmove_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         if (!prop_name.equals( "KO" )) //$NON-NLS-1$
            throwUnexpectedProperty( prop_name );
         if (!text.isEmpty())
            logUnexpectedValue( prop_name, text );
         m_curr_branch.add( PropertyID.NOTE, ImmutablePair.of( m_node_num, ImmutablePair.of( Constants.ILLEGAL_NOTE, "" ) ) ); //$NON-NLS-1$
      }
   };

   /**
    * property: "KM"
    */
   private PropertyReader m_komi_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         BigDecimal value = readDecimal( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         if (!isDuplicateInfoProp( PropertyID.KOMI, prop_name, value.toString() ))
            m_record.getGameProps().set( PropertyID.KOMI, value );
      }
   };

   /**
    * property: "LB"
    */
   private PropertyReader m_label_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Set<ImmutablePair<Short,String>> labels = readAnnotatedPointList( prop_name );
         m_curr_branch.add( PropertyID.LABEL, labels );
      }
   };

   /**
    * property: "CR", "DD", "MA", "SL", "SQ" or "TR"
    */
   private PropertyReader m_markpnts_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Set<Short> pnts = readPointList( prop_name );
         Notations note = Notations.X;
         switch (prop_name)
         {
            case "CR": note = Notations.CIRCLE;              break; //$NON-NLS-1$
            case "DD": note = Notations.DIMMED;              break; //$NON-NLS-1$
            case "MA": note = Notations.X;                   break; //$NON-NLS-1$
            case "SL": note = Notations.SELECTED;            break; //$NON-NLS-1$
            case "SQ": note = Notations.SQUARE;              break; //$NON-NLS-1$
            case "TR": note = Notations.TRIANGLE;            break; //$NON-NLS-1$
            default:   throwUnexpectedProperty( prop_name ); break;
         }
         m_curr_branch.add( PropertyID.MARK, ImmutablePair.of( note, pnts ) );
      }
   };

   /**
    * property: "B" or "W"
    */
   private PropertyReader m_move_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         short pnt = Board.INVALID_POINT;
         if (!text.isEmpty())
         {
            try
            {
               pnt = Convert.stringToPoint( text, m_lexer.getLineNumber() );
            }
            catch (ParseException ex)
            {
               rethrowWithInputPath( ex );
            }
         }
         Colors color = Colors.DONT_CARE;
         if (prop_name.equals( "B" ))       //$NON-NLS-1$
            color = Colors.BLACK;
         else if (prop_name.equals( "W" ))  //$NON-NLS-1$
            color = Colors.WHITE;
         else
            throwUnexpectedProperty( prop_name );

         // As long as we haven't yet seen a white move property, buffer up the
         // black moves until we do.
         //
         // Once we see a white move, then if there is only one black move that
         // was saved, this had to have been a move and not a handicap stone,
         // so set it in the branch as a move.
         //
         // Otherwise, copy all the saved black moves into a STONES
         // property to set the handicap; this is taken to be black's
         // first turn. Then process the white move.
         //
         // From then on, no need to buffer up black moves, and the
         // PositionProperties error checking will do the right thing.
         //
         if (m_white_move == Board.INVALID_POINT)
         {
            if (color == Colors.BLACK)
            {
               m_black_moves.add( pnt );
            }
            else
            {
               if (m_black_moves.isEmpty() && !m_black_stones)
               {
                  throw new PropertyException( PropertyID.MOVE, Colors.WHITE.toString() );
               }
               else if (m_black_moves.size() == 1)
               {
                  m_curr_branch.add( PropertyID.MOVE, ImmutablePair.of( Colors.BLACK, m_black_moves.iterator().next() ) );
               }
               else if (m_black_moves.size() > 1)
               {
                  m_curr_branch.add( PropertyID.STONES, ImmutablePair.of( Colors.BLACK, m_black_moves ) );
               }
               m_black_moves.clear();
               m_white_move = pnt;
               m_curr_branch.add( PropertyID.MOVE, ImmutablePair.of( color, pnt ) );
            }
         }
         else
         {
            m_curr_branch.add( PropertyID.MOVE, ImmutablePair.of( color, pnt ) );
         }
      }
   };

   /**
    * property: "BM", "DO", "IT", or "TE"
    */
   private PropertyReader m_movegrade_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Grades grade = Grades.BAD;
         long sgf_value = 0;
         switch (prop_name)
         {
            case "BM": //$NON-NLS-1$
               sgf_value = readNumber( prop_name );
               if (sgf_value == 1L)
                  grade = Grades.BAD;
               else if (sgf_value == 2L)
                  grade = Grades.VERY_BAD;
               else
                  logUnexpectedValue( prop_name, Long.toString( sgf_value ) );
               break;
            case "TE": //$NON-NLS-1$
               sgf_value = readNumber( prop_name );
               if (sgf_value == 1L)
                  grade = Grades.GOOD;
               else if (sgf_value == 2L)
                  grade = Grades.VERY_GOOD;
               else
                  logUnexpectedValue( prop_name, Long.toString( sgf_value ) );
               break;
            case "DO": //$NON-NLS-1$
               String txt1 = readRawValue( prop_name );
               if (!txt1.isEmpty())
                  logUnexpectedValue( prop_name, txt1 );
               grade = Grades.DOUBTFUL;
               break;
            case "IT": //$NON-NLS-1$
               String txt2 = readRawValue( prop_name );
               if (!txt2.isEmpty())
                  logUnexpectedValue( prop_name, txt2 );
               grade = Grades.INTERESTING;
               break;
            default:
               throwUnexpectedProperty( prop_name );
               break;
         }
         m_curr_branch.add( PropertyID.GRADE, ImmutablePair.of( Colors.DONT_CARE, ImmutablePair.of( GradeTypes.MOVE, grade ) ) );
      }
   };

   /**
    * property: "OB" or "OW"
    */
   private PropertyReader m_movesleft_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         long count = readNumber( prop_name );
         Colors color = Colors.DONT_CARE;
         switch (prop_name)
         {
            case "OB": color = Colors.BLACK;                 break; //$NON-NLS-1$
            case "OW": color = Colors.WHITE;                 break; //$NON-NLS-1$
            default:   throwUnexpectedProperty( prop_name ); break;
         }
         m_curr_branch.add( PropertyID.REMAINING_MOVES, ImmutablePair.of( color, (int) count ) );
      }
   };

   /**
    * property: "HO", "MN", "N", or "V"
    */
   private PropertyReader m_note_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         String label = ""; //$NON-NLS-1$
         switch (prop_name)
         {
            case "HO": //$NON-NLS-1$
               if (text.length() != 1)
                  throwUnexpectedValue( prop_name, text );
               Convert.charToEmphasis( text.charAt( 0 ), m_lexer.getLineNumber() );
               label = Constants.HOTSPOT_NOTE;
               break;
            case "MN": //$NON-NLS-1$
               Convert.stringToNumber( text, m_lexer.getLineNumber() );
               label = Constants.NUMBER_NOTE;
               break;
            case "N": //$NON-NLS-1$
               label = Constants.NAME_NOTE;
               break;
            case "V": //$NON-NLS-1$
               Convert.stringToDecimal( text, m_lexer.getLineNumber() );
               label = Constants.VALUE_NOTE;
               break;
            default:
               throwUnexpectedProperty( prop_name );
               break;
         }
         m_curr_branch.add( PropertyID.NOTE, ImmutablePair.of( m_node_num, ImmutablePair.of( label, text ) ) );
      }
   };

   /**
    * property: "DM", "GB", "GW", or "UC"
    */
   private PropertyReader m_posgrade_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Colors color = Colors.DONT_CARE;
         Grades grade = Grades.BAD;
         int sgf_value = (int) readNumber( prop_name );
         if (sgf_value != 1 && sgf_value != 2)
         {
            logUnexpectedValue( prop_name, Integer.toString( sgf_value ) );
            return;
         }
         switch (prop_name)
         {
            case "DM": //$NON-NLS-1$
               grade = Grades.EVEN;
               break;
            case "GB": //$NON-NLS-1$
               color = Colors.BLACK;
               if (sgf_value == 1)
                  grade = Grades.GOOD;
               else
                  grade = Grades.VERY_GOOD;
               break;
            case "GW": //$NON-NLS-1$
               color = Colors.WHITE;
               if (sgf_value == 1)
                  grade = Grades.GOOD;
               else
                  grade = Grades.VERY_GOOD;
               break;
            case "UC": //$NON-NLS-1$
               grade = Grades.UNCLEAR;
               break;
            default:
               throwUnexpectedProperty( prop_name );
               break;
         }
         m_curr_branch.add( PropertyID.GRADE, ImmutablePair.of( color, ImmutablePair.of( GradeTypes.POSITION, grade ) ) );
      }
   };

   /**
    * property: "BR", "BT, "PB", "PW", "WR" or "WT"
    */
   private PropertyReader m_playertext_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         int prop_id = PropertyID.MAX_STD_ID;
         Colors player = Colors.DONT_CARE;
         switch (prop_name)
         {
            case "BR": //$NON-NLS-1$
               player = Colors.BLACK;
               prop_id = PropertyID.PLAYER_RANK;
               break;
            case "BT": //$NON-NLS-1$
               player = Colors.BLACK;
               prop_id = PropertyID.TEAM_NAME;
               break;
            case "PB": //$NON-NLS-1$
               player = Colors.BLACK;
               prop_id = PropertyID.PLAYER_NAME;
               break;
            case "PW": //$NON-NLS-1$
               player = Colors.WHITE;
               prop_id = PropertyID.PLAYER_NAME;
               break;
            case "WR": //$NON-NLS-1$
               player = Colors.WHITE;
               prop_id = PropertyID.PLAYER_RANK;
               break;
            case "WT": //$NON-NLS-1$
               player = Colors.WHITE;
               prop_id = PropertyID.TEAM_NAME;
               break;
            default:
               throwUnexpectedProperty( prop_name );
               break;
         }
         GameProperties props = m_record.getGameProps();
         if (!props.getPlayerProperty( prop_id, player ).isEmpty())
            logDuplicateProperty( prop_name, text );
         else if (!text.isEmpty())
            props.set( prop_id, ImmutablePair.of( player, text ) );
      }
   };

   /**
    * property: "PL"
    */
   private PropertyReader m_playerturn_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         m_curr_branch.add( PropertyID.TURN, readColor( prop_name ) );
      }
   };

   /**
    * property: "PM"
    */
   private PropertyReader m_printing_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         long value = readNumber( prop_name );
         if (value < 0L || value > 2L)
         {
            logUnexpectedValue( prop_name, Long.toString( value ) );
            return;
         }
         PrintSettings settings = Convert.charToPrintSettings( (char) value, m_lexer.getLineNumber() );
         m_curr_branch.add( PropertyID.PRINT_SETTINGS, settings );
      }
   };

   /**
    * property: "RE"
    */
   private PropertyReader m_result_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         try
         {
            if (!isDuplicateInfoProp( PropertyID.RESULT, prop_name, text ))
               m_record.getGameProps().set( PropertyID.RESULT, Convert.stringToResult( text, m_lexer.getLineNumber() ) );
         }
         catch (ParseException ex)
         {
            m_messages.add( ex.getMessage() );
         }
      }
   };

   /**
    * property: "RO"
    */
   private PropertyReader m_round_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         Round round = Convert.stringToRound( text, m_lexer.getLineNumber() );
         if (!isDuplicateInfoProp( PropertyID.ROUND, prop_name, text ))
            m_record.getGameProps().set( PropertyID.ROUND, round );
      }
   };

   /**
    * property: "SZ"
    */
   private PropertyReader m_size_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         long size = readNumber( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         if (!isDuplicateInfoProp( PropertyID.SIZE, prop_name, Long.toString( size ) ))
         {
            try
            {
               m_record.getGameProps().set( PropertyID.SIZE, (int) size );
               Board.setSize( (int) size );
            }
            catch (Exception ex)
            {
               throw new ParseException( m_input_path, m_lexer.getLineNumber(), ex.getMessage(), Long.toString( size ) );
            }
         }
      }
   };

   /**
    * property: "AB", "AE", or "AW"
    */
   private PropertyReader m_stones_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Set<Short> stones = readPointList( prop_name );
         Colors color = Colors.EMPTY;
         switch (prop_name)
         {
            case "AB": //$NON-NLS-1$
               color = Colors.BLACK;
               if (stones.size() > 0)
                  m_black_stones = true;
               break;
            case "AE": //$NON-NLS-1$
               color = Colors.EMPTY;
               break;
            case "AW": //$NON-NLS-1$
               color = Colors.WHITE;
               break;
            default:
               throwUnexpectedProperty( prop_name );
               break;
         }
         GameProperties info_props = m_record.getGameProps();
         PositionProperties main_branch = m_record.getMainBranch();
         if (color == Colors.BLACK && Game.isValidHandicap( stones.size() ) &&
             !main_branch.has( PropertyID.MOVE ) && !main_branch.has( PropertyID.STONES ) && !info_props.has( PropertyID.HANDICAP ))
            info_props.set( PropertyID.HANDICAP, ImmutablePair.of( Handicaps.UNKNOWN, stones.size() ) );
         m_curr_branch.add( PropertyID.STONES, ImmutablePair.of( color, stones ) );
      }
   };

   /**
    * property: "TB" or "TW"
    */
   private PropertyReader m_territory_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Set<Short> points = readPointList( prop_name );
         Colors color = Colors.EMPTY;
         switch (prop_name)
         {
            case "TB": color = Colors.BLACK;                 break; //$NON-NLS-1$
            case "TW": color = Colors.WHITE;                 break; //$NON-NLS-1$
            default:   throwUnexpectedProperty( prop_name ); break;
         }
         m_curr_branch.add( PropertyID.TERRITORY, ImmutablePair.of( color, points ) );
      }
   };

   /**
    * property: "BL" or "WL"
    */
   private PropertyReader m_timeleft_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         BigDecimal time = readDecimal( prop_name );
         Colors color = Colors.DONT_CARE;
         if (prop_name.equals( "BL" ))      //$NON-NLS-1$
            color = Colors.BLACK;
         else if (prop_name.equals( "WL" )) //$NON-NLS-1$
            color = Colors.WHITE;
         else
            throwUnexpectedProperty( prop_name );
         m_curr_branch.add( PropertyID.REMAINING_TIME, ImmutablePair.of( color, time ) );
      }
   };

   /**
    * property: "TM"
    */
   private PropertyReader m_timelimit_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         long secs = readNumber( prop_name );
         if (!m_branch_stack.isEmpty())
         {
            logUnexpectedProperty( prop_name );
            return;
         }
         if (!isDuplicateInfoProp( PropertyID.TIME_LIMIT, prop_name, Long.toString( secs ) ))
         {
            if (secs <= 0)
            {
               logUnexpectedValue( prop_name, Long.toString( secs ) );
               return;
            }
            m_record.getGameProps().set( PropertyID.TIME_LIMIT, (int) secs );
         }
      }
   };

   /**
    * property: "ST"
    */
   private PropertyReader m_varstyle_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         String text = readRawValue( prop_name );
         if (text.length() != 1)
            throwUnexpectedValue( prop_name, text );
         if (!isDuplicateInfoProp( PropertyID.RENDERING_STYLE, prop_name, text ))
            m_record.getGameProps().set( PropertyID.RENDERING_STYLE, Convert.charToRenderingStyle( text.charAt( 0 ), m_lexer.getLineNumber() ) );
      }
   };

   /**
    * property: "VW"
    */
   private PropertyReader m_view_reader = new PropertyReader()
   {
      @Override
      public void read( String prop_name ) throws IOException
      {
         Set<Short> points = readPointList( prop_name );
         m_curr_branch.add( PropertyID.NOTE, ImmutablePair.of( m_node_num, ImmutablePair.of( Constants.VISIBILITY_NOTE, "" ) ) ); //$NON-NLS-1$
         if (points.size() > 0)
            m_curr_branch.add( PropertyID.VISIBLE, points );
      }
   };

   private void readBranch() throws IOException
   {
      PositionProperties branch = new PositionProperties( m_record == null );
      if (m_record == null)
         m_record = new Record( branch );
      else
         m_branch_stack.push( m_curr_branch );
      m_curr_branch = branch;

      boolean right_paren = false;
      while (m_lexer.hasMore() && !right_paren)
      {
         m_lexer.consumeWhitespace();
         if (!m_lexer.hasMore())
            return;
         ImmutablePair<Integer,String> token = m_lexer.getNext();
         switch (token.getLeft())
         {
            case SEMI_COLON_TOKEN_ID:
               m_curr_branch.add( PropertyID.NOTE, ImmutablePair.of( ++m_node_num, ImmutablePair.of( Constants.NODE_NOTE, "" ) ) ); //$NON-NLS-1$
               readNode();
               break;
            case LEFT_PAREN_TOKEN_ID:
               readBranch();
               break;
            case RIGHT_PAREN_TOKEN_ID:
               // TODO: check if no nodes read, and if none, add empty node
               right_paren = true;
               break;
            default:
               throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5030" ), token.getRight() ); //$NON-NLS-1$
         }
      }
      if (!right_paren)
         throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5031" ) ); //$NON-NLS-1$

      if (!m_branch_stack.isEmpty())
      {
         branch = m_branch_stack.pop();
         branch.add( PropertyID.BRANCH, m_curr_branch );
         m_curr_branch = branch;
      }
   }

   private void readNode() throws IOException
   {
      boolean move_prop = false;
      String prop_name = ""; //$NON-NLS-1$
      while (m_lexer.hasMore())
      {
         m_lexer.consumeWhitespace();
         if (!m_lexer.hasMore())
            throwUnexpectedEndOfInput();
         ImmutablePair<Integer,String> token = m_lexer.getNext();
         switch (token.getLeft())
         {
            case LEFT_PAREN_TOKEN_ID:
            case RIGHT_PAREN_TOKEN_ID:
            case SEMI_COLON_TOKEN_ID:
               m_lexer.pushBack( token );
               return;
            case LEFT_BRACKET_TOKEN_ID:
               if (prop_name.isEmpty())
                  throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5032" ), token.getRight() ); //$NON-NLS-1$
               readProperty( prop_name );
               prop_name = ""; //$NON-NLS-1$
               break;
            case Lexer.TEXT_TOKEN_ID:
               prop_name = token.getRight().toUpperCase();
               if (prop_name.equals( "B" ) || prop_name.equals( "W" )) //$NON-NLS-1$ //$NON-NLS-2$
               {
                  if (move_prop)
                     throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5033" ), token.getRight() ); //$NON-NLS-1$
                  else
                     move_prop = true;
               }
               break;
            default:
               throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5034" ), token.getRight() ); //$NON-NLS-1$
         }
      }
   }

   private void readNonStandardProperty( String prop_name ) throws IOException
   {
      m_messages.add( Resources.formatString( "msg.5035", m_lexer.getLineNumber(), prop_name ) ); //$NON-NLS-1$
      String text = readRawListValue( prop_name );
      int id = Integer.MAX_VALUE;
      if (!m_nonstd_ids.containsKey( prop_name ))
      {
         id = m_next_id++;
         m_nonstd_ids.put( prop_name, id );
      }
      else
      {
         id = m_nonstd_ids.get( prop_name );
      }
      ImmutablePair<String,String> value = ImmutablePair.of( prop_name, text );
      if (m_branch_stack.isEmpty())
         m_record.getGameProps().set( id, value );
      else
         m_curr_branch.add( id, value );
   }

   private void readProperty( String prop_name ) throws IOException
   {
      initStandardPropsTable();
      if (!m_lexer.hasMore())
         throwUnexpectedEndOfInput();
      ImmutablePair<Boolean,PropertyReader> reader = m_prop_readers.get( prop_name );
      if (reader == null)
      {
         readNonStandardProperty( prop_name );
      }
      else
      {
         m_value_required = reader.getLeft();
         reader.getRight().read( prop_name );
      }
   }

   private Set<ImmutablePair<Short,String>> readAnnotatedPointList( String prop_name ) throws IOException
   {
      HashSet<ImmutablePair<Short,String>> points = new HashSet<ImmutablePair<Short,String>>();
      m_lexer.enable( COLON_TOKEN_ID, false );
      try
      {
         while (m_lexer.hasMore())
         {
            ImmutablePair<Integer,String> token = m_lexer.getNext();
            if (token.getLeft() == RIGHT_BRACKET_TOKEN_ID)
               break;
            if (token.getLeft() != Lexer.TEXT_TOKEN_ID)
               throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5036" ), token.getRight() ); //$NON-NLS-1$
            ImmutablePair<Short,String> ann_pnt = Convert.stringToAnnotatedPoint( token.getRight(), m_lexer.getLineNumber() );
            for (ImmutablePair<Short,String> entry : points)
            {
               if (entry.getLeft() == ann_pnt.getLeft())
                  throw new PropertyException( Resources.formatString( "msg.5037", token.getRight() ) ); //$NON-NLS-1$
            }
            points.add( ann_pnt );
            if (!m_lexer.hasMore())
               throwUnexpectedEndOfInput();
            token = m_lexer.getNext();
            if (token.getLeft() != RIGHT_BRACKET_TOKEN_ID)
               throwRightBracketExpected( token.getRight() );
            if (!m_lexer.hasMore())
               throwUnexpectedEndOfInput();
            m_lexer.consumeWhitespace();
            token = m_lexer.getNext();
            while (token.getLeft() == Lexer.EOL_TOKEN_ID)
            {
               if (!m_lexer.hasMore())
                  throwUnexpectedEndOfInput();
               token = m_lexer.getNext();
            }
            if (token.getLeft() != LEFT_BRACKET_TOKEN_ID)
            {
               m_lexer.pushBack( token );
               break;
            }
         }
      }
      finally
      {
         m_lexer.enable( COLON_TOKEN_ID, true );
      }
      return points;
   }

   private Colors readColor( String prop_name ) throws IOException
   {
      String text = readRawValue( prop_name );
      if (text.length() != 1)
         throwUnexpectedValue( prop_name, text );
      char ch = Character.toLowerCase( text.charAt( 0 ) );
      switch (ch)
      {
         case 'b': return Colors.BLACK;
         case 'w': return Colors.WHITE;
         default:
            throwUnexpectedValue( prop_name, text );
            break;
      }
      return Colors.DONT_CARE;  // just to satisfy compiler, we never actually get here
   }

   private BigDecimal readDecimal( String prop_name ) throws IOException
   {
      String text = readRawValue( prop_name );
      BigDecimal value = BigDecimal.ZERO;
      try
      {
         value = new BigDecimal( text );
      }
      catch (NumberFormatException ex)
      {
         logUnexpectedValue( prop_name, text );
      }
      return value;
   }

   private long readNumber( String prop_name ) throws IOException
   {
      String text = readRawValue( prop_name );
      long value = 0L;
      try
      {
         value = Long.parseLong( text );
      }
      catch (NumberFormatException ex)
      {
         logUnexpectedValue( prop_name, text );
      }
      return value;
   }

   private Set<Short> readPointList( String prop_name ) throws IOException
   {
      HashSet<Short> points = new HashSet<Short>();
      m_lexer.enable( COLON_TOKEN_ID, false );
      try
      {
         while (m_lexer.hasMore())
         {
            ImmutablePair<Integer,String> token = m_lexer.getNext();
            if (token.getLeft() == RIGHT_BRACKET_TOKEN_ID)
               break;
            if (token.getLeft() != Lexer.TEXT_TOKEN_ID)
               throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5036" ), token.getRight() ); //$NON-NLS-1$
            boolean fail = false;
            try
            {
               int orig_count = points.size();
               ImmutablePair<Short,Short> pair = Convert.stringToPointPair( token.getRight(), m_lexer.getLineNumber() );
               int new_count = Layers.expandRect( pair, points );
               if (points.size() != orig_count + new_count)
               {
                  fail = true;
                  throw new PropertyException( Resources.formatString( "msg.5038", token.getRight() ) ); //$NON-NLS-1$
               }
            }
            catch (ParseException ex)
            {
               if (!fail)
               {
                  short pnt = Board.INVALID_POINT;
                  try
                  {
                     pnt = Convert.stringToPoint( token.getRight(), m_lexer.getLineNumber() );
                  }
                  catch (ParseException ex2)
                  {
                     rethrowWithInputPath( ex2 );
                  }
                  if (points.contains( pnt ))
                     throw new PropertyException( Resources.formatString( "msg.5039", token.getRight() ) ); //$NON-NLS-1$
                  points.add( pnt );
               }
               else
               {
                  rethrowWithInputPath( ex );
               }
            }
            if (!m_lexer.hasMore())
               throwUnexpectedEndOfInput();
            token = m_lexer.getNext();
            if (token.getLeft() != RIGHT_BRACKET_TOKEN_ID)
               throwRightBracketExpected( token.getRight() );
            if (!m_lexer.hasMore())
               throwUnexpectedEndOfInput();
            m_lexer.consumeWhitespace();
            token = m_lexer.getNext();
            while (token.getLeft() == Lexer.EOL_TOKEN_ID)
            {
               if (!m_lexer.hasMore())
                  throwUnexpectedEndOfInput();
               token = m_lexer.getNext();
            }
            if (token.getLeft() != LEFT_BRACKET_TOKEN_ID)
            {
               m_lexer.pushBack( token );
               break;
            }
         }
      }
      finally
      {
         m_lexer.enable( COLON_TOKEN_ID, true );
      }
      return points;
   }

   private String readRawListValue( String prop_name ) throws IOException
   {
      StringBuilder buffer = new StringBuilder();
      m_lexer.enable( RAW_LIST_PROP_DISABLED_TOKENS, false );
      boolean first = true;
      try
      {
         while (m_lexer.hasMore())
         {
            m_lexer.enable( PAREN_TOKENS, false );
            ImmutablePair<Integer,String> token = m_lexer.getNext();
            if (token.getLeft() == RIGHT_BRACKET_TOKEN_ID)
               break;
            if (token.getLeft() != Lexer.TEXT_TOKEN_ID)
               throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5040" ), token.getRight() ); //$NON-NLS-1$
            if (first)
               first = false;
            else
               buffer.append( "][" ); //$NON-NLS-1$
            buffer.append( token.getRight() );
            if (!m_lexer.hasMore())
               throwUnexpectedEndOfInput();
            token = m_lexer.getNext();
            if (token.getLeft() != RIGHT_BRACKET_TOKEN_ID)
               throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5041" ), token.getRight() ); //$NON-NLS-1$
            m_lexer.enable( PAREN_TOKENS, true );
            if (!m_lexer.hasMore())
               throwUnexpectedEndOfInput();
            m_lexer.consumeWhitespace();
            if (!m_lexer.hasMore())
               throwUnexpectedEndOfInput();
            token = m_lexer.getNext();
            while (token.getLeft() == Lexer.EOL_TOKEN_ID)
            {
               if (!m_lexer.hasMore())
                  throwUnexpectedEndOfInput();
               token = m_lexer.getNext();
            }
            if (token.getLeft() != LEFT_BRACKET_TOKEN_ID)
            {
               m_lexer.pushBack( token );
               break;
            }
         }
      }
      finally
      {
         m_lexer.enable( PAREN_TOKENS, true );
         m_lexer.enable( RAW_LIST_PROP_DISABLED_TOKENS, true );
      }
      return buffer.toString();
   }

   private String readRawValue( String prop_name ) throws IOException
   {
      ImmutablePair<Integer,String> token = null;
      String tk_text = ""; //$NON-NLS-1$
      m_lexer.enable( TEXT_PROP_DISABLED_TOKENS, false );
      try
      {
         token = m_lexer.getNext();
         if (token.getLeft() == RIGHT_BRACKET_TOKEN_ID)
         {
            if (m_value_required)
               m_messages.add( Resources.formatString( "msg.5042", m_lexer.getLineNumber(), prop_name ) ); //$NON-NLS-1$
            return ""; //$NON-NLS-1$
         }
         else if (token.getLeft() != Lexer.TEXT_TOKEN_ID)
         {
            String msg = Resources.formatString( "msg.5043", prop_name, token.getLeft() ); //$NON-NLS-1$
            throw new ParseException( m_input_path, m_lexer.getLineNumber(), msg, token.getRight() );
         }
         else
         {
            tk_text = token.getRight();
         }
      }
      finally
      {
         m_lexer.enable( TEXT_PROP_DISABLED_TOKENS, true );
      }
      if (!m_lexer.hasMore())
         throwUnexpectedEndOfInput();
      token = m_lexer.getNext();
      if (token.getLeft() != RIGHT_BRACKET_TOKEN_ID)
         throwRightBracketExpected( token.getRight() );
      return tk_text;
   }

   private void initStandardPropsTable()
   {
      if (m_prop_readers != null)
         return;
      m_prop_readers = new HashMap<String,ImmutablePair<Boolean,PropertyReader>>();

      m_prop_readers.put( "AB", ImmutablePair.of( true,  m_stones_reader ) );        //$NON-NLS-1$
      m_prop_readers.put( "AE", ImmutablePair.of( true,  m_stones_reader ) );        //$NON-NLS-1$
      m_prop_readers.put( "AN", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "AP", ImmutablePair.of( true,  m_appprop_reader ) );       //$NON-NLS-1$
      m_prop_readers.put( "AR", ImmutablePair.of( true,  m_connector_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "AW", ImmutablePair.of( true,  m_stones_reader ) );        //$NON-NLS-1$
      m_prop_readers.put( "B",  ImmutablePair.of( false, m_move_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "BL", ImmutablePair.of( true,  m_timeleft_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "BM", ImmutablePair.of( true,  m_movegrade_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "BR", ImmutablePair.of( true,  m_playertext_reader ) );    //$NON-NLS-1$
      m_prop_readers.put( "BT", ImmutablePair.of( true,  m_playertext_reader ) );    //$NON-NLS-1$
      m_prop_readers.put( "C",  ImmutablePair.of( false, m_comment_reader ) );       //$NON-NLS-1$
      m_prop_readers.put( "CA", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "CP", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "CR", ImmutablePair.of( true,  m_markpnts_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "DD", ImmutablePair.of( true,  m_markpnts_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "DM", ImmutablePair.of( true,  m_posgrade_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "DO", ImmutablePair.of( false, m_movegrade_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "DT", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "EV", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "FF", ImmutablePair.of( true,  m_filefmt_reader ) );       //$NON-NLS-1$
      m_prop_readers.put( "FG", ImmutablePair.of( false, m_figure_reader ) );        //$NON-NLS-1$
      m_prop_readers.put( "GB", ImmutablePair.of( true,  m_posgrade_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "GC", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "GM", ImmutablePair.of( true,  m_gametype_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "GN", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "GW", ImmutablePair.of( true,  m_posgrade_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "HA", ImmutablePair.of( true,  m_handicap_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "HO", ImmutablePair.of( true,  m_note_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "IT", ImmutablePair.of( false, m_movegrade_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "KM", ImmutablePair.of( true,  m_komi_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "KO", ImmutablePair.of( false, m_illegalmove_reader ) );   //$NON-NLS-1$
      m_prop_readers.put( "LB", ImmutablePair.of( true,  m_label_reader ) );         //$NON-NLS-1$
      m_prop_readers.put( "LN", ImmutablePair.of( true,  m_connector_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "MA", ImmutablePair.of( true,  m_markpnts_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "MN", ImmutablePair.of( true,  m_note_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "N",  ImmutablePair.of( true,  m_note_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "OB", ImmutablePair.of( true,  m_movesleft_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "ON", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "OT", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "OW", ImmutablePair.of( true,  m_movesleft_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "PB", ImmutablePair.of( true,  m_playertext_reader ) );    //$NON-NLS-1$
      m_prop_readers.put( "PC", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "PL", ImmutablePair.of( true,  m_playerturn_reader ) );    //$NON-NLS-1$
      m_prop_readers.put( "PM", ImmutablePair.of( true,  m_printing_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "PW", ImmutablePair.of( true,  m_playertext_reader ) );    //$NON-NLS-1$
      m_prop_readers.put( "RE", ImmutablePair.of( true,  m_result_reader ) );        //$NON-NLS-1$
      m_prop_readers.put( "RO", ImmutablePair.of( true,  m_round_reader ) );         //$NON-NLS-1$
      m_prop_readers.put( "RU", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "SL", ImmutablePair.of( true,  m_markpnts_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "SO", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "SQ", ImmutablePair.of( true,  m_markpnts_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "ST", ImmutablePair.of( true,  m_varstyle_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "SZ", ImmutablePair.of( true,  m_size_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "TB", ImmutablePair.of( true,  m_territory_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "TE", ImmutablePair.of( true,  m_movegrade_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "TM", ImmutablePair.of( true,  m_timelimit_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "TR", ImmutablePair.of( true,  m_markpnts_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "TW", ImmutablePair.of( true,  m_territory_reader ) );     //$NON-NLS-1$
      m_prop_readers.put( "UC", ImmutablePair.of( true,  m_posgrade_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "US", ImmutablePair.of( true,  m_gameinfo_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "V",  ImmutablePair.of( true,  m_note_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "VW", ImmutablePair.of( false, m_view_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "W",  ImmutablePair.of( false, m_move_reader ) );          //$NON-NLS-1$
      m_prop_readers.put( "WL", ImmutablePair.of( true,  m_timeleft_reader ) );      //$NON-NLS-1$
      m_prop_readers.put( "WR", ImmutablePair.of( true,  m_playertext_reader ) );    //$NON-NLS-1$
      m_prop_readers.put( "WT", ImmutablePair.of( true,  m_playertext_reader ) );    //$NON-NLS-1$
   }

   private boolean isDuplicateInfoProp( int prop_id, String prop_name, String prop_value )
   {
      boolean dup = m_record.getGameProps().has( prop_id );
      if (dup)
         logDuplicateProperty( prop_name, prop_value );
      return dup;
   }

   private void logDuplicateProperty( String prop_name, String prop_value )
   {
      m_messages.add( Resources.formatString( "msg.5044", m_lexer.getLineNumber(), prop_name, prop_value ) ); //$NON-NLS-1$
   }

   private void logUnexpectedProperty( String prop_name )
   {
      m_messages.add( Resources.formatString( "msg.5045", m_lexer.getLineNumber(), prop_name ) ); //$NON-NLS-1$
   }

   private void logUnexpectedValue( String prop_name, String prop_value )
   {
      m_messages.add( Resources.formatString( "msg.5046", m_lexer.getLineNumber(), prop_name, prop_value ) ); //$NON-NLS-1$
   }

   private void rethrowWithInputPath( ParseException orig )
   {
      throw new ParseException( m_input_path, orig.getLineNumber(), orig.getDetails(), orig.getToken() );
   }

   private void throwRightBracketExpected( String text )
   {
      throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5041" ), text ); //$NON-NLS-1$
   }

   private void throwUnexpectedEndOfInput()
   {
      throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.getString( "msg.5047" ) ); //$NON-NLS-1$
   }

   private void throwUnexpectedProperty( String prop_name )
   {
      throw new IllegalStateException( Resources.formatString( "msg.5048", m_input_path, m_lexer.getLineNumber(), prop_name ) ); //$NON-NLS-1$
   }

   private void throwUnexpectedValue( String prop_name, String text )
   {
      throw new ParseException( m_input_path, m_lexer.getLineNumber(), Resources.formatString( "msg.5049", prop_name ), text ); //$NON-NLS-1$
   }

   private static final int    LEFT_PAREN_TOKEN_ID    = 0;
   private static final int    RIGHT_PAREN_TOKEN_ID   = 1;
   private static final int    LEFT_BRACKET_TOKEN_ID  = 2;
   private static final int    RIGHT_BRACKET_TOKEN_ID = 3;
   private static final int    COLON_TOKEN_ID         = 4;
   private static final int    SEMI_COLON_TOKEN_ID    = 5;

   private static final ImmutablePair<Integer,String> LEFT_PAREN    = ImmutablePair.of( LEFT_PAREN_TOKEN_ID,    "(" ); //$NON-NLS-1$
   private static final ImmutablePair<Integer,String> RIGHT_PAREN   = ImmutablePair.of( RIGHT_PAREN_TOKEN_ID,   ")" ); //$NON-NLS-1$
   private static final ImmutablePair<Integer,String> LEFT_BRACKET  = ImmutablePair.of( LEFT_BRACKET_TOKEN_ID,  "[" ); //$NON-NLS-1$
   private static final ImmutablePair<Integer,String> RIGHT_BRACKET = ImmutablePair.of( RIGHT_BRACKET_TOKEN_ID, "]" ); //$NON-NLS-1$
   private static final ImmutablePair<Integer,String> COLON         = ImmutablePair.of( COLON_TOKEN_ID,         ":" ); //$NON-NLS-1$
   private static final ImmutablePair<Integer,String> SEMI_COLON    = ImmutablePair.of( SEMI_COLON_TOKEN_ID,    ";" ); //$NON-NLS-1$

   private static final int[] PAREN_TOKENS = new int[]
   {
      LEFT_PAREN_TOKEN_ID, RIGHT_PAREN_TOKEN_ID, Lexer.EOL_TOKEN_ID
   };
   private static final int[] RAW_LIST_PROP_DISABLED_TOKENS = new int[]
   {
      Lexer.WS_TOKEN_ID, COLON_TOKEN_ID,
      Lexer.EOL_TOKEN_ID, Lexer.DECIMAL_TOKEN_ID, Lexer.INTEGER_TOKEN_ID
   };
   private static final int[] TEXT_PROP_DISABLED_TOKENS = new int[]
   {
      Lexer.WS_TOKEN_ID, LEFT_PAREN_TOKEN_ID, RIGHT_PAREN_TOKEN_ID, LEFT_BRACKET_TOKEN_ID,
      COLON_TOKEN_ID, SEMI_COLON_TOKEN_ID, Lexer.EOL_TOKEN_ID, Lexer.DECIMAL_TOKEN_ID,
      Lexer.INTEGER_TOKEN_ID
   };

   private static final String   UNKNOWN_PLAYER  = Resources.getString( "msg.5050" ); //$NON-NLS-1$

   private static final String[] CAPABILITIES    = new String[]
   {
      com.dakengo.libs.application.Constants.MODE_CAP_PREFIX + com.dakengo.libs.application.Constants.READER_CAP_SUFFIX,
      com.dakengo.libs.application.Constants.FORMAT_CAP_PREFIX + Constants.FORMAT
   };

   private Lexer                      m_lexer;
   private String                     m_input_path       = ""; //$NON-NLS-1$
   private List<String>               m_messages         = new ArrayList<String>();
   private Record                     m_record;
   private PositionProperties         m_curr_branch;
   private Deque<PositionProperties>  m_branch_stack     = new ArrayDeque<PositionProperties>();
   private Map<String,Integer>        m_nonstd_ids       = new HashMap<String,Integer>();
   private int                        m_node_num         = -1;
   private int                        m_sgf_version      = Constants.DEFAULT_SGF_VERSION;
   private int                        m_next_id          = PropertyID.MAX_STD_ID + 1;
   private short                      m_white_move       = Board.INVALID_POINT;     // members to help deal with SGF where several black moves
   private Set <Short>                m_black_moves      = new HashSet<Short>();    // in a row are used to place handicap stones (as on KGS)
   private boolean                    m_black_stones     = false;
   private boolean                    m_value_required   = false;
   private Map<String,ImmutablePair<Boolean,PropertyReader>> m_prop_readers;
}
