//
// SGFWriter.java
//
// Copyright (c) 2010-2013, Jack D. Unrue
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.dakengo.plugins.record.sgf;

import java.io.IOException;
import java.io.Writer;
import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import net.xeoh.plugins.base.annotations.PluginImplementation;
import net.xeoh.plugins.base.annotations.Capabilities;
import org.apache.commons.lang3.tuple.ImmutablePair;

import com.dakengo.libs.game.Colors;
import com.dakengo.libs.game.Game;
import com.dakengo.libs.game.Result;
import com.dakengo.libs.record.GameProperties;
import com.dakengo.libs.record.PositionProperties;
import com.dakengo.libs.record.PropertyException;
import com.dakengo.libs.record.PropertyID;
import com.dakengo.libs.record.PropertyVisitor;
import com.dakengo.libs.record.Record;
import com.dakengo.libs.record.RecordWriter;

/**
 * Writes out game records in the SGF format.
 * 
 * @author <a href="mailto:jdunrue@gmail.com">Jack D. Unrue</a>
 * 
 * @see <a href="http://www.red-bean.com/sgf/">SGF File Format<a>
 */
@PluginImplementation
public class SGFWriter implements RecordWriter
{
   public SGFWriter()
   {
   }

   @Override
   public void setOutput( Writer writer )
   {
      m_writer = writer;
   }

   @Override
   public void write( Record rec ) throws IOException
   {
      if (m_writer == null)
         throw new IllegalStateException( Resources.getString( "msg.5051" ) ); //$NON-NLS-1$

      GameProperties props = rec.getGameProps();
      if (props.getPlayerName( Colors.BLACK ).isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.5025" ) ); //$NON-NLS-1$
      if (props.getPlayerName( Colors.WHITE ).isEmpty())
         throw new IllegalArgumentException( Resources.getString( "msg.5025" ) ); //$NON-NLS-1$

      m_buffer = new StringBuilder();
      m_column = 1;
      m_print_semi = true;
      m_app_name = "";    //$NON-NLS-1$
      m_app_version = ""; //$NON-NLS-1$
      initStandardPropsTable();

      m_visitor.startBranch( null );
      // Record.visit() is implemented to traverse the game info props first,
      // but since the writer needs to be sure that the game props are processed first,
      // be explicit about it here instead of making an assumption.
      //
      prettyPrint( "GM", "1" );   // Go game records are game type '1'        //$NON-NLS-1$ //$NON-NLS-2$
      prettyPrint( "FF", Integer.toString( Constants.DEFAULT_SGF_VERSION ) ); //$NON-NLS-1$
      rec.visit( m_visitor, true );
      m_visitor.finishBranch( null );

      m_writer.write( m_buffer.toString() );
      m_writer.flush();
   }

   /**
    * Returns an array of strings specifying the capabilities of this plugin.
    */
   @Capabilities
   public String[] getCapabilities()
   {
      return CAPABILITIES;
   }

   private interface PropertyWriter
   {
      void write( int id, Object value );
   }

   /**
    * property: "KM"
    */
   private PropertyWriter m_decimal_writer = new PropertyWriter()
   {
      @Override
      public void write( int unused, Object value )
      {
         prettyPrint( "KM", Convert.decimalToString( (BigDecimal) value ) ); //$NON-NLS-1$
      }
   };

   /**
    * property: "B" or "W"
    */
   @SuppressWarnings( "unchecked" )
   private PropertyWriter m_move_writer = new PropertyWriter()
   {
      @Override
      public void write( int unused, Object value )
      {
         String prop_name = ""; //$NON-NLS-1$
         ImmutablePair<Colors,Short> move = (ImmutablePair<Colors,Short>) value;
         if (move.getLeft() == Colors.BLACK)
            prop_name = "B"; //$NON-NLS-1$
         else
            prop_name = "W"; //$NON-NLS-1$
         prettyPrint( prop_name, Convert.pointToString( move.getRight() ) );
      }
   };

   /**
    * property: "SZ"
    */
   private PropertyWriter m_number_writer = new PropertyWriter()
   {
      @Override
      public void write( int unused, Object value )
      {
         prettyPrint( "SZ", ((Integer) value).toString() ); //$NON-NLS-1$
      }
   };

   /**
    * property: "HO", "KO", "MN", "N", or "V"
    */
   @SuppressWarnings( "unchecked" )
   private PropertyWriter m_note_writer = new PropertyWriter()
   {
      @Override
      public void write( int unused, Object value )
      {
         ImmutablePair<Integer,ImmutablePair<String,String>> n_val = (ImmutablePair<Integer,ImmutablePair<String,String>>) value;
         String note = n_val.getRight().getLeft();
         String text = n_val.getRight().getRight();
         switch (note)
         {
            case Constants.HOTSPOT_NOTE:
               prettyPrint( "HO", text ); //$NON-NLS-1$
               break;
            case Constants.ILLEGAL_NOTE:
               prettyPrint( "KO", text ); //$NON-NLS-1$
               break;
            case Constants.NAME_NOTE:
               prettyPrint( "N", text ); //$NON-NLS-1$
               break;
            case Constants.NODE_NOTE:
               m_print_semi = true;
               break;
            case Constants.NUMBER_NOTE:
               prettyPrint( "MN", Integer.toString( n_val.getLeft() ) ); //$NON-NLS-1$
               break;
            case Constants.VALUE_NOTE:
               prettyPrint( "V", text ); //$NON-NLS-1$
               break;
            default:
               throw new PropertyException( PropertyID.NOTE, note );
         }
      }
   };

   /**
    * property: "BR", "BT, "PB", "PW", "WR" or "WT"
    */
   @SuppressWarnings( "unchecked" )
   private PropertyWriter m_player_text_writer = new PropertyWriter()
   {
      @Override
      public void write( int prop_id, Object value )
      {
         String prop_name = ""; //$NON-NLS-1$
         ImmutablePair<Colors,String> p_val = (ImmutablePair<Colors,String>) value;
         Game.checkColor( p_val.getLeft() );
         switch (prop_id)
         {
            case PropertyID.PLAYER_NAME:
               prop_name = (p_val.getLeft() == Colors.BLACK ? "PB" : "PW"); //$NON-NLS-1$ //$NON-NLS-2$
               break;
            case PropertyID.PLAYER_RANK:
               prop_name = (p_val.getLeft() == Colors.BLACK ? "BR" : "WR"); //$NON-NLS-1$ //$NON-NLS-2$
               break;
            case PropertyID.TEAM_NAME:
               prop_name = (p_val.getLeft() == Colors.BLACK ? "BT" : "WT"); //$NON-NLS-1$ //$NON-NLS-2$
               break;
            default:
               throwUnexpectedPropertyID( prop_id );
               break;
         }
         prettyPrint( prop_name, Convert.transformText( p_val.getRight(), true ) );
      }
   };

   /**
    * property: "RE"
    */
   private PropertyWriter m_result_writer = new PropertyWriter()
   {
      @Override
      public void write( int unused, Object value )
      {
         prettyPrint( "RE", ((Result) value).toString() ); //$NON-NLS-1$
      }
   };

   /**
    * property: "AB", "AE", or "AW"
    */
   @SuppressWarnings( "unchecked" )
   private PropertyWriter m_stones_writer = new PropertyWriter()
   {
      @Override
      public void write( int unused, Object value )
      {
         String prop_name = ""; //$NON-NLS-1$
         ImmutablePair<Colors,Set<Short>> s_val = (ImmutablePair<Colors,Set<Short>>) value;
         switch (s_val.getLeft())
         {
            case EMPTY:
               prop_name = "AE"; //$NON-NLS-1$
               break;
            case BLACK:
               prop_name = "AB"; //$NON-NLS-1$
               break;
            case WHITE:
               prop_name = "AW"; //$NON-NLS-1$
               break;
            default:
               throw new IllegalArgumentException( Resources.formatString( "msg.5052", s_val.getLeft() ) ); //$NON-NLS-1$
         }
         prettyPrint( prop_name, Convert.pointsToString( s_val.getRight() ) );
      }
   };

   /**
    * property: "AP", "C", "CA", or "DT"
    */
   private PropertyWriter m_text_writer = new PropertyWriter()
   {
      @Override
      public void write( int prop_id, Object value )
      {
         switch (prop_id)
         {
            case PropertyID.APP_NAME:
               m_app_name = Convert.transformText( (String) value, true );
               break;
            case PropertyID.APP_VERSION:
               m_app_version = Convert.transformText( (String) value, true );
               break;
            case PropertyID.CHARSET:
               prettyPrint( "CA", (String) value );                                //$NON-NLS-1$
               break;
            case PropertyID.COMMENT:
               prettyPrint( "C", Convert.transformText( (String) value, false ) ); //$NON-NLS-1$
               break;
            case PropertyID.DATE:
               prettyPrint( "DT", Convert.transformText( (String) value, true ) ); //$NON-NLS-1$
               break;
            default:
               throwUnexpectedPropertyID( prop_id );
               break;
         }
      }
   };

   private PropertyVisitor m_visitor = new PropertyVisitor()
   {
      @Override
      public boolean startBranch( PositionProperties branch )
      {
         if (!m_app_name.isEmpty() || !m_app_version.isEmpty())
         {
            StringBuilder buffer = new StringBuilder();
            buffer.append( m_app_name );
            if (!m_app_name.isEmpty())
               buffer.append( ':' );
            buffer.append( m_app_version );
            prettyPrint( "AP", buffer.toString() ); //$NON-NLS-1$
            m_app_name = "";                        //$NON-NLS-1$
            m_app_version = "";                     //$NON-NLS-1$
         }
         maybeNewline( 1 );
         m_buffer.append( '(' );
         m_column++;
         m_print_semi = true;
         return true;
      }

      @Override
      public void finishBranch( PositionProperties branch )
      {
         maybeNewline( 1 );
         m_buffer.append( ')' );
         m_column++;
      }

      @SuppressWarnings( "unchecked" )
      @Override
      public void accept( int id, Object value )
      {
         if (id > PropertyID.MAX_STD_ID)
         {
            ImmutablePair<String,String> tmp = (ImmutablePair<String,String>) value;
            prettyPrint( tmp.getLeft(), tmp.getRight() );
            return;
         }
         if (!m_prop_writers.containsKey( id ))
            throw new IllegalArgumentException( Resources.formatString( "msg.5053", id ) ); //$NON-NLS-1$
         (m_prop_writers.get( id )).write( id, value );
      }
   };

   private void maybeNewline( int new_len )
   {
      if (m_column + new_len >= MAX_COLUMN)
      {
         m_column = 1;
         m_buffer.append( EOL );
      }
   }

   private void prettyPrint( String prop_name, String prop_val )
   {
      int new_len = prop_name.length() + prop_val.length() + 2;
      if (m_print_semi)
         new_len++;
      maybeNewline( new_len );
      if (m_print_semi)
         m_buffer.append( ';' );
      m_print_semi = false;
      m_buffer.append( prop_name );
      m_buffer.append( '[' );
      m_buffer.append( prop_val );
      m_buffer.append( ']' );
      m_column += new_len;
   }

   private void initStandardPropsTable()
   {
      if (m_prop_writers != null)
         return;
      m_prop_writers = new HashMap<Integer,PropertyWriter>();

      m_prop_writers.put( PropertyID.APP_NAME,     m_text_writer );
      m_prop_writers.put( PropertyID.APP_VERSION,  m_text_writer );
      m_prop_writers.put( PropertyID.CHARSET,      m_text_writer );
      m_prop_writers.put( PropertyID.COMMENT,      m_text_writer );
      m_prop_writers.put( PropertyID.DATE,         m_text_writer );
      m_prop_writers.put( PropertyID.KOMI,         m_decimal_writer );
      m_prop_writers.put( PropertyID.MOVE,         m_move_writer );
      m_prop_writers.put( PropertyID.NOTE,         m_note_writer );
      m_prop_writers.put( PropertyID.PLAYER_NAME,  m_player_text_writer );
      m_prop_writers.put( PropertyID.PLAYER_RANK,  m_player_text_writer );
      m_prop_writers.put( PropertyID.RESULT,       m_result_writer );
      m_prop_writers.put( PropertyID.SIZE,         m_number_writer );
      m_prop_writers.put( PropertyID.STONES,       m_stones_writer );
      m_prop_writers.put( PropertyID.TEAM_NAME,    m_player_text_writer );
   }

   private void throwUnexpectedPropertyID( int id )
   {
      throw new IllegalArgumentException( Resources.formatString( "msg.5054", id ) ); //$NON-NLS-1$
   }

   private static final String   EOL          = System.getProperty( "line.separator" ); //$NON-NLS-1$
   private static final int      MAX_COLUMN   = 78;

   private static final String[] CAPABILITIES = new String[]
   {
      com.dakengo.libs.application.Constants.MODE_CAP_PREFIX + com.dakengo.libs.application.Constants.WRITER_CAP_SUFFIX,
      com.dakengo.libs.application.Constants.FORMAT_CAP_PREFIX + Constants.FORMAT
   };

   private StringBuilder m_buffer;
   private Writer        m_writer;
   private int           m_column      = 1;
   private boolean       m_print_semi  = true;
   private String        m_app_name    = ""; //$NON-NLS-1$
   private String        m_app_version = ""; //$NON-NLS-1$
   private Map<Integer,PropertyWriter> m_prop_writers;
}
